// ==============================================================
//                 ORBITER MODULE: bc_pad
//
// bc_pad.cpp
// Control module for bc_pad vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "gcAPI.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

//const double ANTENNA_OPERATING_SPEED = 0.8;
//const double CHUTE_OPERATING_SPEED = 0.03;
//const double PB_FUELMASS = 750.0;
//const double PB_ISP = 5e4;
//const double PB_MAXHOVERTH = 0;
//const double TUBE_COL = 1.0;

static const DWORD ntdvtx_geardown = 3;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(    0,-3.05,50), 1e6, 1e6, 0.7, 0.7},
	{_V(-43.3,-3.05,-25), 1e6, 1e6, 0.7, 0.7},
	{_V( 43.3,-3.05,-25), 1e6, 1e6, 0.7, 0.7}
};

//const VECTOR3 pist_trans = _V(0,-11.557,-11.877);

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
class bc_pad: public VESSEL3 {
public:
	//enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } chute_status;
	bc_pad (OBJHANDLE hVessel, int flightmodel)
		: VESSEL3 (hVessel, flightmodel) {}
	//void DefineAnimations (void);
	//void RevertAntenna (void);
	//void RevertChute (void);
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//void clbkPostStep (double simt, double simdt, double mjd);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);

	//THRUSTER_HANDLE th_hover, th_steam[4];
	//ANIMATIONCOMPONENT_HANDLE Section0, Section1, Section2, Section3, Section4, Section5;
	ATTACHMENTHANDLE ss_attach;


private:
	//UINT anim_chute, anim_pist, anim_trans; 
	//UINT anim_umb1_rot, anim_umb1_scale,anim_umb2_rot, anim_umb2_scale,anim_umb3_rot, anim_umb3_scale, anim_umb_col;
	//double chute_proc, pist_proc, trans_proc, umb_proc;
	int light_status, lox_status, launch_status;
	SpotLight *slca1_light;
	SpotLight *slca2_light;
	double ps_level, timer, gravity, ps_level2, timer2, gcadjust;
};
// Define animation sequences for moving parts
// --------------------------------------------------------------
//void bc_pad::DefineAnimations (void)
//{
//	// 1. UMBILICALS
//	//anim_ant = CreateAnimation (0);
//	//static UINT HiGainAnt1Grp[3] = {7,8,9};
//
//	//static MGROUP_ROTATE  HiGainAnt1 (0, HiGainAnt1Grp, 3, _V(3.125,0,3.915), _V(0,1,0), (float)(PI*0.611));
//	//AddAnimationComponent (anim_ant, 0.2, 1,   &HiGainAnt1);
//
//
//
//	// 2. TOWER 
//	anim_chute = CreateAnimation (0);
//	static UINT ChuteGrp1[4] = {4,12,13,18};
//
//	static MGROUP_ROTATE  ChuteFold1 (0, ChuteGrp1, 4, _V(0,12.748,-4.411),_V(1,0,0), -15*RAD);
//	Section0 = AddAnimationComponent (anim_chute, 0, 1,   &ChuteFold1);
//
//	// claws
//	static UINT ChuteGrp2[1] = {1};
//
//	static MGROUP_ROTATE  ChuteFold2 (0, ChuteGrp2, 1, _V(-1.361,0,-1.968),_V(0,1,0), -50*RAD);
//	AddAnimationComponent (anim_chute, 0, 0.05,   &ChuteFold2, Section0);
//
//	static UINT ChuteGrp3[1] = {0};
//
//	static MGROUP_ROTATE  ChuteFold3 (0, ChuteGrp3, 1, _V(1.361,0,-1.968),_V(0,1,0), 50*RAD);
//	AddAnimationComponent (anim_chute, 0, 0.05,   &ChuteFold3, Section0);
//
//
//	// pistons
//	anim_pist = CreateAnimation (0);
//	static UINT ChuteGrp4[1] = {14};
//
//	static MGROUP_ROTATE  ChuteFold4 (0, ChuteGrp4, 1, _V(0,17.856,-18.536),_V(-1,0,0), 90*RAD);
//	Section1 = AddAnimationComponent (anim_pist, 0, 1,   &ChuteFold4);
//
//	anim_trans = CreateAnimation (0);
//	static UINT ChuteGrp5[1] = {15};
//
//	VECTOR3 pist_trans = _V(0,-14.139,-14.145);
//
//	static MGROUP_TRANSLATE  ChuteFold5 (0, ChuteGrp5, 1, pist_trans);
//	AddAnimationComponent (anim_trans, 0, 1,   &ChuteFold5, Section1);
//
//	///// umb1
//	
//	static UINT UmbGrp1[1] = {5};
//
//	anim_umb1_rot = CreateAnimation (0);
//	static MGROUP_ROTATE  Umb1 (0, UmbGrp1, 0, _V(0,71.67,-4.26),_V(1,0,0), -15*RAD);
//	Section1 = AddAnimationComponent (anim_umb1_rot, 0, 1,   &Umb1, Section0);
//
//	anim_umb1_scale = CreateAnimation (0);
//	static MGROUP_SCALE  Umb1s (0, UmbGrp1, 0, _V(0,71.67,-4.26),_V(1,1,0.15));
//	Section2 = AddAnimationComponent (anim_umb1_scale, 0, 1, &Umb1s, Section1);
//
//	static MGROUP_ROTATE  Umb2 (0, UmbGrp1, 0, _V(0,71.67,-4.26),_V(1,0,0), 6*RAD);
//	Section3 = AddAnimationComponent (anim_umb1_rot, 0, 1,   &Umb2, Section2);
//
//	/////////////// umb2
//
//	static UINT UmbGrp2[1] = {3};
//
//	anim_umb2_rot = CreateAnimation (0);
//	static MGROUP_ROTATE  Umb12 (0, UmbGrp2, 0, _V(0,58.49,-3.74),_V(1,0,0), -15*RAD);
//	Section1 = AddAnimationComponent (anim_umb2_rot, 0, 1,   &Umb12, Section0);
//
//	anim_umb2_scale = CreateAnimation (0);
//	static MGROUP_SCALE  Umb1s2 (0, UmbGrp2, 0, _V(0,58.49,-3.74),_V(1,1,0.15));
//	Section2 = AddAnimationComponent (anim_umb2_scale, 0, 1, &Umb1s2, Section1);
//
//	static MGROUP_ROTATE  Umb22 (0, UmbGrp2, 0, _V(0,58.49,-3.74),_V(1,0,0), 6*RAD);
//	Section4 = AddAnimationComponent (anim_umb2_rot, 0, 1,   &Umb22, Section2);
//
//	///////////// umb3
//
//	static UINT UmbGrp3[1] = {2};
//
//	anim_umb3_rot = CreateAnimation (0);
//	static MGROUP_ROTATE  Umb13 (0, UmbGrp3, 0, _V(0,49.45,-2.96),_V(1,0,0), -15*RAD);
//	Section1 = AddAnimationComponent (anim_umb3_rot, 0, 1,   &Umb13, Section0);
//
//	anim_umb3_scale = CreateAnimation (0);
//	static MGROUP_SCALE  Umb1s3 (0, UmbGrp3, 0, _V(0,49.45,-2.90),_V(1,1,0.15));
//	Section2 = AddAnimationComponent (anim_umb3_scale, 0, 1, &Umb1s3, Section1);
//
//	static MGROUP_ROTATE  Umb23 (0, UmbGrp3, 0, _V(0,49.45,-2.96),_V(1,0,0), 1*RAD);
//	Section5 = AddAnimationComponent (anim_umb3_rot, 0, 1,   &Umb23, Section2);
//
//	/// collapse umb
//
//	anim_umb_col = CreateAnimation (0);
//
//	static UINT UmbGrp4[1] = {5};
//	static MGROUP_ROTATE  Umb14 (0, UmbGrp4, 1, _V(0,71.67,-4.26),_V(1,0,0), 45*RAD);
//	AddAnimationComponent (anim_umb_col, 0, 1,   &Umb14, Section3);
//
//	static UINT UmbGrp5[1] = {3};
//	static MGROUP_ROTATE  Umb15 (0, UmbGrp5, 1, _V(0,58.49,-3.74),_V(1,0,0), 45*RAD);
//	AddAnimationComponent (anim_umb_col, 0, 1,   &Umb15, Section4);
//
//	static UINT UmbGrp6[1] = {2};
//	static MGROUP_ROTATE  Umb16 (0, UmbGrp6, 1, _V(0,49.45,-2.96),_V(1,0,0), 45*RAD);
//	AddAnimationComponent (anim_umb_col, 0, 1,   &Umb16, Section5);
//}

//void bc_pad::RevertAntenna (void)
//{
//	ant_status = ((ant_status == DOOR_CLOSED || ant_status == DOOR_CLOSING) ?
//		DOOR_OPENING : DOOR_CLOSING);
//}

//void bc_pad::RevertChute (void)
//{
//	chute_status = ((chute_status == DOOR_CLOSED || chute_status == DOOR_CLOSING) ?
//		DOOR_OPENING : DOOR_CLOSING);
//}

// ==============================================================
// bc_pad class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void bc_pad::clbkSetClassCaps (FILEHANDLE cfg)
{
	light_status = 0;
	lox_status = 0;
	launch_status = 0;
	gcadjust = 0;
	ps_level = 0;
	ps_level2 = 0;
	timer = 0;
	timer2 = 0;
	gravity = 9.9;
	gcInitialize();
	// physical specs
	SetEmptyMass (500000);
	SetCW (1, 1, 1, 1);
	SetSize (9);
	SetCrossSections (_V(100,100,100));
	SetPMI (_V(100,100,100));
	SetTouchdownPoints(tdvtx_geardown,ntdvtx_geardown);
	SetRotDrag (_V(1,1,1));

	SetMeshVisibilityMode ((AddMesh ("spacex_starship/bc_pad")), MESHVIS_ALWAYS);
	SetCameraOffset (_V(0,7,-10));

	//PROPELLANT_HANDLE ph_vernier = CreatePropellantResource (PB_FUELMASS);

	PARTICLESTREAMSPEC contrail_hover1 = {
		0, 0.5, 15, 0.5, 15, 2.0, 1.5, 0.4, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		//PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		//PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC contrail_hover1a = {
		0, 0.3, 20, 5, 0.25, 2.0, 0.5, 0.75, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	PARTICLESTREAMSPEC contrail_hover1b = {
		0, 0.5, 20, 8, 0.15, 0.6, 0.5, 1.25, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	///// for d3d9
	PARTICLESTREAMSPEC contrail_hover2 = {
		0, 0.6, 15, 0.5, 15, 3.0, 3, 0.4, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		//PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		//PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC contrail_hover2a = {
		0, 0.4, 20, 5, 0.25, 2.0, 1, 0.75, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	PARTICLESTREAMSPEC contrail_hover2b = {
		0, 0.6, 20, 8, 0.15, 0.6, 1, 1.25, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};


	if (gcEnabled() == true)
	{
		AddParticleStream (&contrail_hover2, _V(0,4,0), _V(0,-1,0), &ps_level);
		AddParticleStream (&contrail_hover2a, _V(0,8, -4.45), _V(0,0,-1), &ps_level);
		AddParticleStream (&contrail_hover2a, _V(-4.3,8, 1.153), _V(-0.966,0,0.259), &ps_level);
		AddParticleStream (&contrail_hover2a, _V(4.3,8, 1.153), _V(0.966,0,0.259), &ps_level);
		AddParticleStream (&contrail_hover2b, _V(0,27, -4.45), _V(0,0,-1), &ps_level2);
		gcadjust = 0.005;
	}
	else 
	{
		AddParticleStream (&contrail_hover1, _V(0,4,0), _V(0,-1,0), &ps_level);
		AddParticleStream (&contrail_hover1a, _V(0,8, -4.45), _V(0,0,-1), &ps_level);
		AddParticleStream (&contrail_hover1a, _V(-4.3,8, 1.153), _V(-0.966,0,0.259), &ps_level);
		AddParticleStream (&contrail_hover1a, _V(4.3,8, 1.153), _V(0.966,0,0.259), &ps_level);
		AddParticleStream (&contrail_hover1b, _V(0,27, -4.45), _V(0,0,-1), &ps_level2);
	}
	


	//AddExhaustStream (th_hover, _V(0,40,-1.8), &contrail_hover);
	//AddExhaustStream (th_hover, _V(0,62.2,-1.8), &contrail_hover);

	//AddExhaustStream (th_steam, _V(0,8,0), &contrail_steam);


	// ***************** thruster definitions *******************

	COLOUR4 col_d = {1,1,0.9,0};
	COLOUR4 col_s = {1,1,0.9,0};
	COLOUR4 col_a = {0,0,0,0};
	COLOUR4 col_white = {1,1,1,0};
	slca1_light = (SpotLight*)AddSpotLight(_V(40,25,40), _V(-0.612,0.5,-0.612), 5000, 3.5e-4, 0, 3.5e-4, RAD*90, RAD*150, col_d, col_s, col_a);
	slca1_light->Activate(false);

	slca2_light = (SpotLight*)AddSpotLight(_V(-40,25,-40), _V(0.612,0.5,0.612), 5000, 3.5e-4, 0, 3.5e-4, RAD*90, RAD*150, col_d, col_s, col_a);
	slca2_light->Activate(false);

	ss_attach = CreateAttachment (false, _V( 0, 3.43, 0), _V( 0,1,0), _V(0,0,-1), "SS");

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new bc_pad (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (bc_pad*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void bc_pad::clbkPostCreation ()
{
	OBJHANDLE hGref = GetGravityRef();
	double gref_mass = oapiGetMass(hGref);
	VECTOR3 rel_pos = _V(0,0,0);
	GetRelativePos(hGref, rel_pos);
	double rad = length(rel_pos);
	gravity = GGRAV*gref_mass/(rad*rad);

}


void bc_pad::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	sprintf (cbuf, "%d %d %d %0.1f",  launch_status, light_status, lox_status, timer);
	oapiWriteScenario_string (scn, "STTS", cbuf);

}
void bc_pad::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "STTS", 4)) {
			sscanf (line+4, "%d%d%d%lf", &launch_status, &light_status, &lox_status, &timer);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
}


bool AddPayload(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	OBJHANDLE hPld = oapiGetObjectByName(cbuf);
	if (hPld == NULL)return false;
	else
	{
	VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
	//DWORD ad = vme->AttachmentCount(false);
	//if (ad == 0) return false;

	ATTACHMENTHANDLE hPldAtt;
	hPldAtt = vme->GetAttachmentHandle(false, 0);
	if(vme->GetAttachmentStatus(hPldAtt) != NULL)return false;

	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hPld);
	DWORD ac = v->AttachmentCount(true);
	if (ac == 0) return false;

	ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(true, 0);
	if(v->GetAttachmentStatus(hAtt) != NULL)return false;

	
	if (vme->AttachChild(hPld, hPldAtt, hAtt) == true) return true;
	}

	return false;
}


int bc_pad::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) {

  } else { // unmodified keys
    switch (key) {
	//case OAPI_KEY_G: // move mss
	//	RevertChute();
	//	return 1;
	//case OAPI_KEY_B: // move umbilicals
	//	RevertAntenna();
	//	return 1;

	case OAPI_KEY_K: // lights on/off
		if (light_status == 0) light_status = 1;
		else if (light_status == 1) light_status = 0;
		return 1;

	case OAPI_KEY_V: // LOX vent on/off
		if (lox_status == 0)lox_status = 1;
		else lox_status = 0;
		return 1;

//	case OAPI_KEY_M: // noise sup water
//		SetThrusterLevel (th_steam[2], 1);
//		SetThrusterLevel (th_steam[3], 1);
//		return 1;

	case OAPI_KEY_P:
		oapiOpenInputBox("Enter Launcher by Name",AddPayload, 0, 50, 0);
		return 1;
	}
  }
  return 0;
}
void bc_pad::clbkPreStep(double SimT, double SimDT, double MJD)
{
//	ENGINESTATUS es;
	OBJHANDLE hV = NULL;
	hV = GetAttachmentStatus(ss_attach);

	///////////// lox vents ///////////////////
	if (lox_status == 0)
	{
	if(hV == NULL) ps_level = 0;
	else ps_level = 1;
	}
	else ps_level = 0;

	if(oapiGetTimeAcceleration() > 10) ps_level = 0;

	if(ps_level == 1)
	{
		if(oapiRand() > (0.99+gcadjust) && ps_level2 == 0)ps_level2 = 1;
		else if(ps_level2 == 1)
		{
			timer2 = timer2 + SimDT;
			if(timer2 > 6)
			{
			timer2 = 0;
			ps_level2 = 0;
			}
		}
	}
	else ps_level2 = 0;
	
	/////////////// launch status and detach //////////
	
	if(launch_status == 0)
	{
	if (hV != NULL) launch_status = 1;	
	}
	if(launch_status == 1)
	{
		VESSEL *v = oapiGetVesselInterface (hV);
		//oapiGetEngineStatus(hV, &es);
		VECTOR3 thrust_vector = _V(0,0,0);
		v->GetThrustVector(thrust_vector);
		double weight = gravity*(v->GetMass());
		double net_z_force = thrust_vector.z - weight;
			if(net_z_force > 0)
			{
			DetachChild(ss_attach, 0.0);
			ps_level = 0;
			launch_status = 2;
			}
	}
	if(launch_status == 2) /// timing lights
		{
			timer = timer + SimDT;
			if(timer > 10)
			{
			launch_status = 0;
			light_status = 1;
			timer = 0;
			}

		}
	
	/////////////////////shadowcheck /  lights

	if(light_status == 0)
	{
	VECTOR3 rsuninv;
	VECTOR3 rsunloc;
	VECTOR3 rsun;
	GetGlobalPos(rsun);
	rsuninv = _V(-rsun.x, -rsun.y, -rsun.z);
	Global2Local(rsuninv,rsunloc);

	if (rsunloc.y < 0 && light_status == 0)
			{
			slca1_light->Activate(true);
			slca2_light->Activate(true);
			}
	else if (rsunloc.y > 0)
			{
			slca1_light->Activate(false);
			slca2_light->Activate(false);
			}
	}
	else if (light_status == 1)
	{
		slca1_light->Activate(false);
		slca2_light->Activate(false);
	}
	
	//////////////////////////////// ANIMATIONS
	
	}

//void bc_pad::clbkPostStep (double simt, double simdt, double mjd)
//{
//	
//	
//}

bool bc_pad::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	int s = hps->H;
	double d = (s*0.00130208);

	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}

	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	//int w1 = (int) (200*d);
	//int w2 = (int)(216*d);
	/*int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);*/

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"K = Night Lights On/Off");
	skp->Text ((10+roxl), (wd+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"V = LOX Vents On/Off");
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"P = Open Launcher Attachment Dialogue Box");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	return true;
	}