// ==============================================================
//
// ss_booster
// Control module for ss_booster vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const DWORD COMM_SEND_STR = 2300;
const DWORD COMM_SEND_AFS = 2301;
const DWORD COMM_SEND_EDLO = 2304;
const DWORD COMM_SEND_BBT = 2305;
const double ISP0 = 3200;
const double ISPV = 3490;
//const double MAINTH0 = 846200;
const double MAINTH1 = 2200000;//923300;
const double RCSTH = 764051;//38000; 2xeng 10deg gimbal
const double RCSTH2 = 3600;
const double FUELMASS = 3400000;//396500;//384000
const double DRYMASS = 180000;//22200;//19000;
const double FIN_SPEED = 0.5;
//const double RCS2MASS = 1000;
const double RCS2ISPV = 3000;
const double RCS2ISP0 = 3000;
const double M_E = 2.71828182845904523536;

const double mu = 398600439968871.2;
const double rt = 6371010;
#define GREEN RGB(125, 255, 125)

OBJHANDLE hGref;

static const DWORD ntdvtx_geardown = 7;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(    0, -8, -34.8), 1e7, 8e6, 1, 1},
	{_V(-6.9,4,-34.8), 1e7, 8e6, 1, 1},
	{_V( 6.9,4,-34.8), 1e7, 8e6, 1, 1},
	{_V(    0, 8, -34.8), 1e7, 8e6, 1},
	{_V(-6.9,-4,-34.8), 1e7, 8e6, 1},
	{_V( 6.9,-4,-34.8), 1e7, 8e6, 1},
	{_V(   0 ,0 ,   20), 1e7, 8e6, 1}
};


void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.1, 0.1, 0, 0.1, 0.1, 0.1, 0, -0.1, -0.1, -0.1, 0, -0.1, -0.1, -0.1, 0};
	//static const double CL[nabsc] = {0, 0.1, 0.12, 0.1, 0, 0.1, 0.12, 0.1, 0, -0.1, -0.12, -0.1, 0, -0.1, -0.12, -0.1, 0};
	
	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	//*cd = 0.55*(0.4 + oapiGetInducedDrag (*cl, 0.02, 0.2));
	*cd = 0.55*(0.4 + oapiGetInducedDrag (*cl, 0.02, 0.2) + oapiGetWaveDrag (M, 2, 4.5, 6, 0.2));
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.1, 0.1, 0, 0.1, 0.1, 0.1, 0, -0.1, -0.1, -0.1, 0, -0.1, -0.1, -0.1, 0};
	//static const double CL[nabsc] = {0, 0.1, 0.12, 0.1, 0, 0.1, 0.12, 0.1, 0, -0.1, -0.12, -0.1, 0, -0.1, -0.12, -0.1, 0};
	

	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	//*cd = 0.55*(0.4 + oapiGetInducedDrag (*cl, 0.02, 0.2));
	*cd = 0.55*(0.4 + oapiGetInducedDrag (*cl, 0.02, 0.2) + oapiGetWaveDrag (M, 2, 4.5, 6, 0.2));
}

// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class ss_booster: public VESSEL3 {
public:
	ss_booster (OBJHANDLE hVessel, int flightmodel);
	~ss_booster ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPostCreation();
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	bool clbkLoadGenericCockpit ();
	//void clbkVisualCreated (VISHANDLE vis, int refcount);
	
private:

	
	ATTACHMENTHANDLE att_catch, att_pad;
	THRUSTER_HANDLE th_main[3],th_rcs[12], th_group[9], th_rcs2[8], th_boom[2], th_dust;
	THGROUP_HANDLE htgrpmain, htgrppup,htgrppdown,htgrpbright, htgrpbleft, htgrpyright, htgrpyleft;
	PROPELLANT_HANDLE hProp;//, hRCSProp;
	OBJHANDLE bbtgtchk;
	OBJHANDLE bbchk;
	//ANIMATIONCOMPONENT_HANDLE Section0, Section1, Section2, Section3, Section4;
	PSTREAM_HANDLE at_trail;
	CTRLSURFHANDLE hElevator, hRudder, hAileron_l, hAileron_r;
	//SURFHANDLE bsx1, bsx2;//, bs3, bs4, bs5;
	MESHHANDLE hMhull;//, hMintstg;
	//DEVMESHHANDLE dhMeshx;
	SpotLight *spt_light;
	//PointLight *pnt_light;
	EXHAUSTSPEC main2_exhaust[3];
	EXHAUSTSPEC main1_exhaust[13];
	EXHAUSTSPEC main0_exhaust[33];

	double timetoimp;
	int reentry_correct_sw;

	void ConfigMain(void);
	void ConfigRCS(void);
	void RevertRCS(void);
	void DefineAnimations (void);
	void DoLand(void);
	void DoRAT(void);
	void DoLAT(void);
	void DoTWU(void);
	void DoHeadsUp(void);
	void GoToFlightVector();
	void SetCamDir();
	void DoBoostBack();
	//void DoBoostBack2();
	void KillAllRCS();
	//void DoCorrection();
	void DoReentryCorrection();
	//void ApplySkin();
	void DoBoom();
	void DoThrustVector();
	void DoLandingRCS();
	void RollToAZ();

	char refname[256];

	double step_time, old_ap_dist, bias, timer, yhdg, boom_sw;// skin_sw, boom_sw;//pitch_angle,
	int dp_status, eng_status, fins_deploy_status, landed_sw;
	int cam_status, boostback_sw, rat_sw, afs_status, rot_sw, rcs_sw, rc_sw;
	int alti_mode_sw2;

	UINT  anim_fins_rot_a, anim_fins_rot_b;//anim_fins_deploy,
	UINT anim_fins_bank;
	double  fins_deploy_proc, fins_rot_a_proc, fins_rot_b_proc, fins_bank_proc;
	double fb_dv_req, pfact , vrot_lim, impact_angle, impact_angle2;
	double dust_level;

	MGROUP_TRANSFORM  *FinsRotate[4];//*FinsDeploy[4],
	MGROUP_TRANSFORM *FinsRotateBank[4];
	
	VECTOR3 azim_local, vazim1, vazim2, flight_vector, yhrz;

	int swt, dust_sw, rchud_sw, land_sw, edlo_sw, mom_sw;
	double dotprod;
	int afs_status_mem, dp_status_mem, rat_sw_mem, boostback_sw_mem;
	
	VECTOR3 torq_vect;
	double pu_level, pd_level, yr_level, yl_level, mth_level;
	double pst_lvl_top, pst_lvl_btm;

	VECTOR3 main2_exhaust_dir;

	VECTOR3 catch_offset;
	bool tgt_att_valid;
};

ss_booster::ss_booster (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

ss_booster::~ss_booster ()
{
	int i;
	for (i = 0; i < 4; i++)
	{
//	delete FinsDeploy[i];
	delete FinsRotate[i];
	delete FinsRotateBank[i];
	}

	//delete bsx1;
	//delete bsx2;
	//oapiReleaseTexture (bsx2);
}

void ss_booster::DefineAnimations (void)
{
	
	///////////////////////////
	// fin deploy
	//anim_fins_deploy = CreateAnimation (0.0);

	//static UINT Finsgrp0[1] = {22};
	//FinsDeploy[0] = new MGROUP_ROTATE(0, Finsgrp0, 0, _V(-4.163, 2.403, 33.655), _V(0.5,0.866,0), PI05);
	//Section1 = AddAnimationComponent (anim_fins_deploy, 0, 1, FinsDeploy[0]);

	//static UINT Finsgrp1[1] = {17};
	//FinsDeploy[1] = new MGROUP_ROTATE(0, Finsgrp1, 0, _V(4.163, -2.403, 33.655), _V(0.5,0.866,0), -PI05);
	//Section2 = AddAnimationComponent (anim_fins_deploy, 0, 1, FinsDeploy[1]);

	//static UINT Finsgrp2[1] = {16};
	//FinsDeploy[2] = new MGROUP_ROTATE(0, Finsgrp2, 0, _V(4.163, 2.403, 33.655), _V(-0.5,0.866,0), -PI05);
	//Section3 = AddAnimationComponent (anim_fins_deploy, 0, 1, FinsDeploy[2]);

	//static UINT Finsgrp3[1] = {11};
	//FinsDeploy[3] = new MGROUP_ROTATE(0, Finsgrp3, 0, _V(-4.163, -2.403, 33.655), _V(-0.5,0.866,0), PI05);
	//Section4 = AddAnimationComponent (anim_fins_deploy, 0, 1, FinsDeploy[3]);


	///// fins rotate // OLD FOLDED
	/*anim_fins_rot_a = CreateAnimation (0.5);

	FinsRotate[0] = new MGROUP_ROTATE(0, Finsgrp0, 1, _V(-4.163, 2.403, 33.655), _V(0,0,1), PI05);
	AddAnimationComponent (anim_fins_rot_a, 0, 1, FinsRotate[0], Section1);

	FinsRotate[1] = new MGROUP_ROTATE(0, Finsgrp1, 1, _V(4.163, -2.403, 33.655), _V(0,0,1), -PI05);
	AddAnimationComponent (anim_fins_rot_a, 0, 1, FinsRotate[1], Section2);

	anim_fins_rot_b = CreateAnimation (0.5);

	FinsRotate[2] = new MGROUP_ROTATE(0, Finsgrp2, 1, _V( 4.163, 2.403, 33.655), _V(0,0,1), PI05);
	AddAnimationComponent (anim_fins_rot_b, 0, 1, FinsRotate[2], Section3);

	FinsRotate[3] = new MGROUP_ROTATE(0, Finsgrp3, 1, _V(-4.163, -2.403, 33.655), _V(0,0,1), -PI05);
	AddAnimationComponent (anim_fins_rot_b, 0, 1, FinsRotate[3], Section4);*/

	////////// new fixed fins
	static UINT Finsgrp0[1] = {18};
	static UINT Finsgrp1[1] = {19};
	static UINT Finsgrp2[1] = {20};
	static UINT Finsgrp3[1] = {11};

	anim_fins_rot_a = CreateAnimation (0.5);

	FinsRotate[0] = new MGROUP_ROTATE(0, Finsgrp0, 1, _V(-3.992, 2.305, 33.655), _V(0.866,-0.5,0), PI05);
	AddAnimationComponent (anim_fins_rot_a, 0, 1, FinsRotate[0]);

	FinsRotate[1] = new MGROUP_ROTATE(0, Finsgrp1, 1, _V(3.992, -2.305, 33.655), _V(0.866,-0.5,0), PI05);
	AddAnimationComponent (anim_fins_rot_a, 0, 1, FinsRotate[1]);

	anim_fins_rot_b = CreateAnimation (0.5);

	FinsRotate[2] = new MGROUP_ROTATE(0, Finsgrp2, 1, _V( 3.992, 2.305, 33.655), _V(0.866,0.5,0), -PI05);
	AddAnimationComponent (anim_fins_rot_b, 0, 1, FinsRotate[2]);

	FinsRotate[3] = new MGROUP_ROTATE(0, Finsgrp3, 1, _V(-3.992, -2.305, 33.655), _V(0.866,0.5,0), -PI05);
	AddAnimationComponent (anim_fins_rot_b, 0, 1, FinsRotate[3]);

	anim_fins_bank = CreateAnimation (0.5);

	static UINT Finsbankgrp0[1] = {18};
	static UINT Finsbankgrp1[1] = {19};
	static UINT Finsbankgrp2[1] = {20};
	static UINT Finsbankgrp3[1] = {11};

	FinsRotateBank[0] = new MGROUP_ROTATE(0, Finsbankgrp0, 1, _V(-3.992, 2.305, 33.655), _V(0.866,-0.5,0), (30*RAD));
	AddAnimationComponent (anim_fins_bank, 0, 1, FinsRotateBank[0]);

	FinsRotateBank[1] = new MGROUP_ROTATE(0, Finsbankgrp1, 1, _V(3.992, -2.305, 33.655), _V(0.866,-0.5,0), (-30*RAD));
	AddAnimationComponent (anim_fins_bank, 0, 1, FinsRotateBank[1]);

	FinsRotateBank[2] = new MGROUP_ROTATE(0, Finsbankgrp2, 1, _V( 3.992, 2.305, 33.655), _V(0.866,0.5,0), (-30*RAD));
	AddAnimationComponent (anim_fins_bank, 0, 1, FinsRotateBank[2]);

	FinsRotateBank[3] = new MGROUP_ROTATE(0, Finsbankgrp3, 1, _V(-3.992, -2.305, 33.655), _V(0.866,0.5,0), (30*RAD));
	AddAnimationComponent (anim_fins_bank, 0, 1, FinsRotateBank[3]);

}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void ss_booster::clbkSetClassCaps (FILEHANDLE cfg)
{
	strcpy(refname, "NULL");
	//dhMeshx = NULL;

	fins_deploy_proc = 0;
	fins_rot_a_proc = 0.5;
	fins_rot_b_proc= 0.5;
	fins_bank_proc = 0.5;
	fins_deploy_status = 0;

	boom_sw = 0;
	step_time = 0;
	cam_status = 0;
	old_ap_dist = 0;
	bias = 0;
	timer = 0;
	//pitch_angle = 90*RAD;
	dp_status = 0;
	afs_status = 0;
	//intfair_status = 0;
	eng_status = 0;
	//calc_sw = 0;
	rot_sw = 0;
	rcs_sw = 0;
	rc_sw = 0;
	dust_sw = 0;
	rchud_sw = 0;
	vrot_lim = 5*RAD;
	azim_local = _V(0,-1,0);
	vazim1 = _V(0,1,0);
	vazim2 = _V(0,1,0);
	flight_vector = _V(0,1,0);
	yhrz = _V(0,0,0);
	yhdg = 0;
	boostback_sw = 0;
	rat_sw = 0;
	fb_dv_req = 0;
	pfact = 0;
	timetoimp = 100;
	impact_angle = 70*RAD;
	impact_angle2 = 65*RAD;
	swt = 0;
	dotprod = 0;
	land_sw = 0;
	landed_sw  = 0;
//	skin_sw = 0;
	reentry_correct_sw = 0;
	dust_level = 0;

	afs_status_mem = 0;
	dp_status_mem = 0;
	rat_sw_mem = 0;
	boostback_sw_mem = 0;

	edlo_sw = 0;
	mom_sw = 0;

	torq_vect = _V(0,0,0);

	pu_level = 0;
	pd_level = 0;
	yr_level = 0;
	yl_level = 0;
	mth_level = 0;

	main2_exhaust_dir = _V(0,0,1);
	alti_mode_sw2 = 0;

	pst_lvl_top = 0;
	pst_lvl_btm = 0;

	catch_offset = _V(0,0,0);
	tgt_att_valid = true;

	DefineAnimations ();
	// physical vessel parameters
	SetSize (35);
	SetEmptyMass (DRYMASS);
	SetPMI(_V(360,360,10));
	SetCrossSections (_V(635,635,83));
	SetCW(0.5,4,1,1);
	SetRotDrag (_V(0.25,0.25,0.1));
	SetTouchdownPoints ( tdvtx_geardown, ntdvtx_geardown);

	// camera parameters
	SetCameraOffset (_V(0,5,16.0));
	SetCameraRotationRange(PI,PI,(0.4*PI),(0.4*PI));

	hMhull = oapiLoadMesh("spacex_starship/ss_booster");
	SetMeshVisibilityMode(AddMesh(hMhull),MESHVIS_ALWAYS);

	//bsx2 = oapiLoadTexture ("Falcon9R_V12/f9_hull_logo_char.dds", true);
	//bs4 = oapiLoadTexture ("Falcon9R_V12/f9logo_char.dds", true);
	// propellant resources
	
	hProp = CreatePropellantResource (FUELMASS,FUELMASS);
	//hRCSProp = CreatePropellantResource (RCS2MASS,RCS2MASS);

	// ***************** thruster definitions *******************

	

	th_rcs2[0] = CreateThruster (_V( 3.267, 3.267, 60), _V(-0.7071,-0.7071, 0), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);
	th_rcs2[1] = CreateThruster (_V(-3.267, 3.267, 60), _V( 0.7071,-0.7071, 0), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);
	th_rcs2[2] = CreateThruster (_V(-3.267,-3.267, 60), _V( 0.7071, 0.7071, 0), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);
	th_rcs2[3] = CreateThruster (_V( 3.267,-3.267, 60), _V(-0.7071, 0.7071, 0), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);

	th_rcs2[4] = CreateThruster (_V( 4.531, 0.764, 5.278), _V( 0,-0.342,0.9397), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);
	th_rcs2[5] = CreateThruster (_V( 4.531,-0.764, 5.278), _V( 0, 0.342,0.9397), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);
	th_rcs2[6] = CreateThruster (_V(-4.531, 0.764, 5.278), _V( 0,-0.342,0.9397), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);
	th_rcs2[7] = CreateThruster (_V(-4.531,-0.764, 5.278), _V( 0, 0.342,0.9397), RCSTH2, hProp, RCS2ISPV, RCS2ISP0);
	
	////////////////////////////

	th_rcs[0] = CreateThruster (_V( 0, 2, -35), _V( 1,0,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[1] = CreateThruster (_V( 0, 2, -35), _V(-1,0,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[2] = CreateThruster (_V( 0,-2, -35), _V( 1,0,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[3] = CreateThruster (_V( 0,-2, -35), _V(-1,0,0), RCSTH, hProp, ISPV, ISP0);

	th_rcs[4] = CreateThruster (_V( 0, 4, -35), _V( 1,0,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[5] = CreateThruster (_V( 0, 4, -35), _V(-1,0,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[6] = CreateThruster (_V( 0,-4, -35), _V( 1,0,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[7] = CreateThruster (_V( 0,-4, -35), _V(-1,0,0), RCSTH, hProp, ISPV, ISP0);

	th_rcs[8] = CreateThruster (_V( 2, 0, -35), _V(0, 1,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[9] = CreateThruster (_V( 2, 0, -35), _V(0,-1,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[10] = CreateThruster (_V(-2,0, -35), _V(0, 1,0), RCSTH, hProp, ISPV, ISP0);
	th_rcs[11] = CreateThruster (_V(-2,0, -35), _V(0,-1,0), RCSTH, hProp, ISPV, ISP0);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[11];
	htgrppup = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[10];
	htgrppdown = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	htgrpyright = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	htgrpyleft = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[4];
	th_group[1] = th_rcs[7];
	htgrpbright = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[5];
	th_group[1] = th_rcs[6];
	htgrpbleft = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);


	th_main[0] = CreateThruster (_V( 0, 0, -35), _V( 0,0,1), 33*MAINTH1, hProp, ISPV, ISP0);
	th_main[1] = CreateThruster (_V( 0, 0, -35), _V( 0,0,1), 13*MAINTH1, hProp, ISPV, ISP0);
	th_main[2] = CreateThruster (_V( 0, 0, -35), _V( 0,0,1), 3*MAINTH1, hProp, ISPV, ISP0);

	th_group[0] = th_main[0];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("spacex_starship/KeroExhaust3");
	//SURFHANDLE exhaust_rcs = oapiRegisterExhaustTexture ("spacex_starship/Exhaust_ss");
	SURFHANDLE exhaust_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");

	double exlenrcs = 8;
	double exwidrcs = 0.5;

	AddExhaust(th_rcs2[0],exlenrcs,exwidrcs, _V( 3.267, 3.267, 29.45), _V( 0.7071, 0.7071, 0), exhaust_rcs);
	AddExhaust(th_rcs2[1],exlenrcs,exwidrcs, _V(-3.267, 3.267, 29.45), _V(-0.7071, 0.7071, 0), exhaust_rcs);
	AddExhaust(th_rcs2[2],exlenrcs,exwidrcs, _V(-3.267,-3.267, 29.45), _V(-0.7071,-0.7071, 0), exhaust_rcs);
	AddExhaust(th_rcs2[3],exlenrcs,exwidrcs, _V( 3.267,-3.267, 29.45), _V( 0.7071,-0.7071, 0), exhaust_rcs);

	AddExhaust(th_rcs2[4],exlenrcs,exwidrcs,exhaust_rcs);
	AddExhaust(th_rcs2[5],exlenrcs,exwidrcs,exhaust_rcs);
	AddExhaust(th_rcs2[6],exlenrcs,exwidrcs,exhaust_rcs);
	AddExhaust(th_rcs2[7],exlenrcs,exwidrcs,exhaust_rcs);





	double exlen = 50;
	double exwid = 3.7;


	MATRIX3 rm;
	VECTOR3 exhaust_pos;

	///// main 0
	VECTOR3 exhaust_0pos[33];

	///central
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_0pos[i] = mul(rm,_V(0,-0.833,-34.9));

	main0_exhaust[i].th = th_main[0];
	main0_exhaust[i].level = NULL;
	main0_exhaust[i].lpos = &exhaust_0pos[i];
	main0_exhaust[i].ldir = NULL;
	main0_exhaust[i].lsize = 1.5*exlen;
	main0_exhaust[i].wsize = exwid;
	main0_exhaust[i].lofs = 0;
	main0_exhaust[i].modulate = 0.5;
	main0_exhaust[i].tex = exhaust;
	main0_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main0_exhaust[i].id = 0;
	AddExhaust (&main0_exhaust[i]);
	}


	///// inner

	
	for (int i = 3; i < 13; i++)
	{
	rm = rotm(_V(0,0,1),(36*RAD*i));
	exhaust_0pos[i] = mul(rm,_V(-0.352,-2.233,-34.9));

	main0_exhaust[i].th = th_main[0];
	main0_exhaust[i].level = NULL;
	main0_exhaust[i].lpos = &exhaust_0pos[i];
	main0_exhaust[i].ldir = NULL;
	main0_exhaust[i].lsize = 1.5*exlen;
	main0_exhaust[i].wsize = exwid;
	main0_exhaust[i].lofs = 0;
	main0_exhaust[i].modulate = 0.5;
	main0_exhaust[i].tex = exhaust;
	main0_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main0_exhaust[i].id = 0;
	AddExhaust (&main0_exhaust[i]);
	}

	///// outer
	for (int i = 13; i < 33; i++)
	{
	rm = rotm(_V(0,0,1),(18*RAD*i));
	exhaust_0pos[i] = mul(rm,_V(0,4.13,-34.253));

	main0_exhaust[i].th = th_main[0];
	main0_exhaust[i].level = NULL;
	main0_exhaust[i].lpos = &exhaust_0pos[i];
	main0_exhaust[i].ldir = NULL;
	main0_exhaust[i].lsize = exlen;
	main0_exhaust[i].wsize = exwid;
	main0_exhaust[i].lofs = 0;
	main0_exhaust[i].modulate = 0.5;
	main0_exhaust[i].tex = exhaust;
	main0_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main0_exhaust[i].id = 0;
	AddExhaust (&main0_exhaust[i]);
	}

	


	///// main 1
	VECTOR3 exhaust_1pos[13];

	///central
	for (int i = 0; i < 3; i++)
	{
	//int i = 8;
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_1pos[i] = mul(rm,_V(0,-0.833,-34.9));

	main1_exhaust[i].th = th_main[1];
	main1_exhaust[i].level = NULL;
	main1_exhaust[i].lpos = &exhaust_1pos[i];
	main1_exhaust[i].ldir = NULL;
	main1_exhaust[i].lsize = exlen;
	main1_exhaust[i].wsize = exwid;
	main1_exhaust[i].lofs = 0;
	main1_exhaust[i].modulate = 0.5;
	main1_exhaust[i].tex = exhaust;
	main1_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main1_exhaust[i].id = 0;
	AddExhaust (&main1_exhaust[i]);
	}

	///// inner
	for (int i = 3; i < 13; i++)
	{
	rm = rotm(_V(0,0,1),(36*RAD*i));
	exhaust_1pos[i] = mul(rm,_V(-0.352,-2.233,-34.9));

	main1_exhaust[i].th = th_main[1];
	main1_exhaust[i].level = NULL;
	main1_exhaust[i].lpos = &exhaust_1pos[i];
	main1_exhaust[i].ldir = NULL;
	main1_exhaust[i].lsize = exlen;
	main1_exhaust[i].wsize = exwid;
	main1_exhaust[i].lofs = 0;
	main1_exhaust[i].modulate = 0.5;
	main1_exhaust[i].tex = exhaust;
	main1_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main1_exhaust[i].id = 0;
	AddExhaust (&main1_exhaust[i]);
	}

	

	/////// main 2

	VECTOR3 exhaust_2pos[3];
	
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_2pos[i] = mul(rm,_V(0,-0.833,-34.9));

	main2_exhaust[i].th = th_main[2];
	main2_exhaust[i].level = NULL;
	main2_exhaust[i].lpos = &exhaust_2pos[i];
	main2_exhaust[i].ldir = &main2_exhaust_dir;
	main2_exhaust[i].lsize = exlen;
	main2_exhaust[i].wsize = exwid;
	main2_exhaust[i].lofs = 0;
	main2_exhaust[i].modulate = 0.5;
	main2_exhaust[i].tex = exhaust;
	main2_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main2_exhaust[i].id = 0;

	AddExhaust (&main2_exhaust[i]);
	}


	SURFHANDLE exhaust2 = oapiRegisterParticleTexture ("spacex_starship/Contrail4x");
	SURFHANDLE exhaust2x = oapiRegisterParticleTexture ("spacex_starship/Contrail3");
	SURFHANDLE exhaustrcs = oapiRegisterParticleTexture ("Contrail1");

	PARTICLESTREAMSPEC exhaust_main1 = {
		0, 1.0, 8, 200, 0.01, 0.05, 2.6, 0.3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 0,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exhaust2
	};

	PARTICLESTREAMSPEC exhaust_main2 = {
		0, 1.0, 8, 200, 0.01, 0.05, 2.5, 0.3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 0,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exhaust2x
	};

	PARTICLESTREAMSPEC exhaust_rcs2 = {
		0, 0.1, 20, 100, 0.1, 0.075, 40, 10, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0.5, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 0.4, 0.8,
		exhaustrcs
	};



	//SURFHANDLE pexhaustrcs = oapiRegisterParticleTexture ("Contrail3");
	//PARTICLESTREAMSPEC exhaust_rcs2 = {
	//	0, 0.2, 30, 20, 0.025, 0.035, -0.2, 0, PARTICLESTREAMSPEC::EMISSIVE,
	//	PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
	//	PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	//	pexhaustrcs
	//};

	PARTICLESTREAMSPEC atmostrail = {
		0, 7, 25, 2, 3, 1, -0.125, 5, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLIN, 0.5, 1,
		exhaustrcs
	};


	///// main0
	//// inner
	for (int i = 0; i < 10; i++)
	{
	rm = rotm(_V(0,0,1),(36*RAD*i));
	exhaust_pos = mul(rm,_V(-0.352,-2.233,-34.9));
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main2);
	}
	//// outer
	for (int i = 0; i < 20; i++)
	{
	rm = rotm(_V(0,0,1),(18*RAD*i));
	exhaust_pos = mul(rm,_V(0,4.13,-34.253));
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main2);
	}
	/// central
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_pos = mul(rm,_V(0,-0.833,-34.9));
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main2);
	}

	///// main1
	/// inner
	for (int i = 0; i < 10; i++)
	{
	rm = rotm(_V(0,0,1),(36*RAD*i));
	exhaust_pos = mul(rm,_V(-0.352,2.233,-35.11));
	AddExhaustStream(th_main[1],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[1],exhaust_pos,&exhaust_main2);
	}
	/// central
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_pos = mul(rm,_V(0,-0.833,-34.9));
	AddExhaustStream(th_main[1],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[1],exhaust_pos,&exhaust_main2);
	}

	//////////// main 2
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_pos = mul(rm,_V(0,-0.833,-34.9));
	AddExhaustStream(th_main[2],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[2],exhaust_pos,&exhaust_main2);
	//// double ups!
	AddExhaustStream(th_main[2],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[2],exhaust_pos,&exhaust_main2);
	}
	
	///////////////////////////////////////////////////////

	at_trail = AddExhaustStream(th_main[2],_V(0,0,-35.11),&atmostrail);

	//// rcs exhaust
	AddExhaustStream(th_rcs2[0], _V(3.267, 3.267, 29.45), &exhaust_rcs2);
	AddExhaustStream(th_rcs2[1], _V(-3.267, 3.267, 29.45), &exhaust_rcs2);
	AddExhaustStream(th_rcs2[2], _V(-3.267,-3.267, 29.45), &exhaust_rcs2);
	AddExhaustStream(th_rcs2[3], _V(3.267,-3.267, 29.45), &exhaust_rcs2);

	AddExhaustStream(th_rcs2[4], &exhaust_rcs2);
	AddExhaustStream(th_rcs2[5], &exhaust_rcs2);
	AddExhaustStream(th_rcs2[6], &exhaust_rcs2);
	AddExhaustStream(th_rcs2[7], &exhaust_rcs2);
	
	///////////////////////////// boom p streams

	th_boom[0] = CreateThruster (_V( 0, 0, 0), _V( 0,0,1), 0.001, hProp, 2, 1);
	th_boom[1] = CreateThruster (_V( 0, 0, 0), _V( 0,0,1), 0.001, hProp, 2, 1);

	SURFHANDLE boom1 = oapiRegisterParticleTexture ("Contrail4");
	SURFHANDLE boom2 = oapiRegisterParticleTexture ("Contrail3");

	PARTICLESTREAMSPEC boom_pst1 = {
		0, 8, 100, 150, 2, 2, 10, 4, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		boom1
	};

	PARTICLESTREAMSPEC boom_pst2 = {
		0, 8, 80, 100, 2, 1, 15, 2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		boom2
	};

	AddExhaustStream(th_boom[0],_V(0,0,0),&boom_pst1);
	AddExhaustStream(th_boom[1],_V(0,0,0),&boom_pst2);


	///////////////////////////////////////////////////////
	/////AirFoils

	CreateAirfoil (LIFT_VERTICAL, _V(0,0,0.1), VLiftCoeff, 50, 420, 0.16);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,0.1), HLiftCoeff, 50, 420, 0.16); //90
	//CreateVariableDragElement (1, 12, _V(0, 0, 1)); // landing gear drag + stabiliser
	CreateVariableDragElement (&fins_deploy_proc, 18, _V(0, 0, 1)); // grid fins 3
	////////

	//local light
	COLOUR4 col_d = {0.125,0.1,0.06125,0};
	COLOUR4 col_s = {0,0,0,0};
	COLOUR4 col_a = {0,0,0,0};
	COLOUR4 col_white = {1,1,1,0};
	spt_light = (SpotLight*)AddSpotLight(_V(0,0,-32), _V(0,0,-1),5000, 1e-1, 0, 1e-7, RAD*170, RAD*180, col_d, col_s, col_a);
	//pnt_light = (PointLight*)AddPointLight(_V(0,0,-22),20000, 5e-7, 0, 5e-7, col_d, col_s, col_a);
	//pnt_light->Activate(false);
	spt_light->Activate(false);

	/////////////////////////////// landing / takeoff dust

	th_dust = CreateThruster (_V(0, 0, -90), _V(0, 0, 1), 0, hProp, 1000000);

		SURFHANDLE exhaustdust = oapiRegisterParticleTexture ("Contrail1");

		PARTICLESTREAMSPEC dust_pstream = {
		0, 8, 20, 200, 0.5, 1.25, 20, 6, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 1, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 0.7e-7, 1,
		exhaustdust
	};

	AddExhaustStream(th_dust, &dust_pstream);

	///////// maxq pstreams ///////////////////////
	///max q pstream
	SURFHANDLE exhaust3 = oapiRegisterParticleTexture ("spacex_starship/contrail");
	PARTICLESTREAMSPEC maxqtrail = {
		0, 8, 20, 170, 0.01, 0.04, 7, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_LIN, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 0.3, 0.3,
		exhaust3
	};
	AddParticleStream(&maxqtrail,_V(0,0,30),_V(0,0,-1), &pst_lvl_top);
	AddParticleStream(&maxqtrail,_V(0,0,-30),_V(0,0,1), &pst_lvl_btm);
	
	///////// attachment ///////////////////
	att_catch = CreateAttachment(true,_V(0,0,33.33),_V(0,0,-1),_V(0,1,0),"ARM", false);
	att_pad = CreateAttachment(true,_V(0,0,-34),_V(0,0,-1),_V(0,1,0),"PAD", false);
}

void ss_booster::clbkPostCreation ()
{
 hGref = GetGravityRef();
 OBJHANDLE bbtgt = oapiGetObjectByName(refname);
 if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, refname);
 if (bbtgt == NULL)
 {
	boostback_sw = 0;
	strcpy(refname, "NULL");
	afs_status = 0;
 }

 if (fins_deploy_status == 2)
	{
	hElevator = CreateControlSurface3 (AIRCTRL_ELEVATOR, 16, 3, _V(0,0,35), AIRCTRL_AXIS_XPOS, 0.5);
	hRudder = CreateControlSurface3 (AIRCTRL_RUDDER, 16, 3, _V(0,0,35), AIRCTRL_AXIS_YPOS, 0.5);
	hAileron_l = CreateControlSurface3 (AIRCTRL_AILERON, 2, 3, _V(-4,0,35), AIRCTRL_AXIS_XPOS, 0.5);
	hAileron_r = CreateControlSurface3 (AIRCTRL_AILERON, 2, 3, _V(4,0,35), AIRCTRL_AXIS_XNEG, 0.5);
	}

  if (eng_status > 0) ConfigMain();
 


//if (skin_sw == 2) oapiSetTexture(hMhull,1,bsx2);
//  if (skin_sw == 2) ApplySkin();

}


// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new ss_booster (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (ss_booster*)vessel;
}


void ss_booster::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	for (int i = 0; i < (int)strlen(refname); i++)
	{
		if (refname[i] == ' ')
		{
			refname[i] = '+';
			break;
		}
	}

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	//sprintf (cbuf, "%d %0.4f %0.4f %0.4f %0.4f", leg_status, legs_proc, pist_rot_proc, pist_trans_proc, skin_sw);
	//oapiWriteScenario_string (scn, "LEGS", cbuf);

	sprintf (cbuf, "%d %0.4f %0.4f %0.4f", fins_deploy_status, fins_deploy_proc, fins_rot_a_proc, fins_rot_b_proc);
	oapiWriteScenario_string (scn, "FINS", cbuf);

	//sprintf (cbuf, "%d %d %d %d %d %d", eng_status, intfair_status, dp_status, boostback_sw, rat_sw, afs_status);
	sprintf (cbuf, "%d %d %d %d %d %d %d %0.4f %d %d", eng_status, dp_status, boostback_sw, rat_sw, afs_status, edlo_sw, mom_sw, boom_sw, rc_sw, reentry_correct_sw);
	oapiWriteScenario_string (scn, "APST", cbuf);

	sprintf (cbuf, "%s", refname);
	oapiWriteScenario_string (scn, "LPRM", cbuf);


}
void ss_booster::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
	//	if (!strnicmp (line, "LEGS", 4)) {
	//		sscanf (line+4, "%d%lf%lf%lf%lf", &leg_status, &legs_proc, &pist_rot_proc, &pist_trans_proc, &skin_sw);
	//	} else 
			if (!strnicmp (line, "FINS", 4)) {
		sscanf (line+4, "%d%lf%lf%lf", &fins_deploy_status, &fins_deploy_proc, &fins_rot_a_proc, &fins_rot_b_proc);
		} else if (!strnicmp (line, "APST", 4)) {
		sscanf (line+4, "%d%d%d%d%d%d%d%lf%d%d", &eng_status, &dp_status, &boostback_sw, &rat_sw, &afs_status, &edlo_sw, &mom_sw, &boom_sw, &rc_sw, &reentry_correct_sw);
		} else if (!strnicmp (line, "LPRM", 4)) {
			sscanf (line+4, "%s", &refname);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }


	//SetAnimation(anim_fins_deploy, fins_deploy_proc);
	SetAnimation(anim_fins_rot_a, fins_rot_a_proc);
	SetAnimation(anim_fins_rot_b, fins_rot_b_proc);

	for (int i = 0; i < (int)strlen(refname); i++)
	{
		if (refname[i] == '+')
		{
			refname[i] = ' ';
			break;
		}
	}

}



bool SetBBTgt(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	OBJHANDLE bbtgt = oapiGetObjectByName(cbuf);
	if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, cbuf);
	if (bbtgt == NULL)return false;
	else 
		{
		VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
		vme->SendBufferedKey(COMM_SEND_BBT, false, cbuf);
		return true;
		}
}

int ss_booster::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
	/////////////////////////////////

	if (key == COMM_SEND_STR) {
			strcpy(refname,kstate);
			return 1;
       }

	if (key == COMM_SEND_AFS) {
			
			bbtgtchk = oapiGetObjectByName(refname);
			if (bbtgtchk == NULL)bbtgtchk = oapiGetBaseByName(GetGravityRef(), refname);
			if (bbtgtchk == NULL) afs_status = 0;
			else afs_status = 1;
			return 1;
       }

	if (key == COMM_SEND_EDLO) {
			edlo_sw = 1;
			return 1;
       }

	if (key == COMM_SEND_BBT) {
			sscanf (kstate, "%s", &refname);
			return 1;
       }


	//////////////////////////////////
if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
	

	} 
  }

  else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) {

		//case OAPI_KEY_G: // legs
		//if (leg_status == 0)
		//	{
		//	SetTouchdownPoints ( tdvtx_geardown, ntdvtx_geardown);
		//	leg_status = 1;
		//	}
		//return 1;



		//case OAPI_KEY_M: // descent pilot
		//if (dp_status == 0)
		//	{
		//	rat_sw = 0;
		//	dp_status = 1;
		//	boostback_sw = 0;
		//	KillAllRCS();
		//	}
		//else if (dp_status == 1)
		//	{
		//	dp_status = 0;
		//	KillAllRCS();
		//	}
		//return 1;
		
		

		case OAPI_KEY_J: // deploy grid fins
		if (fins_deploy_status == 1)fins_deploy_status = 3;
		else if (fins_deploy_status == 2)fins_deploy_status = 3;
		else fins_deploy_status = 1;
		return 1;

		case OAPI_KEY_P: // configure main eng
		if (eng_status == 0) eng_status = 1;
		else if (eng_status == 1) eng_status = 2;
		else if (eng_status == 2) eng_status = 0;
		ConfigMain();
		return 1;

		case OAPI_KEY_K:
		hGref = GetGravityRef();
		oapiOpenInputBox("Enter Flyback Target Name",SetBBTgt, 0, 50, 0);
		return 1;

		case OAPI_KEY_M:
		if (edlo_sw == 0)edlo_sw = 1;
		else edlo_sw = 0;
		return 1;

		case OAPI_KEY_B: //
			bbchk = oapiGetObjectByName(refname);
			if (bbchk == NULL)bbchk = oapiGetBaseByName(GetSurfaceRef(), refname);
			if (bbchk == NULL)afs_status = 0;
			
			else if (afs_status_mem == 0)
			{
			if (afs_status == 0)
				{
			afs_status = 1;
			dp_status = 0;
			rat_sw = 0;
			boostback_sw = 0;
			KillAllRCS();
				}
			else
				{
			afs_status_mem = afs_status;
			afs_status = 0;
			dp_status_mem = dp_status;
			dp_status = 0;
			rat_sw_mem = rat_sw;
			rat_sw = 0;
			boostback_sw_mem = boostback_sw;
			boostback_sw = 0;
			KillAllRCS();
				}
			}
			else if (afs_status_mem > 0)
			{
				if (afs_status == 0)
				{
			afs_status = afs_status_mem;
			afs_status_mem = 0;
			dp_status = dp_status_mem;
			dp_status_mem = 0;
			rat_sw = rat_sw_mem;
			rat_sw_mem = 0;
			boostback_sw = boostback_sw_mem;
			boostback_sw_mem = 0;
			KillAllRCS();
				}
				else
				{
			afs_status_mem = afs_status;
			afs_status = 0;
			dp_status_mem = dp_status;
			dp_status = 0;
			rat_sw_mem = rat_sw;
			rat_sw = 0;
			boostback_sw_mem = boostback_sw;
			boostback_sw = 0;
			KillAllRCS();
				}
			}
		return 1;

		case OAPI_KEY_E:
			if (cam_status == 0)cam_status = 1;
			else if (cam_status == 1)cam_status = 0;
			SetCamDir();
		return 1;

		//case OAPI_KEY_V:
		//if (boostback_sw == 0)
		//	{
		//	rat_sw = 0;
		//	dp_status = 0;
		//	boostback_sw = 1;
		//	KillAllRCS();
		//	}
		//else if (boostback_sw > 0)
		//{
		//	boostback_sw = 0;
		//	KillAllRCS();
		//}
		//return 1;

		//case OAPI_KEY_N:
		//if (rat_sw == 0)
		//{
		//	rat_sw = 1;
		//	dp_status = 0;
		//	boostback_sw = 0;
		//	KillAllRCS();
		//}
		//else if (rat_sw > 0)
		//{
		//	rat_sw = 0;
		//	KillAllRCS();
		//}
		//return 1;
	}
  }
  return 0;
}




void ss_booster::clbkPreStep(double SimT, double SimDT, double MJD)
{

SetThrusterDir(th_main[eng_status], _V(0,0,1));
double getalt = GetAltitude(ALTMODE_GROUND);
rchud_sw = 0;

if (getalt < 5)
{
	if(GroundContact() == true && boom_sw == 0) boom_sw = 1;
}
if (boom_sw > 0 && boom_sw < 2)DoBoom();

////// char skin
//if (getalt < 100000)
//{
//	if (skin_sw < 1)
//	{
//		if (GetThrusterGroupLevel(THGROUP_MAIN) > 0)
//		{
//			VECTOR3 aspdloc = _V(0,0,0);
//			GetShipAirspeedVector(aspdloc);
//			if (aspdloc.z < 0)
//			{
//				skin_sw = skin_sw+(0.5*SimDT);
//				if (skin_sw > 1)
//				{
//					skin_sw = 2; // finished
//					ApplySkin();
//				}
//			}
//		}
//
//	}
//}
//////////// 
///// heat experiment
	//double asp = GetAirspeed ();
	//double atmden = GetAtmDensity();
	//double r_nose = 2;
	//double heatflux2 = atmden*asp*asp*asp*0.5;
	//double heatflux = 0.000183*asp*asp*asp*sqrt(atmden/r_nose);
	
///////////////

double dynp = GetDynPressure();
if(dynp > 20000)
	{
	VECTOR3 aspd_loc;
	GetAirspeedVector(FRAME_LOCAL,aspd_loc);
	normalise(aspd_loc);
	dynp = dynp*aspd_loc.z;
	if(dynp < 0)
		{
		pst_lvl_btm = -((dynp-20000)/10000);
		if (pst_lvl_btm > 1)pst_lvl_btm = 1;
		else if (pst_lvl_btm < 0)pst_lvl_btm = 0;
		pst_lvl_top = 0;
		}
	else
		{
		pst_lvl_top = -((dynp-20000)/10000);
		if (pst_lvl_top > 1)pst_lvl_top = 1;
		else if (pst_lvl_top < 0)pst_lvl_top = 0;
		pst_lvl_btm = 0;

		}
	
	}
	else 
	{
	pst_lvl_top = 0;
	pst_lvl_btm = 0;
	}

////// landing / takeoff dust pstream  /////////////////////////////

	double altdust = GetAltitude(ALTMODE_GROUND);
	if(altdust < 90)
	{
	double sin_pitch = sin(GetPitch());
	if(sin_pitch > 0.7) 
		{
		altdust = altdust/sin_pitch;
		dust_level = GetThrusterGroupLevel(THGROUP_MAIN);
		SetThrusterRef(th_dust, _V(0,0,-altdust));
		}
	else dust_level = 0;
	}
	else dust_level = 0;
	SetThrusterLevel(th_dust, dust_level);

step_time = SimDT;///////
double simdt = SimDT;

double mth_lev = GetThrusterLevel(th_main[eng_status]);
if (mth_lev == 0 && rcs_sw == 0) ConfigRCS();
else if (mth_lev > 0 && rcs_sw == 1) RevertRCS();

//sprintf(oapiDebugString(), "rcs_sw %d", rcs_sw);

////////////////engine light
if (mth_lev > 0)
{
	double rndm = 0.15*(oapiRand());
	rndm = (mth_lev*0.9) - rndm;
	if (rndm < 0)rndm = 0;
	else if (rndm > 1)rndm = 1;
	spt_light->Activate(true);
	spt_light->SetIntensity(rndm);

}
else spt_light->Activate(false);

if (afs_status == 1)
	{
	timer = timer+SimDT;
	if (timer > 2)
		{
		afs_status = 2;
		timer = 0;
		}
	}
else if(afs_status == 2)
	{
	if (edlo_sw == 1) 
		{
		boostback_sw = 0;		
		afs_status = 3;
		}
	else if (boostback_sw == 0)boostback_sw = 1;
	}
else if(afs_status == 3)
	{
	if (rat_sw == 0 && GetNavmodeState(NAVMODE_KILLROT) == false)rat_sw = 1;
	if (getalt > 15000 && rc_sw == 0) //30000
		{
			VECTOR3 haspv;
			GetAirspeedVector(FRAME_HORIZON, haspv);
			if (haspv.y < 0)
			{
			if (fins_deploy_status == 0)fins_deploy_status = 1;
			if (getalt < 100000)DoReentryCorrection();
			}
		}
	if (rc_sw == 0 && getalt < 50000) // 30000
		{
			rc_sw = 1;
			SetThrusterGroupLevel(THGROUP_MAIN, 0);
			mth_level = 0;
		}
	if (getalt < 50000 && rat_sw ==1) //30000
		{
		rat_sw = 0;
		afs_status = 4;
		}

	}
else if(afs_status == 4)
	{
	if (dp_status == 0)dp_status = 1;
	}

//if (calc_sw == 0)
//{
//CalcEmptyMass();
//calc_sw = 1;
//}

if (dp_status == 1)
{
	DoLand();
	if (fins_deploy_status == 0)fins_deploy_status = 1;
}

//if (boostback_sw == 1)
//{
//	DoBoostBack();
//	DoBoostBack2();
//}
//if (boostback_sw == 2)
//{
//	DoBoostBack();
//}

if (boostback_sw > 0)DoBoostBack();

if (rat_sw == 1)DoRAT();


//sprintf(oapiDebugString(), "csw %d st %0.1f", crash_sw, oapiGetSimTime());
//sprintf(oapiDebugString(), "%d %0.4f %0.4f", (sa1_stbd_elb_sts + sa1_stbd_scale_sts + sa1_port_elb_sts + sa1_port_scale_sts),sa1_port_rot_proc,sa1_stbd_rot_proc);

////////////////// ANIMATIONS ///////////////

///////////// fins deploy

if (fins_deploy_status == 1)
{
double da = SimDT * FIN_SPEED;
fins_deploy_proc = fins_deploy_proc+da;
if (fins_deploy_proc > 1)
	{
		fins_deploy_proc = 1;
		fins_deploy_status = 2;
		hElevator = CreateControlSurface3 (AIRCTRL_ELEVATOR, 16, 3, _V(0,0,45), AIRCTRL_AXIS_XPOS, 0.5);
		hRudder = CreateControlSurface3 (AIRCTRL_RUDDER, 16, 3, _V(0,0,45), AIRCTRL_AXIS_YPOS, 0.5);
		hAileron_l = CreateControlSurface3 (AIRCTRL_AILERON, 2, 3, _V(-4,0,35), AIRCTRL_AXIS_XPOS, 0.5);
		hAileron_r = CreateControlSurface3 (AIRCTRL_AILERON, 2, 3, _V(4,0,35), AIRCTRL_AXIS_XNEG, 0.5);
	}
//SetAnimation(anim_fins_deploy, fins_deploy_proc);
}
//else if (fins_deploy_status == 2)SetAnimation(anim_fins_deploy, fins_deploy_proc);
else if (fins_deploy_status == 3)
{
	if (fins_deploy_proc == 1)
	{
		ClearControlSurfaceDefinitions ();
	}
double da = SimDT * FIN_SPEED;
fins_deploy_proc = fins_deploy_proc-da;
if (fins_deploy_proc < 0)
	{
		fins_deploy_proc = 0;
		fins_deploy_status = 0;
	}
//SetAnimation(anim_fins_deploy, fins_deploy_proc);
}

///////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//////// fins rotate

if (fins_deploy_status > 1)
{
//double pitch_up_level = GetThrusterGroupLevel(THGROUP_ATT_PITCHUP);
//double pitch_down_level = GetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN);
//double yaw_right_level = GetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT);
//double yaw_left_level = GetThrusterGroupLevel(THGROUP_ATT_YAWLEFT);

double pitch_up_level = abs(pu_level);
double pitch_down_level = abs(pd_level);
double yaw_right_level = abs(yr_level);
double yaw_left_level = abs(yl_level);

if(rcs_sw == 0)
{
pitch_up_level = 0;
pitch_down_level = 0;
yaw_right_level = 0;
yaw_left_level = 0;
}

VECTOR3 torqe_vector = _V((yaw_right_level-yaw_left_level),(pitch_up_level-pitch_down_level),0);

if (fins_deploy_status == 2)
{

double atmfact = ((GetDynPressure()-0000)*0.00005);
if(atmfact > 1)atmfact = 1;
else if(atmfact < 0)atmfact = 0;
atmfact = atmfact*0.8;
atmfact = 1-atmfact;

//sprintf(oapiDebugString(), "hf %0.1fkw hf2 %0.1fkw", heatflux/1000, heatflux2/1000);

if (rcs_sw == 0)
	{
	SetControlSurfaceLevel(AIRCTRL_ELEVATOR, torqe_vector.y*0.3);
	SetControlSurfaceLevel(AIRCTRL_RUDDER, torqe_vector.x*0.3);
	}
else
	{
	double csl_x = 0.6*torqe_vector.x*atmfact;
	double csl_y = 0.6*torqe_vector.y*atmfact;
	//if(abs(csl_x) < 0.05)csl_x = 0;
	//if(abs(csl_y) < 0.05)csl_y = 0;
	SetControlSurfaceLevel(AIRCTRL_ELEVATOR, csl_y);
	SetControlSurfaceLevel(AIRCTRL_RUDDER, csl_x);
	}
}
double torque_factor = 80*length(torqe_vector);
if (rcs_sw == 0)torque_factor = 40*length(torqe_vector);
if (torque_factor > 1)torque_factor = 1;
normalise(torqe_vector);

double torque_angle = atan2(torqe_vector.y, torqe_vector.x);
double ta_adjusted = torque_angle+(PI/4);
if (ta_adjusted > PI2) ta_adjusted = ta_adjusted - PI2;
else if (ta_adjusted < 0) ta_adjusted = ta_adjusted + PI2;

double req_fin_pitch = 0.5 + (sin(ta_adjusted)*torque_factor*0.5);
double req_fin_yaw = 0.5 + (cos(ta_adjusted)*torque_factor*0.5);

if ((pitch_up_level+pitch_down_level+yaw_right_level+yaw_left_level) == 0)
{
	req_fin_pitch = 0.5;
	req_fin_yaw = 0.5;
}

if (fins_deploy_status == 3)
{
	req_fin_pitch = 0.5;
	req_fin_yaw = 0.5;
}

double db = SimDT*FIN_SPEED;

if (abs(req_fin_pitch-fins_rot_a_proc) < db)fins_rot_a_proc = req_fin_pitch;
else if (req_fin_pitch < fins_rot_a_proc)fins_rot_a_proc = fins_rot_a_proc-db;
else if (req_fin_pitch > fins_rot_a_proc)fins_rot_a_proc = fins_rot_a_proc+db;
SetAnimation(anim_fins_rot_a, fins_rot_a_proc);

if (abs(req_fin_yaw-fins_rot_b_proc) < db)fins_rot_b_proc = req_fin_yaw;
else if (req_fin_yaw < fins_rot_b_proc)fins_rot_b_proc = fins_rot_b_proc-db;
else if (req_fin_yaw > fins_rot_b_proc)fins_rot_b_proc = fins_rot_b_proc+db;
SetAnimation(anim_fins_rot_b, fins_rot_b_proc);

////////////////////////////// bank /////////////////////////////////

double adcs_level_bank = GetControlSurfaceLevel(AIRCTRL_AILERON);
fins_bank_proc = (0.5 * adcs_level_bank) + 0.5;
if(fins_bank_proc < 0)fins_bank_proc = 0;
else if(fins_bank_proc > 1)fins_bank_proc = 1;
SetAnimation(anim_fins_bank, fins_bank_proc);
//sprintf(oapiDebugString(), "fins_bank_proc %0.3f", fins_bank_proc);

}


DoThrustVector ();

if(landed_sw == 1)
	{
	if (GetFlightStatus() == 00) DoLandingRCS();
	else
		{
		SetThrusterLevel(th_rcs2[0], 0);
		SetThrusterLevel(th_rcs2[1], 0);
		SetThrusterLevel(th_rcs2[2], 0);
		SetThrusterLevel(th_rcs2[3], 0);
		}
	}

}

void ss_booster::clbkPostStep (double simt, double simdt, double mjd)
{
	if (GroundContact() == true)landed_sw = 1;
}




bool ss_booster::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);
	int w15 = (int)(424*d);



	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	double m1 = GetEmptyMass();//+GetPropellantMass(hRCSProp);
	double m2 = GetPropellantMass(hProp);
	double test = ((m1+m2)/m1);
	double isp_curr = GetThrusterIsp(th_main[2]);
	double zdv = (isp_curr*(log(test)));

	sprintf(abuf,"Remaining dV %0.1fm/s", zdv);//, GetPropellantMass(hRCSProp) );
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}


	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"K = Enter Landing Target: %s", refname);
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(edlo_sw == 1)sprintf(abuf,"M = Set Boostback Mode: LAND ONLY");
	else sprintf(abuf,"M = Set Boostback Mode: BOOSTBACK+LAND");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	if (afs_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	if (edlo_sw == 1) sprintf(abuf,"B = Enable Auto LAND ONLY");
	else sprintf(abuf,"B = Enable Auto BOOSTBACK+LAND");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	if (edlo_sw == 1) sprintf(abuf,"B = Auto LAND ONLY Active");
	else sprintf(abuf,"B = Auto BOOSTBACK+LAND Active");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	
	if (rchud_sw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Boostback Burn Enabled  : Required dV %0.1f m/s", fb_dv_req);
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else if (rchud_sw == 2)
	{
	skp->SetTextColor(0x006666FF);
	char abuf[256];
	sprintf(abuf,"Boostback Burn Active : Required dV %0.1f m/s", fb_dv_req);
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else if (rchud_sw == 3)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Re-Entry Attitude Enabled");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else if (rchud_sw == 4)
	{
	//skp->SetTextColor(0x0066FFFF);
	//char abuf[256];
	//sprintf(abuf,"Re-Entry Burn Enabled");
	//skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else if (rchud_sw == 5)
	{
	//skp->SetTextColor(0x006666FF);
	//char abuf[256];
	//sprintf(abuf,"Re-Entry Burn Active");
	//skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else if (rchud_sw == 6)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Aero.Steering/Landing Burn Enabled");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else if (rchud_sw == 7)
	{
	skp->SetTextColor(0x006666FF);
	char abuf[256];
	sprintf(abuf,"Landing Burn Active");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	

	

	if (eng_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"P = Engine Selection : 33");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}

	else if (eng_status == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"P = Engine Selection : 13");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}

	else if (eng_status == 2)
	{
	skp->SetTextColor(0x006666FF);
	char abuf[256];
	sprintf(abuf,"P = Engine Selection : 3");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}

	/*if (fins_deploy_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Deploy Grid Fins");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}
	else if (fins_deploy_status == 1)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Stow Grid Fins");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}
	else if (fins_deploy_status == 2)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Stow Grid Fins");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}
	else if (fins_deploy_status == 3)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Deploy Grid Fins");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}*/

	/*if (leg_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Deploy Landing Gear");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}*/

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"E = Set View +Z/-Z");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}

	return true;
}



void ss_booster::RevertRCS (void)
{
	SetThrusterGroupLevel(htgrppup,0);
	SetThrusterGroupLevel(htgrppdown,0);
	SetThrusterGroupLevel(htgrpbright,0);
	SetThrusterGroupLevel(htgrpbleft,0);
	SetThrusterGroupLevel(htgrpyright,0);
	SetThrusterGroupLevel(htgrpyleft,0);
	SetThrusterLevel(th_rcs2[0], 0);
	SetThrusterLevel(th_rcs2[1], 0);
	SetThrusterLevel(th_rcs2[2], 0);
	SetThrusterLevel(th_rcs2[3], 0);
	SetThrusterLevel(th_rcs2[4], 0);
	SetThrusterLevel(th_rcs2[5], 0);
	SetThrusterLevel(th_rcs2[6], 0);
	SetThrusterLevel(th_rcs2[7], 0);

	DelThrusterGroup(htgrppup,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpbright,THGROUP_ATT_BANKRIGHT, false);
	DelThrusterGroup(htgrpbleft,THGROUP_ATT_BANKLEFT, false);
	DelThrusterGroup(htgrpyright,THGROUP_ATT_YAWRIGHT, false);
	DelThrusterGroup(htgrpyleft,THGROUP_ATT_YAWLEFT, false);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[11];
	htgrppup = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[10];
	htgrppdown = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	htgrpyright = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	htgrpyleft = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[4];
	th_group[1] = th_rcs[7];
	htgrpbright = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[5];
	th_group[1] = th_rcs[6];
	htgrpbleft = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	rcs_sw = 0;
	vrot_lim = 5*RAD;
}

void ss_booster::ConfigRCS (void)
{
	SetThrusterGroupLevel(htgrppup,0);
	SetThrusterGroupLevel(htgrppdown,0);
	SetThrusterGroupLevel(htgrpbright,0);
	SetThrusterGroupLevel(htgrpbleft,0);
	SetThrusterGroupLevel(htgrpyright,0);
	SetThrusterGroupLevel(htgrpyleft,0);

	DelThrusterGroup(htgrppup,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpbright,THGROUP_ATT_BANKRIGHT, false);
	DelThrusterGroup(htgrpbleft,THGROUP_ATT_BANKLEFT, false);
	DelThrusterGroup(htgrpyright,THGROUP_ATT_YAWRIGHT, false);
	DelThrusterGroup(htgrpyleft,THGROUP_ATT_YAWLEFT, false);

	th_group[0] = th_rcs2[2];
	th_group[1] = th_rcs2[3];
	htgrppup = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs2[0];
	th_group[1] = th_rcs2[1];
	htgrppdown = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs2[4];
	th_group[1] = th_rcs2[7];
	htgrpbright = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs2[5];
	th_group[1] = th_rcs2[6];
	htgrpbleft = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs2[1];
	th_group[1] = th_rcs2[2];
	htgrpyright = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs2[0];
	th_group[1] = th_rcs2[3];
	htgrpyleft = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	rcs_sw = 1;
	vrot_lim = 5*RAD;
}

void ss_booster::ConfigMain (void)
{
	double old_tg_lev = GetThrusterGroupLevel(htgrpmain);
	SetThrusterGroupLevel(htgrpmain,0);



	DelThrusterGroup(htgrpmain,THGROUP_MAIN, false);

	if (eng_status == 0) 
	{
		th_group[0] = th_main[0];
		SetThrusterLevel(th_main[1],0);
		SetThrusterLevel(th_main[2],0);
		int i;
		for (i = 0; i < 12; i++)
		{
		SetThrusterMax0(th_rcs[i],RCSTH*1);
		}
	}
	else if (eng_status == 1)
	{
		th_group[0] = th_main[1];
		SetThrusterLevel(th_main[0],0);
		SetThrusterLevel(th_main[2],0);
		int i;
		for (i = 0; i < 12; i++)
		{
		SetThrusterMax0(th_rcs[i],RCSTH*0.25);
		}
	}
	else if (eng_status == 2)
	{
		th_group[0] = th_main[2];
		SetThrusterLevel(th_main[0],0);
		SetThrusterLevel(th_main[1],0);
		int i;
		for (i = 0; i < 12; i++)
		{
		SetThrusterMax0(th_rcs[i],RCSTH*0.15);
		}
	}
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);

	SetThrusterGroupLevel(htgrpmain,old_tg_lev);
}

void ss_booster::KillAllRCS()
{
SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
}


void ss_booster::DoLand(void)
{
OBJHANDLE htgt_vessel = NULL;
htgt_vessel = oapiGetObjectByName(refname);
if(htgt_vessel != NULL)
{
	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(htgt_vessel);
	DWORD ac = v->AttachmentCount(false);
	if(ac == 0) tgt_att_valid = false;
	else
	{
	for (DWORD j = 0; j < ac; j++)
		{
		ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(false, j);
		const char *id = v->GetAttachmentId (hAtt);
		if (!strnicmp (id, "BST", 3))
			{
			VECTOR3 attpos, attdir, attrot, attpos_hrz;
			v->GetAttachmentParams(hAtt, attpos, attdir, attrot);
			attpos.x = -11.6;
			attpos.z = -4.8;
			v->HorizonRot(attpos,attpos_hrz);
			double v_alt = v->GetAltitude();
			catch_offset = attpos_hrz;
			catch_offset.y = v_alt + attpos_hrz.y - 78.33;
			//sprintf(oapiDebugString(), "catch_offset %0.1f %0.1f %0.1f", catch_offset.x, catch_offset.y,catch_offset.z);
			break;
			}
		else if(j == (ac-1))
			{
				tgt_att_valid = false;
				catch_offset = _V(0,0,0);
			}
		}
	}
}
else
{
	catch_offset = _V(0,0,0);
	tgt_att_valid = false;
}
	
	//sprintf(oapiDebugString(), "DoLand");
double m2 = GetPropellantMass(hProp);
if (m2 > 140000 && eng_status == 2)
	{
	eng_status = 1;
	ConfigMain();
	}
else if (m2 > 140000 && eng_status == 0)
	{
	eng_status = 1;
	ConfigMain();
	}
else if (m2 < 140000 && eng_status == 1)
	{
	eng_status = 2;
	ConfigMain();
	}
else if (m2 < 140000 && eng_status == 0)
	{
	eng_status = 2;
	ConfigMain();
	}

if(GetAttachmentStatus(att_catch) != NULL)landed_sw = 1; // caught - shutdown

VECTOR3 weightv = _V(0,0,0);
double weight = 0;
GetWeightVector(weightv);
weight = length(weightv);

VECTOR3 haspdv = _V(0,0,0);
GetGroundspeedVector(FRAME_HORIZON, haspdv);

VECTOR3 aspdv = _V(0,0,0);
GetGroundspeedVector(FRAME_LOCAL, aspdv);
double air_spd = length(haspdv);

	VECTOR3 z_axis_ves = _V(0,0,1);
	VECTOR3 z_axis_hrz = _V(0,0,1);
	HorizonRot(z_axis_ves, z_axis_hrz);
	normalise(z_axis_hrz);
	double ves_pitch = PI05 - acos(dotp(z_axis_hrz, _V(0,1,0)));

double pitch = ves_pitch;
double cosp = (cos(pitch));
double sinp = (sin(pitch));
double pf = (1/(sin(pitch)));

//double bank = GetBank();
//double cosb = (cos(bank));
//double bf = (1/(cos(bank)));

double alt = GetAltitude(ALTMODE_GROUND);
if(tgt_att_valid == true)alt = alt - catch_offset.y;

double mass = 0;
VECTOR3 drag = _V(0,0,0);
GetDragVector(drag);
VECTOR3 drag_hrz = _V(0,0,0);
HorizonRot(drag,drag_hrz);
double th_max = GetThrusterMax(th_main[eng_status]);

	double m1 = GetEmptyMass();//+GetPropellantMass(hRCSProp);
	mass = m1+m2;

	double test = ((m1+m2)/m1);
	double isp_curr = GetThrusterIsp(th_main[eng_status]);
	double dv = (isp_curr*(log(test)));

	/*if (alt < 15000 && alt > 4000)
	{
		if (m2 > 4000 && correct_sw == 0) 
		{
			DoCorrection();
		}
		else
		{	
		SetThrusterLevel(th_main[eng_status],0);
		KillAllRCS();
		correct_sw = 1;
		rot_sw = 0;
		
		}
	}
	else if (alt < 4000)
	{
	correct_sw = 1;
	}*/

	if (alt < 50000)// && correct_sw == 1)
	{
		//if(GetAttachmentStatus(att_catch) != NULL)landed_sw = 1; // caught - shutdown
		if (GroundContact() == true)landed_sw = 1; // ensure shutdown on landing
		if (landed_sw == 1)
		{
		double englvl = GetThrusterGroupLevel(THGROUP_MAIN);
		if (englvl > 0.2)englvl = 0.2; // ensure bloody SHUTDOWN, I said!
		englvl = englvl - (step_time*1.0); //0.5
		SetThrusterGroupLevel(THGROUP_MAIN, englvl);
		mth_level = englvl;
		if(englvl < 0)
			{
			SetThrusterGroupLevel(THGROUP_MAIN, 0);
			mth_level = 0;
			KillAllRCS();
			dp_status = 0;
			afs_status = 0;
			fins_deploy_status = 3;
			}
			return;
		}

	/*if (alt < 300 && leg_status == 0) 
		{
		leg_status = 1;
		SetTouchdownPoints ( tdvtx_geardown, ntdvtx_geardown);
		}*/
	if (alt > 100) DoLAT();//50
	else if (alt > 20)DoTWU();//40
	if (alt > 20000) DoHeadsUp();
	else SetControlSurfaceLevel (AIRCTRL_AILERON, 0, false);


	double alt_factor = m2/26000;
	if (alt_factor < 1.4)alt_factor = 1.4;

	//double accfact = 1.0;

	double start_alt = (100+(900*alt_factor));//-250
	if(alt > start_alt && alt < 5000)
	{
	double time = (-2*(alt-start_alt)/(haspdv.y+30));

	double acc = ((-(haspdv.y+30))/time);//accfact*((-(haspdv.y+30))/time);

	double max_acc = 0.75*(((sinp*th_max)-weight+drag_hrz.y)/mass);//0.75

	double force = ((mass*acc)+ weight - drag_hrz.y);

	double englvl = min(1, pf*force/th_max);

	englvl = max(0,englvl);
	englvl = min(0.5,englvl);//0.8
	
	double sao = (m2 - 28000)*0.05;
	if(sao > 0)sao = 0;
	double wspeed;
	oapiGetWindVector(GetGravityRef(), 0, 0, 4000, 0, &wspeed);
	if(wspeed < 0.1)sao = sao - 200;

	if (GetThrusterGroupLevel(THGROUP_MAIN) > 0)max_acc = 0;
	if (max_acc < acc)
		{
		SetThrusterGroupLevel(THGROUP_MAIN, englvl);
		mth_level = englvl;
		}
	else if (alt < (2000+sao))/////////////////!!!!!!!!!!!!!!!!
		{
		SetThrusterGroupLevel(THGROUP_MAIN, englvl);
		mth_level = englvl;
		}
	else
		{
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
		mth_level = 0;
		}
	if(haspdv.y > 0)
		{
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
		mth_level = 0;
		}
	}

	else if ( alt < 5000 && alt > ((alt_factor*300)-100))
	{
	double time = (-2*(alt-((alt_factor*300)-100))/(haspdv.y+15));


	double acc = ((-(haspdv.y+15))/time);//accfact*((-(haspdv.y+15))/time);

	double force = ((mass*acc)+ weight - drag_hrz.y);
	
	double englvl = min(1, pf*force/th_max);

	englvl = max(0,englvl);
	//englvl = min(0.8,englvl);
	//sprintf(oapiDebugString(), "f %0.4f t %0.4f",force, (th_max*englvl*sinp));
	if(englvl > 0.8)englvl = 0.8;

	SetThrusterGroupLevel(THGROUP_MAIN, englvl);
	mth_level = englvl;
	if(haspdv.y > 0)
		{
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
		mth_level = 0;
		}
	}
	else if (alt > (1*100)  && alt < 5000)//50
	{
	double time = (-2*(alt-(1*50))/(haspdv.y+5));//50


	double acc = ((-(haspdv.y+5))/time);//accfact*((-(haspdv.y+5))/time);

	double force = ((mass*acc)+ weight - drag_hrz.y);
	
	double englvl = min(1, pf*force/th_max);

	englvl = max(0,englvl);
	if (drag_hrz.y < 0)englvl = 0;
	SetThrusterGroupLevel(THGROUP_MAIN, englvl);
	mth_level = englvl;
	
	}
	else if (alt > 70  && alt < 5000) //41
	{
	double time = (-2*(alt-50)/(haspdv.y+3));//41

	if (time < 0)landed_sw = 1;

	double acc = ((-(haspdv.y+3))/time);

	double force = ((mass*acc)+ weight - drag_hrz.y);
	
	double englvl = min(1, pf*force/th_max);

	englvl = max(0,englvl);
	//englvl = min(0.8,englvl);

	if (drag_hrz.y < 0)englvl = 0;
	SetThrusterGroupLevel(THGROUP_MAIN, englvl);
	mth_level = englvl;
	//sprintf(oapiDebugString(), "vvel %0.2f alt %0.1f", haspdv.y, alt);
	}
	else if (alt > 36  && alt < 5000)
	{
	double time = (-2*(alt-45)/(haspdv.y+2.0));
	if((alt-45) < 0.1) time = 0.1;

	double acc = ((-(haspdv.y+2.0))/time);

	if (haspdv.y > -1)acc = 0;

	double force = ((mass*acc)+ weight - drag_hrz.y);
	
	double englvl = min(1, pf*force/th_max);

	englvl = max(0,englvl);
	//englvl = min(0.8,englvl);

	if (drag_hrz.y < 0)englvl = 0;

	SetThrusterGroupLevel(THGROUP_MAIN, englvl);
	mth_level = englvl;

//	sprintf(oapiDebugString(), "vvel %0.2f alt %0.1f", haspdv.y, alt);
	}
	else if (alt < 5000)
		{
	//sprintf(oapiDebugString(), "here");
	landed_sw = 1;
	double englvl = GetThrusterGroupLevel(THGROUP_MAIN);
	englvl = englvl - (step_time*0.25);
	SetThrusterGroupLevel(THGROUP_MAIN, englvl);
	mth_level = englvl;
	if(englvl < 0)
		{
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
		mth_level = 0;
		KillAllRCS();
		dp_status = 0;
		afs_status = 0;
		fins_deploy_status = 3;
		}
		}
	}
}


void ss_booster::DoRAT(void)
{
	if(rcs_sw == 1)
	{
		VECTOR3 aspdv;
		GetGroundspeedVector(FRAME_HORIZON, aspdv);
		HorizonInvRot(-aspdv, flight_vector);
		normalise(flight_vector);
		GoToFlightVector();
		rchud_sw = 3;
		return;
	}
	
	double angoa = GetAOA();
	double slipage = GetSlipAngle();

	if(angoa > 0)angoa = angoa-PI;
	else angoa = angoa+PI;

	if(slipage > 0)slipage = slipage-PI;
	else slipage = slipage+PI;
	slipage = -slipage;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[0]);
	double max_torque = 2*max_rcs*35;

	if (rcs_sw == 1)
	{
		max_rcs = GetThrusterMax(th_rcs2[0]);
		max_torque = 2*max_rcs*60*0.7071;
	}

	double inertia = 360*mass;

	double finfact = 1;//(GetDynPressure()/40000);
	if (fins_deploy_status > 0 && rcs_sw == 1 && GetAltitude() < 60000) finfact = 0.05;

	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double max_acc = max_torque/inertia;
	double max_drot = max_acc*step_time;

	vrot_lim = 3*RAD;

	double reqxv = 0.9*(sqrt(abs(angoa*max_acc*2)));
	if (reqxv > vrot_lim)reqxv = vrot_lim;
	if (angoa > 0)reqxv = -reqxv;
	double req_drotx = reqxv - avel.x;
	double thx_lvl = req_drotx/max_drot;
	if((angoa*DEG) < 1 && (angoa*DEG) > -1)
	{
	reqxv = 1*-angoa;
	double acc_reqx = (reqxv-avel.x)/0.1;
	thx_lvl = acc_reqx/max_acc;
	}
	thx_lvl = thx_lvl*finfact;

	double reqyv = 0.9*(sqrt(abs(slipage*max_acc*2)));
	if (reqyv > vrot_lim)reqyv = vrot_lim;
	if (slipage > 0)reqyv = -reqyv;
	double req_droty = reqyv - avel.y;
	double thy_lvl = req_droty/max_drot;
	if((slipage*DEG) < 1 && (slipage*DEG) > -1)
	{
	reqyv = 1*-slipage;
	double acc_reqy = (reqyv-avel.y)/0.1;
	thy_lvl = acc_reqy/max_acc;
	}
	thy_lvl = thy_lvl*finfact;

	//sprintf(oapiDebugString(), "slip %0.1f avely %0.1f", slipage*DEG, avel.y*DEG);
	//sprintf(oapiDebugString(), "angoa %0.6f thy_lvl %0.1f thx_lvl %0.1f",angoa*DEG, thy_lvl, thx_lvl);

	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;

	//sprintf(oapiDebugString(), "angoa %0.6f thy_lvl %0.1f thx_lvl %0.1f",angoa*DEG, thy_lvl, thx_lvl);

	double tacc = oapiGetTimeAcceleration();
	if(oapiGetTimeAcceleration() > 2)
	{
		thx_lvl = thx_lvl/tacc;//*0.5;
		thy_lvl = thy_lvl/tacc;//*0.5;
	}


	if(thx_lvl > 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		pu_level = thx_lvl;
		pd_level = 0;
	}
	else if(thx_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		pd_level = -thx_lvl;
		pu_level = 0;
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		pu_level = 0;
		pd_level = 0;
	}

	if(thy_lvl > 0)
	{		
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		yr_level = 0;
		yl_level = thy_lvl;
	}
	else if(thy_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		yl_level = 0;
		yr_level = -thy_lvl;
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		yr_level = 0;
		yl_level = 0;
	}

	rchud_sw = 3;
}

void ss_booster::DoHeadsUp(void)
{
	double zavel_lim = (20 * RAD);
	VECTOR3 avel;
	GetAngularVel(avel);
	double bank = GetBank();
	double req_z_avel = -bank * 0.25;
	double req_z_acc = req_z_avel - avel.z;
	if(req_z_acc > zavel_lim)req_z_acc = zavel_lim;
	else if(req_z_acc < -zavel_lim)req_z_acc = -zavel_lim;
	double adcs_level = -req_z_acc/1.25;
	if(adcs_level > 1)adcs_level = 1;
	else if(adcs_level < -1)adcs_level = -1;

	SetControlSurfaceLevel (AIRCTRL_AILERON, adcs_level, false);


}

void ss_booster::DoTWU(void)
{
	VECTOR3 wv = _V(0,0,0);
	GetWeightVector(wv);
	wv = -wv;

	//sprintf(oapiDebugString(), "here");
	
	double angoa = atan2(-wv.y, wv.z);
	double slipage = atan2(-wv.x, wv.z);
	slipage = -slipage;

	if(angoa > PI)angoa = (PI*2)-angoa;
	else if(angoa < -PI)angoa = (PI*2)+angoa;

	if(slipage > PI)slipage = (PI*2)-slipage;
	else if(slipage < -PI)slipage = (PI*2)+slipage;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[0]);
	double max_torque = 2*max_rcs*35;

	if (rcs_sw == 1)
	{
		max_rcs = GetThrusterMax(th_rcs2[0]);
		max_torque = max_rcs*60*0.7071;
	}

	double inertia = 360*mass;

	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double max_acc = max_torque/inertia;
	double max_drot = max_acc*step_time;

	double reqxv = 0.9*(sqrt(abs(angoa*max_acc*2)));
	if (reqxv > vrot_lim)reqxv = vrot_lim;
	if (angoa > 0)reqxv = -reqxv;
	double req_drotx = reqxv - avel.x;
	double thx_lvl = req_drotx/max_drot;
	if((angoa*DEG) < 1 && (angoa*DEG) > -1)
	{
	reqxv = 1*-angoa;
	double acc_reqx = (reqxv-avel.x)/0.1;
	thx_lvl = acc_reqx/max_acc;
	}

	double reqyv = 0.9*(sqrt(abs(slipage*max_acc*2)));
	if (reqyv > vrot_lim)reqyv = vrot_lim;
	if (slipage > 0)reqyv = -reqyv;
	double req_droty = reqyv - avel.y;
	double thy_lvl = req_droty/max_drot;
	if((slipage*DEG) < 1 && (slipage*DEG) > -1)
	{
	reqyv = 1*-slipage;
	double acc_reqy = (reqyv-avel.y)/0.1;
	thy_lvl = acc_reqy/max_acc;
	}

	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;


	if(thx_lvl > 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		pu_level = thx_lvl;
		pd_level = 0;
	}
	else if(thx_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		pd_level = -thx_lvl;
		pu_level = 0;
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		pu_level = 0;
		pd_level = 0;
	}

	if(thy_lvl > 0)
	{		
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		yr_level = 0;
		yl_level = thy_lvl;
	}
	else if(thy_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		yl_level = 0;
		yr_level = -thy_lvl;
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		yr_level = 0;
		yl_level = 0;
	}


}


void ss_booster::GoToFlightVector()
{
if(GetAltitude() < 1000) vrot_lim = 2*RAD;
	
	DeactivateNavmode(NAVMODE_KILLROT);

	double angoa = atan2(-flight_vector.y, flight_vector.z);
	double slipage = atan2(flight_vector.x, flight_vector.z);

	if(angoa > PI)angoa = (PI*2)-angoa;
	else if(angoa < -PI)angoa = (PI*2)+angoa;

	if(slipage > PI)slipage = (PI*2)-slipage;
	else if(slipage < -PI)slipage = (PI*2)+slipage;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[0]);
	double max_torque = 2*max_rcs*35;

	if (rcs_sw == 1)
	{
		max_rcs = GetThrusterMax(th_rcs2[0]);
		max_torque = 1*max_rcs*60;
	}

	double inertia = 360*mass;

	double finfact = 1;
	if (fins_deploy_status > 0 && rcs_sw == 1) finfact = 0.02; // 0.5

	
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);

	double max_acc = max_torque/inertia;
	double max_drot = max_acc*step_time;

	double reqxv = 0.9*(sqrt(abs(angoa*max_acc*2)));
	if (reqxv > vrot_lim)reqxv = vrot_lim;
	if (angoa > 0)reqxv = -reqxv;
	double req_drotx = reqxv - avel.x;
	double thx_lvl = req_drotx/max_drot;
	if((angoa*DEG) < 1 && (angoa*DEG) > -1)
	{
	reqxv = 1*-angoa;
	double acc_reqx = (reqxv-avel.x)/0.2;
	thx_lvl = acc_reqx/max_acc;
	}
	thx_lvl = thx_lvl*finfact;


	double reqyv = 0.9*(sqrt(abs(slipage*max_acc*2)));
	if (reqyv > vrot_lim)reqyv = vrot_lim;
	if (slipage > 0)reqyv = -reqyv;
	double req_droty = reqyv - avel.y;
	double thy_lvl = req_droty/max_drot;
	if((slipage*DEG) < 1 && (slipage*DEG) > -1)
	{
	reqyv = 1*-slipage;
	double acc_reqy = (reqyv-avel.y)/0.2;
	thy_lvl = acc_reqy/max_acc;
	}
	thy_lvl = thy_lvl*finfact;

	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;

//	double tacc = oapiGetTimeAcceleration();
//	if(tacc < 1)tacc = 1;
//	thx_lvl = thx_lvl/tacc;
//	thy_lvl = thy_lvl/tacc;

	double th_level = 0;

	if(thx_lvl > 0)
	{
		th_level = thx_lvl;
		if(th_level < 0.3)th_level = 0;
		if(rcs_sw == 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, th_level);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
		pu_level = thx_lvl;
		pd_level = 0;
	}
	else if(thx_lvl < 0)
	{
		th_level = thx_lvl;
		if(th_level > -0.3)th_level = 0;
		if(rcs_sw == 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -th_level);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		}
		pd_level = -thx_lvl;
		pu_level = 0;
	}
	else 
	{
		if(rcs_sw == 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
		pu_level = 0;
		pd_level = 0;
	}

	if(thy_lvl > 0)
	{	
		th_level = thy_lvl;
		if(th_level < 0.3)th_level = 0;
		if(rcs_sw == 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, th_level);
		}
		yr_level = 0;
		yl_level = thy_lvl;
	}
	else if(thy_lvl < 0)
	{
		th_level = thy_lvl;
		if(th_level > -0.3)th_level = 0;
		if(rcs_sw == 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -th_level);
		}
		yl_level = 0;
		yr_level = -thy_lvl;
	}
	else 
	{
		if(rcs_sw == 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		}
		yr_level = 0;
		yl_level = 0;
	}
	///////////////////////////////////////////////////////////////
	if (rcs_sw == 1)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);

	VECTOR3 fv_nml = flight_vector;
	normalise(fv_nml);

	VECTOR3 fv_z_normal = crossp( _V(0,0,1), fv_nml);
	normalise(fv_z_normal);

	double fv_z_angle = acos(dotp(fv_nml, _V(0,0,1)));

	MATRIX3 rm = rotm(_V(0,0,1), (-45*RAD));

	double req_av = 0.9*(sqrt(fv_z_angle*max_acc*2));
	if(req_av > vrot_lim) req_av = vrot_lim;


	if((fv_z_angle*DEG) < 2)
	{
	req_av = fv_z_angle;
	}

	VECTOR3 avel_error;

	avel_error.x = (fv_z_normal.x * req_av) + avel.x;
	avel_error.y = (fv_z_normal.y * req_av) + avel.y;
	avel_error.z = 0;

	if(afs_status < 4)finfact = 1;
	else finfact = 0.01;

	VECTOR3 avel_acc_req;
	avel_acc_req.x	= finfact*avel_error.x/(max_drot*4);
	avel_acc_req.y	= finfact*avel_error.y/(max_drot*4);
	avel_acc_req.z	= 0;



	VECTOR3 thrust_pointer = crossp(avel_acc_req, _V(0,0,1));
	//normalise(thrust_pointer);

	thrust_pointer = mul(rm, thrust_pointer);

	//sprintf(oapiDebugString(), "x %0.4f y %0.4f z %0.4f", thrust_pointer.x, thrust_pointer.y,thrust_pointer.z);

	//sprintf(oapiDebugString(), "x %0.4f y %0.4f z %0.4f angle %0.4f", fv_z_normal.x, fv_z_normal.y, fv_z_normal.z, fv_z_angle*DEG);

	

	//sprintf(oapiDebugString(), "x %0.4f y %0.4f avelx %0.4f avely %0.4f", avel_acc_req.x, avel_acc_req.y, avel.x, avel.y);

	//VECTOR3 avel_acc_req_rot = mul(rm,avel_acc_req);

	//VECTOR3 avel_acc_req_rot_nmd = avel_acc_req_rot;
	//normalise(avel_acc_req_rot_nmd);

	//VECTOR3 exp_vect = mul(rm,_V(0,1,0));
	


	double x_level = thrust_pointer.x;
	if(x_level > 1)x_level = 1;
	else if(x_level < -1)x_level = -1;
	//x_level = x_level * finfact;
	if(abs(x_level) < 0.75)x_level = 0;

	double y_level = thrust_pointer.y;
	if(y_level > 1)y_level = 1;
	else if(y_level < -1)y_level = -1;
	//y_level = y_level * finfact;
	if(abs(y_level) < 0.75)y_level = 0;

	//sprintf(oapiDebugString(), "x %0.4f y %0.4f x_level %0.4f y_level %0.4f", avel_acc_req.x, avel_acc_req.y, x_level, y_level);

	if (x_level > 0)
		{
		SetThrusterLevel(th_rcs2[2], x_level);
		SetThrusterLevel(th_rcs2[0], 0);
		}
	else if (x_level < 0)
		{
		SetThrusterLevel(th_rcs2[2], 0);
		SetThrusterLevel(th_rcs2[0], -x_level);
		}
	else
		{
		SetThrusterLevel(th_rcs2[0], 0);
		SetThrusterLevel(th_rcs2[2], 0);
		}

	if (y_level > 0)
		{
		SetThrusterLevel(th_rcs2[3], y_level);
		SetThrusterLevel(th_rcs2[1], 0);
		}
	else if (y_level < 0)
		{
		SetThrusterLevel(th_rcs2[3], 0);
		SetThrusterLevel(th_rcs2[1], -y_level);
		}
	else
		{
		SetThrusterLevel(th_rcs2[1], 0);
		SetThrusterLevel(th_rcs2[3], 0);
		}


	}
	
}


bool ss_booster::clbkLoadGenericCockpit ()
{
	SetCamDir();
return true;
}

void ss_booster::SetCamDir()
{
	VECTOR3 camdefdir = _V(0,0,1);
	if(cam_status == 1)camdefdir = _V(0,0,-1);
	SetCameraDefaultDirection(camdefdir, 0);
	oapiCameraAttach(GetHandle(), 2);
}


/////////////////////// BOOSTBACKS ///////////////////////

void ss_booster::DoBoostBack (void)
{
if (eng_status != 1)
{
eng_status = 1;
ConfigMain();
}

if(boostback_sw == 1)
{
timer = timer+step_time;
if(timer > 4)
	{
		boostback_sw = 3; // one engine on // 2
		timer = 0;
	}
}

else if(boostback_sw == 2)
{
timer = timer+step_time;
if(timer > 1.5)
	{
		boostback_sw = 3; // one engine off
		timer = 0;
	}
}

	VECTOR3 haspdv = _V(0,0,0);
	GetGroundspeedVector(FRAME_HORIZON,haspdv);

	

	double emp_mass = GetEmptyMass();//+GetPropellantMass(hRCSProp);
	double vmass = GetMass();
	double mass_fct = (emp_mass*emp_mass)/(vmass*vmass);
	if (mass_fct < 0.4 )mass_fct = 0.4;
	double tgt_height_offset = 3850+(2020*mass_fct);//-(1.59-sin(2*((90*RAD)-impact_angle))));//8550

	double fpang = atan2(haspdv.x,haspdv.z);
	double sinfpang = sin(fpang);
	double fpfact = -2000*sinfpang;
	tgt_height_offset = tgt_height_offset + fpfact;

	double aoi_fct = 15000*(sin((65*RAD)-impact_angle));
	tgt_height_offset = tgt_height_offset + aoi_fct ;/////////////
	
	double ref_fct = ((vmass - emp_mass)/26000) + 0.8;
	if (ref_fct > 2.1)ref_fct = 2.1;
	else if (ref_fct < 1)ref_fct = 1;
	tgt_height_offset = tgt_height_offset*ref_fct*1;////*2.9
	tgt_height_offset = tgt_height_offset + 500;
	if(tgt_height_offset < 8500)tgt_height_offset = 8500;
	else if(tgt_height_offset > 10500)tgt_height_offset = 10500;
	
	//sprintf(oapiDebugString(), "tho %0.1f impa %0.1f", tgt_height_offset, impact_angle*DEG);



	OBJHANDLE hearth=GetSurfaceRef();
	OBJHANDLE htgt_vessel = oapiGetObjectByName(refname);
	if (htgt_vessel == NULL)htgt_vessel = oapiGetBaseByName(hearth,refname);
	if (htgt_vessel == NULL)
	{
		boostback_sw = 0;
		return;
	}
	DeactivateNavmode(NAVMODE_KILLROT);
	

		VECTOR3 relvel = _V(0,0,0);
		GetRelativeVel(hearth,relvel);
		

	double ves_lng = 0;
	double ves_lat = 0;
	double ves_rad = 0;
	if (oapiGetObjectType(htgt_vessel) == OBJTP_VESSEL)
	{
	oapiGetEquPos(htgt_vessel, &ves_lng, &ves_lat, &ves_rad);
	
	}

	else
	{
		oapiGetBaseEquPos(htgt_vessel, &ves_lng, &ves_lat, &ves_rad);
		
	}
	double lng_off = PI2*(timetoimp+0.0)/86164;


	double tgt_lng = ves_lng+lng_off;
	double tgt_lat = ves_lat;
	if (tgt_lng > PI2) tgt_lng = tgt_lng - PI2;
	double tgt_rad = ves_rad + tgt_height_offset;

	

	VECTOR3 tgt_pos_global = _V(0,0,0);
	oapiEquToGlobal(hearth, tgt_lng, tgt_lat, tgt_rad, &tgt_pos_global);

	

	VECTOR3 earth_pos = _V(0,0,0);
	oapiGetGlobalPos(hearth, &earth_pos);

	VECTOR3 tgt_vector_rele = tgt_pos_global - earth_pos;
	
	VECTOR3 my_vector_rele = _V(0,0,0);
	GetRelativePos(hearth, my_vector_rele);

	VECTOR3 tgt_dist = my_vector_rele - tgt_vector_rele;
	double r1 = length(my_vector_rele);
	double r2 = length(tgt_vector_rele);
	double d = length(tgt_dist);
	double s = 0.5*(r1 + r2 + d);
	double GM = GGRAV*oapiGetMass(hearth);
	double h = (r1-r2);
	

	double v2 = (2*GM*(s-r1))/(r1*s);
	double v = sqrt(v2);


	double tantheta = sqrt(((s-d)*(d-h))/(s*(d+h)));
	double theta = atan(tantheta);


	double vx = v*cos(theta);
	double vy = v*sin(theta);

		double grdspd = sqrt((length(relvel)*length(relvel))-(haspdv.y*haspdv.y));
		double rds = (rt + GetAltitude());
		double ang_vel = grdspd/(1*rds);//(orb_grnd_spd+vx)/(2*rds);
		double acc_cent_fact = 1;
		double acc_cent = ang_vel*ang_vel*rds*acc_cent_fact;
		double gravt = GM/(rds*rds);
		double g = acc_cent - gravt;


	double tti = (sqrt((s*d*(s-d))/(2*GM)))*(1 + ((s/(sqrt(d*(s-d)))) *(atan(sqrt(d/(s-d))))));
	timetoimp = tti;


	normalise(tgt_vector_rele);
	normalise(my_vector_rele);

	dotprod = dotp(tgt_vector_rele,my_vector_rele);
	if (dotprod > 1)dotprod = 1;
	else if (dotprod < -1)dotprod = -1;

	double tgt_angle = acos(dotprod);

	VECTOR3 normal_me_tgt_vects = _V(0,0,0);
	normal_me_tgt_vects = crossp(tgt_vector_rele,my_vector_rele);
	normalise(normal_me_tgt_vects);

	VECTOR3 tangential_vector = _V(0,0,0);
	tangential_vector = crossp(my_vector_rele, normal_me_tgt_vects);
	normalise(tangential_vector);

	double tgt_range = tgt_angle * (rt+((GetAltitude()+tgt_height_offset)/2));
	//tgt_range = tgt_range + 1000;

	VECTOR3 vertvel_glob = _V((vy*my_vector_rele.x),(vy*my_vector_rele.y),(vy*my_vector_rele.z));
	
	


	double final_vel_xz = tgt_range/timetoimp;
	VECTOR3 final_vel_xz_glob = _V(0,0,0);
	final_vel_xz_glob.x = tangential_vector.x*vx;
	final_vel_xz_glob.y = tangential_vector.y*vx;
	final_vel_xz_glob.z = tangential_vector.z*vx;

	

	VECTOR3 final_vel_vect = final_vel_xz_glob+vertvel_glob;

	double fvvh = GetAltitude();//-tgt_height_offset;
	double finalvv = length(vertvel_glob)+sqrt(2*-g*fvvh);
	double finalvh = length(final_vel_xz_glob);
	impact_angle = atan2(finalvv,finalvh);
	//impact_angle = PI05-theta;
	
	

	VECTOR3 req_dv_vector = _V(0,0,0);
	req_dv_vector.x = final_vel_vect.x - relvel.x;
	req_dv_vector.y = final_vel_vect.y - relvel.y;
	req_dv_vector.z = final_vel_vect.z - relvel.z;

	double thmain_max = GetThrusterMax(th_main[eng_status]);
	double acc_lim = 58.86;
	double max_eng_acc = thmain_max/vmass;
	double acc_lim_factor = acc_lim/max_eng_acc;
	if(acc_lim_factor > 1)acc_lim_factor = 1;

	double dvmain = length(req_dv_vector);
	double fraction = exp(dvmain/GetThrusterIsp(th_main[eng_status]));
	double fuelmass = vmass - (vmass/fraction);
	double englevel = acc_lim_factor;//GetThrusterLevel(th_main[eng_status]);
	double fuelrate = thmain_max*englevel/GetThrusterIsp(th_main[eng_status]);
	double dvtime = fuelmass/fuelrate;
	double vvel_off = 0.75*dvtime*-g;
	if (vvel_off < 0) vvel_off = 0;

	vertvel_glob = _V((vvel_off*my_vector_rele.x),(vvel_off*my_vector_rele.y),(vvel_off*my_vector_rele.z));

	req_dv_vector.x = req_dv_vector.x + vertvel_glob.x;
	req_dv_vector.y = req_dv_vector.y + vertvel_glob.y;
	req_dv_vector.z = req_dv_vector.z + vertvel_glob.z;


	VECTOR3 mypos_glob = _V(0,0,0);
	GetGlobalPos(mypos_glob);

	
	req_dv_vector = req_dv_vector + mypos_glob;

	VECTOR3 req_dv_vector_loc = _V(0,0,0);
	Global2Local(req_dv_vector,req_dv_vector_loc);

	/////////////

	fb_dv_req = length(req_dv_vector_loc);

	flight_vector = req_dv_vector_loc;
	vrot_lim = 10*RAD;
	//GoToFlightVector();

	VECTOR3 req_dv_vector_norm = req_dv_vector_loc;
	normalise(req_dv_vector_norm);
	VECTOR3 curr_att_loc = _V(0,0,1);

	dotprod = dotp(curr_att_loc,req_dv_vector_norm);
	if (dotprod > 1)dotprod = 1;
	else if (dotprod < -1)dotprod = -1;

	double pointing_error = acos(dotprod);

	

	double max_dv = cos(pointing_error)*step_time*acc_lim_factor*thmain_max/vmass;
	double eng_lvl = flight_vector.z/max_dv;
	if (eng_lvl > acc_lim_factor)eng_lvl = acc_lim_factor;
	else if (eng_lvl < 0)eng_lvl = 0;

	double pe_fact = abs(fb_dv_req*tan(pointing_error));

	
	//if ((pe_fact) < 300 && flight_vector.z > 4)
	if(abs(pointing_error) < (10*RAD) && flight_vector.z > 2)
	{
		if (fb_dv_req < 50)
			{
			eng_status = 2;
			ConfigMain();
			boostback_sw = 4;
			}

		SetThrusterGroupLevel(THGROUP_MAIN, eng_lvl);
		mth_level = eng_lvl;
		rchud_sw = 2; // flyback burn active
		vrot_lim = 5*RAD;
	}
	else if (boostback_sw == 4)
	{

		eng_status = 2;
		ConfigMain();
		SetThrusterGroupLevel(THGROUP_MAIN, eng_lvl);
		mth_level = eng_lvl;
		rchud_sw = 2; // flyback burn active
		vrot_lim = 3*RAD;
	}
	else if (boostback_sw == 2)
	{
	eng_status = 2;
	ConfigMain();
	vrot_lim = 10*RAD;
	SetThrusterGroupLevel(THGROUP_MAIN, 0.4);
	mth_level = 0.4;
	rchud_sw = 1;
	}
	else if(abs(pointing_error) < (60*RAD) && length(flight_vector) > 2)
	{
	eng_status = 2;
	ConfigMain();
	vrot_lim = 10*RAD;
	mth_level = 0.5;
	SetThrusterGroupLevel(THGROUP_MAIN, mth_level);
	rchud_sw = 2; // flyback burn active
	}
	else 
	{	
	SetThrusterGroupLevel(THGROUP_MAIN, 0);
	mth_level = 0;
	rchud_sw = 1; // flyback burn enabled
	}

	GoToFlightVector();

	if (abs(flight_vector.z) < 2 && fb_dv_req < 16)
	//if (length(flight_vector) < 4)
	{
	KillAllRCS();
	SetThrusterGroupLevel(THGROUP_MAIN,0);
	mth_level = 0;
	if (afs_status == 2)afs_status = 3;
	//ActivateNavmode(NAVMODE_KILLROT);
	boostback_sw = 0;
	}
	
	//sprintf(oapiDebugString(), "flight_vector %0.1f %0.1f %0.1f boostback_sw %d", flight_vector.x,flight_vector.y,flight_vector.z,boostback_sw);
}

//void ss_booster::DoBoostBack2 (void)
//{
//if (boostback_sw == 0) return;
//
//if (eng_status != 1)
//{
//eng_status = 1;
//ConfigMain();
//}
//
//	
//	OBJHANDLE hearth=GetSurfaceRef();
//	OBJHANDLE htgt_vessel = oapiGetObjectByName(refname);
//	if (htgt_vessel == NULL)htgt_vessel = oapiGetBaseByName(hearth,refname);
//	if (htgt_vessel == NULL)
//	{
//		boostback_sw = 0;
//		return;
//	}
//	DeactivateNavmode(NAVMODE_KILLROT);
//	
//
//	VECTOR3 haspdv = _V(0,0,0);
//	GetGroundspeedVector(FRAME_HORIZON,haspdv);
//	double GM = GGRAV*oapiGetMass(hearth);
//
//	if (haspdv.y > 0)
//		{
//		boostback_sw = 2;
//		return;
//		}
//
//	double emp_mass = GetEmptyMass()+GetPropellantMass(hRCSProp);
//	double vmass = GetMass();
//	double mass_fct = (emp_mass*emp_mass)/(vmass*vmass);
//	if (mass_fct < 0.4 )mass_fct = 0.4;
//	double tgt_height_offset = 5550+(4020*mass_fct);//-(1.59-sin(2*((90*RAD)-impact_angle))));//9000 //5750
//
//	double fpang = atan2(haspdv.x,haspdv.z);
//	double sinfpang = sin(fpang);
//	double fpfact = -2000*sinfpang;
//	tgt_height_offset = tgt_height_offset + fpfact;
//
//	double aoi_fct = 15000*(sin((65*RAD)-impact_angle));
//	tgt_height_offset = tgt_height_offset + aoi_fct ;/////////////
//	
//	double ref_fct = ((vmass - emp_mass)/7500) + 0.8;//10000
//	if (ref_fct > 3)ref_fct = 3;
//	else if (ref_fct < 1)ref_fct = 1;
//	tgt_height_offset = tgt_height_offset*ref_fct;////*2.9
//
//	//centrifugal acc
//		VECTOR3 relvel = _V(0,0,0);
//		GetRelativeVel(hearth,relvel);
//		double orb_vel = length(relvel);
//		double orb_grnd_spd = sqrt((orb_vel*orb_vel)-(haspdv.y*haspdv.y));
//		double rds = (rt + GetAltitude());
//		double ang_vel = orb_grnd_spd/rds;
//		double acc_cent_fact = 1;
//		double acc_cent = ang_vel*ang_vel*rds*acc_cent_fact;
//		double gravt = GM/(rds*rds);
//		double g = acc_cent - gravt;
//
//	double aa = g/2;
//	double bb = haspdv.y;
//	double cc = GetAltitude()-tgt_height_offset;
//
//	double timetoimp = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);
//
//	double ves_lng = 0;
//	double ves_lat = 0;
//	double ves_rad = 0;
//	if (oapiGetObjectType(htgt_vessel) == OBJTP_VESSEL)
//	{
//		oapiGetEquPos(htgt_vessel, &ves_lng, &ves_lat, &ves_rad);
//	}
//	
//	else
//	{
//		oapiGetBaseEquPos(htgt_vessel, &ves_lng, &ves_lat, &ves_rad);
//	}
//	double lng_off = PI2*timetoimp/86164;
//
//	double tgt_lng = ves_lng+lng_off;
//	double tgt_lat = ves_lat;
//	if (tgt_lng > PI2) tgt_lng = tgt_lng - PI2;
//	double tgt_rad = ves_rad + tgt_height_offset;
//
//	VECTOR3 tgt_pos_global = _V(0,0,0);
//	oapiEquToGlobal(hearth, tgt_lng, tgt_lat, tgt_rad, &tgt_pos_global);
//
//	VECTOR3 earth_pos = _V(0,0,0);
//	oapiGetGlobalPos(hearth, &earth_pos);
//
//
//	VECTOR3 tgt_vector_rele = tgt_pos_global - earth_pos;
//	normalise(tgt_vector_rele);
//
//	VECTOR3 my_vector_rele = _V(0,0,0);
//	GetRelativePos(hearth, my_vector_rele);
//	normalise(my_vector_rele);
//
//	dotprod = dotp(tgt_vector_rele,my_vector_rele);
//	if (dotprod > 1)dotprod = 1;
//	else if (dotprod < -1)dotprod = -1;
//	
//	double tgt_angle = acos(dotprod);
//
//	VECTOR3 normal_me_tgt_vects = _V(0,0,0);
//	normal_me_tgt_vects = crossp(tgt_vector_rele,my_vector_rele);
//	normalise(normal_me_tgt_vects);
//
//	VECTOR3 tangential_vector = _V(0,0,0);
//	tangential_vector = crossp(my_vector_rele, normal_me_tgt_vects);
//	normalise(tangential_vector);
//
//	double tgt_range = tgt_angle * (rt+((GetAltitude()+tgt_height_offset)/2));
//
//	VECTOR3 vertvel_glob = _V((bb*my_vector_rele.x),(bb*my_vector_rele.y),(bb*my_vector_rele.z));
//	
//
//	double final_vel_xz = tgt_range/timetoimp;
//	VECTOR3 final_vel_xz_glob = _V(0,0,0);
//	final_vel_xz_glob.x = tangential_vector.x*final_vel_xz;
//	final_vel_xz_glob.y = tangential_vector.y*final_vel_xz;
//	final_vel_xz_glob.z = tangential_vector.z*final_vel_xz;
//
//	
//
//	VECTOR3 final_vel_vect = final_vel_xz_glob+vertvel_glob;
//
//	double fvvh = GetAltitude()-tgt_height_offset;
//	double finalvv = length(vertvel_glob)+sqrt(2*-g*fvvh);
//	double finalvh = length(final_vel_xz_glob);
//	impact_angle2 = atan2(finalvv,finalvh);
//	//sprintf(oapiDebugString(), "impang2 %0.1f", impact_angle2*DEG);
//
//
//	VECTOR3 req_dv_vector = _V(0,0,0);
//	req_dv_vector.x = final_vel_vect.x - relvel.x;
//	req_dv_vector.y = final_vel_vect.y - relvel.y;
//	req_dv_vector.z = final_vel_vect.z - relvel.z;
//
//	VECTOR3 mypos_glob = _V(0,0,0);
//	GetGlobalPos(mypos_glob);
//
//	req_dv_vector = req_dv_vector + mypos_glob;
//
//	VECTOR3 req_dv_vector_loc = _V(0,0,0);
//	Global2Local(req_dv_vector,req_dv_vector_loc);
//
//
//	if (length(req_dv_vector_loc) > fb_dv_req)
//	{	
//		boostback_sw = 2;
//		return;
//	}
//	else fb_dv_req = length(req_dv_vector_loc);
//
//	flight_vector = req_dv_vector_loc;
//	GoToFlightVector();
//
//	VECTOR3 req_dv_vector_norm = req_dv_vector_loc;
//	normalise(req_dv_vector_norm);
//	VECTOR3 curr_att_loc = _V(0,0,1);
//
//	double pointing_error = acos(dotp(curr_att_loc,req_dv_vector_norm));
//
//
//	double max_dv = step_time*(GetThrusterMax(th_main[eng_status])/vmass);
//	double eng_lvl = flight_vector.z/max_dv;
//	if (eng_lvl > 1)eng_lvl = 1;
//	else if (eng_lvl < 0)eng_lvl = 0;
//
//	double pe_fact = abs(fb_dv_req*tan(pointing_error));
//
//	if (pe_fact < 25 && flight_vector.z > 4)
//	{
//		SetThrusterGroupLevel(THGROUP_MAIN, eng_lvl);
//		mth_level = eng_lvl;
//		rchud_sw = 2;
//	}
//	else 
//	{
//	SetThrusterGroupLevel(THGROUP_MAIN, 0);
//	mth_level = 0;
//	rchud_sw = 1;
//	}
//
//	if (abs(flight_vector.z) < 4 && length(flight_vector) < 8)
//	{
//	KillAllRCS();
//	SetThrusterGroupLevel(THGROUP_MAIN,0);
//	mth_level = 0;
//	if (afs_status == 2)afs_status = 3;
//	//ActivateNavmode(NAVMODE_KILLROT);
//	boostback_sw = 0;
//	}
//}

//void ss_booster::DoCorrection (void)
//{
//if (mom_sw == 1)/// 3 eng
//{
//if (eng_status != 1)
//	{
//	eng_status = 1;
//	ConfigMain();
//	}
//}
//else if (edlo_sw == 0)
//{
//if (eng_status != 2)
//	{
//	eng_status = 2;
//	ConfigMain();
//	}
//}
//else
//{
//if (eng_status != 1)
//	{
//	eng_status = 1;
//	ConfigMain();
//	}
//}
//
//OBJHANDLE htgt_vessel = oapiGetObjectByName(refname);
//if (htgt_vessel == NULL)htgt_vessel = oapiGetBaseByName(GetGravityRef(),refname);
//if (htgt_vessel == NULL) 
//{
//	//sprintf(oapiDebugString(), "here3  cs %d", correct_sw);
//	correct_sw = 1;
//	return;
//}
//
//double emp_mass = GetEmptyMass()+GetPropellantMass(hRCSProp);
//double vmass = GetMass();
//double tgt_height_offset = 100;//250
//
//VECTOR3 tgt_glob_pos = _V(0,0,0);
//oapiGetGlobalPos(htgt_vessel, &tgt_glob_pos);
//
//VECTOR3 tgt_loc_pos = _V(0,0,0);
//Global2Local(tgt_glob_pos,tgt_loc_pos);
//
//VECTOR3 tgt_hrz_pos = _V(0,0,0);
//HorizonRot(tgt_loc_pos,tgt_hrz_pos);
//
//VECTOR3 haspdv = _V(0,0,0);
//GetGroundspeedVector (FRAME_HORIZON,haspdv);
//
//double alti = GetAltitude();
//
//
//	OBJHANDLE hearth=GetSurfaceRef();
//	double rds = rt+alti;
//    double GM = GGRAV*oapiGetMass(hearth);
//	double gravt = GM/(rds*rds);
//	double g = -gravt;
//	double aa = g/2;
//	double bb = haspdv.y;
//	double cc = alti-tgt_height_offset;
//	double timetoimp = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);
//
//
//	double k_factor = (30000 - alti)/20000; // 0 -1
//	if (k_factor > 1)k_factor = 1;
//	k_factor = 1.85 - (1.0*k_factor);
//	double k = 75*9.81/k_factor;//120
//	//double k = 75*9.81/1.3;
//	 double gsvx = sqrt((haspdv.x*haspdv.x)+(haspdv.z*haspdv.z));
//	 double expo = -k*timetoimp/vmass;
//
//	 double  range = vmass*gsvx/k*(1-(pow(M_E, expo)));
//
//	 double grdspd_az = atan2(haspdv.x, haspdv.z);
//	 double tgt_az = atan2(tgt_hrz_pos.x,tgt_hrz_pos.z);
//	 double tgt_angle = grdspd_az - tgt_az;
//
//	 if (tgt_angle > PI)tgt_angle = tgt_angle-PI2;
//	 if (tgt_angle < -PI)tgt_angle = tgt_angle+PI2;
//
//	double tgt_range = sqrt((tgt_hrz_pos.x*tgt_hrz_pos.x)+(tgt_hrz_pos.z*tgt_hrz_pos.z));
//	tgt_range = tgt_range*cos(tgt_angle);
//
//
//	//sprintf(oapiDebugString(), "rng %0.1f tgtrng %0.1f tgtang %0.1f", range, tgt_range, tgt_angle*DEG);
//	if (range < tgt_range)
//	{
//		correct_sw = 1;
//		rot_sw = 0;
//		return;
//	}
//	//sprintf(oapiDebugString(), "rng %0.1f tgtrng %0.1f correct_sw %0.1f", range, tgt_range, correct_sw);
//
//
//	VECTOR3 req_fl_vec_hrz =  haspdv;
//
//	if(abs(tgt_angle) < (1*RAD)) rot_sw = 1;
//	else if(abs(tgt_angle) > (80*RAD)) rot_sw = 1;
//
//	if (rot_sw == 0)
//	{
//	VECTOR3 norm_flt_pln_hrz = crossp(_V(0,1,0),haspdv);
//	VECTOR3 rot_vec = crossp(norm_flt_pln_hrz,haspdv);
//	normalise(rot_vec);
//
//	/////// more efficient if no correction!
//	double rot_angle = 0 * -tgt_angle;
//
//	double ct = cos(rot_angle);
//	double st = sin(rot_angle);
//
//	double ux = rot_vec.x;
//	double uy = rot_vec.y;
//	double uz = rot_vec.z;
//	double ux2 = ux*ux;
//	double uy2 = uy*uy;
//	double uz2 = uz*uz;
//
//	MATRIX3 RG;
//
//	RG.m11 = ux2+(1-ux2)*ct;
//	RG.m12 = ux*uy*(1-ct)-uz*st;
//	RG.m13 = ux*uz*(1-ct)+uy*st;
//	RG.m21 = ux*uy*(1-ct)+uz*st;
//	RG.m22 = uy2+(1-uy2)*ct;
//	RG.m23 = uy*uz*(1-ct)-ux*st;
//	RG.m31 = ux*uz*(1-ct)-uy*st;
//	RG.m32 = uy*uz*(1-ct)+ux*st;
//	RG.m33 = uz2+(1-uz2)*ct;
//
//	req_fl_vec_hrz = mul(RG, haspdv);
//	}
//
//
//	VECTOR3 req_fv_loc = _V(0,0,0);
//	HorizonInvRot(-req_fl_vec_hrz,req_fv_loc);
//	flight_vector = req_fv_loc;
//	GoToFlightVector();
//
//	VECTOR3 req_fv_loc_norm = req_fv_loc;
//	normalise(req_fv_loc_norm);
//	VECTOR3 curr_att_loc = _V(0,0,1);
//
//	dotprod = dotp(curr_att_loc,req_fv_loc_norm);
//	if (dotprod > 1)dotprod = 1;
//	else if (dotprod < -1)dotprod = -1;
//
//	double pointing_error = acos(dotprod);
//
//	//sprintf(oapiDebugString(), "pt err %0.1f rng %0.1f tgtrng %0.1f rot_sw %d", pointing_error*DEG, range, tgt_range, rot_sw);
//
//	if (pointing_error < (10*RAD) && range > tgt_range)
//	{
//	SetThrusterLevel(th_main[eng_status],1);
//	if ((edlo_sw+mom_sw) > 0)SetThrusterLevel(th_main[eng_status],0.7);
//	}
//	else 
//	{
//	SetThrusterLevel(th_main[eng_status],0);
//	}
//
//	//// hud info
//	rchud_sw = 4;
//	if (swt == 1)rchud_sw = 5;
//	
//}

void ss_booster::DoReentryCorrection (void)
{
	/////////////// disabled 
	rc_sw = 1;
	return;
	/////////////////////////////////////////////

OBJHANDLE htgt_vessel = oapiGetObjectByName(refname);
if (htgt_vessel == NULL)htgt_vessel = oapiGetBaseByName(GetGravityRef(),refname);
if (htgt_vessel == NULL) return;

//if (mom_sw == 1)/// 3 eng
//{
if (eng_status != 1)
	{
	eng_status = 1;
	ConfigMain();
	}
//}

//else if (edlo_sw == 0)/// 1 eng
//{
//if (eng_status != 2)
//	{
//	eng_status = 2;
//	ConfigMain();
//	}
//}
//else/// 3 eng
//{
//if (eng_status != 1)
//	{
//	eng_status = 1;
//	ConfigMain();
//	}
//}


double alti = GetAltitude();
double emp_mass = GetEmptyMass();//+GetPropellantMass(hRCSProp);
double vmass = GetMass();
double current_fuel_mass = vmass - emp_mass;
double tgt_height_offset = 700 + (edlo_sw * 800);

VECTOR3 tgt_glob_pos = _V(0,0,0);
oapiGetGlobalPos(htgt_vessel, &tgt_glob_pos);

VECTOR3 tgt_loc_pos = _V(0,0,0);
Global2Local(tgt_glob_pos,tgt_loc_pos);

VECTOR3 tgt_hrz_pos = _V(0,0,0);
HorizonRot(tgt_loc_pos,tgt_hrz_pos);

VECTOR3 haspdv = _V(0,0,0);
GetGroundspeedVector(FRAME_HORIZON,haspdv);

double momentum = length(haspdv)*(vmass-emp_mass)/7;
if (mom_sw == 0)
{
if (momentum > (1000*10000)) //1500*20000
	{
	mom_sw = 1;
	if (eng_status != 1)
		{
		eng_status = 1;
		ConfigMain();
		}
	}
}

double gsvx = sqrt((haspdv.x*haspdv.x)+(haspdv.z*haspdv.z));

//double alti = GetAltitude();
OBJHANDLE hearth=GetSurfaceRef();

VECTOR3 relvelhz = _V(0,0,0);
GetRelativeVel(hearth, relvelhz);

double rvh = length(relvelhz);
rvh = sqrt((rvh*rvh) - (haspdv.y*haspdv.y));
double acc_cfg = rvh*rvh/(alti+rt); // centrifugal acc

	double rds = rt+alti;
    double GM = GGRAV*oapiGetMass(hearth);
	double gravt = GM/(rds*rds);
	gravt = gravt - acc_cfg;
	double g = -gravt;
	double aa = g/2;
	//double aaa = -g;
	double bb = haspdv.y;
	double cc = alti-tgt_height_offset;
	double timetoimp = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);
	double timetoimp2 = (-bb-(sqrt((bb*bb)-(4*aa*(cc-25000)))))/(2*aa);
	if (timetoimp2 < 0)timetoimp2 = 0;
	else if (((bb*bb)-(4*aa*(cc-25000))) < 0)timetoimp2 = 0;

	 double range1 = gsvx*timetoimp2;  // above 25km
	 
	 double k = 75*9.81/0.25; // 2
	 double expo = -k*(timetoimp-timetoimp2)/vmass;
	 double range2 = vmass*gsvx/k*(1-(pow(M_E, expo))); //below 25km
	 double range = (range1+range2);


	 double grdspd_az = atan2(haspdv.x, haspdv.z);
	 double tgt_az = atan2(tgt_hrz_pos.x,tgt_hrz_pos.z);
	 double tgt_angle = grdspd_az - tgt_az;

	 if (tgt_angle > PI)tgt_angle = tgt_angle-PI2;
	 if (tgt_angle < -PI)tgt_angle = tgt_angle+PI2;

	double tgt_range = 250 + sqrt((tgt_hrz_pos.x*tgt_hrz_pos.x)+(tgt_hrz_pos.z*tgt_hrz_pos.z));

	// get pitch
	VECTOR3 z_axis_ves = _V(0,0,1);
	VECTOR3 z_axis_hrz = _V(0,0,1);
	HorizonRot(z_axis_ves, z_axis_hrz);
	normalise(z_axis_hrz);
	double ves_pitch = PI05 - acos(dotp(z_axis_hrz, _V(0,1,0)));

	double cosp = cos(ves_pitch);//cos(GetPitch());
	double sinp = sin(ves_pitch);//sin(GetPitch());
	//sprintf(oapiDebugString(), "req_rdv %0.4f cosp %0.4f", req_rdv, cosp);

	double th_isp = GetThrusterIsp(th_main[eng_status]);
	double th_max = GetThrusterMax(th_main[eng_status]);
	if ((edlo_sw+mom_sw) > 0)th_max = th_max*0.7; // limit 3 eng to 70%

	double renglvl = 0;

	//if(swt == 0)// ready to start burn
	{
	double fuelmass = current_fuel_mass-22000;//3800; //4000
	double avail_dv = th_isp*log(vmass/(vmass-fuelmass));
	//double gspeed = length(haspdv);
	//if(avail_dv > gspeed) avail_dv = gspeed;
	double fuelrate = th_max/th_isp;
	double burntime = fuelmass/fuelrate;
	

	////////// dv check
	VECTOR3 finv = haspdv + _V(0,(burntime * g),0);
	if(avail_dv > length(finv))
	{
	double dvmain = length(finv);
	double fraction = exp(dvmain/GetThrusterIsp(th_main[eng_status]));
	fuelmass = vmass - (vmass/fraction);
	burntime = fuelmass/fuelrate;
	finv = haspdv + _V(0,(burntime * g),0); 
	
	// 2nd iteration
	dvmain = length(finv);
	fraction = exp(dvmain/GetThrusterIsp(th_main[eng_status]));
	fuelmass = vmass - (vmass/fraction);
	burntime = fuelmass/fuelrate;
	finv = haspdv + _V(0,(burntime * g),0);
	avail_dv = length(finv);
	dvmain = length(finv);
	fraction = exp(dvmain/GetThrusterIsp(th_main[eng_status]));
	fuelmass = vmass - (vmass/fraction);
	burntime = fuelmass/fuelrate;
	}

	double burntime_frac = burntime*0.05;
	double fuelmass_frac = fuelmass*0.05;
	double avail_dv_frac = th_isp*log(vmass/(vmass-fuelmass_frac));
	double av_gravt = 9.81;//0.5*(9.81 + gravt);
	double cos_hdg_err = abs(cos(1.8*tgt_angle));//2
	double gsvx_current = gsvx;
	double v_vel_current = haspdv.y;
	double cosp_current = cosp;
	double sinp_current = sinp;
	double final_vel_h = gsvx_current - (cosp_current*avail_dv_frac*cos_hdg_err);
	double final_vel_v = v_vel_current + (sinp_current*avail_dv_frac)-(av_gravt*burntime_frac);
	double final_ves_pitch = atan2(-final_vel_v,final_vel_h);
	

	//double av_vel_h = (gsvx + final_vel_h)*0.5;
	//double av_vel_v = (haspdv.y + final_vel_v)*0.5;

	double acc_h = (final_vel_h - gsvx_current)/burntime_frac;
	double acc_v = (final_vel_v - v_vel_current)/burntime_frac;

	double final_tgt_range = tgt_range;// - ((burntime*gsvx) + (acc_h*burntime*burntime*0.5));
	double final_ves_alt = alti;// + ((burntime*haspdv.y) + (acc_v*burntime*burntime*0.5));
	
	int i;
	for (i = 0; i < 20; i++)
	{
		avail_dv_frac = th_isp*log((vmass-(fuelmass_frac*i))/(vmass-(fuelmass_frac*(i+1))));
		final_vel_h = gsvx_current - (cosp_current*avail_dv_frac*cos_hdg_err);
		final_vel_v = v_vel_current + (sinp_current*avail_dv_frac)-(av_gravt*burntime_frac);
		final_ves_pitch = atan2(-final_vel_v,final_vel_h);

		acc_h = (final_vel_h - gsvx_current)/burntime_frac;
		acc_v = (final_vel_v - v_vel_current)/burntime_frac;
		final_tgt_range = final_tgt_range - ((burntime_frac*gsvx_current) + (acc_h*burntime_frac*burntime_frac*0.5));
		final_ves_alt = final_ves_alt + ((burntime_frac*v_vel_current) + (acc_v*burntime_frac*burntime_frac*0.5));

		gsvx_current = final_vel_h;
		v_vel_current = final_vel_v;
		cosp_current = cos(final_ves_pitch);
		sinp_current = sin(final_ves_pitch);
	}

	bb = final_vel_v;
	cc = final_ves_alt - tgt_height_offset;
	timetoimp = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);
	timetoimp2 = (-bb-(sqrt((bb*bb)-(4*aa*(cc-25000)))))/(2*aa);
	if (timetoimp2 < 0)timetoimp2 = 0;
	else if (((bb*bb)-(4*aa*(cc-25000))) < 0)timetoimp2 = 0;

	range1 = final_vel_h*timetoimp2;  // above 25km
	 
	k = 75*9.81/0.25; // 0.25
	expo = -k*(timetoimp-timetoimp2)/(emp_mass+22000);//4000
	range2 = (emp_mass+22000)*final_vel_h/k*(1-(pow(M_E, expo)));//4000 //below 25km
	double final_range = (range1+range2);

	VECTOR3 gspd_vect_loc = _V(0,0,1);
	GetGroundspeedVector(FRAME_LOCAL,gspd_vect_loc);
	normalise(gspd_vect_loc);
	double ptng_err = acos(dotp(gspd_vect_loc, _V(0,0,-1)));

	if(ptng_err < 5*RAD)
		{
		if(final_range < final_tgt_range)reentry_correct_sw = 1;
		if(final_range > (final_tgt_range) && reentry_correct_sw == 1)swt = 1;
		if(edlo_sw == 0)
			{
			if(final_ves_alt < (20000))swt = 1;
			if(final_tgt_range < (2000)) swt = 1;
			}
		}

	////sprintf(oapiDebugString(), "avail_dv %0.1f burntime %0.1f av_ves_pitch %0.1f final_tgt_range %0.1f final_range %0.1f range %0.1f tgt_range %0.1f g %0.1f final_ves_alt %0.1f cos_hdg_err %0.1f swt %d ptng_err %0.1f", avail_dv, burntime,av_ves_pitch*DEG, final_tgt_range, final_range,range, tgt_range, g, final_ves_alt, cos_hdg_err, swt, DEG*ptng_err);
	//sprintf(oapiDebugString(), "avail_dv %0.1f burntime %0.1f final_tgt_range %0.1f final_range %0.1f range %0.1f tgt_range %0.1f g %0.1f final_ves_alt %0.1f", avail_dv, burntime, final_tgt_range, final_range,range, tgt_range, g, final_ves_alt);
	}


	renglvl = swt;
	if((edlo_sw+swt) == 2)renglvl = 0.7; // 3 eng throttle70%
	else if((mom_sw+swt) == 2)renglvl = 0.7; // 3 eng throttle70%

	//hud_flight_range = range;
	//hud_range = tgt_range;

	if (range < tgt_range)// burn end
	{
		SetThrusterLevel(th_main[eng_status],0);
		mth_level = 0;
		rc_sw = 1;
		//correct_sw = 1;
		return;
	}
	else if (GetPropellantMass(hProp) < 22000)// 3800// burn end
	{
		SetThrusterLevel(th_main[eng_status],0);
		mth_level = 0;
		rc_sw = 1;
		//correct_sw = 1;
		return;
	}
	

	VECTOR3 req_fl_vec_hrz =  haspdv;
	int rot_sw2 = 0;
	if(abs(tgt_angle) < (0.05*RAD)) rot_sw2 = 1;
	else if(abs(tgt_angle) > (90*RAD)) rot_sw2 = 1;
	if (alti < 35000)rot_sw2 = 1;
	if(swt == 0)rot_sw2 = 1;

	if (rot_sw2 == 0)
	{
	VECTOR3 norm_flt_pln_hrz = crossp(_V(0,1,0),haspdv);
	normalise(norm_flt_pln_hrz);
	VECTOR3 rot_vec = crossp(norm_flt_pln_hrz,haspdv);
	normalise(rot_vec);
	//////////
	//rot_vec = _V(0,-1,0);

	double rot_angle = -5 * tgt_angle; //10

	double ct = cos(rot_angle);
	double st = sin(rot_angle);

	double ux = rot_vec.x;
	double uy = rot_vec.y;
	double uz = rot_vec.z;
	double ux2 = ux*ux;
	double uy2 = uy*uy;
	double uz2 = uz*uz;

	MATRIX3 RG;

	RG.m11 = ux2+(1-ux2)*ct;
	RG.m12 = ux*uy*(1-ct)-uz*st;
	RG.m13 = ux*uz*(1-ct)+uy*st;
	RG.m21 = ux*uy*(1-ct)+uz*st;
	RG.m22 = uy2+(1-uy2)*ct;
	RG.m23 = uy*uz*(1-ct)-ux*st;
	RG.m31 = ux*uz*(1-ct)-uy*st;
	RG.m32 = uy*uz*(1-ct)+ux*st;
	RG.m33 = uz2+(1-uz2)*ct;

	req_fl_vec_hrz = mul(RG, haspdv);
	}


	VECTOR3 req_fv_loc = _V(0,0,0);
	HorizonInvRot(-req_fl_vec_hrz,req_fv_loc);
	flight_vector = req_fv_loc;
	GoToFlightVector();

	VECTOR3 req_fv_loc_norm = req_fv_loc;
	normalise(req_fv_loc_norm);
	VECTOR3 curr_att_loc = _V(0,0,1);

	dotprod = dotp(curr_att_loc,req_fv_loc_norm);
	if (dotprod > 1)dotprod = 1;
	else if (dotprod < -1)dotprod = -1;

	double pointing_error = acos(dotprod);

	if (pointing_error < (80*RAD) && range > tgt_range)
	{
	SetThrusterLevel(th_main[eng_status],renglvl);
	mth_level = renglvl;
	}
	else 
	{
	SetThrusterLevel(th_main[eng_status],0);
	mth_level = 0;
	}
	rat_sw = 0;

	//// hud info
	rchud_sw = 4;
	if (swt == 1)rchud_sw = 5;
}


void ss_booster::DoLAT (void)
{
double machnum = GetMachNumber();
OBJHANDLE htgt_vessel = oapiGetObjectByName(refname);
if (htgt_vessel == NULL) htgt_vessel = oapiGetBaseByName(GetGravityRef(),refname);
if (htgt_vessel == NULL) 
{
	DoRAT();
	return;
}
if (machnum > 6)
{
	DoRAT();
	return;
}

double lon,lat,rad = 0;
oapiGetEquPos(GetHandle(), &lon, &lat, &rad);
double windspeed = 0;
VECTOR3 windvector = oapiGetWindVector(GetSurfaceRef(),lon, lat, (rad-6371010), 0, &windspeed );
if (length(windvector) == 0) windvector = _V(0,0.00001,0);

double alti = GetAltitude() - catch_offset.y;

double vmass = GetMass();
double tgt_height_offset = 65+(2.2*alti/3);
if(alti > 5000)tgt_height_offset = 65+(7.5*alti/8);
//if(GetThrusterGroupLevel(THGROUP_MAIN) > 0)tgt_height_offset = 65+(2*alti/3);
//if(edlo_sw == 1 && alti > 5000) tgt_height_offset = 1*alti/2;

VECTOR3 tgt_glob_pos = _V(0,0,0);
oapiGetGlobalPos(htgt_vessel, &tgt_glob_pos);

VECTOR3 tgt_loc_pos = _V(0,0,0);
Global2Local(tgt_glob_pos,tgt_loc_pos);

VECTOR3 tgt_hrz_pos = _V(0,0,0);
HorizonRot(tgt_loc_pos,tgt_hrz_pos);

//tgt_hrz_pos.y = tgt_hrz_pos.y - catch_offset.y;
if(tgt_att_valid == true)tgt_hrz_pos = tgt_hrz_pos + catch_offset;

//oapiGetEquPos(htgt_vessel, &lon, &lat, &rad);
//double tgt_windspeed = 0;
//VECTOR3 tgt_windvector = oapiGetWindVector(GetSurfaceRef(),lon, lat, (rad-6371010+1000), 0, &tgt_windspeed );
//if (length(tgt_windvector) == 0) tgt_windvector = _V(0,0.00001,0);
//sprintf(oapiDebugString(), "tgt_windvector %0.1f %0.1f %0.1f", tgt_windvector.x,tgt_windvector.y,tgt_windvector.z);

VECTOR3 haspdv = _V(0,0,0);
GetGroundspeedVector (FRAME_HORIZON,haspdv);


	 double gsvx = sqrt((haspdv.x*haspdv.x)+(haspdv.z*haspdv.z));
	

	VECTOR3 haspdv_nm = haspdv;
	normalise(haspdv_nm);
	VECTOR3 tgt_vec_nm = _V(tgt_hrz_pos.x,-(alti-tgt_height_offset), tgt_hrz_pos.z);
	normalise(tgt_vec_nm);

	double tgt_vec_length = sqrt((tgt_hrz_pos.x*tgt_hrz_pos.x)+(tgt_hrz_pos.z*tgt_hrz_pos.z));
	double tgt_vec_length2 = tgt_vec_length;
	tgt_vec_length = tgt_vec_length + (alti*0.001*10);//125;
	VECTOR3 tgt_hrz_pos_adj = tgt_hrz_pos;
	tgt_hrz_pos_adj.y = 0;
	normalise(tgt_hrz_pos_adj);
	tgt_hrz_pos_adj.x = tgt_hrz_pos_adj.x*tgt_vec_length;
	tgt_hrz_pos_adj.y = 0;
	tgt_hrz_pos_adj.z = tgt_hrz_pos_adj.z*tgt_vec_length;

	VECTOR3 tgt_vec_nm2 = _V(tgt_hrz_pos_adj.x,-(alti*0.59), tgt_hrz_pos_adj.z);//0.55
	if(edlo_sw == 1)tgt_vec_nm2 = _V(tgt_hrz_pos_adj.x,-(alti*0.55), tgt_hrz_pos_adj.z);//0.55
	//VECTOR3 tgt_vec_nm2 = _V(tgt_hrz_pos_adj.x,-(alti*0.66), tgt_hrz_pos_adj.z);//0.55
	normalise(tgt_vec_nm2);

	double dotprod = dotp(tgt_vec_nm,_V(0,-1,0));
	if (dotprod > 1)dotprod = 1;
	else if (dotprod < -1)dotprod = -1;
	 double tgt_angle = acos(dotprod);

	dotprod = dotp(haspdv_nm,_V(0,-1,0));
	if (dotprod > 1)dotprod = 1;
	else if (dotprod < -1)dotprod = -1;
	 double aspd_angle = acos(dotprod);

	 if (tgt_angle > PI)tgt_angle = tgt_angle-PI2;
	 if (tgt_angle < -PI)tgt_angle = tgt_angle+PI2;

	double tgt_range = sqrt((tgt_hrz_pos.x*tgt_hrz_pos.x)+(tgt_hrz_pos.z*tgt_hrz_pos.z));

	VECTOR3 req_fl_vec_hrz =  _V(0,-1,0);

	if(GetThrusterGroupLevel(THGROUP_MAIN) == 0)
	{

	req_fl_vec_hrz = haspdv_nm;
	
	dotprod = dotp(tgt_vec_nm2,haspdv_nm);
	if (dotprod > 1)dotprod = 1;
	else if (dotprod < -1)dotprod = -1;
	double tgt_angle_az = acos(dotprod);
	VECTOR3 rot_vec = crossp(tgt_vec_nm2,haspdv_nm);
	normalise(rot_vec);
	

	

	double mach_rot_factor = 0.5*(6 - machnum);
	if(mach_rot_factor > 1)mach_rot_factor = 1;
	else if(mach_rot_factor < 0)mach_rot_factor = 0;

	double rot_angle = -3.5 * tgt_angle_az*(1.0+windspeed/40); // 3.5

	if(rot_angle > (20*RAD))rot_angle = 20*RAD;
	else if(rot_angle < (-20*RAD))rot_angle = -20*RAD;

	rot_angle = rot_angle*mach_rot_factor;

	double ct = cos(rot_angle);
	double st = sin(rot_angle);

	double ux = rot_vec.x;
	double uy = rot_vec.y;
	double uz = rot_vec.z;
	double ux2 = ux*ux;
	double uy2 = uy*uy;
	double uz2 = uz*uz;

	MATRIX3 RG;

	RG.m11 = ux2+(1-ux2)*ct;
	RG.m12 = ux*uy*(1-ct)-uz*st;
	RG.m13 = ux*uz*(1-ct)+uy*st;
	RG.m21 = ux*uy*(1-ct)+uz*st;
	RG.m22 = uy2+(1-uy2)*ct;
	RG.m23 = uy*uz*(1-ct)-ux*st;
	RG.m31 = ux*uz*(1-ct)-uy*st;
	RG.m32 = uy*uz*(1-ct)+ux*st;
	RG.m33 = uz2+(1-uz2)*ct;

	req_fl_vec_hrz = mul(RG, haspdv_nm);

	//if(alti < 2000)req_fl_vec_hrz =  _V(0,-1,0);

	rchud_sw = 6;
	}

	else
	{
	VECTOR3 rot_vec = crossp(tgt_vec_nm,haspdv_nm);
	normalise(rot_vec);

	dotprod = dotp(tgt_vec_nm,haspdv_nm);
	if (dotprod > 1)dotprod = 1;
	else if (dotprod < -1)dotprod = -1;
	tgt_angle = 0.75*acos(dotprod);//0.75
	double rot_angle = tgt_angle;

	int windrot_sw = 0;

	int alti_mode_sw = 600;
	if (windspeed > 0.1)alti_mode_sw = 600;
//	if (alti < alti_mode_sw)	//300

	double time_to_tgt = tgt_vec_length2/gsvx;
	if(time_to_tgt < ((tgt_angle*DEG*2)/4) && alti < 700) alti_mode_sw2 = 1;

	if(alti_mode_sw2 == 1)
	{
//	sprintf(oapiDebugString(), "alti_mode_sw2 %d", alti_mode_sw2);
	rot_angle = -0.02 * gsvx * aspd_angle * 0.5;// 0.8
	rot_vec = crossp(haspdv_nm,_V(0,-1,0));
	normalise(rot_vec);
	windrot_sw = 1;
	}


	double ct = cos(rot_angle);
	double st = sin(rot_angle);

	double ux = rot_vec.x;
	double uy = rot_vec.y;
	double uz = rot_vec.z;
	double ux2 = ux*ux;
	double uy2 = uy*uy;
	double uz2 = uz*uz;

	MATRIX3 RG;

	RG.m11 = ux2+(1-ux2)*ct;
	RG.m12 = ux*uy*(1-ct)-uz*st;
	RG.m13 = ux*uz*(1-ct)+uy*st;
	RG.m21 = ux*uy*(1-ct)+uz*st;
	RG.m22 = uy2+(1-uy2)*ct;
	RG.m23 = uy*uz*(1-ct)-ux*st;
	RG.m31 = ux*uz*(1-ct)-uy*st;
	RG.m32 = uy*uz*(1-ct)+ux*st;
	RG.m33 = uz2+(1-uz2)*ct;

	req_fl_vec_hrz = mul(RG, haspdv);

	
		
		
	rot_vec = crossp(req_fl_vec_hrz,windvector);
	normalise(rot_vec);
	
	rot_angle = 30*RAD*windspeed*windspeed/1000; //45
	if (windrot_sw == 1)rot_angle = 20*RAD*windspeed*windspeed/1000;//20
	if (rot_angle > (15*RAD))rot_angle = 15*RAD;
	else if (rot_angle < (-15*RAD))rot_angle = -15*RAD;

	ct = cos(rot_angle);
	st = sin(rot_angle);

	ux = rot_vec.x;
	uy = rot_vec.y;
	uz = rot_vec.z;
	ux2 = ux*ux;
	uy2 = uy*uy;
	uz2 = uz*uz;

	RG.m11 = ux2+(1-ux2)*ct;
	RG.m12 = ux*uy*(1-ct)-uz*st;
	RG.m13 = ux*uz*(1-ct)+uy*st;
	RG.m21 = ux*uy*(1-ct)+uz*st;
	RG.m22 = uy2+(1-uy2)*ct;
	RG.m23 = uy*uz*(1-ct)-ux*st;
	RG.m31 = ux*uz*(1-ct)-uy*st;
	RG.m32 = uy*uz*(1-ct)+ux*st;
	RG.m33 = uz2+(1-uz2)*ct;

	req_fl_vec_hrz = mul(RG, req_fl_vec_hrz);
	rchud_sw = 7;

	RollToAZ();
	}

	
	VECTOR3 req_fv_loc = _V(0,0,0);
	HorizonInvRot(-req_fl_vec_hrz,req_fv_loc);
	flight_vector = req_fv_loc;
	GoToFlightVector();

}

//void ss_booster::ApplySkin ()
//{	
//	//bsx2 = oapiLoadTexture ("Falcon9R_V12/f9_hull_logo_char.dds", true);
//	bsx1 = oapiGetTextureHandle (hMhull, 1);
//	oapiBlt(bsx1, bsx2, 0, 0, 0, 0, 256, 1024);
//	//oapiReleaseTexture (bsx2);
//
//	//if(dhMeshx)oapiSetTexture(dhMeshx,1,bsx2);
//
//}

void ss_booster::DoBoom ()
{
	if(GetAttachmentStatus(att_catch) != NULL)
	{
		boom_sw = 2;
		return;
	}

	VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	GetStatusEx(&vs);
	OBJHANDLE hVessel = GetHandle();
	double longitude, latitude, radius;
	double heading;
	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
	oapiGetHeading(hVessel,&heading);
	vs.surf_lat = latitude;
	vs.surf_lng = longitude;
	vs.surf_hdg = heading;
	vs.status = 1;
	DefSetStateEx (&vs);

	//SetMeshVisibilityMode(0,MESHVIS_NEVER);
	//SetMeshVisibilityMode(1,MESHVIS_NEVER);

	boom_sw = boom_sw + step_time;
	if (boom_sw < 2)
	{
		SetThrusterLevel(th_boom[0],1);
		SetThrusterLevel(th_boom[1],1);
	}
	else
	{
		boom_sw = 2;
		SetThrusterLevel(th_boom[0],0);
		SetThrusterLevel(th_boom[1],0);
	}

}

void ss_booster::DoThrustVector ()
{
if (mth_level == 0) return;

double abs_pitch_level = pu_level - pd_level;
double abs_yaw_level = yr_level - yl_level;

//sprintf(oapiDebugString(), "apl %0.5f ayl %0.5f", abs_pitch_level, abs_yaw_level);

double main_max_th = GetThrusterMax(th_main[eng_status]);
double rcs_max_th = GetThrusterMax(th_rcs[0]) * 2;
double main_level = mth_level;

double abs_main_th = main_max_th * main_level;
double abs_pitch_th = rcs_max_th * abs_pitch_level;
double abs_yaw_th = rcs_max_th * abs_yaw_level;

VECTOR3 main_vector = _V(-abs_yaw_th,-abs_pitch_th,abs_main_th);
double ratio_main = length(main_vector) / length(_V(0,0,abs_main_th));
normalise(main_vector);

double new_th_main_level = ratio_main * main_level;
if (new_th_main_level > 1)new_th_main_level = 1;
else if (new_th_main_level < 0)new_th_main_level = 0;

SetThrusterDir(th_main[eng_status], main_vector);
main2_exhaust_dir = main_vector;
main2_exhaust[0].ldir = &main2_exhaust_dir;

double ga = acos(dotp(main_vector, _V(0,0,1)));
//sprintf(oapiDebugString(), "gimbal angle %0.6f", ga*DEG);
SetThrusterLevel(th_main[eng_status], new_th_main_level);

SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);


}

void ss_booster::DoLandingRCS()
{
	VECTOR3 surf_norm_hrz;
	VECTOR3 surf_norm_ves;
	surf_norm_hrz = GetSurfaceNormal();
	HorizonInvRot(surf_norm_hrz,surf_norm_ves);
	MATRIX3 rm = rotm(_V(0,0,1),(45*RAD));
	surf_norm_ves = mul(rm, surf_norm_ves);
	normalise(surf_norm_ves);
	double y_angle = atan2(surf_norm_ves.x,surf_norm_ves.z);
	double x_angle = atan2(surf_norm_ves.y,surf_norm_ves.z);

	//sprintf(oapiDebugString(), "y_angle %0.3f x_angle %0.3f", DEG*y_angle, DEG*x_angle);
	//sprintf(oapiDebugString(), "surf_norm_ves.x %0.3f surf_norm_ves.y %0.3f surf_norm_ves.z %0.3f", surf_norm_ves.x, surf_norm_ves.y, surf_norm_ves.z);

	VECTOR3 ang_vel = _V(0,0,0);
	GetAngularVel(ang_vel);
	ang_vel = mul(rm, ang_vel);

	if (DEG*(abs(y_angle) + abs(x_angle)) < 1.5)
	{
	SetThrusterLevel(th_rcs2[0], 0);
	SetThrusterLevel(th_rcs2[1], 0);
	SetThrusterLevel(th_rcs2[2], 0);
	SetThrusterLevel(th_rcs2[3], 0);
	return;	
	}

	else if (DEG*(abs(ang_vel.y) + abs(ang_vel.x)) < 0.2)
	{
	SetThrusterLevel(th_rcs2[0], 0);
	SetThrusterLevel(th_rcs2[1], 0);
	SetThrusterLevel(th_rcs2[2], 0);
	SetThrusterLevel(th_rcs2[3], 0);
	return;	
	}

	double y_level_req = DEG*(y_angle + (2*ang_vel.y));
	if(y_level_req > 1) y_level_req = 1;
	else if(y_level_req < -1) y_level_req = -1;

	double x_level_req = DEG*(x_angle - (2*ang_vel.x));
	if(x_level_req > 1) x_level_req = 1;
	else if(x_level_req < -1) x_level_req = -1;

	if(x_level_req > 0)
	{
		SetThrusterLevel(th_rcs2[2], x_level_req);
		SetThrusterLevel(th_rcs2[0], 0);
	}
	else if(x_level_req < 0)
	{
		SetThrusterLevel(th_rcs2[2], 0);
		SetThrusterLevel(th_rcs2[0], -x_level_req);
	}

	if(y_level_req > 0)
	{
		SetThrusterLevel(th_rcs2[1], y_level_req);
		SetThrusterLevel(th_rcs2[3], 0);
	}
	else if(y_level_req < 0)
	{
		SetThrusterLevel(th_rcs2[1], 0);
		SetThrusterLevel(th_rcs2[3], -y_level_req);
	}

	//sprintf(oapiDebugString(), "x_level_req %0.3f y_level_req %0.3f x_angle %0.3f y_angle %0.3f ", x_level_req, y_level_req, x_angle*DEG, y_angle*DEG);

}

//void ss_booster::clbkVisualCreated (VISHANDLE vis, int refcount)
//{
//dhMeshx = GetDevMesh(vis, 0); 
//if(skin_sw == 2)ApplySkin();
//}

void ss_booster::RollToAZ()
{
OBJHANDLE htgt_vessel = oapiGetObjectByName(refname);
VESSEL *v = oapiGetVesselInterface (htgt_vessel);
VECTOR3 piv_pos_global;
v->Local2Global(_V(-0.024,0,-16.454),piv_pos_global);
VECTOR3 piv_pos_local;
Global2Local(piv_pos_global, piv_pos_local);
VECTOR3 piv_pos_hrz;
HorizonRot(piv_pos_local, piv_pos_hrz);

VECTOR3 piv_pos_hrz_lin = piv_pos_hrz;
piv_pos_hrz_lin.y = 0;
HorizonInvRot(piv_pos_hrz_lin,piv_pos_local);

double roll_angle = atan2(-piv_pos_local.x,piv_pos_local.y);
//if(piv_pos_local.y < 0)roll_angle = atan2(-piv_pos_local.x,piv_pos_local.y);
if(roll_angle < -PI)roll_angle = roll_angle + PI2;
else if(roll_angle > PI)roll_angle = roll_angle - PI2;

double th_max = GetThrusterMax(th_rcs[4]);
double max_torque = 2*4*th_max;
VECTOR3 pmi;
GetPMI(pmi);
double vmass = GetMass();
double inertia_z = pmi.z*vmass;
double max_acc = max_torque/inertia_z;
double max_dvrotz = max_acc*step_time;

double roll_lim = 0.9*roll_angle;
if(roll_lim > (10*RAD))roll_lim = (10*RAD);
else if(roll_lim < (-10*RAD))roll_lim = (-10*RAD);

if(-piv_pos_hrz.y < 200)roll_lim = 0; // stop
else if(-piv_pos_hrz.y < 150)
{
	SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT,0);
	SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT,0);
	return;
}
//if(abs(roll_angle) < (5*RAD))roll_lim = 0; // stop

VECTOR3 avel;
GetAngularVel(avel);

double avel_z_req = (-roll_lim) - avel.z;
double rcs_level = avel_z_req/max_dvrotz;
if(rcs_level > 1)rcs_level = 1;
else if(rcs_level < -1)rcs_level = -1;
//sprintf(oapiDebugString(), "avelz %0.1f roll_angle %0.1f rcs_level %0.1f", avel.z*DEG, roll_angle*DEG, rcs_level);

if(rcs_level > 0)
{
	SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT,rcs_level);
	SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT,0);
}
else if(rcs_level < 0)
{
	SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT,0);
	SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT,-rcs_level);
}
else
{
	SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT,0);
	SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT,0);
}





}