// ==============================================================
//
// ss_tanker
// Control module for ss_tanker vessel class
//
// ==============================================================

#define STRICT 1
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include <math.h>
#include "DlgCtrl.h"
#include "ss_tanker_resource.h"
#include "gcAPI.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

#define CTDN	1
#define SEP		2
#define SEPST	3
#define SEPFR	4


const DWORD COMM_SEND_STR = 2300;
const DWORD COMM_SEND_AFS = 2301;
const DWORD COMM_CALC_MASS = 2302;
const DWORD COMM_CALC_FRE = 2303;
const DWORD COMM_SEND_EDLO = 2304;
const DWORD COMM_SEND_BBT = 2405;

const double STG1ISP0 = 3200;// sl wiki
const double STG1ISPV = 3490;// vac wiki
const double STG1MAINTH0 = 2200000;// each raptor wiki
const double STG1MAINTH1 = 2200000;// each raptor wiki
const double STG1RCSTH = 191000;// 5deg gimbal x1 raptor
const double STG1DRYMASS = 180000;// wiki
const double STG1FUELMASS = 3400000;//spacex;

const double STG2ISP0 = 3100;// guess
const double STG2ISPV = 3680;// vac wiki
const double STG2MAINTH0 = 2200000;// each raptorvac wiki
const double STG2MAINTH1 = 2200000;// each raptorvac wiki
const double STG2RCSTH1 = 3600;// rcs guess
const double STG2RCSTH2 = 570000; // 5deg gimbal x3 raptor
const double STG2FUELMASS = 1175000;// wiki - reserve
const double STG2FUELRESERVE = 25000;// guess
const double STG2DRYMASS = 120000;//wiki;
//const double STG2RCSFUELMASS = 2000; // guess
const double STG2RCSISP = 3000; // guess

const double PLD_FUEL_MASS = 170000; // payload fuel max mass

const VECTOR3 STG1_OFF = _V(0,0,-25);
const VECTOR3 STG1STG2_OFF = _V(0,0,26.779);

//const VECTOR3 ADAPTOR_OFF = _V(0,0,26.646);

const double WING_SPEED = 0.25;// 0.25
const double ENG_SPEED = 1.0;
const double LEG_SPEED = 0.5;
//const double HATCH_SPEED = 0.1;// payload cover
// double PAF_SPEED = 0.1;// paf ring
//VECTOR3 paf_rot_axis = _V(0,1.026,14.045);

VECTOR3 port_canard_pos = _V(-7,0,27);
VECTOR3 stbd_canard_pos = _V( 7,0,27);
VECTOR3 port_wing_pos = _V(-7,0,-9);
VECTOR3 stbd_wing_pos = _V( 7,0,-9);
VECTOR3 hull_wing_pos = _V(0, 0.2165,-0.003);
VECTOR3 hull_wing_reentry_pos = _V(0, 0.2165,-0.003);
VECTOR3 wing_vertical_pos = _V(0, 2, -4);

const double hull_area = 300; //450
const double hull_chord = 45; //45
const double wing_area = 60; //54
const double canard_area = 20; //18
const double aspect_ratio = 3;
const double hull_aspect_ratio = 1;

char refname[256];
char tgtname[256];
//char pldname[256][20];
int afs_status = 0;
double tgt_inc = 0;
double min_inc = 0;
double fuel_reserve = 0;
double pld_mass = 0;
int dlg_status = 0;
int dlgl_status = 0;
//int pld_att_status = 0;
int calc_sw = 0;
//int intfair_status = 0;
int eng_status = 0;
int apop_status = 0;
int edlo_sw = 0;
//int nolegs = 0;
double ap_tgt = 135000;
const double mu = 398600439968871.2;
const double rt = 6371010;
double op_meco_alt = 180000;
double op_tgt = 200000;
double fuel_res_est = 0;
//ATTACHMENTHANDLE pld_attach[20];
ATTACHMENTHANDLE lpd_attach;
OBJHANDLE hMe;
OBJHANDLE hGref;
OBJHANDLE bchk;
PROPELLANT_HANDLE hProp_payload;
double asds_core_lat = 0;
double asds_core_lng = 0;
//double frv_core_lat = 0;
//double frv_core_lng = 0;
NOTEHANDLE hNote;
int lot_hh = 0;
int lot_mm = 0;
int lot_ss = 0;
int lot_status = 0;
double lot_timer = 0;
double acc_lim = 39.24;
OBJHANDLE hTransTarget = NULL;
int tank_index = 0;
double tank_max = 0;
double tank_level = 0;

//VECTOR3 bspos = _V(2.7, 2.7, 13.2)+STG1STG2_OFF;
//VECTOR3 bscol = _V(1,1,1);
//VECTOR3 new_beac_pos = _V(2.7, 2.7, 13.2);

static const DWORD ntdvtx_upper = 7;
static TOUCHDOWNVTX tdvtx_upper[ntdvtx_upper] = {
	{_V(0, -5.5,-16.78), 1e8, 4e6, 1, 1},
	{_V(-4.33, 2.75,-16.78), 1e8, 4e6, 1, 1},
	{_V( 4.33, 2.75,-16.78), 1e8, 4e6, 1, 1},
	{_V( 0, 5.5,-16.78), 1e8, 4e6, 1, 1},
	{_V(-4.33,-2.75,-16.78), 1e8, 4e6, 1, 1},
	{_V( 4.33,-2.75,-16.78), 1e8, 4e6, 1, 1},
	{_V(0 ,0 , 33.2), 1e8, 2e6, 1}
};

static const DWORD ntdvtx_booster = 7;
static TOUCHDOWNVTX tdvtx_booster[ntdvtx_booster] = {
	{_V(   0,-5.3, -60.8), 4e8, 2e7, 0.8, 0.8},
	{_V(-5.3, 5.3, -60.8), 4e8, 2e7, 0.8, 0.8},
	{_V( 5.3, 5.3, -60.8), 4e8, 2e7, 0.8, 0.8},
	{_V(   0, 5.3 ,-60.8), 4e8, 2e7, 0.8},
	{_V(-5.3,-5.3 ,-60.8), 4e8, 2e7, 0.8},
	{_V( 5.3,-5.3 ,-60.8), 4e8, 2e7, 0.8},
	{_V(   0,   0 ,   60), 4e8, 2e7, 1.5}
};

#define GREEN RGB(125, 255, 125)

////// energy/range profile
static const int ndp_5000 = 11;
static const int step_5000 = 200000;
static const double J_5000[ndp_5000] = 
{91445195.2,91558262.3,91674335.4,91792423.9,91909828.6,92018046.8,
92093043.6,92074016.7,91828586.2,91134116.0,89774288.3};

static const int ndp_3000 = 21;
static const int step_3000 = 100000;
static const double J_3000[ndp_3000] = 
{89774288.3,88863789.8,87883043.9,86933007.2,86075333.9,85328727.3,
84711234.9,84202589.9,83780526.7,83421931.9,83105701.8,82808055.0,
82518327.1,82212868.1,81868792.7,81452570.0,80918344.8,80209049.8,
79261468.7,78038578.0,76634238.8};

static const int ndp_1000 = 50;
static const int step_1000 = 20000;
static const double J_1000[ndp_1000] = 
{76634238.8,76344375.4,76052190.0,75760928.7,75469767.7,75179712.7,
74894227.8,74611963.5,74334106.3,74060913.4,73793485.4,73534658.4,
73283710.4,73041576.2,72807627.6,72580888.2,72361126.0,72147088.4,
71937535.7,71730259.7,71524934.3,71319320.1,71111908.9,70900973.3,
70684365.0,70461132.0,70227453.6,69982060.2,69722294.5,69445334.6,
69152212.9,68839870.0,68510072.7,68167092.0,67817254.4,67465992.8,
67117506.4,66774708.2,66438874.4,66112627.8,65799282.0,65499541.5,
65013431.3,64740112.4,64476683.6,64218715.3,63961155.1,63428145.1,
//71737535.7,71530259.7,71324934.3,71119320.1,70911908.9,70700973.3,
//70484365.0,70261132.0,70027453.6,69782060.2,69522294.5,69245334.6,
//68952212.9,68639870.0,68310072.7,67967092.0,67617254.4,67265992.8,
//66917506.4,66574708.2,66238874.4,65912627.8,65599282.0,65299541.5,
//64813431.3,64540112.4,64276683.6,64018715.3,63761155.1,63228145.1,
62899404.1,62589045.5};

typedef struct {
	HINSTANCE hDLL;
} GDIParams;
GDIParams g_Param;
//#define LOADBMP(id) (LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (id)))


//void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
//{
//	static const double step = RAD*22.5;
//	static const double istep = 1.0/step;
//	static const int nabsc = 17;
//	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0};
//
//
//	aoa += PI;
//	int idx = max (0, min (15, (int)(aoa*istep)));
//	double d = aoa*istep - idx;
//	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
//	*cm = 0.0;
//	*cd = 0.85*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
//	
//	
//}
//
//// 2. horizontal lift component (vertical stabiliser and body)
//
//void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
//{
//	static const double step = RAD*22.5;
//	static const double istep = 1.0/step;
//	static const int nabsc = 17;
//	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0};
//
//	beta += PI;
//	int idx = max (0, min (15, (int)(beta*istep)));
//	double d = beta*istep - idx;
//	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
//	*cm = 0.0;
//	*cd = 0.85*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
//	
//}

void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.2, 0.4, 0.2, 0,-0.2,-0.4,-0.2, 0, 0.2, 0.4, 0.2, 0,-0.2,-0.4,-0.2, 0};
	
	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0;// *cm = CM[idx] + (CM[idx+1]-CM[idx])*d;
	*cd = 0.5*(0.12 + oapiGetInducedDrag (*cl, 0.16, 0.2));
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.2, 0.4, 0.2, 0,-0.2,-0.4,-0.2, 0, 0.2, 0.4, 0.2, 0,-0.2,-0.4,-0.2, 0};
	
	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 0.5*(0.12 + oapiGetInducedDrag (*cl, 0.16, 0.2));
	
}

// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class ss_tanker: public VESSEL3 {
public:
	ss_tanker (OBJHANDLE hVessel, int flightmodel);
	~ss_tanker ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPostCreation();
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	int MyID;	// this is your unique Id, it will identify your ship in OrbiterSound.
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	bool clbkLoadGenericCockpit ();
	void clbkVisualCreated (VISHANDLE vis, int refcount);
	void clbkVisualDestroyed (VISHANDLE vis, int refcount);
	void DefineAnimations (void);
	void clbkDockEvent (int dock, OBJHANDLE mate);
private:

	//char adname [256];

	THRUSTER_HANDLE th_main[3],th_rcs[37], th_group[9], th_dump, th_dust;
	THGROUP_HANDLE htgrpmain;
	PROPELLANT_HANDLE hProp1,hProp2, hProp_reserve;//, hProp_payload; hRCSProp2,
	THGROUP_HANDLE htgrppup2,htgrppdown2,htgrpbright2, htgrpbleft2, htgrpyright2, htgrpyleft2;
	MESHHANDLE hMStg2;
	MESHHANDLE hMhull;

	ANIMATIONCOMPONENT_HANDLE Parent, Section, Section2, Section3, Section4, Section5, Section6, Section7;
	AIRFOILHANDLE hHA, hVA, hPortCanard, hStbdCanard, hPortWing, hStbdWing;
	AIRFOILHANDLE hWing_Vertical;
	AIRFOILHANDLE hVABooster, hHABooster;
	UINT anim_port_can,anim_stbd_can, anim_port_wing, anim_stbd_wing, anim_legs;//, anim_hatch, anim_paf;
	double port_can_proc, stbd_can_proc, port_wing_proc, stbd_wing_proc, legs_proc;
//	double hatch_proc, paf_proc;
	int  legs_status;//, hatch_status, paf_status;
	double wing_trim;
	MGROUP_TRANSFORM *PortCan_trans,*StbdCan_trans, *PortWing_trans, *StbdWing_trans;
	MGROUP_TRANSFORM *Leg1_rot, *Leg1_trans, *Leg2_rot, *Leg2_trans, *Leg3_rot, *Leg3_trans;
	MGROUP_TRANSFORM *Leg4_rot, *Leg4_trans, *Leg5_rot, *Leg5_trans, *Leg6_rot, *Leg6_trans;
//	MGROUP_TRANSFORM *Hatch_trans, *Paf_trans;

	ELEMENTS el;
	ORBITPARAM prm;

	//OBJHANDLE hpd;
	OBJHANDLE bbchk;
	PointLight *spt_light;

	EXHAUSTSPEC upper_exhaust[3];
	EXHAUSTSPEC main2_exhaust[33];

	//SURFHANDLE bs1, bs2;

	//void JettisonPld(void);
	//void JettisonFair(void);
	void JettisonStg(void);
	void ConfigMain(void);
	void ConfigRCS2(void);
	void RevertRCS2(void);
	void CalcEmptyMass(void);
	void DoAscent(void);
	void DoOrbit(void);
	double CalcAzimuth();
	void CalcFlightVector();
	void GoToFlightVector();
	double FinalV();
	double GetOS();
	void RollToAz();
	void RollToAzLanding();
	void SetCamDir();
	void KillAllRCS2();
	//void InitDialog(HWND hWnd);
	void CalcFuelRes();
	void DoLOT();
	void CalcAsdsPos();
	void DoTDPoints();
	void SetADCLevel();
	void ConfigLateralRCS();
	void GetWindVector(void);
	void DoSkydiveTargeting(void);
	void DoBackFlip(void);
	void DoRetroLand(void);
	void GoToFlightVectorSD(void);
	void GoToFlightVectorBF(void);
	void ConfigUpperMain(void);
	void ConfigUpperRaptorSL(void);
	void SetPropReserve(void);
	void DoFuelTransfer(void);
	void DoReentryNT(void);
	void GoToFlightVectorRE(void);
	void DoReentry(void);
	void DoEnergyManagement(void);
	void DoFBWReentry(void);

	double red_fact;
	//double red_fact2;
	DEVMESHHANDLE dhMesh0;
	void SetMeshMat(void);
	void SetUpperExhaust(void);

	double GetAoAbyWT(double wingtrim);
	double GetWTbyAoA(double current_aoa);
	

	double step_time, old_ap_dist, pitch_angle, bias, timer, vrot_lim, pst_lvl, pst_lvl2;
	int op_status, ap_status, rcs_sw2;
	int ap_function, cam_status;// drg, nfr;
	int start_sw, tu_sw;// skin_sw;
	int disable_rcs;
	int skydive_sw, backflip_sw, retroland_sw;
	int roll_ofs_init;
	double hud_tgt_az;
	double hud_gspd_az;
	double hud_tgt_range;
	double hud_ves_range;
	double hud_crossrange_error;
	double hud_hacc_req, hud_hacc_current, hud_hacc_error; 
	int uppermaincfg;
	int uppermainprop;
	//int disable_fair_rec;

	//int he_vent_status;
	//double he_vent_level, he_vent_timer;

	double azimuth;	
	VECTOR3 azim_local, vazim1, vazim2, flight_vector;
	VECTOR3 headsup_vector;
	VECTOR3 wind_vector_tgt;
	double windspeed_tgt;

	double atl_current;

//	SpotLight *spotlight1;
//	BEACONLIGHTSPEC beac;
//	int lamp_status;
	
//	VECTOR3 pld_attach_pos_init[20];
//	VECTOR3 pld_attach_dir_init[20];
//	VECTOR3 pld_attach_rot_init[20];

	double pitch_level;
	double bank_level;
	double yaw_level;
	double js_pu, js_pd, js_bl, js_br, js_yl, js_yr; // joystick inputs
	int adcs_reentry_mode, hud_display_info, adcm_scn;
	VECTOR3 upperexhaustdir;
	int first_frame_sw;
	int de_first_frame_sw;
	int raptor_sl_selection;
	UINT exhaust_id[3];
	VECTOR3 exhaust_2pos[3];
	double dust_level;
	DOCKHANDLE hDock_rear;
	int fuel_trans_sw, fuel_source_sw;

	int reentry_sw;
	int bank_dir_sw, bank_mode_sw;

	double intv, range_0, tgt_dist;
	int data_sw;
	double post_step, tot_energy_old, dst_past_tld_old;
	double bank_correction, digi_dar_old, digi_dar;
	int digi_dar_check_sw, ttb_check_sw;
	int reentry_vspd_sw;

	double old_vSpeed, old_hSpeed, vSpeed_adjust;
	VECTOR3 catch_offset;
	bool tgt_att_valid;
	double az_angle;
	bool tgt_exists;

	int fbw_reentry_sw;
	double fbw_bank, fbw_aoa;

	ATTACHMENTHANDLE arm_attach;

};

ss_tanker::ss_tanker (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

ss_tanker::~ss_tanker ()
{
	delete PortCan_trans;
	delete StbdCan_trans;
	delete PortWing_trans;
	delete StbdWing_trans;
	delete Leg1_rot;
	delete Leg1_trans;
	delete Leg2_rot;
	delete Leg2_trans;
	delete Leg3_rot;
	delete Leg3_trans;
	delete Leg4_rot;
	delete Leg4_trans;
	delete Leg5_rot;
	delete Leg5_trans;
	delete Leg6_rot;
	delete Leg6_trans;
//	delete Hatch_trans;
//	delete Paf_trans;
}

void ss_tanker::DefineAnimations (void)
{
	// canards
	anim_port_can = CreateAnimation (0);
	anim_stbd_can = CreateAnimation (0);

	static UINT CanGrp1[2] = {2,3};
	PortCan_trans = new MGROUP_ROTATE(0, CanGrp1, 2, _V(-5.103,0,23.678), _V(0.30902,0,0.95106), (float)(-60*RAD));
	AddAnimationComponent (anim_port_can, 0, 1, PortCan_trans);

	static UINT CanGrp2[2] = {4,5};
	StbdCan_trans = new MGROUP_ROTATE(0, CanGrp2, 2, _V( 5.103,0,23.678), _V(-0.30902,0,0.95106), (float)(60*RAD));
	AddAnimationComponent (anim_stbd_can, 0, 1, StbdCan_trans);

	// wings
	anim_port_wing = CreateAnimation (0);
	anim_stbd_wing = CreateAnimation (0);

	static UINT WngGrp1[2] = {44,45};
	PortWing_trans = new MGROUP_ROTATE(0, WngGrp1, 2, _V(-5.821,0,1.031), _V(0,0,1), (float)(-70*RAD));
	AddAnimationComponent (anim_port_wing, 0, 1, PortWing_trans);

	static UINT WngGrp2[2] = {46,47};
	StbdWing_trans = new MGROUP_ROTATE(0, WngGrp2, 2, _V( 5.821,0,1.031), _V(0,0,1), (float)(70*RAD));
	AddAnimationComponent (anim_stbd_wing, 0, 1, StbdWing_trans);

	//Legs

	///////////////////////// leg 1//////////////////////////////////////////
	VECTOR3 rot_point = _V(3.972,0,-16.427);
	VECTOR3 rot_axis = _V(0,-1,0);
	VECTOR3 trans_ofs = _V(0,0,0.9);
	

	anim_legs = CreateAnimation (0);

	static UINT LegGrp1[2] = {25,26};
	Leg1_rot = new MGROUP_ROTATE(0, LegGrp1, 2, rot_point, rot_axis, (float)(195*RAD));
	Parent = AddAnimationComponent (anim_legs, 0, 0.7, Leg1_rot);

	static UINT LegGrp7[1] = {27};
	Leg1_trans = new MGROUP_TRANSLATE(0, LegGrp7, 1, trans_ofs);
	AddAnimationComponent (anim_legs, 0.7, 1.0, Leg1_trans, Parent);

	/////////////////// leg 2 ///////////////////////////

	double rot_angle = -60*RAD;
	MATRIX3 rm = rotm(_V(0,0,1), rot_angle);
	rot_point = mul(rm, rot_point);
	rot_axis = mul(rm, rot_axis);

	static UINT LegGrp2[2] = {28,29};
	Leg2_rot = new MGROUP_ROTATE(0, LegGrp2, 2, rot_point, rot_axis, (float)(195*RAD));
	Parent = AddAnimationComponent (anim_legs, 0, 0.7, Leg2_rot);

	static UINT LegGrp8[1] = {30};
	Leg2_trans = new MGROUP_TRANSLATE(0, LegGrp8, 1, trans_ofs);
	AddAnimationComponent (anim_legs, 0.7, 1.0, Leg2_trans, Parent);

	//////////////////// leg 3 //////////////////////////////

	//rot_angle = -120*RAD;
	rm = rotm(_V(0,0,1), rot_angle);
	rot_point = mul(rm, rot_point);
	rot_axis = mul(rm, rot_axis);

	static UINT LegGrp3[2] = {31,32};
	Leg3_rot = new MGROUP_ROTATE(0, LegGrp3, 2, rot_point, rot_axis, (float)(195*RAD));
	Parent = AddAnimationComponent (anim_legs, 0, 0.7, Leg3_rot);

	static UINT LegGrp9[1] = {33};
	Leg3_trans = new MGROUP_TRANSLATE(0, LegGrp9, 1, trans_ofs);
	AddAnimationComponent (anim_legs, 0.7, 1.0, Leg3_trans, Parent);

	/////////////////// leg 4 ////////////////////////////////

	//rot_angle = -180*RAD;
	rm = rotm(_V(0,0,1), rot_angle);
	rot_point = mul(rm, rot_point);
	rot_axis = mul(rm, rot_axis);

	static UINT LegGrp4[2] = {34,35};
	Leg4_rot = new MGROUP_ROTATE(0, LegGrp4, 2, rot_point, rot_axis, (float)(195*RAD));
	Parent = AddAnimationComponent (anim_legs, 0, 0.7, Leg4_rot);

	static UINT LegGrp10[1] = {36};
	Leg4_trans = new MGROUP_TRANSLATE(0, LegGrp10, 1, trans_ofs);
	AddAnimationComponent (anim_legs, 0.7, 1.0, Leg4_trans, Parent);

	/////////////////////// leg 5 //////////////////////////////////

	//rot_angle = -240*RAD;
	rm = rotm(_V(0,0,1), rot_angle);
	rot_point = mul(rm, rot_point);
	rot_axis = mul(rm, rot_axis);

	static UINT LegGrp5[2] = {19,20};
	Leg5_rot = new MGROUP_ROTATE(0, LegGrp5, 2, rot_point, rot_axis, (float)(195*RAD));
	Parent = AddAnimationComponent (anim_legs, 0, 0.7, Leg5_rot);

	static UINT LegGrp11[1] = {21};
	Leg5_trans = new MGROUP_TRANSLATE(0, LegGrp11, 1, trans_ofs);
	AddAnimationComponent (anim_legs, 0.7, 1.0, Leg5_trans, Parent);

	///////////////////////// leg 6 //////////////////////////////////

	//rot_angle = -300*RAD;
	rm = rotm(_V(0,0,1), rot_angle);
	rot_point = mul(rm, rot_point);
	rot_axis = mul(rm, rot_axis);

	static UINT LegGrp6[2] = {22,23};
	Leg6_rot = new MGROUP_ROTATE(0, LegGrp6, 2, rot_point, rot_axis, (float)(195*RAD));
	Parent = AddAnimationComponent (anim_legs, 0, 0.7, Leg6_rot);

	static UINT LegGrp12[1] = {24};
	Leg6_trans = new MGROUP_TRANSLATE(0, LegGrp12, 1, trans_ofs);
	AddAnimationComponent (anim_legs, 0.7, 1.0, Leg6_trans, Parent);

	// payload hatch //////////////////////
	//anim_hatch = CreateAnimation (0);

	//static UINT HatchGrp1[2] = {22,59};
	//Hatch_trans = new MGROUP_ROTATE(0, HatchGrp1, 2, _V(0,4.483,13.121), _V(-1,0,0), (float)(60*RAD));
	//AddAnimationComponent (anim_hatch, 0, 1, Hatch_trans);

	/////// paf ring
	//anim_paf = CreateAnimation (0);

	//static UINT PafGrp1[1] = {1};
	//Paf_trans = new MGROUP_ROTATE(0, PafGrp1, 1, paf_rot_axis, _V(-1,0,0), (float)(30*RAD));
	//AddAnimationComponent (anim_paf, 0, 1, Paf_trans);

}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void ss_tanker::clbkSetClassCaps (FILEHANDLE cfg)
{
	strcpy(refname, "NULL");
	strcpy(tgtname, "NULL");

	/*for (int i = 0; i < 20; i++)
	{
	strcpy(pldname[i], "NULL");
	pld_attach_pos_init[i] = _V(0,0,0);
	pld_attach_dir_init[i] = _V(0,0,1);
	pld_attach_rot_init[i] = _V(0,1,0);
	}*/
	step_time = 0;
	ap_status = 0;
	op_status = 0;
	cam_status = 0;
	ap_function = 0;
	old_ap_dist = 0;
	bias = 0;
	pitch_angle = 90*RAD;
	//intfair_status = 0;
	//drg = 0;
	//nfr = 0;
	//eng_status = 0;
	//calc_sw = 0;
	rcs_sw2 = 0;
	vrot_lim = 5*RAD;
	azimuth = 0;
	azim_local = _V(0,-1,0);
	vazim1 = _V(0,1,0);
	vazim2 = _V(0,1,0);
	flight_vector = _V(0,0,1);
	headsup_vector = _V(0,1,0);
	wind_vector_tgt = _V(0,0,0);
	windspeed_tgt = 0;

	timer = 0;
	start_sw = 0;
	pst_lvl = 0;
	pst_lvl2 = 0;
	tu_sw = 0;
	atl_current = 1;
	//skin_sw = 0;
	//disable_fair_rec = 0;

	port_can_proc = 0;
	stbd_can_proc = 0;
	port_wing_proc = 0;
	stbd_wing_proc = 0;
	legs_proc = 0;
	legs_status = 0;
	wing_trim = 0;

	//hatch_proc = 0;
	//hatch_status = 0;
	//paf_proc = 0;
	//paf_status = 0;

	red_fact = 0;
	//red_fact2 = 0;
	//lamp_status = 0;

	js_pu = 0;
	js_pd = 0;
	js_bl = 0;
	js_br = 0;
	js_yl = 0;
	js_yr = 0;

	pitch_level = 0;
	bank_level = 0;
	yaw_level = 0;

	adcs_reentry_mode = 0;
	hud_display_info = 0;
	adcm_scn = 0;

	disable_rcs = 0;

	skydive_sw = 0;
	backflip_sw = 0;
	retroland_sw = 0;
	roll_ofs_init = 0;

	hud_tgt_az = 0;
	hud_gspd_az = 0;
	hud_tgt_range = 0;
	hud_ves_range = 0;
	hud_hacc_req = 0;
	hud_hacc_current = 0;
	hud_hacc_error = 0; 

	uppermaincfg = 0;
	uppermainprop = 0;

	upperexhaustdir = _V(0,0,1);
	exhaust_2pos[0] = _V(0,0,0);

	first_frame_sw = 0;
	de_first_frame_sw = 0;

	raptor_sl_selection = 3;

	dust_level = 0;

	fuel_trans_sw = 0;
	fuel_source_sw = 0;

	reentry_sw = 0;
	bank_dir_sw = 1;
	bank_mode_sw = 0;

	data_sw = 0;
	intv = 200000;
	range_0 = 5000000;
	tgt_dist = 0;
	post_step = 0;
	tot_energy_old = 0;
	dst_past_tld_old = 0;
	bank_correction = 0;
	digi_dar_old = 0;
	digi_dar = 0;
	digi_dar_check_sw = 0;

	reentry_vspd_sw = 0;

	old_vSpeed = 0;
	old_hSpeed = 0;
	vSpeed_adjust = 0;

	catch_offset = _V(0,0,0);
	tgt_att_valid = true;
	tgt_exists = false;
	az_angle = 0;

	fbw_reentry_sw = 0;
	fbw_bank = 0;
	fbw_aoa = 70;

	gcInitialize();

	DefineAnimations();
	// physical vessel parameters
	SetSize (60);
	SetEmptyMass (STG1DRYMASS+STG2DRYMASS);
	SetPMI(_V(1000,1000,10));
	SetCrossSections (_V(900,900,60));
	SetCW(0.5,1,1,1);
	SetRotDrag (_V(0.5,0.5,0.1));
	SetTouchdownPoints ( tdvtx_booster, ntdvtx_booster);

	// camera parameters
	SetCameraOffset (_V(0,5,0));
	SetCameraRotationRange(PI,PI,(0.4*PI),(0.4*PI));
	
	hMStg2 = oapiLoadMeshGlobal("spacex_starship/ss_tanker");
	SetMeshVisibilityMode(AddMesh(hMStg2, &STG1STG2_OFF),MESHVIS_ALWAYS);
	
	hMhull = oapiLoadMeshGlobal("spacex_starship/ss_booster");
	SetMeshVisibilityMode(AddMesh(hMhull, &STG1_OFF),MESHVIS_ALWAYS);


	/////// charred skin
	/*if(oapiReadItem_int (cfg, "Charred", skin_sw) == true)
	{
	if(skin_sw == 1)
		{
		bs2 = oapiLoadTexture ("Falcon9R_V12/f9_hull_logo_char.dds", true);
		bs1 = oapiGetTextureHandle (hMhull, 1);
		oapiBlt(bs1, bs2, 0, 0, 0, 0, 256, 1024);
		oapiReleaseTexture (bs2);
		}
	}*/

	// propellant resources
	
	
	hProp2 = CreatePropellantResource (STG2FUELMASS,STG2FUELMASS);
	//hRCSProp2 = CreatePropellantResource (STG2RCSFUELMASS,STG2RCSFUELMASS);
	hProp_reserve = CreatePropellantResource (STG2FUELRESERVE,STG2FUELRESERVE);
	hProp_payload = CreatePropellantResource(PLD_FUEL_MASS, 0);
	hProp1 = CreatePropellantResource (STG1FUELMASS,STG1FUELMASS);
	SetDefaultPropellantResource(hProp1);

	// ***************** thruster definitions *******************
	
	//booster rcs
	th_rcs[0] = CreateThruster (_V( 0, 0, -60), _V( 1,0,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // yl
	th_rcs[1] = CreateThruster (_V( 0, 0, -60), _V(-1,0,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // yr
	th_rcs[2] = CreateThruster (_V( 0, 0, -60), _V(0,-1,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // pup
	th_rcs[3] = CreateThruster (_V( 0, 0, -60), _V(0, 1,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // pdn
	th_rcs[4] = CreateThruster (_V( 0, 4, -60), _V(-1,0,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // bl
	th_rcs[5] = CreateThruster (_V( 0,-4, -60), _V( 1,0,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // bl
	th_rcs[6] = CreateThruster (_V( 0, 4, -60), _V( 1,0,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // br
	th_rcs[7] = CreateThruster (_V( 0,-4, -60), _V(-1,0,0), 4*STG1RCSTH, hProp1, STG1ISPV, STG1ISP0); // br



	////// stg2 main

	th_rcs[8] = CreateThruster (_V( 0, 0, -16), _V( 1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); //yl
	th_rcs[9] = CreateThruster (_V( 0, 0, -16), _V(-1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); //  yr
	th_rcs[10] = CreateThruster (_V( 0, 0, -16), _V( 0,-1,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // pup
	th_rcs[11] = CreateThruster (_V( 0, 0, -16), _V( 0,1,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // pd
	th_rcs[12] = CreateThruster (_V( 0, 1, -16), _V(-1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // bl
	th_rcs[13] = CreateThruster (_V( 0,-1, -16), _V( 1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // bl
	th_rcs[14] = CreateThruster (_V( 0,-1, -16), _V(-1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // br
	th_rcs[15] = CreateThruster (_V( 0, 1, -16), _V( 1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // br

	// stg2 rcs

	th_rcs[16] = CreateThruster (_V(0,0,-16), _V( 1,0,0), 2*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // yl
	th_rcs[17] = CreateThruster (_V(0,0, 32), _V(-1,0,0), STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP);// yl
	th_rcs[18] = CreateThruster (_V(0,0,-16), _V(-1,0,0), 2*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // yr
	th_rcs[19] = CreateThruster (_V(0,0, 32), _V( 1,0,0), STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP);// yr
	th_rcs[20] = CreateThruster (_V(0,0,-16), _V(0,-1,0), 2*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // pup
	th_rcs[21] = CreateThruster (_V(0,0, 32), _V(-0.7071, 0.7071,0), STG2RCSTH1*0.7071, hProp2, STG2RCSISP, STG2RCSISP);// pup
	th_rcs[22] = CreateThruster (_V(0,0, 32), _V(0.7071, 0.7071,0), STG2RCSTH1*0.7071, hProp2, STG2RCSISP, STG2RCSISP);// pup
	th_rcs[23] = CreateThruster (_V(0,0, 32), _V(0,-1,0), STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // pdn
	th_rcs[24] = CreateThruster (_V(0,0,-16), _V(0.7071,0.7071,0), 2*STG2RCSTH1*0.7071, hProp2, STG2RCSISP, STG2RCSISP);// pdn
	th_rcs[25] = CreateThruster (_V(0,0,-16), _V(-0.7071,0.7071,0), 2*STG2RCSTH1*0.7071, hProp2, STG2RCSISP, STG2RCSISP);// pdn
	th_rcs[26] = CreateThruster (_V(5,0,0), _V(0, 1,0), 2*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // bl
	th_rcs[27] = CreateThruster (_V(-5,0,0), _V(0,-1,0), 2*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP);// bl
	th_rcs[28] = CreateThruster (_V(5,0,0), _V(0,-1,0), 2*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // br
	th_rcs[29] = CreateThruster (_V(-5,0,0), _V(0, 1,0), 2*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP);// br

	////// upper translation
	th_rcs[30] = CreateThruster (_V(0,0,0), _V(0,0, 1), 4*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // fwd
	th_rcs[31] = CreateThruster (_V(0,0,0), _V(0,0,-1), 4*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP);// bck
	th_rcs[32] = CreateThruster (_V(0,0,0), _V(0.7071, 0.7071,0), 4*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // up
	th_rcs[33] = CreateThruster (_V(0,0,0), _V(-0.7071, 0.7071,0), 4*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // up
	th_rcs[34] = CreateThruster (_V(0,0,0), _V(0,-1,0), 6*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP);// dn
	th_rcs[35] = CreateThruster (_V(0,0,0), _V(-1,0,0), 4*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP); // left
	th_rcs[36] = CreateThruster (_V(0,0,0), _V( 1,0,0), 4*STG2RCSTH1, hProp2, STG2RCSISP, STG2RCSISP);// right


	//// booster rot
	th_group[0] = th_rcs[2];
	htgrppup2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[3];
	htgrppdown2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[1];
	htgrpyright2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[0];
	htgrpyleft2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[6];
	th_group[1] = th_rcs[7];
	htgrpbright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[4];
	th_group[1] = th_rcs[5];
	htgrpbleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);


	// mains
	th_main[0] = CreateThruster (_V( 0, 0, -60), _V( 0,0,1), 33*STG1MAINTH0, hProp1, STG1ISPV, STG1ISP0);
	th_main[1] = CreateThruster (_V( 0, 0, -16), _V( 0,0,1), 3*STG2MAINTH1, hProp2, STG1ISPV, STG1ISP0);
	th_main[2] = CreateThruster (_V( 0, 0, -16), _V( 0,0,1), 3*STG2MAINTH1, hProp2, STG2ISPV, STG2ISP0);
	th_dump = CreateThruster (_V( 0, 0, -16), _V( 0,0,1), 0.1, hProp2, 0.0001, 0.0001);

	th_group[0] = th_main[0];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("spacex_starship/KeroExhaust3");
	SURFHANDLE exhaustupper = oapiRegisterExhaustTexture ("Exhaust2");
	SURFHANDLE exhaustrcs = oapiRegisterExhaustTexture ("spacex_starship/Exhaust_ss");


	double exlen = 50;
	double exwid = 3.7;

	///// booster

	MATRIX3 rm;
	VECTOR3 exhaust_pos2[33];

	///// outer 20 ring ///// /////
	for (int i = 13; i < 33; i++)
	{
	rm = rotm(_V(0,0,1),(18*RAD*i));
	exhaust_pos2[i] = mul(rm,_V(0,4.13,-59.143));

	main2_exhaust[i].th = th_main[0];
	main2_exhaust[i].level = NULL;
	main2_exhaust[i].lpos = &exhaust_pos2[i];
	main2_exhaust[i].ldir = NULL; //&main2_exhaust_dir;
	main2_exhaust[i].lsize = exlen*1.5; // 1
	main2_exhaust[i].wsize = exwid;
	main2_exhaust[i].lofs = 0;
	main2_exhaust[i].modulate = 0.9;
	main2_exhaust[i].tex = exhaust;
	main2_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main2_exhaust[i].id = 0;
	AddExhaust (&main2_exhaust[i]);
	}

	// inner 10 ring ///////////////
	for (int i = 3; i < 13; i++)
	{
	rm = rotm(_V(0,0,1),(36*RAD*i));
	//exhaust_pos2[i] = mul(rm,_V(0,4.13,-59.143));
	exhaust_pos2[i] = mul(rm,_V(-0.352,-2.233,-59.143));

	main2_exhaust[i].th = th_main[0];
	main2_exhaust[i].level = NULL;
	main2_exhaust[i].lpos = &exhaust_pos2[i];
	main2_exhaust[i].ldir = NULL; //&main2_exhaust_dir;
	main2_exhaust[i].lsize = exlen*1.875; // 1.25
	main2_exhaust[i].wsize = exwid;
	main2_exhaust[i].lofs = 0;
	main2_exhaust[i].modulate = 0.8;
	main2_exhaust[i].tex = exhaust;
	main2_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main2_exhaust[i].id = 0;
	AddExhaust (&main2_exhaust[i]);
	}

	//////////// centre 3 ////////////////////////
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_pos2[i] = mul(rm,_V(0,-0.833,-60));

	main2_exhaust[i].th = th_main[0];
	main2_exhaust[i].level = NULL;
	main2_exhaust[i].lpos = &exhaust_pos2[i];
	main2_exhaust[i].ldir = NULL; //&main2_exhaust_dir;
	main2_exhaust[i].lsize = 2.25*exlen; //1.5
	main2_exhaust[i].wsize = exwid;
	main2_exhaust[i].lofs = 0;
	main2_exhaust[i].modulate = 0.9;
	main2_exhaust[i].tex = exhaust;
	main2_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	main2_exhaust[i].id = 0;
	AddExhaust (&main2_exhaust[i]);
	}

	/// upper main /////////////////////////////////////

	VECTOR3 exhaust_pos;
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_pos = mul(rm,_V(0,-3.141,-17.5));
	AddExhaust(th_main[2],exlen*0.75,0.5*exwid,exhaust_pos,_V(0,0,-1),exhaustupper);
	}

	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_2pos[i] = mul(rm,_V(0,0.981,-16.5));
	upper_exhaust[i].th = th_main[1];
	upper_exhaust[i].level = NULL;
	upper_exhaust[i].lpos = &exhaust_2pos[i];
	upper_exhaust[i].ldir = &upperexhaustdir;
	upper_exhaust[i].lsize = exlen*0.75;
	upper_exhaust[i].wsize = exwid;
	upper_exhaust[i].lofs = 0;
	upper_exhaust[i].modulate = 0.3;
	upper_exhaust[i].tex = exhaust;
	upper_exhaust[i].flags = EXHAUST_CONSTANTPOS;
	upper_exhaust[i].id = 0;
	}


	exhaust_id[0] = AddExhaust (&upper_exhaust[0]);
	exhaust_id[1] = AddExhaust (&upper_exhaust[1]);
	exhaust_id[2] = AddExhaust (&upper_exhaust[2]);

	//////////// upper rcs /////////////////////////

	double exlenrcs = 3;
	double exwidrcs = 0.3;

	AddExhaust(th_rcs[16],exlenrcs,exwidrcs,_V(-4.437,0.83,-16.078),_V(-1,0,0),exhaustrcs);
	AddExhaust(th_rcs[16],exlenrcs,exwidrcs,_V(-4.437,0.83,-16.318),_V(-1,0,0),exhaustrcs);
	AddExhaust(th_rcs[17],exlenrcs,exwidrcs,_V(0.894,0.147,32.819),_V(1,0,0),exhaustrcs);
	AddExhaust(th_rcs[17],exlenrcs,exwidrcs,_V(1.102,0.147,32.556),_V(1,0,0),exhaustrcs);

	AddExhaust(th_rcs[18],exlenrcs,exwidrcs,_V(4.437,0.83,-16.078),_V(1,0,0),exhaustrcs);
	AddExhaust(th_rcs[18],exlenrcs,exwidrcs,_V(4.437,0.83,-16.318),_V(1,0,0),exhaustrcs);
	AddExhaust(th_rcs[19],exlenrcs,exwidrcs,_V(-0.894,0.147,32.819),_V(-1,0,0),exhaustrcs);
	AddExhaust(th_rcs[19],exlenrcs,exwidrcs,_V(-1.102,0.147,32.556),_V(-1,0,0),exhaustrcs);

	AddExhaust(th_rcs[20],exlenrcs,exwidrcs,_V(4.683,0.523,-15.554),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[20],exlenrcs,exwidrcs,_V(-4.683,0.523,-15.554),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[20],exlenrcs,exwidrcs,_V(4.683,0.523,-15.814),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[20],exlenrcs,exwidrcs,_V(-4.683,0.523,-15.814),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[21],exlenrcs,exwidrcs,_V(1.373,0.142,32.156),_V(0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[21],exlenrcs,exwidrcs,_V(-1.373,0.142,32.156),_V(-0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[22],exlenrcs,exwidrcs,_V(1.536,0.142,31.89),_V(0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[22],exlenrcs,exwidrcs,_V(-1.536,0.142,31.89),_V(-0.7071,-0.7071,0),exhaustrcs);

	AddExhaust(th_rcs[23],exlenrcs,exwidrcs,_V(0.158,1.217,32.392),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[23],exlenrcs,exwidrcs,_V(-0.158,1.217,32.392),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[24],exlenrcs,exwidrcs,_V(4.344,1.192,-16.625),_V(0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[24],exlenrcs,exwidrcs,_V(-4.344,1.192,-16.625),_V(-0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[25],exlenrcs,exwidrcs,_V(4.475,0.666,-16.625),_V(0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[25],exlenrcs,exwidrcs,_V(-4.475,0.666,-16.625),_V(-0.7071,-0.7071,0),exhaustrcs);

	AddExhaust(th_rcs[27],exlenrcs,exwidrcs,_V(-4.683,0.523,-15.554),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[27],exlenrcs,exwidrcs,_V(-4.683,0.523,-15.814),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[27],exlenrcs,exwidrcs,_V(-0.158,1.217,32.292),_V(0,1,0),exhaustrcs);

	AddExhaust(th_rcs[28],exlenrcs,exwidrcs,_V(4.683,0.523,-15.554),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[28],exlenrcs,exwidrcs,_V(4.683,0.523,-15.814),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[28],exlenrcs,exwidrcs,_V(0.158,1.217,32.292),_V(0,1,0),exhaustrcs);


	AddExhaust(th_rcs[30],exlenrcs,exwidrcs,_V(4.607,0.151,-16.465),_V(0,0,-1),exhaustrcs);
	AddExhaust(th_rcs[30],exlenrcs,exwidrcs,_V(4.607,-0.151,-16.465),_V(0,0,-1),exhaustrcs);
	AddExhaust(th_rcs[30],exlenrcs,exwidrcs,_V(-4.607,0.151,-16.465),_V(0,0,-1),exhaustrcs);
	AddExhaust(th_rcs[30],exlenrcs,exwidrcs,_V(-4.607,-0.151,-16.465),_V(0,0,-1),exhaustrcs);
	AddExhaust(th_rcs[31],exlenrcs,exwidrcs,_V(0.131,0.623,33.174),_V(0,0,1),exhaustrcs);
	AddExhaust(th_rcs[31],exlenrcs,exwidrcs,_V(-0.131,0.623,33.174),_V(0,0,1),exhaustrcs);
	AddExhaust(th_rcs[31],exlenrcs,exwidrcs,_V(0.131,0.907,32.811),_V(0,0,1),exhaustrcs);
	AddExhaust(th_rcs[31],exlenrcs,exwidrcs,_V(-0.131,0.907,32.811),_V(0,0,1),exhaustrcs);

	AddExhaust(th_rcs[32],exlenrcs,exwidrcs,_V(4.344,1.192,-16.625),_V(0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[32],exlenrcs,exwidrcs,_V(4.475,0.666,-16.625),_V(0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[32],exlenrcs,exwidrcs,_V(1.373,0.142,32.156),_V(0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[32],exlenrcs,exwidrcs,_V(1.536,0.142,31.89),_V(0.7071,-0.7071,0),exhaustrcs);

	AddExhaust(th_rcs[33],exlenrcs,exwidrcs,_V(-4.344,1.192,-16.625),_V(-0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[33],exlenrcs,exwidrcs,_V(-4.475,0.666,-16.625),_V(-0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[33],exlenrcs,exwidrcs,_V(-1.373,0.142,32.156),_V(-0.7071,-0.7071,0),exhaustrcs);
	AddExhaust(th_rcs[33],exlenrcs,exwidrcs,_V(-1.536,0.142,31.89),_V(-0.7071,-0.7071,0),exhaustrcs);

	AddExhaust(th_rcs[34],exlenrcs,exwidrcs,_V(0.158,1.217,32.392),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[34],exlenrcs,exwidrcs,_V(-0.158,1.217,32.392),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[34],exlenrcs,exwidrcs,_V(4.683,0.523,-15.554),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[34],exlenrcs,exwidrcs,_V(-4.683,0.523,-15.554),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[34],exlenrcs,exwidrcs,_V(4.683,0.523,-15.814),_V(0,1,0),exhaustrcs);
	AddExhaust(th_rcs[34],exlenrcs,exwidrcs,_V(-4.683,0.523,-15.814),_V(0,1,0),exhaustrcs);

	AddExhaust(th_rcs[35],exlenrcs,exwidrcs,_V(4.437,0.83,-16.078),_V(1,0,0),exhaustrcs);
	AddExhaust(th_rcs[35],exlenrcs,exwidrcs,_V(4.437,0.83,-16.318),_V(1,0,0),exhaustrcs);
	AddExhaust(th_rcs[35],exlenrcs,exwidrcs,_V(0.894,0.147,32.819),_V(1,0,0),exhaustrcs);
	AddExhaust(th_rcs[35],exlenrcs,exwidrcs,_V(1.102,0.147,32.556),_V(1,0,0),exhaustrcs);
	AddExhaust(th_rcs[36],exlenrcs,exwidrcs,_V(-4.437,0.83,-16.078),_V(-1,0,0),exhaustrcs);
	AddExhaust(th_rcs[36],exlenrcs,exwidrcs,_V(-4.437,0.83,-16.318),_V(-1,0,0),exhaustrcs);
	AddExhaust(th_rcs[36],exlenrcs,exwidrcs,_V(-0.894,0.147,32.819),_V(-1,0,0),exhaustrcs);
	AddExhaust(th_rcs[36],exlenrcs,exwidrcs,_V(-1.102,0.147,32.556),_V(-1,0,0),exhaustrcs);

	////////// upper rcs pstreams
//	SURFHANDLE pexhaustrcs = oapiRegisterParticleTexture ("Contrail1");
//	PARTICLESTREAMSPEC exhaust_rcs = {
//		0, 0.15, 30, 70, 0.05, 0.05, 30, 15, PARTICLESTREAMSPEC::EMISSIVE,
//		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
//		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
//		pexhaustrcs
//	};

	SURFHANDLE pexhaustrcs = oapiRegisterParticleTexture ("Contrail3");
	PARTICLESTREAMSPEC exhaust_rcs = {
		0, 0.2, 30, 20, 0.025, 0.035, -0.2, 0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		pexhaustrcs
	};

	AddExhaustStream(th_rcs[16], _V(-4.437,0.83,-16.078),&exhaust_rcs);
	AddExhaustStream(th_rcs[16], _V(-4.437,0.83,-16.318),&exhaust_rcs);
	AddExhaustStream(th_rcs[17], _V(0.894,0.147,32.819),&exhaust_rcs);
	AddExhaustStream(th_rcs[17], _V(1.102,0.147,32.556),&exhaust_rcs);

	AddExhaustStream(th_rcs[18], _V(4.437,0.83,-16.078),&exhaust_rcs);
	AddExhaustStream(th_rcs[18], _V(4.437,0.83,-16.318),&exhaust_rcs);
	AddExhaustStream(th_rcs[19], _V(-0.894,0.147,32.819),&exhaust_rcs);
	AddExhaustStream(th_rcs[19], _V(-1.102,0.147,32.556),&exhaust_rcs);

	AddExhaustStream(th_rcs[20], _V(4.683,0.523,-15.554),&exhaust_rcs);
	AddExhaustStream(th_rcs[20], _V(-4.683,0.523,-15.554),&exhaust_rcs);
	AddExhaustStream(th_rcs[20], _V(4.683,0.523,-15.814),&exhaust_rcs);
	AddExhaustStream(th_rcs[20], _V(-4.683,0.523,-15.814),&exhaust_rcs);
	AddExhaustStream(th_rcs[21], _V(1.373,0.142,32.156),&exhaust_rcs);
	AddExhaustStream(th_rcs[22], _V(-1.373,0.142,32.156),&exhaust_rcs);
	AddExhaustStream(th_rcs[21], _V(1.536,0.142,31.89),&exhaust_rcs);
	AddExhaustStream(th_rcs[22], _V(-1.536,0.142,31.89),&exhaust_rcs);

	AddExhaustStream(th_rcs[23], _V(0.158,1.217,32.392),&exhaust_rcs);
	AddExhaustStream(th_rcs[23], _V(-0.158,1.217,32.392),&exhaust_rcs);
	AddExhaustStream(th_rcs[25], _V(4.344,1.192,-16.625),&exhaust_rcs);
	AddExhaustStream(th_rcs[24], _V(-4.344,1.192,-16.625),&exhaust_rcs);
	AddExhaustStream(th_rcs[25], _V(4.475,0.666,-16.625),&exhaust_rcs);
	AddExhaustStream(th_rcs[24], _V(-4.475,0.666,-16.625),&exhaust_rcs);


	AddExhaustStream(th_rcs[27], _V(-4.683,0.523,-15.554),&exhaust_rcs);
	AddExhaustStream(th_rcs[27], _V(-4.683,0.523,-15.814),&exhaust_rcs);
	AddExhaustStream(th_rcs[27], _V(-0.158,1.217,32.292),&exhaust_rcs);

	AddExhaustStream(th_rcs[28], _V(4.683,0.523,-15.554),&exhaust_rcs);
	AddExhaustStream(th_rcs[28], _V(4.683,0.523,-15.814),&exhaust_rcs);
	AddExhaustStream(th_rcs[28], _V(0.158,1.217,32.292),&exhaust_rcs);

	AddExhaustStream(th_rcs[30], _V(4.607,0.151,-16.465),&exhaust_rcs);
	AddExhaustStream(th_rcs[30], _V(4.607,-0.151,-16.465),&exhaust_rcs);
	AddExhaustStream(th_rcs[30], _V(-4.607,0.151,-16.465),&exhaust_rcs);
	AddExhaustStream(th_rcs[30], _V(-4.607,-0.151,-16.465),&exhaust_rcs);
	AddExhaustStream(th_rcs[31], _V(0.131,0.623,33.174),&exhaust_rcs);
	AddExhaustStream(th_rcs[31], _V(-0.131,0.623,33.174),&exhaust_rcs);
	AddExhaustStream(th_rcs[31], _V(0.131,0.907,32.811),&exhaust_rcs);
	AddExhaustStream(th_rcs[31], _V(-0.131,0.907,32.811),&exhaust_rcs);

	AddExhaustStream(th_rcs[32], _V(-4.344,1.192,-16.625),&exhaust_rcs);
	AddExhaustStream(th_rcs[32], _V(-4.475,0.666,-16.625),&exhaust_rcs);
	AddExhaustStream(th_rcs[32], _V(-1.373,0.142,32.156),&exhaust_rcs);
	AddExhaustStream(th_rcs[32], _V(-1.536,0.142,31.89),&exhaust_rcs);

	AddExhaustStream(th_rcs[33], _V(4.344,1.192,-16.625),&exhaust_rcs);
	AddExhaustStream(th_rcs[33], _V(4.475,0.666,-16.625),&exhaust_rcs);
	AddExhaustStream(th_rcs[33], _V(1.373,0.142,32.156),&exhaust_rcs);
	AddExhaustStream(th_rcs[33], _V(1.536,0.142,31.89),&exhaust_rcs);

	AddExhaustStream(th_rcs[34], _V(0.158,1.217,32.392),&exhaust_rcs);
	AddExhaustStream(th_rcs[34], _V(-0.158,1.217,32.392),&exhaust_rcs);
	AddExhaustStream(th_rcs[34], _V(4.683,0.523,-15.554),&exhaust_rcs);
	AddExhaustStream(th_rcs[34], _V(-4.683,0.523,-15.554),&exhaust_rcs);
	AddExhaustStream(th_rcs[34], _V(4.683,0.523,-15.814),&exhaust_rcs);
	AddExhaustStream(th_rcs[34], _V(-4.683,0.523,-15.814),&exhaust_rcs);

	AddExhaustStream(th_rcs[35], _V(4.437,0.83,-16.078),&exhaust_rcs);
	AddExhaustStream(th_rcs[35], _V(4.437,0.83,-16.318),&exhaust_rcs);
	AddExhaustStream(th_rcs[35], _V(0.894,0.147,32.819),&exhaust_rcs);
	AddExhaustStream(th_rcs[35], _V(1.102,0.147,32.556),&exhaust_rcs);
	AddExhaustStream(th_rcs[36], _V(-4.437,0.83,-16.078),&exhaust_rcs);
	AddExhaustStream(th_rcs[36], _V(-4.437,0.83,-16.318),&exhaust_rcs);
	AddExhaustStream(th_rcs[36], _V(-0.894,0.147,32.819),&exhaust_rcs);
	AddExhaustStream(th_rcs[36], _V(-1.102,0.147,32.556),&exhaust_rcs);


	////// p streams
	SURFHANDLE exhaust2 = oapiRegisterParticleTexture ("spacex_starship/Contrail4x");

	SURFHANDLE exhaust2x = oapiRegisterParticleTexture ("spacex_starship/Contrail3");


	PARTICLESTREAMSPEC exhaust_main1 = {
		0, 1.0, 8, 200, 0.05, 0.075, 2.6, 0.3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 0,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exhaust2
	};

	PARTICLESTREAMSPEC exhaust_main2 = {
		0, 1.0, 8, 200, 0.05, 0.075, 2.5, 0.3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 0,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exhaust2x
	};

	///max q pstream
	SURFHANDLE exhaust3 = oapiRegisterParticleTexture ("spacex_starship/contrail");
	PARTICLESTREAMSPEC maxqtrail = {
		0, 6, 10, 170, 0.01, 0.25, 7, 0.5, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 1, 0,
		PARTICLESTREAMSPEC::ATM_FLAT, 0.25, 0.25,
		exhaust3
	};
	AddParticleStream(&maxqtrail,_V(0,0,47),_V(0,0,-1), &pst_lvl);

	////// condensation
	PARTICLESTREAMSPEC condensation = {
		0, 5, 4, 0.5, 0.25, 5, 0.0, 0.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0.5, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		exhaust3
	};

	///// fuel dump
	PARTICLESTREAMSPEC fuel_dump = {
		0, 3.0, 4, 150, 0.5, 3, 20.0, 10.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0.5, 0.5,
		PARTICLESTREAMSPEC::ATM_FLAT, 0.5, 0.5,
		exhaust3
	};

	///// he_vent
//	PARTICLESTREAMSPEC he_vent = {
//		0, 0.04, 10, 10, 0.05, 0.05, 4, 1, PARTICLESTREAMSPEC::EMISSIVE,
//		PARTICLESTREAMSPEC::LVL_SQRT, 0.5, 0.5,
//		PARTICLESTREAMSPEC::ATM_FLAT, 0.5, 0.5,
//		exhaust3
//	};
//	AddParticleStream(&he_vent,_V(-0.46,0,-5.964),_V(-0.7071,0,-0.7071), &he_vent_level);


	if (gcEnabled() == false)
	{
	AddParticleStream(&condensation,_V(0,0,0),_V(0,0,-1), &pst_lvl2);
	AddParticleStream(&condensation,_V(0,0,34),_V(0,0,-1), &pst_lvl2);
	}


	///// main0
	//// inner
	for (int i = 0; i < 10; i++)
	{
	rm = rotm(_V(0,0,1),(36*RAD*i));
	exhaust_pos = mul(rm,_V(-0.352,-2.233,-60.5));
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main2);
	}
	//// outer
	for (int i = 0; i < 20; i++)
	{
	rm = rotm(_V(0,0,1),(18*RAD*i));
	exhaust_pos = mul(rm,_V(0,4.13,-59.643));
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main2);
	}
	/// central
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_pos = mul(rm,_V(0,-0.833,-60.5));
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main1);
	AddExhaustStream(th_main[0],exhaust_pos,&exhaust_main2);
	}

	//////// fuel dump
	for (int i = 0; i < 3; i++)
	{
	rm = rotm(_V(0,0,1),(120*RAD*i));
	exhaust_pos = mul(rm,_V(0,0.981,-18));
	AddExhaustStream(th_dump, exhaust_pos, &fuel_dump);
	}
	///////////////////////////////////////////////////////
	/////AirFoils

	hVABooster = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeff,0, 40, 300, 0.16);
	hHABooster = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeff,0, 40, 300, 0.16);

	////////
	/*char att_params[256];
	char tag_nbr[256];
	VECTOR3 att_pos = _V(0,0,0);
	VECTOR3 att_dir = _V(0,0,1);
	VECTOR3 att_rot = _V(0,1,0);

	for (int i = 0; i < 20; i++)
	{
	
	sprintf (tag_nbr, "PayloadAttach%d", i);
	if(oapiReadItem_string(cfg, tag_nbr, att_params) == TRUE)
		{
			sscanf (att_params, "%lf%lf%lf%lf%lf%lf%lf%lf%lf", &att_pos.x, &att_pos.y, &att_pos.z, &att_dir.x, &att_dir.y, &att_dir.z, &att_rot.x, &att_rot.y, &att_rot.z);
			pld_attach[i] = CreateAttachment (false, att_pos, att_dir, att_rot, "SH");
		}
	else 
		{
		att_pos = _V( 0, 0, 14.232) + STG1STG2_OFF;
		att_dir = _V( 0,0,1);
		att_rot = _V(0,1,0);
		pld_attach[i] = CreateAttachment (false, att_pos, att_dir, att_rot, "SH");
		}
	pld_attach_pos_init[i] = att_pos;
	pld_attach_dir_init[i] = att_dir;
	pld_attach_rot_init[i] = att_rot;
	}*/

	


	lpd_attach = CreateAttachment (true, _V( 0, 0, -56), _V( 0,0,-1), _V(0,1,0), "LP");
	arm_attach = CreateAttachment (true, _V( 0, 0, 50.55), _V( 0,0,-1), _V(0,1,0), "ARM");
	hDock_rear = CreateDock(_V(0,4.545,0), _V(0,1,0), _V(0,0,1));

	//////////
	//local light
	COLOUR4 col_d = {0.125,0.1,0.06125,0};
	COLOUR4 col_s = {0,0,0,0};
	COLOUR4 col_a = {0,0,0,0};
	COLOUR4 col_white = {1,1,1,0};
	spt_light = (PointLight*)AddPointLight(_V(0,0,-61),5000, 1e-1, 0, 1e-6, col_d, col_s, col_a);
	spt_light->Activate(false);

	/*COLOUR4 col_d2 = {1,1,1,0};
	COLOUR4 col_s2 = {1,1,1,0};
	COLOUR4 col_a2 = {0,0,0,0};
	
	spotlight1 = (SpotLight*)AddSpotLight(bspos, _V(0,0,1),200, 1, 0.2, 2.5e-5, (120*RAD), (180*RAD), col_d2, col_s2, col_a2);
	spotlight1->Activate(false);

	beac.shape = BEACONSHAPE_STAR;
	beac.pos = &bspos;
	beac.col = &bscol;
	beac.size = 0.5;
	beac.falloff = 0.5;
	beac.period = 0;
	beac.duration = 0;
	beac.tofs = 0;
	beac.active = false;
	AddBeacon (&beac);*/

	/////////////////////////////// landing / takeoff dust

	th_dust = CreateThruster (_V(0, 0, -70), _V(0, 0, 1), 0, hProp2, 1000000);

		SURFHANDLE exhaustdust = oapiRegisterParticleTexture ("Contrail1");

		PARTICLESTREAMSPEC dust_pstream = {
		0, 8, 20, 200, 0.5, 1.25, 20, 5, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 1, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 0.7e-7, 1,
		exhaustdust
	};

	AddExhaustStream(th_dust, &dust_pstream);

}

void ss_tanker::clbkPostCreation ()
{
 hGref = GetGravityRef();
 hMe = GetHandle();
 OBJHANDLE bbtgt = oapiGetObjectByName(refname);
 if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, refname);
 if (bbtgt == NULL)
 {
	strcpy(refname, "NULL");
 }

 OBJHANDLE landtgt = oapiGetObjectByName(tgtname);
 if (landtgt == NULL)landtgt = oapiGetBaseByName(hGref, tgtname);
 if (landtgt == NULL)
 {
	strcpy(tgtname, "NULL");
 }

 GetWindVector();
 
 if (eng_status == 1)
  {
	  DelMesh(1);
	  ConfigMain();
	  if(rcs_sw2 == 1)ConfigLateralRCS();
	  DoTDPoints();
  }
 else SetADCtrlMode(0);

  CalcEmptyMass();
  CalcFuelRes();
 
	double lat, lng, rds;
	GetEquPos(lng,lat,rds);
	min_inc = lat;
	if (abs(tgt_inc) < abs(lat)) tgt_inc = lat;

	///////////// sound
	MyID=ConnectToOrbiterSoundDLL(GetHandle());

	RequestLoadVesselWave(MyID,CTDN,"Sound\\Vessel\\countdown1.wav",DEFAULT);
	RequestLoadVesselWave(MyID,SEP,"Sound\\Vessel\\undocking.wav",BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave(MyID,SEPST,"Sound\\_CustomVesselsSounds\\spacex_starship\\stgsep.wav",DEFAULT);
	//RequestLoadVesselWave(MyID,SEPFR,"Sound\\_CustomVesselsSounds\\Falcon9R_V12\\frsep.wav",DEFAULT);
	
	SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	//SoundOptionOnOff3(MyID,PLAYWHENATTITUDEMODECHANGE,FALSE);
	//SoundOptionOnOff3(MyID,PLAYGPWS,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	//SoundOptionOnOff3(MyID,PLAYRADARBIP,FALSE);

	/*for (int i = 0; i < 20; i++)
	{
	OBJHANDLE hPayload = GetAttachmentStatus(pld_attach[i]);
	if(hPayload)pld_att_status = pld_att_status+1;
	}*/

	///////// forward lamp
	/*if (eng_status == 1 && lamp_status == 1)
	{
	beac.active = true;
	spotlight1->Activate(true);
	}
	else
	{
	lamp_status = 0;
	beac.active = false;
	spotlight1->Activate(false);
	}*/

	//ConfigLateralRCS();

	if(uppermainprop == 1)SetPropReserve();
	if(raptor_sl_selection < 3)ConfigUpperRaptorSL();
	
}


// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new ss_tanker (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (ss_tanker*)vessel;
}


void ss_tanker::clbkSaveState (FILEHANDLE scn)
{
	adcm_scn = GetADCtrlMode();

	char cbuf[256];

	for (int i = 0; i < (int)strlen(refname); i++)
	{
		if (refname[i] == ' ')
		{
			refname[i] = '+';
			break;
		}
	}

	for (int i = 0; i < (int)strlen(tgtname); i++)
	{
		if (tgtname[i] == ' ')
		{
			tgtname[i] = '+';
			break;
		}
	}

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	sprintf (cbuf, "%d", raptor_sl_selection);
	oapiWriteScenario_string (scn, "RSL", cbuf);

	sprintf (cbuf, "%0.4f", acc_lim);
	oapiWriteScenario_string (scn, "ACL", cbuf);

	sprintf (cbuf, "%d %d", uppermaincfg, uppermainprop);
	oapiWriteScenario_string (scn, "UEC", cbuf);

	sprintf (cbuf, "%d %d %d %s", skydive_sw, backflip_sw, retroland_sw, tgtname);
	oapiWriteScenario_string (scn, "TGT", cbuf);

	sprintf (cbuf, "%d", disable_rcs);
	oapiWriteScenario_string (scn, "DLR", cbuf);

	sprintf (cbuf, "%d %d %d", adcs_reentry_mode, adcm_scn, reentry_vspd_sw);
	oapiWriteScenario_string (scn, "RNT", cbuf);

	sprintf (cbuf, "%d %0.2f %0.2f", fbw_reentry_sw, fbw_bank, fbw_aoa);
	oapiWriteScenario_string (scn, "FBW", cbuf);

	sprintf (cbuf, "%d", hud_display_info);
	oapiWriteScenario_string (scn, "DSP", cbuf);

	//sprintf (cbuf, "%d %0.4f", paf_status, paf_proc);
	//oapiWriteScenario_string (scn, "PAF", cbuf);

	//sprintf (cbuf, "%d %0.4f", hatch_status, hatch_proc);
	//oapiWriteScenario_string (scn, "HAT", cbuf);


	sprintf (cbuf, "%0.4f %0.4f", port_can_proc, stbd_can_proc);
	oapiWriteScenario_string (scn, "CAN", cbuf);

	sprintf (cbuf, "%0.4f %0.4f %0.4f", port_wing_proc, stbd_wing_proc, wing_trim);
	oapiWriteScenario_string (scn, "WNG", cbuf);

	sprintf (cbuf, "%d %0.4f", legs_status, legs_proc);
	oapiWriteScenario_string (scn, "LEG", cbuf);

	sprintf (cbuf, "%d %d %d %d %d %d %d", eng_status, ap_status, ap_function, op_status, afs_status, start_sw, edlo_sw);
	oapiWriteScenario_string (scn, "SOST", cbuf);

	sprintf (cbuf, "%0.7f %0.1f %0.1f %0.1f %0.1f %s %0.4f", tgt_inc, op_meco_alt, op_tgt, fuel_reserve, ap_tgt, refname, pitch_angle);
	oapiWriteScenario_string (scn, "SOPR", cbuf);

	sprintf (cbuf, "%d %d %d %d %0.4f", lot_hh, lot_mm, lot_ss, lot_status, timer);
	oapiWriteScenario_string (scn, "LOTM", cbuf);

	//sprintf (cbuf, "%0.4f %0.4f %d", red_fact, red_fact2, lamp_status);
	//oapiWriteScenario_string (scn, "NOZL", cbuf);fuel_trans_sw

	sprintf (cbuf, "%d %d %0.4f %0.4f %d", fuel_trans_sw, tank_index, tank_max, tank_level, fuel_source_sw);
	oapiWriteScenario_string (scn, "TRAN", cbuf);

	sprintf (cbuf, "%0.4f %d %d", red_fact, reentry_sw, bank_dir_sw);
	oapiWriteScenario_string (scn, "NOZL", cbuf);
}

void ss_tanker::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "RSL", 3)) {
			sscanf_s (line+3, "%d", &raptor_sl_selection);
		}else if (!strnicmp (line, "ACL", 3)) {
			sscanf_s (line+3, "%lf", &acc_lim);
		}else if (!strnicmp (line, "UEC", 3)) {
			sscanf_s (line+3, "%d%d", &uppermaincfg, &uppermainprop);
		}else if (!strnicmp (line, "TGT", 3)) {
			sscanf (line+3, "%d%d%d%s", &skydive_sw, &backflip_sw, &retroland_sw, &tgtname);
		}else if (!strnicmp (line, "DLR", 3)) {
			sscanf_s (line+3, "%d", &disable_rcs);
		}else if (!strnicmp (line, "RNT", 3)) {
			sscanf_s (line+3, "%d%d%d", &adcs_reentry_mode, &adcm_scn, &reentry_vspd_sw);
		}else if (!strnicmp (line, "FBW", 3)) {
			sscanf_s (line+3, "%d%lf%lf", &fbw_reentry_sw, &fbw_bank, &fbw_aoa);
		}else if (!strnicmp (line, "DSP", 3)) {
			sscanf_s (line+3, "%d", &hud_display_info);
		//}else if (!strnicmp (line, "PAF", 3)) {
		//	sscanf_s (line+3, "%d%lf", &paf_status, &paf_proc);
		//}else if (!strnicmp (line, "HAT", 3)) {
		//	sscanf_s (line+3, "%d%lf", &hatch_status, &hatch_proc);
		}else if (!strnicmp (line, "CAN", 3)) {
            sscanf (line+3, "%lf%lf", &port_can_proc, &stbd_can_proc);
		}else if (!strnicmp (line, "WNG", 3)) {
            sscanf (line+3, "%lf%lf%lf", &port_wing_proc, &stbd_wing_proc, &wing_trim);
		}else if (!strnicmp (line, "LEG", 3)) {
			sscanf_s (line+3, "%d%lf", &legs_status, &legs_proc);
		}else if (!strnicmp (line, "SOST", 4)) {
		sscanf (line+4, "%d%d%d%d%d%d%d", &eng_status, &ap_status, &ap_function, &op_status, &afs_status, &start_sw, &edlo_sw);
		} else if (!strnicmp (line, "SOPR", 4)) {
			sscanf (line+4, "%lf%lf%lf%lf%lf%s%lf",&tgt_inc, &op_meco_alt, &op_tgt, &fuel_reserve, &ap_tgt, &refname, &pitch_angle);
		} else if (!strnicmp (line, "LOTM", 4)) {
			sscanf (line+4, "%d%d%d%d%lf",&lot_hh, &lot_mm, &lot_ss, &lot_status, &timer);
		} else if (!strnicmp (line, "TRAN", 4)) {
			sscanf (line+4, "%d%d%lf%lf%d",&fuel_trans_sw, &tank_index, &tank_max, &tank_level, &fuel_source_sw);
		} else if (!strnicmp (line, "NOZL", 4)) {
			sscanf (line+4, "%lf%d%d",&red_fact, &reentry_sw, &bank_dir_sw);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }


	SetAnimation (anim_port_can, port_can_proc);
	SetAnimation (anim_stbd_can, stbd_can_proc);
	SetAnimation (anim_port_wing, port_wing_proc);
	SetAnimation (anim_stbd_wing, stbd_wing_proc);
	SetAnimation (anim_legs, legs_proc);
//	SetAnimation (anim_hatch, hatch_proc);
//	SetAnimation (anim_paf, paf_proc);

	for (int i = 0; i < (int)strlen(refname); i++)
	{
		if (refname[i] == '+')
		{
			refname[i] = ' ';
			break;
		}
	}

	////////////////

	for (int i = 0; i < (int)strlen(tgtname); i++)
	{
		if (tgtname[i] == '+')
		{
			tgtname[i] = ' ';
			break;
		}
	}

	SetADCtrlMode(adcm_scn);

	//CalcFuelRes();
}

void ss_tanker::clbkDockEvent (int dock, OBJHANDLE mate)
{
	hTransTarget = mate;
	if(de_first_frame_sw == 1)
	{
		if(hTransTarget != NULL)
		{
		VESSEL2* vtgt = (VESSEL2*)oapiGetVesselInterface(hTransTarget);
		PROPELLANT_HANDLE hDefaultProp = vtgt->GetDefaultPropellantResource();
		int tank_count = vtgt->GetPropellantCount();
		for (int i = 0; i < tank_count; i++)
			{
			PROPELLANT_HANDLE hTgtProp = vtgt->GetPropellantHandleByIndex(i);
			if(hTgtProp == hDefaultProp)
				{
				tank_index = i;
				tank_max = vtgt->GetPropellantMaxMass(hDefaultProp);
				tank_level = vtgt->GetPropellantMass(hDefaultProp);
				break;
				}
			}

		}
	else 
		{
		tank_index = 0;
		tank_max = 0;
		tank_level = 0;
		}
	}
}

void ss_tanker::DoFuelTransfer()
{
	if(hTransTarget == NULL)
	{
		fuel_trans_sw = 0;
		tank_index = 0;
		return;
	}
	else
	{
	VESSEL2* vtgt = (VESSEL2*)oapiGetVesselInterface(hTransTarget);
	PROPELLANT_HANDLE hTgtProp = vtgt->GetPropellantHandleByIndex(tank_index);
	tank_level = vtgt->GetPropellantMass(hTgtProp);
	tank_max = vtgt->GetPropellantMaxMass(hTgtProp);
	double delta_prop = (step_time*200);
	double pld_mass_2 = GetPropellantMass(hProp_payload);
	if(fuel_source_sw == 1)pld_mass_2 = GetPropellantMass(hProp2);

	if(delta_prop > pld_mass_2)
		{
		delta_prop = pld_mass_2;
		fuel_trans_sw = 0;
		}
	if(delta_prop > (tank_max - tank_level))
		{
		delta_prop = (tank_max - tank_level);
		fuel_trans_sw = 0;
		}


	vtgt->SetPropellantMass(hTgtProp,(tank_level + delta_prop));
	if(fuel_source_sw == 0)SetPropellantMass(hProp_payload, (pld_mass_2 - delta_prop));
	else if(fuel_source_sw == 1)SetPropellantMass(hProp2, (pld_mass_2 - delta_prop));
	tank_level = tank_level+delta_prop;
	}

}

bool SetTANK(void *id, char *str, void *data)
{
	if(hTransTarget == NULL)return false;
	float new_tank;
	sscanf(str, "%f", &new_tank);
	int new_tank_id = int(new_tank);
	VESSEL2* vtgt = (VESSEL2*)oapiGetVesselInterface(hTransTarget);
	int tank_count = vtgt->GetPropellantCount();
	if(new_tank_id > (tank_count-1))return false;
	else if(new_tank_id < 0)return false;
	else tank_index = new_tank_id;
	PROPELLANT_HANDLE hTgtProp = vtgt->GetPropellantHandleByIndex(tank_index);
	tank_max = vtgt->GetPropellantMaxMass(hTgtProp);
	tank_level = vtgt->GetPropellantMass(hTgtProp);
	return true;
}

bool SetINC(void *id, char *str, void *data)
{
	float new_inc;
	sscanf(str, "%f", &new_inc);
	if (abs(new_inc) < abs(min_inc*DEG))return false;
	else 
		{
		tgt_inc = new_inc*RAD;
		return true;
		}
}

bool SetACC(void *id, char *str, void *data)
{
	float new_acc;
	sscanf(str, "%f", &new_acc);
	acc_lim = new_acc;
	return true;
		
}

bool SetRF(void *id, char *str, void *data)
{
	float new_rf;
	sscanf(str, "%f", &new_rf);
	if (new_rf < 0)return false;
	else if (new_rf > 7*80000)return false;
	else 
		{
		fuel_reserve = new_rf;
	//	if (ap_tgt > (450000 - (3*fuel_reserve)))ap_tgt = (450000 - (3*fuel_reserve));
	//	if (ap_tgt < 100000)ap_tgt = 100000;
		return true;
		}
}

bool SetAP(void *id, char *str, void *data)
{
	float new_ap;
	sscanf(str, "%f", &new_ap);
	if (new_ap < 100)return false;
	else if (new_ap > 450)return false;
	else 
		{
		ap_tgt = new_ap*1000;
	//	if (fuel_reserve > ((450000 - ap_tgt)/3)) fuel_reserve = ((450000 - ap_tgt)/3);
	//	if (fuel_reserve < 0)fuel_reserve = 0;
		return true;
		}
}

bool SetBBTgt(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	if (!strnicmp(cbuf, "NULL", 4))
	{
		strcpy(refname,cbuf);
		afs_status = 0;
		return true;
	}
	OBJHANDLE bbtgt = oapiGetObjectByName(cbuf);
	if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, cbuf);
	if (bbtgt == NULL)return false;
	else 
		{
		strcpy(refname,cbuf);
		return true;
		}
}

bool SetTGT(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	OBJHANDLE landtgt = oapiGetObjectByName(cbuf);
	if (landtgt == NULL)landtgt = oapiGetBaseByName(hGref, cbuf);
	if (landtgt == NULL)return false;
	else 
		{
		VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
		vme->SendBufferedKey(COMM_SEND_BBT, false, cbuf);
		return true;
		}
}

bool SetOPE(void *id, char *str, void *data)
{
	float new_ope;
	sscanf(str, "%f", &new_ope);
	if (new_ope < 160)return false;
	else 
		{
			if (new_ope > op_tgt/1000)
			{
			op_meco_alt = op_tgt;
			op_tgt = new_ope*1000;
			} 
			else op_meco_alt = new_ope*1000;
			//ap_tgt = 0.67*op_meco_alt;
			if (((op_meco_alt*0.001)-200) > 0)
			{
			ap_tgt = 1000*((5.05*(sqrt((op_meco_alt*0.001)-200))) + 125);
			}
			else if (((op_meco_alt*0.001)-200) < 0)
			{
			ap_tgt = (op_meco_alt*0.625);
			}
			else ap_tgt = 125000;

		//	if (ap_tgt > (450000 - (3*fuel_reserve)))ap_tgt = (450000 - (3*fuel_reserve));
			if (ap_tgt < 100000)ap_tgt = 100000;
		return true;
		}
}

bool SetOAP(void *id, char *str, void *data)
{
	float new_oap;
	sscanf(str, "%f", &new_oap);
	if (new_oap < 160)return false;
	else 
		{
			if (new_oap < op_meco_alt/1000)
			{
			op_tgt = op_meco_alt;
			op_meco_alt = new_oap*1000;
			} 
			else op_tgt = new_oap*1000;
			//ap_tgt = 0.67*op_meco_alt;
			if (((op_meco_alt*0.001)-200) > 0)
			{
			ap_tgt = 1000*((5.05*(sqrt((op_meco_alt*0.001)-200))) + 125);
			}
			else if (((op_meco_alt*0.001)-200) < 0)
			{
			ap_tgt = (op_meco_alt*0.625);
			}
			else ap_tgt = 125000;

		//	if (ap_tgt > (450000 - (3*fuel_reserve)))ap_tgt = (450000 - (3*fuel_reserve));
			if (ap_tgt < 100000)ap_tgt = 100000;
		return true;
		}
}

bool SetPLD(void *id, char *str, void *data)
{
	float new_fuel;
	sscanf(str, "%f", &new_fuel);
	if(new_fuel > 170000)return false;
	else if(new_fuel < 0)return false;
	else {
	VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
	vme->SetPropellantMass(hProp_payload, new_fuel);
	pld_mass = new_fuel;
	return true;
	}
	/*if(pld_att_status > 19)return false;
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	OBJHANDLE hPld = oapiGetObjectByName(cbuf);
	if (hPld == NULL)return false;
	else
	{
	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hPld);
	DWORD ac = v->AttachmentCount(true);
	if (ac == 0) return false;
	for (DWORD j = 0; j < ac; j++)
		{
		ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(true, j);
		OBJHANDLE hOp = v->GetAttachmentStatus(hAtt);
		if (hOp == NULL)
			{
			VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(hMe);
			if (vme->AttachChild(hPld, pld_attach[pld_att_status], hAtt) == true)
				{
					calc_sw = 0;
					strcpy(pldname[pld_att_status], cbuf);
					pld_att_status = pld_att_status + 1;
					return true;
				}
			else return false;
			}
		else if (j == ac-1) return false;
		}
	
	}*/
	
}

BOOL CALLBACK DlgProc (HWND, UINT, WPARAM, LPARAM);

int ss_tanker::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

	HWND hDbox;

////////////// internal message string //////////
	if (key == COMM_SEND_BBT) {
			sscanf (kstate, "%s", &tgtname);
			GetWindVector(); // get wind at new tgt
			return 1;
       }

//////////////////////////////////////////////

if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
	

	} 
  }

  else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		case OAPI_KEY_1:
		raptor_sl_selection = 1;
		ConfigUpperRaptorSL();
		return 1;

		case OAPI_KEY_2:
		raptor_sl_selection = 2;
		ConfigUpperRaptorSL();
		return 1;

		case OAPI_KEY_3:
		raptor_sl_selection = 3;
		ConfigUpperRaptorSL();
		return 1;

		case OAPI_KEY_A:
		RESETKEY (kstate, OAPI_KEY_A);
		if(adcs_reentry_mode == 0) adcs_reentry_mode = 1;
		else adcs_reentry_mode = 0;
		return 1;

		case OAPI_KEY_N:
		if(uppermaincfg == 0) uppermaincfg = 1;
		else if(uppermaincfg == 1) uppermaincfg = 2;
		else if(uppermaincfg == 2) uppermaincfg = 0;
		ConfigUpperMain();
		return 1;

		case OAPI_KEY_J:
		if(uppermainprop == 0) uppermainprop = 1;
		else uppermainprop = 0;
		SetPropReserve();
		return 1;


		case OAPI_KEY_K:
		hGref = GetGravityRef();
		oapiOpenInputBox("Enter Landing Target Name",SetTGT, 0, 50, 0);
		return 1;

		case OAPI_KEY_V: 
		if (retroland_sw == 0)
		{
			retroland_sw = 1;
			backflip_sw = 0;
			skydive_sw = 0;
		//	prop_sw = 1;
		//	SetEngProp();
		}
		else 
		{
			retroland_sw = 0;
			KillAllRCS2();
		}
		return 1;

		case OAPI_KEY_B: 
		if (skydive_sw == 0)
		{
			skydive_sw = 1;
			retroland_sw = 0;
		//	prop_sw = 1;
		//	SetEngProp();
		}
		else 
		{
			skydive_sw = 0;
			KillAllRCS2();
		}
		return 1;
	} 
  }

else { // unmodified keys
    switch (key) {

		case OAPI_KEY_U: 
		if(reentry_sw == 0)
		{
			reentry_sw = 1;
			fbw_reentry_sw = 0;
			adcs_reentry_mode = 1;
			SetADCtrlMode(0);
			disable_rcs = 1;
			ConfigLateralRCS();
		}
		else if(reentry_sw == 1)
		{
			reentry_sw = 0;
			KillAllRCS2();
		}
		return 1;

		case OAPI_KEY_F: 
		if(fbw_reentry_sw == 0)
		{
			fbw_reentry_sw = 1;
			reentry_sw = 0;
			adcs_reentry_mode = 1;
			SetADCtrlMode(0);
			disable_rcs = 1;
			ConfigLateralRCS();
		}
		else if(fbw_reentry_sw == 1)
		{
			fbw_reentry_sw = 0;
			KillAllRCS2();
		}
		return 1;

		case OAPI_KEY_D:
		if (eng_status == 1)
		{
		//RESETKEY (kstate, OAPI_KEY_D);
		if(disable_rcs == 0) disable_rcs = 1;
		else disable_rcs = 0;
		ConfigLateralRCS();
		}
		return 1;

		case OAPI_KEY_G: 
		if (eng_status == 1)
		{
		if (legs_status  == 0)legs_status = 1;
		else if (legs_status  == 1)legs_status = 3;
		else if (legs_status  == 2)legs_status = 3;
		else if (legs_status  == 3)legs_status = 1;
		}
		return 1;

		case OAPI_KEY_C: // hud display
		RESETKEY (kstate, OAPI_KEY_C);
		if(hud_display_info == 0)hud_display_info = 1; // orbital
		else if(hud_display_info == 1)hud_display_info = 2; // reentry
		else if(hud_display_info == 2)hud_display_info = 3; // landing
		else if(hud_display_info == 3)hud_display_info = 0; // launch
		return 1;
		
		case OAPI_KEY_DELETE: 
		wing_trim = 0;
		return 1;

		case OAPI_KEY_J: // fuel transfer on/off
		if(fuel_trans_sw == 0)fuel_trans_sw = 1;
		else fuel_trans_sw = 0;
		return 1;

		case OAPI_KEY_W:
		oapiOpenInputBox("Enter Tank Index", SetTANK, 0, 50, 0);
		return 1;

		//case OAPI_KEY_P: // add payload
		//hMe = GetHandle();
		//if(pld_att_status == 0)oapiOpenInputBox("Enter Payload Name", SetPLD, 0, 50, 0);
		//return 1;

		//case OAPI_KEY_P:
		//
		//return 1;

		//case OAPI_KEY_M:
		//if (dlg_status == 0)
		//{
		//oapiOpenInputBox("Enter Apogee at Stage1 MECO1 100-450 (km alt.)",SetAP, 0, 50, 0);
		//dlg_status = 1;
		//}
		//else if (dlg_status == 1)
		//{
		//oapiOpenInputBox("Enter Flyback Fuel Reserve 0-80000 (kg)",SetRF, 0, 50, 0);
		//dlg_status = 0;
		//}
		//return 1;

		case OAPI_KEY_V: // ascent pilot
		if(lot_status != 1 && lot_status != 2)
		{
		if (eng_status == 0)
			{
			if(ap_status == 0)ap_status = 1;
			else ap_status = 0;
			op_status = 0;
			KillAllRCS2();
			}
		else if (eng_status == 1)
			{
			if(op_status == 0)op_status = 1;
			else op_status = 0;
			ap_status = 0;
			KillAllRCS2();
			}
		apop_status = ap_status + op_status;
		}
		return 1;

		case OAPI_KEY_E:
		if (cam_status == 0)cam_status = 1;
		else if (cam_status == 1)cam_status = 2;
		else if (cam_status == 2)cam_status = 3;
		else if (cam_status == 3)cam_status = 4;
		else if (cam_status == 4)cam_status = 0;
		SetCamDir();
		return 1;

		case OAPI_KEY_K:
			hDbox = oapiFindDialog (g_Param.hDLL, IDD_CTRL);
			if (hDbox) oapiCloseDialog (hDbox);
			else hDbox = oapiOpenDialogEx(g_Param.hDLL, IDD_CTRL, DlgProc,DLG_CAPTIONCLOSE, this);
		return 1;

		case OAPI_KEY_B:
		if(GetThrusterGroupLevel(THGROUP_MAIN) == 0)
		{
			if(GetThrusterLevel(th_dump) == 0) SetThrusterLevel(th_dump, 1);
			else if(GetThrusterLevel(th_dump) > 0) SetThrusterLevel(th_dump, 0);
		}
		else SetThrusterLevel(th_dump, 0);
		return 1;

		case OAPI_KEY_N:
		if (eng_status == 0)
		{
			JettisonStg();
			if(ap_status == 1)
			{
				ap_status = 0;
				op_status = 1;
			}
		}
		return 1;

		case OAPI_KEY_M:
		if (fuel_source_sw == 0)fuel_source_sw = 1;
		else if (fuel_source_sw == 1)fuel_source_sw = 0;
		return 1;

		/*case OAPI_KEY_M:
			if (eng_status == 1)
			{
			if(lamp_status == 0) 
				{
				lamp_status = 1;
				beac.active = true;
				spotlight1->Activate(true);
				}
			else 
				{
				lamp_status = 0;
				beac.active = false;
				spotlight1->Activate(false);
				}
			}
			else lamp_status = 0;
		return 1;*/

		case COMM_CALC_MASS:
		CalcEmptyMass();
		return 1;

		case COMM_CALC_FRE:
		CalcFuelRes();
		return 1;
	}
  }
  return 0;
}


int ss_tanker::clbkConsumeDirectKey(char *kstate)
{
	if (KEYMOD_ALT (kstate))
	{

		if (KEYDOWN (kstate, OAPI_KEY_PRIOR)) // wing trim
		{
		RESETKEY (kstate, OAPI_KEY_PRIOR);
		wing_trim = wing_trim + (step_time*0.25);
		if(wing_trim > 1)wing_trim = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_NEXT)) // wing trim
		{
		RESETKEY (kstate, OAPI_KEY_NEXT);
		wing_trim = wing_trim - (step_time*0.25);
		if(wing_trim < -1)wing_trim = -1;
		}

	}
	else if (KEYMOD_CONTROL (kstate)) 
	{
		if (KEYDOWN (kstate, OAPI_KEY_PRIOR)) // wing trim
		{
		RESETKEY (kstate, OAPI_KEY_PRIOR);
		wing_trim = wing_trim + (step_time*0.025);
		if(wing_trim > 1)wing_trim = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_NEXT)) // wing trim
		{
		RESETKEY (kstate, OAPI_KEY_NEXT);
		wing_trim = wing_trim - (step_time*0.025);
		if(wing_trim < -1)wing_trim = -1;
		}

	}
	else if (KEYMOD_SHIFT (kstate)) 
	{
		if (KEYDOWN (kstate, OAPI_KEY_LEFT)) // fbw_bank
		{
		RESETKEY (kstate, OAPI_KEY_LEFT);
		fbw_bank = fbw_bank + (step_time*5);
		if(fbw_bank > 180)fbw_bank = fbw_bank - 360;
		}

		if (KEYDOWN (kstate, OAPI_KEY_RIGHT)) // fbw_bank
		{
		RESETKEY (kstate, OAPI_KEY_RIGHT);
		fbw_bank = fbw_bank - (step_time*5);
		if(fbw_bank < -180)fbw_bank = fbw_bank + 360;
		}

		if (KEYDOWN (kstate, OAPI_KEY_UP)) // fbw_aoa
		{
		RESETKEY (kstate, OAPI_KEY_UP);
		fbw_aoa = fbw_aoa + (step_time*5);
		if(fbw_bank > 90)fbw_bank = 90;
		}

		if (KEYDOWN (kstate, OAPI_KEY_DOWN)) // fbw_aoa
		{
		RESETKEY (kstate, OAPI_KEY_DOWN);
		fbw_aoa = fbw_aoa - (step_time*5);
		if(fbw_bank < 0)fbw_bank = 0;
		}
	}
	else
	{
		
		

	}

	return 0;
}


DLLCLBK void InitModule (HINSTANCE hModule)
{
	g_Param.hDLL = hModule;
	oapiRegisterCustomControls (hModule);
	
	
}

DLLCLBK void ExitModule (HINSTANCE hModule)
{
	oapiUnregisterCustomControls (hModule);
	
}
void InitDialog(HWND hWnd)
{
	char cbuf[256];
	char lotbuf[2];
	char lotbuftemp[2];
	strcpy(lotbuftemp, "0");
	
	sprintf (cbuf, "%0.0f", fuel_reserve);
	SetWindowText (GetDlgItem (hWnd, IDC_FUEL_RES), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Enter");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_FUEL_APPLY), cbuf);

	sprintf (cbuf, "%s", refname);
	SetWindowText (GetDlgItem (hWnd, IDC_TGTNM), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Enter");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_TGTNM_APPLY), cbuf);

	sprintf (cbuf, "%s", tgtname);
	SetWindowText (GetDlgItem (hWnd, IDC_TGTNM2), cbuf);
	sprintf (cbuf, "Enter");
	SetWindowText (GetDlgItem (hWnd, IDC_TGTNM2_APPLY), cbuf);

	if (afs_status == 0)sprintf(cbuf,"OFF");
	else sprintf(cbuf,"ENABLED");
	SetWindowText (GetDlgItem (hWnd, IDC_FB_STATUS), cbuf);
	if (eng_status == 0)sprintf (cbuf, "On/Off");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_FB_STATUS_APPLY), cbuf);

	//sprintf (cbuf, "%s", pldname[pld_att_status - 1]);
	sprintf (cbuf, "%0.0f", (pld_mass));
	SetWindowText (GetDlgItem (hWnd, IDC_PLD_NM), cbuf);
	//if (pld_att_status < 20)SetWindowText (GetDlgItem (hWnd, IDC_PLD_NM_APPLY), "Attach");
	SetWindowText (GetDlgItem (hWnd, IDC_PLD_NM_APPLY), "Enter");

	//strcpy (cbuf, "NAME");
	/*if (pld_att_status > 0)
	{
		SetWindowText (GetDlgItem (hWnd, IDC_PLD_JET), pldname[pld_att_status - 1]);
		SetWindowText (GetDlgItem (hWnd, IDC_PLD_JET_APPLY), "Jettison");
	}
	else 
	{
		SetWindowText (GetDlgItem (hWnd, IDC_PLD_JET), "NULL");
		SetWindowText (GetDlgItem (hWnd, IDC_PLD_JET_APPLY), "N/A");
	}*/

	sprintf (cbuf, "%0.0f kg", (pld_mass));
	SetWindowText (GetDlgItem (hWnd, IDC_PLD_MASS), cbuf);


	sprintf (cbuf, "%0.2f", (tgt_inc*DEG));
	SetWindowText (GetDlgItem (hWnd, IDC_INC), cbuf);

	sprintf (cbuf, "%0.2f", (acc_lim));
	SetWindowText (GetDlgItem (hWnd, IDC_ACC), cbuf);

	sprintf (cbuf, "%0.1f", (op_meco_alt*0.001));
	SetWindowText (GetDlgItem (hWnd, IDC_PER), cbuf);

	sprintf (cbuf, "%0.1f", (op_tgt*0.001));
	SetWindowText (GetDlgItem (hWnd, IDC_APO), cbuf);

	sprintf (cbuf, "%0.2f", (ap_tgt*0.001));
	SetWindowText (GetDlgItem (hWnd, IDC_APO1), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Enter");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_APO1_APPLY), cbuf);

	

	//////////////////
	if(lot_status == 1)
	{
		if (apop_status == 0)sprintf (cbuf, "ENABLED");
		else sprintf (cbuf, "ACTIVE");
	}
	else if (apop_status == 0)sprintf (cbuf, "OFF");
	else sprintf (cbuf, "ACTIVE");
	SetWindowText (GetDlgItem (hWnd, IDC_APOP), cbuf);

	/////////// launch timer
	sprintf (lotbuf, "%d", (lot_hh));
	if(lot_hh < 10) 
		{
		strcat(lotbuftemp,lotbuf);
		SetWindowText (GetDlgItem (hWnd, IDC_LOT_HH), lotbuftemp);
		strcpy(lotbuftemp, "0");
		}
	else SetWindowText (GetDlgItem (hWnd, IDC_LOT_HH), lotbuf);

	sprintf (lotbuf, "%d", (lot_mm));
	if(lot_mm < 10)
		{
		strcat(lotbuftemp,lotbuf);
		SetWindowText (GetDlgItem (hWnd, IDC_LOT_MM), lotbuftemp);
		strcpy(lotbuftemp, "0");
		}
	else SetWindowText (GetDlgItem (hWnd, IDC_LOT_MM), lotbuf);

	sprintf (lotbuf, "%d", (lot_ss));
	if(lot_ss < 10)
		{
		strcat(lotbuftemp,lotbuf);
		SetWindowText (GetDlgItem (hWnd, IDC_LOT_SS), lotbuftemp);
		strcpy(lotbuftemp, "0");
		}
	else SetWindowText (GetDlgItem (hWnd, IDC_LOT_SS), lotbuf);

	if(lot_status == 0)sprintf (cbuf, "OFF");
	else if(lot_status == 3)sprintf (cbuf, "OFF");
	else sprintf (cbuf, "ON");
	SetWindowText (GetDlgItem (hWnd, IDC_ALOT), cbuf);

	/////////////////////////////////////////

	//sprintf (cbuf, "N/A");
	//SetWindowText (GetDlgItem (hWnd, IDC_FAIR_APPLY), cbuf);

	if (eng_status == 0)sprintf (cbuf, "Jettison");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_JSTG_APPLY), cbuf);

	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
	v->SendBufferedKey(COMM_CALC_FRE);
	sprintf (cbuf, "%0.0f kg", fuel_res_est);
	SetWindowText (GetDlgItem (hWnd, IDC_FRE), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Apply");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_FRE_APPLY), cbuf);
	
	OBJHANDLE hSLC = v->GetAttachmentStatus(lpd_attach);
	if (hSLC)SetWindowText (GetDlgItem (hWnd, IDC_HDBK), "Apply");
	else SetWindowText (GetDlgItem (hWnd, IDC_HDBK), "N/A");

	if (edlo_sw == 0)sprintf (cbuf, "BOOST+LAND");
	else sprintf (cbuf, "LAND ONLY");
	SetWindowText (GetDlgItem (hWnd, IDC_EDLO), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Change");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_EDLO_APPLY), cbuf);

	/////// asds pos
	sprintf (cbuf, "Lng. %0.3f  Lat. %0.3f", (asds_core_lng*DEG), (asds_core_lat*DEG));
	SetWindowText (GetDlgItem (hWnd, IDC_ASDS), cbuf);

	/////// frv pos
	//sprintf (cbuf, "Lng. %0.3f  Lat. %0.3f", (frv_core_lng*DEG), (frv_core_lat*DEG));
	//SetWindowText (GetDlgItem (hWnd, IDC_FRV), cbuf);
}

bool SetLOT(char *str1, char *str2, char *str3)
{
	float new_hh;
	sscanf(str1, "%f", &new_hh);
	float new_mm;
	sscanf(str2, "%f", &new_mm);
	float new_ss;
	sscanf(str3, "%f", &new_ss);

	if (new_hh > 23) return false;
	else if (new_hh < 0) return false;
	else if (new_mm > 59) return false;
	else if (new_mm < 0) return false;
	else if (new_ss > 59) return false;
	else if (new_ss < 0) return false;
	else
		{
			lot_hh = (int)(new_hh);
			lot_mm = (int)(new_mm);
			lot_ss = (int)(new_ss);
			return true;
		}
	return true;
}

BOOL CALLBACK DlgProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	
	

	switch (uMsg) {
	case WM_INITDIALOG:
		InitDialog(hWnd);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) 
		{
		char cbuf[256];
		char chh[256];
		char cmm[256];
		char css[256];

		case IDCANCEL:
			oapiCloseDialog (hWnd);
			return TRUE;
		case IDC_FUEL_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_FUEL_RES), cbuf, 256);
			SetRF(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;

		case IDC_FRE_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_FRE), cbuf, 256);
			SetRF(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;
		
		case IDC_TGTNM_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_TGTNM), cbuf, 256);
			SetBBTgt(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;

		case IDC_TGTNM2_APPLY:
			//if (eng_status == 0)
			//{
			GetWindowText (GetDlgItem (hWnd, IDC_TGTNM2), cbuf, 256);
			SetTGT(0,cbuf,0);
			//}
			InitDialog(hWnd);
			return TRUE;

		case IDC_PLD_NM_APPLY:
			//if (pld_att_status < 20)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_PLD_NM), cbuf, 256);
			SetPLD(0,cbuf,0);
			//VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			//v->SendBufferedKey(COMM_CALC_MASS);
			InitDialog(hWnd);
			}
			//else if (pld_att_status == 1 && intfair_status == 1)
			//{
			//VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			//v->SendBufferedKey(OAPI_KEY_J);
			//InitDialog(hWnd);
			//}
			return TRUE;

		/*case IDC_PLD_JET_APPLY:
			if (pld_att_status > 0)
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(OAPI_KEY_J);
			InitDialog(hWnd);
			}
			return TRUE;*/
			
		case IDC_FB_STATUS_APPLY:
			if (eng_status == 0)
			{
			bchk = oapiGetObjectByName(refname);
			if (bchk == NULL)bchk = oapiGetBaseByName(hGref, refname);
			if (bchk == NULL)afs_status = 0;
			else if(afs_status == 0)afs_status = 1;
			else afs_status = 0;
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_INC_APPLY:
			GetWindowText (GetDlgItem (hWnd, IDC_INC), cbuf, 256);
			SetINC(0,cbuf,0);
			InitDialog(hWnd);
			return TRUE;

			case IDC_ACC_APPLY:
			GetWindowText (GetDlgItem (hWnd, IDC_ACC), cbuf, 256);
			SetACC(0,cbuf,0);
			InitDialog(hWnd);
			return TRUE;

			case IDC_PER_APPLY:
			GetWindowText (GetDlgItem (hWnd, IDC_PER), cbuf, 256);
			SetOPE(0,cbuf,0);
			InitDialog(hWnd);
			return TRUE;

			case IDC_APO_APPLY:
			GetWindowText (GetDlgItem (hWnd, IDC_APO), cbuf, 256);
			SetOAP(0,cbuf,0);
			InitDialog(hWnd);
			return TRUE;

			case IDC_APO1_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_APO1), cbuf, 256);
			SetAP(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_APOP_APPLY:
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(OAPI_KEY_V);
			}
			InitDialog(hWnd);
			return TRUE;

			//case IDC_FAIR_APPLY:
			//
			//return TRUE;

			case IDC_JSTG_APPLY:
			if(eng_status == 0)
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(OAPI_KEY_N);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_HDBK:
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			OBJHANDLE hLpad = v->GetAttachmentStatus(lpd_attach);
			if(hLpad) 
				{
				v = (VESSEL2*)oapiGetVesselInterface(hLpad);
				v->SendBufferedKey(OAPI_KEY_B);
				}
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_EDLO_APPLY:
			if(eng_status == 0)
			{
			if(edlo_sw == 0) edlo_sw = 1;
			else edlo_sw = 0;
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_LOT_APPLY:
			{
			GetWindowText (GetDlgItem (hWnd, IDC_LOT_HH), chh, 256);
			GetWindowText (GetDlgItem (hWnd, IDC_LOT_MM), cmm, 256);
			GetWindowText (GetDlgItem (hWnd, IDC_LOT_SS), css, 256);
			
			SetLOT(chh,cmm,css);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_ALOT_APPLY:
			{
			if (apop_status > 0 && lot_status == 0)lot_status = 0;
			else if(lot_status == 0)lot_status = 1;
			else if(lot_status == 1)lot_status = 0;
			else if(lot_status == 2)lot_status = 0;
			oapiDelAnnotation (hNote);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_REFRESH:
			InitDialog(hWnd);
			return TRUE;

		}
		
		break;
	}
	return oapiDefDialogProc (hWnd, uMsg, wParam, lParam);
}


void ss_tanker::clbkPreStep(double SimT, double SimDT, double MJD)
{
//////////// fuel payload
	pld_mass = GetPropellantMass(hProp_payload);
	if(fuel_trans_sw == 1)DoFuelTransfer();

////////// launch timer
if(lot_status == 1)DoLOT();
else if(lot_status == 2)DoLOT();

///////// max q trail
double dynp = GetDynPressure();
if(dynp > 22000)
	{
	pst_lvl = (dynp-22000)/3000;
	if (pst_lvl > 1)pst_lvl = 1;
	if(eng_status > 0)pst_lvl = 0;
	}
else pst_lvl = 0;


////////////////////////// l pad condensation
if (GetAttachmentStatus(lpd_attach) == NULL) pst_lvl2 = 0;
else if (oapiGetTimeAcceleration() > 10) pst_lvl2 = 0;
else if(eng_status == 0)pst_lvl2 = 0.7;


step_time = SimDT;///////
double simdt = SimDT;

double mth_lev = GetThrusterLevel(th_main[eng_status]);
if(eng_status == 1)
{
if (mth_lev == 0 && rcs_sw2 == 0) ConfigRCS2();
else if (mth_lev > 0 && rcs_sw2 == 1) RevertRCS2();

//if(mth_lev > 0)he_vent_status = 1;
//else if(mth_lev == 0 && he_vent_status == 1)
//	{
//		he_vent_timer = he_vent_timer + SimDT;
//		if (he_vent_timer < 6) he_vent_level = 1;
//		else
//		{
//			he_vent_level = 0;
//			he_vent_timer = 0;
//			he_vent_status = 0;
//		}
//	}
SetUpperExhaust();
}

if(mth_lev > 0)SetThrusterLevel(th_dump, 0);



if (calc_sw == 0)
{
CalcEmptyMass();
calc_sw = 1;
}

if (ap_status == 1) DoAscent();
else if (op_status == 1) DoOrbit();
apop_status = op_status+ap_status;
////////////////engine light
if (mth_lev > 0 && eng_status == 0)
	{
	double rndm = 0.15*(oapiRand());
	rndm = (mth_lev*0.9) - rndm;
	if (rndm < 0)rndm = 0;
	else if (rndm > 1)rndm = 1;
	spt_light->Activate(true);
	spt_light->SetIntensity(rndm);

	}
	else spt_light->Activate(false);
//////////
	if(eng_status == 1)
	{
	js_pu = 0;
	js_pd = 0;
	js_br = 0;
	js_bl = 0;
	js_yl = 0;
	js_yr = 0;
	// get joystick input
	int rcs_mode = GetAttitudeMode();
	if(rcs_mode != RCS_LIN && oapiGetFocusObject() == GetHandle())
	{
	js_pu = GetManualControlLevel(THGROUP_ATT_PITCHUP,MANCTRL_ROTMODE,MANCTRL_ANYDEVICE);
	js_pd = GetManualControlLevel(THGROUP_ATT_PITCHDOWN,MANCTRL_ROTMODE,MANCTRL_ANYDEVICE);
	js_br = GetManualControlLevel(THGROUP_ATT_BANKRIGHT,MANCTRL_ROTMODE,MANCTRL_ANYDEVICE);
	js_bl = GetManualControlLevel(THGROUP_ATT_BANKLEFT,MANCTRL_ROTMODE,MANCTRL_ANYDEVICE);
	js_yl = GetManualControlLevel(THGROUP_ATT_YAWLEFT,MANCTRL_ROTMODE,MANCTRL_ANYDEVICE);
	js_yr = GetManualControlLevel(THGROUP_ATT_YAWRIGHT,MANCTRL_ROTMODE,MANCTRL_ANYDEVICE);
	}
	//sprintf(oapiDebugString(), "js_pu %0.2f js_pd %0.2f js_br %0.2f js_bl %0.2f js_yl %0.2f js_yr %0.2f", js_pu, js_pd, js_br, js_bl, js_yl, js_yr);
	pitch_level = js_pu - js_pd; // total level from kbd + joystick
	bank_level = js_bl - js_br;
	yaw_level = js_yl - js_yr;

	/////////// autopilots ////////////////////////
	if (skydive_sw > 0) 
	{
		DoSkydiveTargeting();
		retroland_sw = 0;
		reentry_sw = 0;
		fbw_reentry_sw = 0;
		apop_status = 0;
		ap_status = 0;
		op_status = 0;
	}
	if (backflip_sw > 0)
	{
		DoBackFlip();
		retroland_sw = 0;
		reentry_sw = 0;
		fbw_reentry_sw = 0;
		apop_status = 0;
		ap_status = 0;
		op_status = 0;
	}
	if(retroland_sw > 0)
	{
		DoRetroLand();
		backflip_sw = 0;
		skydive_sw = 0;
		fbw_reentry_sw = 0;
		reentry_sw = 0;
		apop_status = 0;
		ap_status = 0;
		op_status = 0;
	}
	
	if(reentry_sw == 1)
	{
		fbw_reentry_sw = 0;
		backflip_sw = 0;
		skydive_sw = 0;
		retroland_sw = 0;
		apop_status = 0;
		ap_status = 0;
		op_status = 0;
		DoReentry();
	}

	if(fbw_reentry_sw == 1)
	{
		reentry_sw = 0;
		backflip_sw = 0;
		skydive_sw = 0;
		retroland_sw = 0;
		apop_status = 0;
		ap_status = 0;
		op_status = 0;
		DoFBWReentry();
	}

	///////////////////////////////
	SetADCLevel();
	

	////////
	SetMeshMat();
	}
	else
	{
	backflip_sw = 0;
	skydive_sw = 0;
	retroland_sw = 0;
	reentry_sw = 0;
	fbw_reentry_sw = 0;
	}

	
////// landing / takeoff dust pstream  /////////////////////////////

	double altdust = GetAltitude(ALTMODE_GROUND);
	if(altdust < 70)
	{
	double sin_pitch = sin(GetPitch());
	if(sin_pitch > 0.7) 
		{
		altdust = altdust/sin_pitch;
		dust_level = GetThrusterGroupLevel(THGROUP_MAIN);
		SetThrusterRef(th_dust, _V(0,0,-altdust));
		}
	else dust_level = 0;
	}
	else dust_level = 0;
	SetThrusterLevel(th_dust, dust_level);
	

//////////////////// ANIMATIONS  ////////////////////////////////////////////
	if (legs_status == 0) legs_proc = 0;
	else if (legs_status == 1) 
	{
		double da = SimDT * LEG_SPEED;
		legs_proc = legs_proc + da;
		if (legs_proc > 1) 
		{
			legs_proc = 1;
			legs_status = 2;
		}
		DoTDPoints();
	}
	else if (legs_status == 2) legs_proc = 1;
	else if (legs_status == 3) 
	{
		double da = SimDT * LEG_SPEED;
		legs_proc = legs_proc - da;
		if (legs_proc < 0) 
		{
			legs_proc = 0;
			legs_status = 0;
		}
		DoTDPoints();
	}
	SetAnimation (anim_legs, legs_proc);


}



void ss_tanker::clbkPostStep (double simt, double simdt, double mjd)
{
post_step = simdt;
/////////////// touchdown shutoffs ////////////////////////
if(backflip_sw == 1)
	{
	if (GroundContact() == true)
		{
		//main_throttle_level = 0;
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
		backflip_sw = 0;
		wing_trim = 0;
		KillAllRCS2();
		ActivateNavmode(NAVMODE_KILLROT);
		}
	}

if(retroland_sw > 0)
	{
	if (GroundContact() == true)
		{
		//main_throttle_level = 0;
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
		retroland_sw = 0;
		wing_trim = 0;
		KillAllRCS2();
		ActivateNavmode(NAVMODE_KILLROT);
		}
	}
//////////// 
de_first_frame_sw = 1;

///////////////////////////////////////////
///////////// ACCEL LIMITER ///////////////
//
//double max_main_thrust = GetThrusterMax(th_main[0]);
//	if(eng_status == 1)
//	{
//		if(uppermaincfg == 0)
//		{
//			max_main_thrust = GetThrusterMax(th_main[1]);
//			max_main_thrust = max_main_thrust + GetThrusterMax(th_main[2]);
//		}
//		else if(uppermaincfg == 1)
//		{
//			max_main_thrust = GetThrusterMax(th_main[1]);
//		}
//		else if(uppermaincfg == 2)
//		{
//			max_main_thrust = GetThrusterMax(th_main[2]);
//		}
//
//	}
//	double max_main_acc = max_main_thrust/GetMass();
//	double max_main_level = acc_lim/max_main_acc;
//	if(GetThrusterGroupLevel(THGROUP_MAIN) > max_main_level)
//	{
//		SetThrusterGroupLevel(THGROUP_MAIN, max_main_level);
//	}
}

void ss_tanker::DoTDPoints (void)
{
	if(eng_status == 1)
	{
	double extension = -2.15 * legs_proc;
	double default_pos = -16.78;

	for(int i=0;i<6;i++)
		{
		tdvtx_upper[i].pos.z = default_pos + extension;
		}
	SetTouchdownPoints(tdvtx_upper, ntdvtx_upper);
	}
}


bool ss_tanker::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);
	//int w15 = (int)(424*d);


	

	if (eng_status == 1)
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	double m1 = GetMass();
	double m2 = GetPropellantMass(hProp2);
	double fraction = (m1/(m1-m2));
	double isp_curr = 0;
	{
		isp_curr = 0.5*GetThrusterIsp(th_main[1]);
		isp_curr = isp_curr + 0.5*GetThrusterIsp(th_main[2]);

	}
	double zdv = (isp_curr*(log(fraction)));

	sprintf(abuf,"Remaining dV = %0.2f m/s", zdv);
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	double m1 = GetMass();
	double m2 = 0;
	if (eng_status == 0)m2 = GetPropellantMass(hProp1);
	else m2 = GetPropellantMass(hProp2);
	double test = (m1/(m1-m2));
	double isp_curr = GetThrusterIsp(th_main[eng_status]);
	double zdv = (isp_curr*(log(test)));

	sprintf(abuf,"Remaining dV = %0.2f m/s", zdv);
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	{
	char abuf[256];
	skp->SetTextColor(0x00FFFFFF);
	if(hud_display_info == 0)sprintf(abuf,"C = Display Info: LAUNCH");
	else if(hud_display_info == 1)sprintf(abuf,"C = Display Info: ORBITAL");
	else if(hud_display_info == 2)sprintf(abuf,"C = Display Info: REENTRY");	
	else if(hud_display_info == 3)sprintf(abuf,"C = Display Info: LANDING");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"E = Set Camera View");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	

if(hud_display_info == 0)// launch
{
	/*{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Next Payload: %s", pldname[pld_att_status-1]);
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}*/

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Fuel Payload: %0.0fkg", (GetPropellantMass(hProp_payload)));
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}


	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Tgt.Orb: Inc.%0.2f ApA.%dkm PeA.%dkm",abs(tgt_inc*DEG),int(op_tgt/1000),int(op_meco_alt/1000));
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}

	if (eng_status == 0)
	{
		{
		skp->SetTextColor(0x00FFFFFF);
		char abuf[256];
		sprintf(abuf,"1st Stg: Fuel Res.%dkg ApA.%dkm",int(fuel_reserve),int(ap_tgt/1000));
		skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
		}

		if (afs_status == 0)
		{
		skp->SetTextColor(0x00FFFFFF);
		char abuf[256];
		if(edlo_sw == 1)sprintf(abuf,"1st Stg: EDL ONLY OFF: TGT. %s", refname);
		else sprintf(abuf,"1st Stg: FLYBACK-EDL OFF: TGT. %s", refname);
		skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
		}
		else
		{
		skp->SetTextColor(0x0066FFFF);
		char abuf[256];
		if(edlo_sw == 1)sprintf(abuf,"1st Stg: EDL ONLY ON: Tgt. %s", refname);
		else sprintf(abuf,"1st Stg: FLYBACK-EDL ON: Tgt. %s", refname);
		skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
		}
	}
	else
	{
		skp->SetTextColor(0x0066FFFF);
		char abuf[256];
		sprintf(abuf,"Booster Jettisoned");
		skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
		skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"K = Launch Control Panel On/Off");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}


	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if (eng_status == 0)sprintf(abuf,"N = Jettison Booster");
	else 
	{
		sprintf(abuf,"Booster Jettisoned");
		skp->SetTextColor(0x0066FFFF);
	}
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}

	if ((ap_status + op_status) > 0)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Launch Autopilot Active");
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"V = Start Launch Autopilot");
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}


}

else if(hud_display_info == 1)// orbital display
{

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Fuel Payload: %0.0fkg", (GetPropellantMass(hProp_payload)));
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	if(fuel_trans_sw > 0)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"J = Fuel Transfer On/Off");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	if(fuel_trans_sw > 0)skp->SetTextColor(0x0066FFFF);

	char dbuf[256];
	if(fuel_source_sw == 0)sprintf(dbuf,"M = Select Source Tank:PAYLOAD", tank_index);
	else if(fuel_source_sw == 1)sprintf(dbuf,"M = Select Source Tank:MAIN", tank_index);
	skp->Text ((10+roxl), (w4+royl), dbuf,strlen(dbuf));

	char abuf[256];
	sprintf(abuf,"W = Select Target Tank by Index: %d", tank_index);
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	char bbuf[256];
	sprintf(bbuf,"Target Tank Total Capacity: %dkg", int(tank_max));
	skp->Text ((10+roxl), (w6+royl), bbuf,strlen(bbuf));
	char cbuf[256];
	sprintf(cbuf,"Target Tank Current Level : %dkg", int(tank_level));
	skp->Text ((10+roxl), (w7+royl), cbuf,strlen(cbuf));
	}

	
	

	/*{
	skp->SetTextColor(0x00FFFFFF);
	if(lamp_status > 0)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"M = Fwd.Lamp ");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}*/


}
else if(hud_display_info == 2)// reentry display
{
	VECTOR3 gspeedv;
	GetGroundspeedVector(FRAME_HORIZON,gspeedv);
	gspeedv.y = 0;
	double gspeed = length(gspeedv);

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(adcs_reentry_mode == 0) sprintf(abuf,"Ctrl+A = Aerofoil Mode: FLIGHT");
	else sprintf(abuf,"Ctrl+A = Aerofoil Mode: REENTRY");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char cbuf[256];
	double wtf = 22;
	if(wing_trim < 0)wtf = 42.5;
	sprintf(cbuf,"Up/Dn/Delete = Trim %0.1f%% : Estimated AoA %0.1f ",(wing_trim*100), 58+(wing_trim * wtf));
	skp->Text ((10+roxl), (w3+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x0066FFFF);
	DWORD adcm = GetADCtrlMode(); 
	char cbuf[256];
	if(adcm == 0) sprintf(cbuf,"Alt+NumPad/ = Aerofoil Steering: DISABLED");
	else 
		{
		skp->SetTextColor(0x00FFFFFF);
		sprintf(cbuf,"Alt+NumPad/ = Aerofoil Steering: ENABLED");
		}
	skp->Text ((10+roxl), (w4+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char cbuf[256];
	if(disable_rcs == 0) sprintf(cbuf,"D = Aft Yaw RCS: ENABLED");
	else 
		{
		sprintf(cbuf,"D = Aft Yaw RCS: DISABLED");
		skp->SetTextColor(0x0066FFFF);
		}
	skp->Text ((10+roxl), (w5+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	if(GetThrusterLevel(th_dump) > 0)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"B = 2nd Stage Main Fuel Dump");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Ctrl+K = Enter Landing Target: %s", tgtname);
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	if (reentry_sw == 1)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"U = Auto Reentry On/Off");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}

	if (reentry_sw == 1)
	{
		{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	if(tgt_exists == true)sprintf(abuf,"Tgt.Range %0.1fkm Crossrange Error %0.1fkm", (hud_tgt_range*0.001), (hud_crossrange_error*0.001));
	else sprintf(abuf,"NO TARGET");
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
		}

		{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	if(tgt_exists == true)
			{
	sprintf(abuf,"Groundspeed %0.1f Required Acc. %0.2f Current Acc. %0.2f",gspeed, hud_hacc_req, hud_hacc_current);
	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
			}
		}
	}

	if (reentry_sw == 0)
	{
	skp->SetTextColor(0x00FFFFFF);
	if (fbw_reentry_sw == 1)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"F = Fly-by-Wire Reentry On/Off [Shift + Arrows = Bank/AoA]");
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}

	if (fbw_reentry_sw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Bank %0.1f AoA %0.1f ", fbw_bank, fbw_aoa);
	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));

	if(tgt_exists == true)
		{
		sprintf(abuf,"Groundspeed %0.1f Required Acc. %0.2f Current Acc. %0.2f ",gspeed, hud_hacc_req, hud_hacc_current);
		skp->Text ((10+roxl), (w11+royl), abuf,strlen(abuf));

		sprintf(abuf,"Tgt.Range %0.1fkm Crossrange Error %0.1fkm", (hud_tgt_range*0.001), (hud_crossrange_error*0.001));
		skp->Text ((10+roxl), (w12+royl), abuf,strlen(abuf));
		}
	else
		{
		sprintf(abuf,"Groundspeed %0.1f Acc. %0.2f", gspeed, hud_hacc_current);
		skp->Text ((10+roxl), (w11+royl), abuf,strlen(abuf));

		sprintf(abuf,"NO TARGET");
		skp->Text ((10+roxl), (w12+royl), abuf,strlen(abuf));
		}
	}

}
else if(hud_display_info == 3)// landing display
{
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(adcs_reentry_mode == 0) sprintf(abuf,"Ctrl+A = Aerofoil Mode: FLIGHT");
	else 
		{
		sprintf(abuf,"Ctrl+A = Aerofoil Mode: REENTRY");
		skp->SetTextColor(0x0066FFFF);
		}
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char cbuf[256];
	double wtf = 22;
	if(wing_trim < 0)wtf = 42.5;
	sprintf(cbuf,"Up/Dn/Delete = Trim %0.1f%% : Estimated AoA %0.1f ",(wing_trim*100), 58+(wing_trim * wtf));
	skp->Text ((10+roxl), (w3+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x0066FFFF);
	DWORD adcm = GetADCtrlMode(); 
	char cbuf[256];
	if(adcm == 0) sprintf(cbuf,"Alt+NumPad/ = Aerofoil Steering: DISABLED");
	else 
		{
		skp->SetTextColor(0x00FFFFFF);
		sprintf(cbuf,"Alt+NumPad/ = Aerofoil Steering: ENABLED");
		}
	skp->Text ((10+roxl), (w4+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char cbuf[256];
	if(disable_rcs == 0) sprintf(cbuf,"D = Aft Yaw RCS: ENABLED");
	else 
		{
		sprintf(cbuf,"D = Aft Yaw RCS: DISABLED");
		skp->SetTextColor(0x0066FFFF);
		}
	skp->Text ((10+roxl), (w5+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Ctrl+J = Use Landing Reserve Fuel: %dkg",int(GetPropellantMass(hProp_reserve)) );
	if(uppermainprop == 1)skp->SetTextColor(0x0066FFFF);
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}

	{
	if(raptor_sl_selection == 3)skp->SetTextColor(0x00FFFFFF);
	else skp->SetTextColor(0x0066FFFF);
	char xbuf[256];
	if(uppermaincfg == 0)sprintf(xbuf,"Ctrl+N = Raptor Vac/SL Selection: Vac+SL(%d)", raptor_sl_selection+3);
	else if(uppermaincfg == 1)
		{
		skp->SetTextColor(0x0066FFFF);
		sprintf(xbuf,"Ctrl+N = Raptor Vac/SL Selection: SL(%d)", raptor_sl_selection);
		}
	else if(uppermaincfg == 2)
		{
		skp->SetTextColor(0x0066FFFF);
		sprintf(xbuf,"Ctrl+N = Raptor Vac/SL Selection: Vac(3)");
		}
	skp->Text ((10+roxl), (w7+royl), xbuf,strlen(xbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char xbuf[256];
	if(raptor_sl_selection == 3)sprintf(xbuf,"Ctrl+1/2/3 = Raptor SL Selection:3");
	else if(raptor_sl_selection == 2)
		{
		skp->SetTextColor(0x0066FFFF);
		sprintf(xbuf,"Ctrl+1/2/3 = Raptor SL Selection:2");
		}
	else if(raptor_sl_selection == 1)
		{
		skp->SetTextColor(0x0066FFFF);
		sprintf(xbuf,"Ctrl+1/2/3 = Raptor SL Selection:1");
		}
	skp->Text ((10+roxl), (w8+royl), xbuf,strlen(xbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Ctrl+K = Enter Landing Target: %s", tgtname);
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}

	{
	char cbuf[256];
	if (backflip_sw > 0)
		{
		sprintf(cbuf,"Ctrl+B = Skydive and Land On : LANDING BURN");
		skp->SetTextColor(0x006666FF);
		}
	else if (skydive_sw == 0) 
		{
		sprintf(cbuf,"Ctrl+B = Skydive and Land Off");
		skp->SetTextColor(0x00FFFFFF);
		}
	else if (skydive_sw > 0)
		{
		sprintf(cbuf,"Ctrl+B = Skydive and Land On: Tgt.Hdg. %d: Tgt.Range %0.1fkm :Flt.Hdg. %d : Flt.Range %0.1fkm", (int)(hud_tgt_az*DEG), (hud_tgt_range*0.001), (int)(hud_gspd_az*DEG) , (hud_ves_range*0.001));
		skp->SetTextColor(0x0066FFFF);
		}
	skp->Text ((10+roxl), (w10+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char cbuf[256];
	if (retroland_sw == 0) sprintf(cbuf,"Ctrl+V = RetroLanding OFF");
	else
		{
		sprintf(cbuf,"Ctrl+V = RetroLanding ON");
		skp->SetTextColor(0x0066FFFF);
		}
	skp->Text ((10+roxl), (w11+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(legs_status == 0) sprintf(abuf,"G = Landing Gear");
	else 
		{
		sprintf(abuf,"G = Landing Gear");
		skp->SetTextColor(0x0066FFFF);
		}
	skp->Text ((10+roxl), (w12+royl), abuf,strlen(abuf));
	}

}

return true;
}

void ss_tanker::RevertRCS2 (void)
{
	KillAllRCS2();

	DelThrusterGroup(htgrppup2,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown2,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpyright2,THGROUP_ATT_YAWRIGHT, false);
	DelThrusterGroup(htgrpyleft2,THGROUP_ATT_YAWLEFT, false);
	DelThrusterGroup(htgrpbright2,THGROUP_ATT_BANKRIGHT, false);
	DelThrusterGroup(htgrpbleft2,THGROUP_ATT_BANKLEFT, false);

	th_group[0] = th_rcs[10];
	htgrppup2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[11];
	htgrppdown2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[9];
	htgrpyright2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[8];
	htgrpyleft2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[14];
	th_group[1] = th_rcs[15];
	htgrpyright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[12];
	th_group[1] = th_rcs[13];
	htgrpyleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	rcs_sw2 = 0;
}

void ss_tanker::ConfigRCS2 (void)
{
	KillAllRCS2();

	DelThrusterGroup(htgrppup2,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown2,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpyright2,THGROUP_ATT_YAWRIGHT, false);
	DelThrusterGroup(htgrpyleft2,THGROUP_ATT_YAWLEFT, false);
	DelThrusterGroup(htgrpbright2,THGROUP_ATT_BANKRIGHT, false);
	DelThrusterGroup(htgrpbleft2,THGROUP_ATT_BANKLEFT, false);

	th_group[0] = th_rcs[20];
	th_group[1] = th_rcs[21];
	th_group[2] = th_rcs[22];
	htgrppup2 = CreateThrusterGroup (th_group, 3, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[23];
	th_group[1] = th_rcs[24];
	th_group[2] = th_rcs[25];
	htgrppdown2 = CreateThrusterGroup (th_group, 3, THGROUP_ATT_PITCHDOWN);

	if(disable_rcs == 0)
	{
	th_group[0] = th_rcs[18];
	th_group[1] = th_rcs[19];
	htgrpyright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[17];
	htgrpyleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);
	}
	else if(disable_rcs == 1)
	{
	th_group[0] = th_rcs[19];
	htgrpyright2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[17];
	htgrpyleft2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);
	}

	th_group[0] = th_rcs[28];
	th_group[1] = th_rcs[29];
	htgrpyright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[26];
	th_group[1] = th_rcs[27];
	htgrpyleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	rcs_sw2 = 1;
}

void ss_tanker::ConfigLateralRCS (void)
{
	if(rcs_sw2 == 1)
	{
	KillAllRCS2();

	
	DelThrusterGroup(htgrpyright2,THGROUP_ATT_YAWRIGHT, false);
	DelThrusterGroup(htgrpyleft2,THGROUP_ATT_YAWLEFT, false);

	if(disable_rcs == 0)
	{
	th_group[0] = th_rcs[18];
	th_group[1] = th_rcs[19];
	htgrpyright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[17];
	htgrpyleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);
	}
	else if(disable_rcs == 1)
	{
	th_group[0] = th_rcs[19];
	htgrpyright2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[17];
	htgrpyleft2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);
	}
	}
}




void ss_tanker::JettisonStg (void)
{
  VESSELSTATUS vs;
  //char name[256];
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (STG1_OFF, vs.rpos);
  GlobalRot (_V(0,0,-7), rvel);
  vs.rvel = vs.rvel + rvel;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  vs.vrot.x = vs.vrot.x + (2*RAD);
  vs.vrot.z = vs.vrot.z;
  vs.vrot.y = vs.vrot.y;
  char myname[256];
  oapiGetObjectName(hMe,myname,30);
  char stgname[256];
  strcpy(stgname, myname);
  strcat(stgname, "_Booster");
  OBJHANDLE hStg1;
  hStg1 = oapiCreateVessel (stgname, "spacex_starship/spacex_booster", vs);
  double fuel_trans = GetPropellantMass(hProp1);
  VESSEL2* v =(VESSEL2*)oapiGetVesselInterface(hStg1);
  //VESSEL *v = oapiGetVesselInterface (hStg1);
 // DWORD nProp = v->GetPropellantCount ();
 // for (DWORD j = 0; j < nProp; j++)
	{
	PROPELLANT_HANDLE ph = v->GetPropellantHandleByIndex(0);// (j);
//	if(j == 0)v->SetPropellantMass(ph,fuel_trans);
	v->SetPropellantMass(ph,fuel_trans);
//	else
//		{
//		double maxfuel = v->GetPropellantMaxMass(ph);
//		v->SetPropellantMass(ph,maxfuel);
//		}
	}
  v->clbkConsumeBufferedKey(COMM_SEND_STR,false,refname);
  if(afs_status == 1)
  {
	v->clbkConsumeBufferedKey(COMM_SEND_AFS,false,refname);
	if(edlo_sw == 1)v->clbkConsumeBufferedKey(COMM_SEND_EDLO,false,refname);
  }

  VESSELSTATUS2 vss;
  memset (&vss, 0, sizeof(vss));
  vss.version = 2;
  GetStatusEx(&vss);
  Local2Rel (STG1STG2_OFF, vss.rpos);
  DefSetStateEx (&vss);

	DelMesh(1);
	eng_status = 1;


	CalcEmptyMass();
	ConfigMain();

	PlayVesselWave(MyID, SEP, 0, 200, 25000);
	PlayVesselWave(MyID, SEPST, 0, 225, 0);

}

void ss_tanker::ConfigMain (void)
{
	KillAllRCS2();
	SetThrusterLevel(th_main[0],0);
	DelThrusterGroup(htgrpmain,THGROUP_MAIN, false);
	DelThrusterGroup(htgrpbleft2,THGROUP_ATT_BANKLEFT, false);
	DelThrusterGroup(htgrpbright2,THGROUP_ATT_BANKRIGHT, false);
	DelThrusterGroup(htgrppup2,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown2,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpyleft2,THGROUP_ATT_YAWLEFT, false);
	DelThrusterGroup(htgrpyright2,THGROUP_ATT_YAWRIGHT, false);

	DelPropellantResource(hProp1);
	//SetPropellantMass(hProp1, 0);
	SetDefaultPropellantResource(hProp2);

	/// upper rotation
	th_group[0] = th_rcs[10];
	htgrppup2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[11];
	htgrppdown2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[9];
	htgrpyright2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[8];
	htgrpyleft2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[14];
	th_group[1] = th_rcs[15];
	htgrpbright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[12];
	th_group[1] = th_rcs[13];
	htgrpbleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);
	
	if (uppermaincfg == 0)
	{
	th_group[0] = th_main[1];
	th_group[1] = th_main[2];
	htgrpmain = CreateThrusterGroup (th_group, 2, THGROUP_MAIN);
	}
	else if (uppermaincfg == 1)
	{
	th_group[0] = th_main[1];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);
	}
	else if (uppermaincfg == 2)
	{
	th_group[0] = th_main[2];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);
	}

	// upper translation
	th_group[0] = th_rcs[30];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[31];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_BACK);

	th_group[0] = th_rcs[32];
	th_group[1] = th_rcs[33];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_UP);

	th_group[0] = th_rcs[34];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[35];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[36];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_RIGHT);

	////////////////////////////////////////

	/*for (int k = 0; k < 20; k++)
	{
	VECTOR3 at_dir, at_rot, at_pos;
	GetAttachmentParams(pld_attach[k],at_pos,at_dir,at_rot);
	at_pos = at_pos - STG1STG2_OFF;
	SetAttachmentParams(pld_attach[k],at_pos,at_dir,at_rot);
	pld_attach_pos_init[k] = pld_attach_pos_init[k] - STG1STG2_OFF;
	}*/

	SetAttachmentParams(lpd_attach, _V(0,0,-16.782), _V(0,0,-1), _V(0,1,0));
	SetAttachmentParams(arm_attach, _V(0,0,23.55), _V(0,0,-1), _V(0,1,0));

	for (UINT i = 0; i < GetMeshCount(); i++)
	{
	ShiftMesh(i,-STG1STG2_OFF);
	}

	SetSize (25);
	SetPMI(_V(100,100,12));
	SetCrossSections (_V(426,426,76));
	SetCW(0.5,1,1,1);
	SetRotDrag (_V(0.5,0.4,0.5));
	SetTouchdownPoints ( tdvtx_upper, ntdvtx_upper);

	//ClearAirfoilDefinitions ();
	DelAirfoil(hVABooster);
	DelAirfoil(hHABooster);
	/////AirFoils
	hVA = CreateAirfoil3 (LIFT_VERTICAL, hull_wing_pos, VLiftCoeff, 0, hull_chord, hull_area, hull_aspect_ratio);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, hull_wing_pos, HLiftCoeff, 0, hull_chord, hull_area, hull_aspect_ratio);
	hPortCanard = CreateAirfoil3 (LIFT_VERTICAL, port_canard_pos, VLiftCoeff, 0, sqrt(canard_area), canard_area, aspect_ratio);
	hStbdCanard = CreateAirfoil3 (LIFT_VERTICAL, stbd_canard_pos, VLiftCoeff, 0, sqrt(canard_area), canard_area, aspect_ratio);
	hPortWing = CreateAirfoil3 (LIFT_VERTICAL, port_wing_pos, VLiftCoeff, 0, sqrt(wing_area), wing_area, aspect_ratio);
	hStbdWing = CreateAirfoil3 (LIFT_VERTICAL, stbd_wing_pos, VLiftCoeff, 0, sqrt(wing_area), wing_area, aspect_ratio);
	hWing_Vertical = CreateAirfoil3 (LIFT_HORIZONTAL, wing_vertical_pos, HLiftCoeff, 0, sqrt(wing_area), wing_area*0.5, aspect_ratio);
	////////////////

	//beac.pos =  &new_beac_pos;
	//spotlight1->SetPosition(new_beac_pos);
}

void ss_tanker::SetPropReserve (void)
{
	if(uppermainprop == 0)
	{
		SetThrusterResource (th_main[1], hProp2);
		SetThrusterResource (th_main[2], hProp2);
		for (int i = 8; i < 37; i++)
		{
		SetThrusterResource(th_rcs[i],hProp2);
		}
		SetThrusterResource (th_dust, hProp2);
	}
	else
	{
		SetThrusterResource (th_main[1], hProp_reserve);
		SetThrusterResource (th_main[2], hProp_reserve);
		for (int i = 8; i < 37; i++)
		{
		SetThrusterResource(th_rcs[i],hProp_reserve);
		}
		SetThrusterResource (th_dust, hProp_reserve);
	}
	//ConfigUpperMain();
}

void ss_tanker::ConfigUpperMain (void)
{
	
	DelThrusterGroup(htgrpmain,THGROUP_MAIN, false);
	
	if (uppermaincfg == 0)
	{
	th_group[0] = th_main[1];
	th_group[1] = th_main[2];
	htgrpmain = CreateThrusterGroup (th_group, 2, THGROUP_MAIN);
	}
	else if (uppermaincfg == 1)
	{
	th_group[0] = th_main[1];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);
	SetThrusterLevel(th_main[2],0);
	}
	else if (uppermaincfg == 2)
	{
	th_group[0] = th_main[2];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);
	SetThrusterLevel(th_main[1],0);
	}

}

void ss_tanker::KillAllRCS2()
{
SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
}

void ss_tanker::CalcEmptyMass(void)
{
	double totalmass = STG2DRYMASS;
	
	if (eng_status == 0) totalmass = totalmass +  STG1DRYMASS;

	/*pld_mass = 0;
	for (int i = 0; i < 20; i++)
	{
	OBJHANDLE hpld = GetAttachmentStatus(pld_attach[i]);
	if (hpld == NULL) continue;
	else
		{
		double new_pld_mass = oapiGetMass(hpld);
		totalmass = totalmass + new_pld_mass;
		pld_mass = pld_mass + new_pld_mass;
		oapiGetObjectName(hpld, pldname[i], 30);
		}
	
	}*/
	SetEmptyMass(totalmass);
}
void ss_tanker::CalcFuelRes(void)
{
	//////// check if landed and tgt valid
	VESSELSTATUS vs;
	GetStatus(vs);
	if (vs.status == 0) 
	{
		fuel_res_est = 0;
		return;
	}
	OBJHANDLE bbtgt = oapiGetObjectByName(refname);
	if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, refname);
	if (bbtgt == NULL)
	{		
		fuel_res_est = 0;
		CalcAsdsPos();
		return;
	}
	///////////////
	VECTOR3 tgt_gpos = _V(0,0,0);
	oapiGetGlobalPos(bbtgt,&tgt_gpos);
	VECTOR3 tgt_lpos = _V(0,0,0);
	Global2Local(tgt_gpos, tgt_lpos);
	VECTOR3 tgt_hpos = _V(0,0,0);
	HorizonRot(tgt_lpos, tgt_hpos);
	tgt_hpos.y = 0;

	double pld_factor = (100000 - pld_mass) * 0.25;

	double apo_factor = (250000 - ap_tgt);
	if (apo_factor < 0)apo_factor = 0;
	apo_factor = (apo_factor*3.5)+250000;

	double apo_factor2 = (125000 - ap_tgt)*0.35;

	double azm_angle = CalcAzimuth();
	VECTOR3 imp_hpos = _V(0,0,0);
	imp_hpos.x = sin(azm_angle)*apo_factor;
	imp_hpos.z = cos(azm_angle)*apo_factor;

	double tgt_imp_dist = dist(imp_hpos,tgt_hpos);
	VECTOR3 imp_mid_pnt = _V((imp_hpos.x*0.25),0,(imp_hpos.z*0.25));
	VECTOR3 imp_mid_to_tgt = tgt_hpos - imp_mid_pnt;
	normalise(imp_mid_to_tgt);
	normalise(imp_mid_pnt);
	double cos_dl = dotp(imp_mid_to_tgt, imp_mid_pnt);

	double dl_factor = -cos_dl*2000;
	
	if (edlo_sw == 0)
	{
	fuel_res_est =  7*(26500 + (tgt_imp_dist*0.01*3) + dl_factor) + apo_factor2 + pld_factor + 1000;
	fuel_res_est = 0.85 * fuel_res_est;
	if (fuel_res_est < 7*29000)fuel_res_est = 7*29000;
	}
	else
	{
	fuel_res_est = 40000; // 150000
	//fuel_res_est =  21500 + (tgt_imp_dist*0.01*0.5);
	//if (fuel_res_est < 22000)fuel_res_est = 22000;
	}
	if (fuel_res_est > 7*80000)fuel_res_est = 7*80000;
	//sprintf(oapiDebugString(), "fre %0.1f tid %0.1f apf %0.1f", fuel_res_est, tgt_imp_dist, apo_factor);

	CalcAsdsPos();

}
void ss_tanker::CalcAsdsPos(void)
{
// stg1 ap at meco in km
	double algo_ap = ap_tgt * 0.001;
	double algo_ap2 = algo_ap * algo_ap;

	double core_range_15600 = 1030*((-2.092883971E-2*algo_ap2) + (6.11044315*algo_ap) + 158.2077267);
	double core_range_6000 = 1030*((-1.639211208E-2*algo_ap2) + (5.162726067*algo_ap) + 255.785993);

	double delta_range_kg = (core_range_15600 - core_range_6000)/9600;
	double pld_15600_diff = 15600 - (0.15*pld_mass);
	double core_range = core_range_15600 - (delta_range_kg*pld_15600_diff);
	//double hi_ap_factor = (ap_tgt - 200000)*4;
	//if(hi_ap_factor > 0)core_range = core_range - hi_ap_factor;

	VESSEL2* v;
	OBJHANDLE hPad = GetAttachmentStatus(lpd_attach);
	if(hPad != NULL) v = (VESSEL2*)oapiGetVesselInterface(hPad);
	else v = (VESSEL2*)oapiGetVesselInterface(GetHandle());

	OBJHANDLE hEarth = v->GetGravityRef();
	VECTOR3 falcon_pos = _V(0,0,0);
	v->GetRelativePos(hEarth,falcon_pos);
	VECTOR3 falcon_pos_nm = falcon_pos;
	normalise(falcon_pos_nm);

	double azm_angle = CalcAzimuth();
	double algo_range = core_range*0.001;
	double algo_range2 = algo_range*algo_range;
	double azm_range_factor = RAD*0.01*((8.654755941E-3*algo_ap2) - (2.433711805*algo_ap) + 258.983414);//(0.15*RAD) + (RAD*algo_ap*0.006);
	azm_angle = azm_angle + azm_range_factor;

	VECTOR3 vel_vector_hz = _V((sin(azm_angle)),0,(cos(azm_angle)));
	VECTOR3 vel_vector_hz_nm = vel_vector_hz;
	normalise(vel_vector_hz_nm);

	VECTOR3 vel_vector_loc = _V(0,0,0);
	v->HorizonInvRot(vel_vector_hz, vel_vector_loc);
	VECTOR3 vel_vector_loc_nm = vel_vector_loc;
	normalise(vel_vector_loc_nm);

	VECTOR3 vel_vector = _V(0,0,0);
	v->GlobalRot(vel_vector_loc, vel_vector);
	VECTOR3 vel_vector_nm = vel_vector;
	normalise(vel_vector_nm);

	VECTOR3 rot_axis = crossp(falcon_pos_nm, vel_vector_nm);
	VECTOR3 rot_axis_nm = rot_axis;
	normalise(rot_axis_nm);

	double core_rot_angle = ((core_range*0.935)-20000)/rt;
	if(edlo_sw == 0) core_rot_angle = (core_range*0.65)/rt;

	double rot_angle = core_rot_angle;
	VECTOR3 rot_vec = rot_axis_nm;

	MATRIX3 RG = rotm(rot_vec,rot_angle);
	VECTOR3 core_asds_vect = mul(RG, falcon_pos);

	if(edlo_sw == 1) rot_angle = ((core_range*0.935)-5000)/rt;

	double frf = ((38700-(fuel_reserve/7))/5200)*0.08;
	RG = rotm(rot_vec,(rot_angle*1.2));
	if(edlo_sw == 0) RG = rotm(rot_vec,(rot_angle*(1.4+frf)));
	if(fuel_reserve == 0) RG = rotm(rot_vec,(rot_angle*2.3));
	VECTOR3 core_fvr_vect = mul(RG, falcon_pos);

	double asds_rad = 0;

	VECTOR3 earth_pos = _V(0,0,0);
	oapiGetGlobalPos(hEarth,&earth_pos);
	core_asds_vect = core_asds_vect+earth_pos;
	oapiGlobalToEqu(hEarth,core_asds_vect, &asds_core_lng, &asds_core_lat, &asds_rad);
	if(fuel_reserve == 0)
	{
		asds_core_lng = 0;
		asds_core_lat = 0;
	}

	//double fvr_rad = 0;
	//core_fvr_vect = core_fvr_vect+earth_pos;
	//oapiGlobalToEqu(hEarth,core_fvr_vect, &frv_core_lng, &frv_core_lat, &fvr_rad);
	
	//(disable_fair_rec == 1)
	//{
	//	frv_core_lng = 0;
	//	frv_core_lat = 0;
	//}
	//sprintf(oapiDebugString(), "asdscore lng %0.3f lat %0.3f core_range %0.3f azm_rng_factor %0.3f", (asds_core_lng*DEG), (asds_core_lat*DEG),core_range*0.001,azm_range_factor*DEG);
	//sprintf(oapiDebugString(), "fvr_core lng %0.3f lat %0.3f frf %0.3f", (frv_core_lng*DEG), (frv_core_lat*DEG), frf);
}

void ss_tanker::DoLOT(void)
{
	oapiDelAnnotation (hNote);
	HWND hDlg;
	double lot_secs = (3600*lot_hh) + (60*lot_mm) + (lot_ss);
	double current_mjd = oapiGetSimMJD();
	double ct_secs_frac = current_mjd - (int)(current_mjd);
	double ct_secs = 86400*ct_secs_frac;
	//double ct_secs_diff = ct_secs - (int)(ct_secs);

	int ct_secs_int = (int)(ct_secs);
	//if(ct_secs_diff > 0) ct_secs_int = ct_secs_int + 1;

	int lot_secs_int = (int) (lot_secs);
	if (lot_secs_int < ct_secs_int)lot_secs_int = lot_secs_int + 86400;
	

	int t_minus = (lot_secs_int - ct_secs_int);
	
	int t_minus_hh = (int)(t_minus/3600);
	int t_minus_mm = (int)((t_minus - (t_minus_hh * 3600))/60);
	int t_minus_ss = (int)(t_minus - (t_minus_hh * 3600) - (t_minus_mm * 60));

char cbuf_hh[2];
char cbuf_mm[2];
char cbuf_ss[2];
char cbuf[256];

if (abs(t_minus_hh) < 10)sprintf(cbuf_hh, "0%d", t_minus_hh);
else sprintf(cbuf_hh, "%d", t_minus_hh);

if (abs(t_minus_mm) < 10)sprintf(cbuf_mm, "0%d", t_minus_mm);
else sprintf(cbuf_mm, "%d", t_minus_mm);

if (abs(t_minus_ss) < 10)sprintf(cbuf_ss, "0%d", t_minus_ss);
else sprintf(cbuf_ss, "%d", t_minus_ss);

sprintf(cbuf,"T - %s:%s:%s", cbuf_hh,cbuf_mm,cbuf_ss);

VECTOR3 col = _V(1,0.9,0);
hNote = oapiCreateAnnotation(true, 3, col);
oapiAnnotationSetPos(hNote,0.36,0.89,0.8,0.99);
oapiAnnotationSetText(hNote, cbuf);

	if (lot_status == 1)
	{
	if ((lot_secs_int - ct_secs_int) < 0) return;
	else if ((lot_secs_int - ct_secs_int) < 11) 
		{
			////// send key press V /////////
			{
			if (eng_status == 0)
				{
				if(ap_status == 0)ap_status = 1;
				else ap_status = 0;
				op_status = 0;
				KillAllRCS2();
				}
			else if (eng_status == 1)
				{
				if(op_status == 0)op_status = 1;
				else op_status = 0;
				ap_status = 0;
				KillAllRCS2();
				}
			apop_status = ap_status + op_status;
			}
			/////////////////////////

		lot_status = 2;
		hDlg = oapiFindDialog (g_Param.hDLL, IDD_CTRL);
		if (hDlg) InitDialog(hDlg);
		}
	}
	if (lot_status == 2)
	{
	lot_timer = lot_timer + step_time;
	if (lot_timer > 10)
		{
		sprintf(cbuf,"T - 00:00:00");
		oapiAnnotationSetText(hNote, cbuf);
		}
	if (lot_timer > 11) 
		{
		lot_status = 3;
		lot_timer = 0;
		oapiDelAnnotation (hNote);
		hDlg = oapiFindDialog (g_Param.hDLL, IDD_CTRL);
		if (hDlg) InitDialog(hDlg);
		return;
		}
	}

}

void ss_tanker::DoAscent(void)
{
if (eng_status == 1)return;


CalcAzimuth();
CalcFlightVector();

double th_max = GetThrusterMax(th_main[eng_status]);
double mass = GetMass();
double maxacc = th_max/mass;
//double acc_lim = 39.24;
//if (drg == 1)acc_lim = 29.43;
double englev = acc_lim/maxacc;
if (englev > 1)englev = 1;

if(GetDynPressure() > 23000)
{
	if(englev > 0.90)englev = 0.90;
}


if(ap_function == 0)
{
OBJHANDLE hLpad;
VESSEL2* v;

if (tu_sw == 0)
{
hLpad = GetAttachmentStatus(lpd_attach);
if (!hLpad) tu_sw = 1;
else 
	{
	tu_sw = 2;
	v = (VESSEL2*)oapiGetVesselInterface(hLpad);
	}
}

if (tu_sw == 1)
	{
	if (timer == 0) PlayVesselWave(MyID, CTDN, 0, 200, 0);
	timer = timer+step_time;
	if (timer > 10) 
		{
		englev = 1;
		timer = 10;
		}
	else englev = 0;
	}
else if (tu_sw == 2)
	{
	if (timer == 0) 
		{
		PlayVesselWave(MyID, CTDN, 0, 200, 0);
		v->SendBufferedKey(OAPI_KEY_M);
		}
	timer = timer+step_time;
	if (timer > 10) timer = 10;
	if (timer < 10) englev = (timer-8)/2;
	if (englev < 0)englev = 0;
	}
}
SetThrusterGroupLevel(THGROUP_MAIN,englev);

if (GetAltitude(ALTMODE_GROUND) > 280 && ap_function == 0)
{
	ap_function = 1;
	timer = 0;
}
if (ap_function > 0) RollToAz();
//if (ap_function == 0) GoToFlightVector();
if (ap_function == 1) GoToFlightVector();
if (ap_function == 2) GoToFlightVector();
if (ap_function == 3) GoToFlightVector();
if (ap_function == 4) GoToFlightVector();
if (ap_function == 5)
	{
	KillAllRCS2();
	SetThrusterGroupLevel(THGROUP_MAIN,0);
	ActivateNavmode(NAVMODE_KILLROT);
	ap_function = 0;
	ap_status = 0;
	JettisonStg();
	op_status = 1;
	}




}

double ss_tanker::CalcAzimuth()
{
	VESSEL2* v;
	OBJHANDLE hPad = GetAttachmentStatus(lpd_attach);
	if(hPad != NULL) v = (VESSEL2*)oapiGetVesselInterface(hPad);
	else v = (VESSEL2*)oapiGetVesselInterface(GetHandle());

	double lon,lat,rad;
	v->GetEquPos(lon,lat,rad);
	if(abs(tgt_inc)<abs(lat)){
		tgt_inc=abs(lat);
	}else{
		tgt_inc=tgt_inc;
	}
	/////////////
	OBJHANDLE hearth=v->GetSurfaceRef();
	VECTOR3 relvel = _V(0,0,0);
	VECTOR3 globpos = _V(0,0,0);
	v->GetRelativeVel(hearth,relvel);
	v->GetGlobalPos(globpos);
	VECTOR3 relvel_glob = relvel+globpos;
	VECTOR3 rel_vel_loc = _V(0,0,0);
	v->Global2Local(relvel_glob,rel_vel_loc);

	VECTOR3 ov_hrz = _V(0,0,0);
	v->HorizonRot(rel_vel_loc, ov_hrz);

	double orb_ground_spd = sqrt((ov_hrz.x*ov_hrz.x)+(ov_hrz.z*ov_hrz.z));

	VECTOR3 orb_ground_vector = _V(ov_hrz.x,0,ov_hrz.z);

	////////
	double tgt_inc_sc = tgt_inc;
	if (v->GetAltitude(ALTMODE_GROUND) > 10000) // bloody winds at launch!!
	{
		if(lat > 0)
		{
		if (ov_hrz.z < -1 && tgt_inc > 0) tgt_inc_sc = -tgt_inc;
		}
		else if(lat < 0)
		{
		if (ov_hrz.z > 1 && tgt_inc < 0) tgt_inc_sc = -tgt_inc;
		}
	}
	double final_azimuth = asin(cos(tgt_inc_sc)/cos(lat));
	final_azimuth+=(90*RAD-final_azimuth)*(1-(tgt_inc_sc/abs(tgt_inc_sc)));
	VECTOR3 final_ground_vector = _V((FinalV()*sin(final_azimuth)),0,(FinalV()*cos(final_azimuth)));

	VECTOR3 req_ground_vector = final_ground_vector - orb_ground_vector;

	azimuth = atan2(req_ground_vector.x,req_ground_vector.z);

	/////////////
	//azimuth=atan2(FinalV()*(sin(asin(cos(tgt_inc)/cos(lat))))-GetOS(),FinalV()*(cos(asin(cos(tgt_inc)/cos(lat)))));
	//azimuth+=(90*RAD-azimuth)*(1-(tgt_inc/abs(tgt_inc)));
	
	vazim1=_V(sin(azimuth),0,cos(azimuth));
	
	v->HorizonInvRot(vazim1,azim_local);
	//sprintf(oapiDebugString(), "ap_function %d", ap_function);
	return azimuth;
}
double ss_tanker::GetOS(){
	OBJHANDLE hearth=GetSurfaceRef();
	VECTOR3 relvel;
	GetRelativeVel(hearth,relvel);
	double os=length(relvel);
	return os;
}

double ss_tanker::FinalV()
{
//double tgtapo = 180000;
//double tgtperi = -6318000;
double tgtapo = 200000;
double tgtperi = 2000000;
double abside=tgtapo;
	if(tgtapo<=260000){
			abside=tgtapo;
		}else{
			abside=tgtperi;}
if (op_status == 1) return sqrt(mu*(2/(op_meco_alt+rt)-1/(0.5*(op_tgt+op_meco_alt+(2*rt)))));
else return sqrt(mu*(2/(abside+rt)-1/(0.5*(tgtapo+tgtperi+2*rt))));

}

void ss_tanker::CalcFlightVector()
{
	double initial_pitch = RAD*(73.0 + (6*((ap_tgt - 125000)*0.00001))); //76

	double fuel_res_fctr = sin(PI05*(fuel_reserve/7)/100000);
	if (fuel_res_fctr < 0)fuel_res_fctr = 0;
	else if (fuel_res_fctr > 1)fuel_res_fctr = 1;
	fuel_res_fctr=3.5*fuel_res_fctr*RAD; // 3

	initial_pitch = initial_pitch + fuel_res_fctr;

	double azmth = CalcAzimuth();
	double azmth_fctr = RAD*(-0.25*(sin(azmth)));

	initial_pitch = initial_pitch + azmth_fctr;

	if (ap_function == 2) 
	{
		VECTOR3 haspdv = _V(0,1,0);
		GetGroundspeedVector (FRAME_HORIZON, haspdv);
		double mass = GetMass();
		VECTOR3 thrustvector = _V(0,0,0);
		GetThrustVector(thrustvector);
		double trst = length(thrustvector);
		double acc = trst/mass;
		double time_avail = ((210)-length(haspdv))/(acc-9.81);
		double pitch_rate = ((pitch_angle)-initial_pitch)/time_avail;
		pitch_angle = pitch_angle - (step_time*pitch_rate);

		//sprintf(oapiDebugString(), "initial_pitch %0.4f pitch_angle %0.4f", initial_pitch*DEG,pitch_angle*DEG);

		if (length(haspdv) > (210))ap_function = 3;


	}
	
	
	else if (ap_function == 3)
	{
		VECTOR3 haspdv = _V(0,1,0);
		GetGroundspeedVector (FRAME_HORIZON, haspdv);
		double xzlngth = sqrt((haspdv.x*haspdv.x)+(haspdv.z*haspdv.z));
		if (haspdv.y < 0)haspdv.y = 0;
		pitch_angle = atan2(haspdv.y,xzlngth);
		if (GetAltitude() > 30000) ap_function = 4;
		//if (length(haspdv) < 200) sprintf(oapiDebugString(), "gspd ptch angle %0.4f ", pitch_angle*DEG);
	}

	else if (ap_function == 4)
	{
		double ap_dist = 0;
		GetApDist(ap_dist);

		double englev = GetThrusterGroupLevel(THGROUP_MAIN);
		double fuel_remaining = GetPropellantMass(hProp1) - fuel_reserve;
		double isp_curr = GetThrusterIsp(th_main[eng_status]);
		double th_max = GetThrusterMax(th_main[eng_status]);
		double fuel_rate = englev*th_max/isp_curr;
		double mintime = fuel_remaining/fuel_rate;
		if (mintime < 0)mintime = 0;

		/////////////////////////

		VECTOR3 haspdv = _V(0,1,0);
		GetGroundspeedVector (FRAME_HORIZON, haspdv);
		//double xzlngth = sqrt((haspdv.x*haspdv.x)+(haspdv.z*haspdv.z));
		if (haspdv.y < 0)haspdv.y = 0;

		///////////////////////////////////////////////////////
		double vmass = GetMass();

		OBJHANDLE hearth=GetSurfaceRef();
		VECTOR3 relvel = _V(0,0,0);
		GetRelativeVel(hearth,relvel);
		double orb_vel = length(relvel);
		double orb_grnd_spd = sqrt((orb_vel*orb_vel)-(haspdv.y*haspdv.y));
		double circ = PI2*(rt+GetAltitude());
		double ang_vel = PI2*orb_grnd_spd/circ;
		VECTOR3 angmomvect = _V(0,(-1*ang_vel),0);
		VECTOR3 velvect = _V(0,0,orb_grnd_spd);
		VECTOR3 corvect = crossp(angmomvect,velvect);
		double acc_cor = length(corvect);

		double sinp = sin(GetPitch());
		double dv_grnd = isp_curr*log(vmass/(vmass-fuel_remaining));
		dv_grnd = dv_grnd*sinp;

		double final_orb_grnd_spd = orb_grnd_spd+dv_grnd;
		velvect = _V(0,0,final_orb_grnd_spd);
		VECTOR3 final_corvect = crossp(angmomvect,velvect);
		double final_acc_cor = length(corvect);

		VECTOR3 weight = _V(0,0,0);
		GetWeightVector(weight);
		
		double gravt = length(weight);
		double gg = (gravt/vmass)-final_acc_cor;
		double aa = (mintime*mintime)/(2*gg);
		double bb = (haspdv.y*mintime/gg)+((mintime*mintime)/2);
		double cc = -(ap_tgt - GetAltitude() - (haspdv.y*mintime) - ((haspdv.y*haspdv.y)/(2*gg)) );

		double req_acc_1 = (-bb+(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);
		double req_acc_2 = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);

		//if (mintime < 4) englev = englev-(step_time/5);
		//if(englev < 0.1)englev = 0.1;

		double curr_acc = th_max*englev/(vmass-(fuel_remaining*0.5));
		
		//if (ap_dist > (rt+ap_tgt))ap_function = 5;
		if (fuel_remaining < (0.5*fuel_rate*step_time))ap_function = 5;
		
		double total_req_acc = req_acc_1+gg-acc_cor;

		double psc = total_req_acc/curr_acc;
		//if (psc > 0.985)psc = 0.985;
		if (psc > 0.9962)psc = 0.9962;
		else if (psc < -0.9962)psc = -0.9962;
		double pangle = asin(psc);

		if (mintime < 6 && mintime > 3)pangle = pangle - (4*RAD);
		pitch_angle = pangle;
		
		//sprintf(oapiDebugString(), "af %d bias %0.4f reqaprt %0.2f aprate %0.2f", ap_function, bias, req_ap_rate, ap_rate);
		//sprintf(oapiDebugString(), "pangle %0.4f ", pangle*DEG);
		
		
	}

		
		vazim2 = _V(vazim1.x, (tan(pitch_angle)), vazim1.z);

	//if (ap_function < 2)vazim2 = _V(0,1,0);
	if (ap_function < 2)
	{
	VECTOR3 haspdv = _V(0,1,0);
	GetGroundspeedVector (FRAME_HORIZON, haspdv);
	vazim2 = _V(-haspdv.x*0.25,haspdv.y,-haspdv.z*0.25);
	}
	HorizonInvRot(vazim2,flight_vector);

		//sprintf(oapiDebugString(), "af %d", ap_function);
}

void ss_tanker::RollToAz()
{
DeactivateNavmode(NAVMODE_KILLROT);
double roll_angle = atan2(azim_local.x,-azim_local.y);
if (roll_angle > PI) roll_angle = roll_angle - PI2;
else if (roll_angle < -PI) roll_angle = roll_angle + PI2;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[0]);
	double max_torque = 2*max_rcs*4;
	VECTOR3 current_pmi;
	GetPMI(current_pmi);
	double inertia = current_pmi.z*mass;

	if (eng_status == 1)
	{
	max_rcs = GetThrusterMax(th_rcs[8]);
	max_torque = 2*max_rcs*1;
	GetPMI(current_pmi);
	inertia = current_pmi.z*mass;
	}
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);


	double max_acc = max_torque/inertia;

	double max_drot = max_acc*step_time;

	double reqzv = 0.9*(sqrt(abs(roll_angle*max_acc*2)));
	if (roll_angle > 0)reqzv = -reqzv;

	double req_drot = reqzv - avel.z;

	double thz_lvl = req_drot/max_drot;


	if((roll_angle*DEG) < 1 && (roll_angle*DEG) > -1)
	{
	reqzv = 1*-roll_angle;
	double acc_reqz = (reqzv-avel.z)/0.1;
	thz_lvl = acc_reqz/max_acc;
	}


	if(thz_lvl > 1)thz_lvl = 1;
	else if(thz_lvl < -1)thz_lvl = -1;


	if(thz_lvl > 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}
	else if(thz_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, -thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}

	if (ap_function == 1)
	{

		//if (roll_angle > -0.05 && roll_angle < 0.05)
		//{
		VECTOR3 aspdv = _V(0,1,0);
		GetHorizonAirspeedVector(aspdv);
		//double azimuth = CalcAzimuth();
		//double az_factor = 30*(1 - sin(azimuth));
		if (aspdv.y > (60)) //55
			{
			vrot_lim = 2*RAD;
			ap_function = 2;
			}
		//}
	}
}

void ss_tanker::RollToAzLanding()
{
	DeactivateNavmode(NAVMODE_KILLROT);
	double roll_angle = az_angle;
	if (roll_angle > PI) roll_angle = roll_angle - PI2;
	else if (roll_angle < -PI) roll_angle = roll_angle + PI2;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[26]);
	double max_torque = 2*max_rcs*5;
	VECTOR3 current_pmi;
	GetPMI(current_pmi);
	double inertia = current_pmi.z*mass;

	
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);


	double max_acc = max_torque/inertia;

	double max_drot = max_acc*step_time;

	double reqzv = 0.9*(sqrt(abs(roll_angle*max_acc*2)));
	if (roll_angle > 0)reqzv = -reqzv;

	double req_drot = reqzv - avel.z;

	double thz_lvl = req_drot/max_drot;


	if((roll_angle*DEG) < 1 && (roll_angle*DEG) > -1)
	{
	reqzv = 1*-roll_angle;
	double acc_reqz = (reqzv-avel.z)/0.1;
	thz_lvl = acc_reqz/max_acc;
	}


	if(thz_lvl > 1)thz_lvl = 1;
	else if(thz_lvl < -1)thz_lvl = -1;


	if(thz_lvl > 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}
	else if(thz_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, -thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}
}

void ss_tanker::GoToFlightVector()
{
	//VECTOR3 wv = _V(0,0,0);
	//GetWeightVector(wv);
	//wv = -wv;

	
	//double angoa = atan2(-wv.y, wv.z);
	//double slipage = atan2(-wv.x, wv.z);
	DeactivateNavmode(NAVMODE_KILLROT);

	//vrot_lim = 5*RAD;


	double angoa = atan2(-flight_vector.y, flight_vector.z);
	double slipage = atan2(flight_vector.x, flight_vector.z);

	if(angoa > PI)angoa = (PI*2)-angoa;
	else if(angoa < -PI)angoa = (PI*2)+angoa;

	if(slipage > PI)slipage = (PI*2)-slipage;
	else if(slipage < -PI)slipage = (PI*2)+slipage;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[0]);
	double max_torque = 2*max_rcs*60;
	VECTOR3 current_pmi;
	GetPMI(current_pmi);
	double inertia_x = current_pmi.x*mass;
	double inertia_y = current_pmi.y*mass;
	
	if (eng_status == 1)
	{
	max_rcs = GetThrusterMax(th_rcs[8]);
	max_torque = 1*max_rcs*16;
	GetPMI(current_pmi);
	inertia_x = current_pmi.x*mass;
	inertia_y = current_pmi.y*mass;
	if (rcs_sw2 == 1)
		{
		max_rcs = GetThrusterMax(th_rcs[16]);
		max_torque = 2*max_rcs*16;
		}
	}

	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double max_acc_x = max_torque/inertia_x;
	double max_drot_x = max_acc_x*step_time;
	double max_acc_y = max_torque/inertia_y;
	if (rcs_sw2+disable_rcs == 2)max_acc_y = 0.5*max_acc_y; // upper aft rcs disabled
	double max_drot_y = max_acc_y*step_time;

	double reqxv = 0.9*(sqrt(abs(angoa*max_acc_x*2)));
	if (reqxv > vrot_lim)reqxv = vrot_lim;
	if (angoa > 0)reqxv = -reqxv;
	double req_drotx = reqxv - avel.x;
	double thx_lvl = req_drotx/max_drot_x;
	if((angoa*DEG) < 1 && (angoa*DEG) > -1)
	{
	reqxv = 1*-angoa;
	double acc_reqx = (reqxv-avel.x)/0.1;
	thx_lvl = acc_reqx/max_acc_x;
	}

	double reqyv = 0.9*(sqrt(abs(slipage*max_acc_y*2)));
	if (reqyv > vrot_lim)reqyv = vrot_lim;
	if (slipage > 0)reqyv = -reqyv;
	double req_droty = reqyv - avel.y;
	double thy_lvl = req_droty/max_drot_y;
	if((slipage*DEG) < 1 && (slipage*DEG) > -1)
	{
	reqyv = 1*-slipage;
	double acc_reqy = (reqyv-avel.y)/0.1;
	thy_lvl = acc_reqy/max_acc_y;
	}

	//sprintf(oapiDebugString(), "slip %0.4f yrot %0.4f reqyv%0.4f", (slipage*DEG),(avel.y*DEG), (reqyv*DEG));


	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;


	if(thx_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
	else if(thx_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	}


	if(thy_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		}
	else if(thy_lvl < 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	}


}

bool ss_tanker::clbkLoadGenericCockpit ()
{
	SetCamDir();
return true;
}

void ss_tanker::SetCamDir()
{
	VECTOR3 camdefdir = _V(0,0,1);
	int cam_tilt = 0;
	VECTOR3 cam_pos_off = _V(0,5,0);
if(eng_status == 0)
	{
		if(cam_status > 0)
		{
		camdefdir = _V(0,0,-1);
		cam_tilt = 0;
		cam_pos_off = _V(0,5,0);
		cam_status = 4;
		}
	}

else{
	if(cam_status == 1)
		{
		camdefdir = _V(0,-0.9397,0.3420);
		cam_tilt = 0;
		cam_pos_off = _V(0,0,0);
		}
	else if(cam_status == 2)
		{
		camdefdir = _V(0,-0.5,-0.866);
		cam_tilt = 0;
		cam_pos_off = _V(0,4,-12);
		}
	else if(cam_status == 3)
		{
		camdefdir = _V(0,0,1);
		cam_tilt = 0;
		cam_pos_off = _V(-5.25,2.25,19);
		}
	else if(cam_status == 4)
		{
		camdefdir = _V(0,0,-1);
		cam_tilt = 0;
		cam_pos_off = _V(-6,0.5,5);
		}
	}
	SetCameraOffset(cam_pos_off);
	SetCameraDefaultDirection(camdefdir, cam_tilt);
	oapiCameraAttach(GetHandle(), 2);
}

void ss_tanker::DoOrbit()
{
	
	
	if (start_sw == 0)
	{
	timer = timer+step_time;
	if (timer < 4) return;
	else
		{
		start_sw = 1;
		timer = 0;
		SetThrusterGroupLevel(THGROUP_MAIN,1);
		}
	}

	


DeactivateNavmode(NAVMODE_KILLROT);
CalcAzimuth();

double fnl_vel = sqrt(mu*(2/(op_meco_alt+rt)-1/(0.5*(op_tgt+op_meco_alt+(2*rt)))));

double emp_mass = GetEmptyMass();
double vmass = GetMass();

OBJHANDLE hearth=GetSurfaceRef();

VECTOR3 haspdv = _V(0,0,0);
GetGroundspeedVector (FRAME_HORIZON, haspdv);
	
VECTOR3 weight = _V(0,0,0);
GetWeightVector(weight);

//centrifugal acc
	VECTOR3 relvel = _V(0,0,0);
	GetRelativeVel(hearth,relvel);
	double orb_vel = length(relvel);
	double orb_grnd_spd = sqrt((orb_vel*orb_vel)-(haspdv.y*haspdv.y));
	double rds = (rt + GetAltitude());
	double ang_vel = orb_grnd_spd/rds;
	double acc_cent_fact = 1;
	double acc_cent = ang_vel*ang_vel*rds*acc_cent_fact;
		

	double gravt = length(weight);
	double g = -(gravt/vmass)+acc_cent;

	double aa = g/2;
	double bb = haspdv.y;
	double cc = GetAltitude()-op_meco_alt;

	//double timetoimp = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);
	double max_th = 0;
	double current_isp = 0;
	DWORD n_th = GetGroupThrusterCount(THGROUP_MAIN);
	int n_th_int = n_th;
	for (int i = 0; i < n_th_int; i++)
	{
		THRUSTER_HANDLE hMain_th = GetGroupThruster(THGROUP_MAIN, i);
		max_th = max_th + GetThrusterMax(hMain_th);
		current_isp = current_isp + (GetThrusterIsp(hMain_th)/n_th_int);
	}
	//double max_th = GetThrusterMax(th_main[1]);
	double max_acc = max_th/vmass;
	//double acc_lim = 50;
	//if (drg > 0) acc_lim = 30;
	double max_eng_lvl = 1;
	if (max_acc > acc_lim) max_eng_lvl = acc_lim/max_acc;
	if (max_eng_lvl < 0.7)max_eng_lvl = 0.7;
	if ((fnl_vel-orb_vel) < 400)
	{
	if (max_eng_lvl > 0.7)max_eng_lvl = 0.7;
	}
	double sinp = sin(GetPitch());
	double cosp = cos(GetPitch());
	
	double dv_req = fnl_vel - orb_grnd_spd;
	double fraction = exp(dv_req/current_isp);
	double fuelmass = vmass - (vmass/fraction);

	double dth = step_time * 0.2;

	double englevel = GetThrusterGroupLevel(THGROUP_MAIN);
	if(start_sw == 1)
	{
	englevel = 1;
	start_sw = 2;
	}
	//sprintf(oapiDebugString(), "englevel %0.1f", englevel);
	
	double fuelrate = max_th*englevel/current_isp;
	double dvtime = fuelmass/fuelrate;
	dvtime = dvtime/cosp;

	GetElements (hearth,el,&prm,0,FRAME_EQU);
	double ap_time = prm.ApT;
	//sprintf(oapiDebugString(), "dvtime %0.1f ap_time %0.1f", dvtime, ap_time);
	double ap_time_limiter = dvtime/ap_time;
	if (ap_time_limiter > 1)ap_time_limiter = 1;
	else if (ap_time_limiter < 0.7)ap_time_limiter = 0.7;
	if (ap_time_limiter < max_eng_lvl) max_eng_lvl = ap_time_limiter;
	

	double vvel_fnl = 0;

	//double acc_req = 2*((vvel_fnl*dvtime)+GetAltitude()-op_meco_alt)/(dvtime*dvtime);
	//double acc_req = 2*2*(op_meco_alt-GetAltitude()-(haspdv.y*dvtime))/(dvtime*dvtime);
	double acc_req = ((6*(op_meco_alt-GetAltitude() - (haspdv.y*dvtime)))+(2*dvtime*haspdv.y))/(dvtime*dvtime);
	
	acc_req = acc_req - g;

	//sprintf(oapiDebugString(), "acc_req %0.1f", acc_req);

	double srp = acc_req/(max_acc*englevel);
	if (srp > 1)srp = 1;
	else if (srp < -1)srp = -1;
	double req_pitch = asin(srp);
	double crp = cos(req_pitch);

	VECTOR3 fv2d = _V(0, srp, crp);
	VECTOR3 fv_hrz = _V(0,0,0);
	fv_hrz.x = sin(azimuth)*crp;
	fv_hrz.y = srp;
	fv_hrz.z = cos(azimuth)*crp;

	double max_dv = step_time*max_acc*englevel;
	double req_englevel = (fnl_vel-orb_vel)/max_dv;
	if (req_englevel < 0)
	{
		op_status = 0;
		englevel = 0;
		KillAllRCS2();
		SetThrusterGroupLevel(THGROUP_MAIN,0);
		ActivateNavmode(NAVMODE_KILLROT);
		return;
	}
	//englevel = (fnl_vel-orb_vel)/max_dv;
	if (req_englevel > max_eng_lvl)req_englevel = max_eng_lvl;
	

	if (englevel > req_englevel)
	{
		englevel = englevel-dth;
		if (englevel < req_englevel) englevel = req_englevel;
	}
	else if (englevel < req_englevel)
	{
		englevel = englevel+dth;
		if (englevel > req_englevel) englevel = req_englevel;
	}
	

	if ((fnl_vel-orb_vel) < 30)
	{
		//flight_vector = flight_vector;
		ActivateNavmode(NAVMODE_KILLROT);
		///if(englevel > 0)englevel = 0.7;
		///if (englevel > max_eng_lvl)englevel = max_eng_lvl;
		//sprintf(oapiDebugString(), "fv-ov %0.1f", (fnl_vel-orb_vel));
	}

	//else if ((fnl_vel-orb_vel) < 400)
	//{
	//	if(englevel > 0)englevel = 0.7;
	//	if (englevel > max_eng_lvl)englevel = max_eng_lvl;
	//	HorizonInvRot(fv_hrz, flight_vector);
	//	GoToFlightVector();
	//}
	else 
	{
	HorizonInvRot(fv_hrz, flight_vector);
	GoToFlightVector();
	}

	if (start_sw == 0)englevel = 0;
	SetThrusterGroupLevel(THGROUP_MAIN,englevel);




}

void ss_tanker::clbkVisualCreated (VISHANDLE vis, int refcount)
{
dhMesh0 = GetDevMesh(vis, 0); // stg2
SetMeshMat();
}

void ss_tanker::clbkVisualDestroyed (VISHANDLE vis, int refcount)
{
dhMesh0 = NULL; // stg2
}

void ss_tanker::SetMeshMat()
{
double main_lvl1 = GetThrusterLevel(th_main[2]);
double drf = step_time*0.1;
if (red_fact < main_lvl1)red_fact = red_fact+drf;
else if (red_fact > main_lvl1)red_fact = red_fact-drf;

if (red_fact > 1)red_fact = 1;
else if (red_fact < 0)red_fact = 0;
double yel_fact = red_fact*0.1;

//double main_lvl2 = GetThrusterLevel(th_main[1]);
//if (red_fact2 < main_lvl2)red_fact2 = red_fact2+drf;
//else if (red_fact2 > main_lvl2)red_fact2 = red_fact2-drf;

//if (red_fact2 > 1)red_fact2 = 1;
//else if (red_fact2 < 0)red_fact2 = 0;
//double yel_fact2 = red_fact2*0.1;

MATERIAL *mat2 = oapiMeshMaterial (hMStg2, 13);
//MATERIAL *mat3 = oapiMeshMaterial (hMStg2, 16);

	
	mat2->emissive.r = red_fact;
	mat2->emissive.g = yel_fact;
	mat2->emissive.b = 0;
	mat2->emissive.a = 1;
	//mat3->emissive.r = red_fact2;
	//mat3->emissive.g = yel_fact2;
	//mat3->emissive.b = 0;
	//mat3->emissive.a = 1;
	if(dhMesh0 != NULL)
	{
		oapiSetMaterial(dhMesh0,13, mat2);
		//oapiSetMaterial(dhMesh0,16, mat3);
	}
}

void ss_tanker::SetUpperExhaust (void)
{
	double p_level = GetThrusterGroupLevel(THGROUP_ATT_PITCHUP) - GetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN);
	double y_level = GetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT) - GetThrusterGroupLevel(THGROUP_ATT_YAWLEFT);

	upperexhaustdir.x = -sin(5*RAD*y_level);
	upperexhaustdir.y = -sin(5*RAD*p_level);
	upperexhaustdir.z = 1;
	normalise(upperexhaustdir);

	upper_exhaust[0].ldir = &upperexhaustdir;
	upper_exhaust[1].ldir = &upperexhaustdir;
	upper_exhaust[2].ldir = &upperexhaustdir;
}

void ss_tanker::ConfigUpperRaptorSL (void)
{
	//SURFHANDLE exhaust = oapiRegisterExhaustTexture ("spacex_starship/KeroExhaust3");
	DelExhaust(exhaust_id[0]);
	DelExhaust(exhaust_id[1]);
	DelExhaust(exhaust_id[2]);

	

	if(raptor_sl_selection == 3)
	{
	SetThrusterMax0(th_main[1],STG2MAINTH0*3);
	exhaust_id[0] = AddExhaust (&upper_exhaust[0]);
	exhaust_id[1] = AddExhaust (&upper_exhaust[1]);
	exhaust_id[2] = AddExhaust (&upper_exhaust[2]);
	}
	else if(raptor_sl_selection == 2)
	{
	SetThrusterMax0(th_main[1],STG2MAINTH0*2);
	exhaust_id[0] = AddExhaust (&upper_exhaust[0]);
	exhaust_id[1] = AddExhaust (&upper_exhaust[1]);
	}
	else if(raptor_sl_selection == 1)
	{
	SetThrusterMax0(th_main[1],STG2MAINTH0*1);
	exhaust_id[0] = AddExhaust (&upper_exhaust[0]);
	}
}

void ss_tanker::SetADCLevel (void)
{
	double pitch_deflection = (60*RAD);
	double min_cos = cos(pitch_deflection);

	double pitch_deflection_wing = (70*RAD);
	double min_cos_wing = cos(pitch_deflection_wing);
	
	double port_can_tgt_cos = 1;
	double stbd_can_tgt_cos = 1;
	double port_wng_tgt_cos = 1;
	double stbd_wng_tgt_cos = 1;

	VECTOR3 aspd;
	GetAirspeedVector(FRAME_LOCAL, aspd);
	if(aspd.y > 0)
	{
		pitch_level = -pitch_level;
		bank_level = -bank_level;
	}
	if(abs(pitch_level) < 0.3) pitch_level = 0;
	if(abs(bank_level) < 0.3) bank_level = 0;
	if(abs(yaw_level) < 0.3) yaw_level = 0;

	

DWORD adcm = GetADCtrlMode(); // aerofoils disabled
if(adcm == 0)
	{
		pitch_level = 0;
		bank_level = 0;
		yaw_level = 0;
	}

if(adcs_reentry_mode == 1)
{
	port_can_tgt_cos = 0.5;
	stbd_can_tgt_cos = 0.5;
	port_wng_tgt_cos = 0.5;
	stbd_wng_tgt_cos = 0.5;

	if(pitch_level > 0) // pitch
	{
	port_can_tgt_cos = 0.5 + (0.250 * pitch_level);
	stbd_can_tgt_cos = 0.5 + (0.250 * pitch_level);
	port_wng_tgt_cos = 0.5;
	stbd_wng_tgt_cos = 0.5;
	}
	else if(pitch_level < 0)
	{
	port_can_tgt_cos = 0.5;
	stbd_can_tgt_cos = 0.5;
	port_wng_tgt_cos = 0.5 - (0.250 * pitch_level);
	stbd_wng_tgt_cos = 0.5 - (0.250 * pitch_level);
	}
	


	//bank
	
	if(bank_level > 0)
	{
	port_can_tgt_cos = port_can_tgt_cos;
	stbd_can_tgt_cos = stbd_can_tgt_cos + (0.150 * bank_level);
	port_wng_tgt_cos = port_wng_tgt_cos;
	stbd_wng_tgt_cos = stbd_wng_tgt_cos + (0.150 * bank_level);
	}
	else if(bank_level < 0)
	{
	port_can_tgt_cos = port_can_tgt_cos - (0.150 * bank_level);
	stbd_can_tgt_cos = stbd_can_tgt_cos;
	port_wng_tgt_cos = port_wng_tgt_cos - (0.150 * bank_level);
	stbd_wng_tgt_cos = stbd_wng_tgt_cos;
	}

	//yaw
	
	if(yaw_level > 0)
	{
	port_can_tgt_cos = port_can_tgt_cos - (0.10 * yaw_level);
	stbd_can_tgt_cos = stbd_can_tgt_cos + (0.10 * yaw_level);
	port_wng_tgt_cos = port_wng_tgt_cos + (0.10 * yaw_level);
	stbd_wng_tgt_cos = stbd_wng_tgt_cos - (0.10 * yaw_level);
	}
	else if(yaw_level < 0)
	{
	port_can_tgt_cos = port_can_tgt_cos - (0.10 * yaw_level);
	stbd_can_tgt_cos = stbd_can_tgt_cos + (0.10 * yaw_level);
	port_wng_tgt_cos = port_wng_tgt_cos + (0.10 * yaw_level);
	stbd_wng_tgt_cos = stbd_wng_tgt_cos - (0.10 * yaw_level);
	}
	//sprintf(oapiDebugString(), "stbd_wng_tgt_cos %0.5f bank_level %0.3f pitch_level %0.3f yaw_level %0.3f", stbd_wng_tgt_cos, bank_level, pitch_level, yaw_level);
}

else
{
	
	if(pitch_level > 0) // pitch
	{
	port_can_tgt_cos = 1;
	stbd_can_tgt_cos = 1;
	port_wng_tgt_cos = 1 - (0.250 * pitch_level);
	stbd_wng_tgt_cos = 1 - (0.250 * pitch_level);
	}
	else if(pitch_level < 0)
	{
	port_can_tgt_cos = 1 + (0.250 * pitch_level);
	stbd_can_tgt_cos = 1 + (0.250 * pitch_level);
	port_wng_tgt_cos = 1;
	stbd_wng_tgt_cos = 1;
	}
	else
	{
	port_can_tgt_cos = 1;
	stbd_can_tgt_cos = 1;
	port_wng_tgt_cos = 1;
	stbd_wng_tgt_cos = 1;
	}


	//bank
	
	if(bank_level > 0)
	{
	port_can_tgt_cos = port_can_tgt_cos - (0.150 * bank_level);
	stbd_can_tgt_cos = stbd_can_tgt_cos;
	port_wng_tgt_cos = port_wng_tgt_cos - (0.150 * bank_level);
	stbd_wng_tgt_cos = stbd_wng_tgt_cos;
	}
	else if(bank_level < 0)
	{
	port_can_tgt_cos = port_can_tgt_cos;
	stbd_can_tgt_cos = stbd_can_tgt_cos + (0.150 * bank_level);
	port_wng_tgt_cos = port_wng_tgt_cos;
	stbd_wng_tgt_cos = stbd_wng_tgt_cos + (0.150 * bank_level);
	}

	//yaw
	
	if(yaw_level > 0)
	{
	port_can_tgt_cos = port_can_tgt_cos + (0.10 * yaw_level);
	stbd_can_tgt_cos = stbd_can_tgt_cos - (0.10 * yaw_level);
	port_wng_tgt_cos = port_wng_tgt_cos - (0.10 * yaw_level);
	stbd_wng_tgt_cos = stbd_wng_tgt_cos + (0.10 * yaw_level);
	}
	else if(yaw_level < 0)
	{
	port_can_tgt_cos = port_can_tgt_cos + (0.10 * yaw_level);
	stbd_can_tgt_cos = stbd_can_tgt_cos - (0.10 * yaw_level);
	port_wng_tgt_cos = port_wng_tgt_cos - (0.10 * yaw_level);
	stbd_wng_tgt_cos = stbd_wng_tgt_cos + (0.10 * yaw_level);
	}
}
	if(adcs_reentry_mode == 0)
	{
		if (wing_trim < 0)wing_trim = 0;
	}

	{ /// set wing trim
		port_wng_tgt_cos = port_wng_tgt_cos - (0.6*wing_trim*((cos(60*RAD)) - (cos(70*RAD))));
		stbd_wng_tgt_cos = stbd_wng_tgt_cos - (0.6*wing_trim*((cos(60*RAD)) - (cos(70*RAD))));
	}
	
	
	{
	if(port_can_tgt_cos < min_cos)port_can_tgt_cos = min_cos; // lock to range 0.5 to 1.0
	else if(port_can_tgt_cos > 1)port_can_tgt_cos = 1;
	if(stbd_can_tgt_cos < min_cos)stbd_can_tgt_cos = min_cos;
	else if(stbd_can_tgt_cos > 1)stbd_can_tgt_cos = 1;

	if(port_wng_tgt_cos < min_cos_wing)port_wng_tgt_cos = min_cos_wing; // lock to range cos(70) to 1.0
	else if(port_wng_tgt_cos > 1)port_wng_tgt_cos = 1;
	if(stbd_wng_tgt_cos < min_cos_wing)stbd_wng_tgt_cos = min_cos_wing;
	else if(stbd_wng_tgt_cos > 1)stbd_wng_tgt_cos = 1;
	}
	

	double port_can_proc_tgt = acos(port_can_tgt_cos)/pitch_deflection;
	double stbd_can_proc_tgt = acos(stbd_can_tgt_cos)/pitch_deflection;
	double port_wing_proc_tgt = acos(port_wng_tgt_cos)/pitch_deflection_wing;
	double stbd_wing_proc_tgt = acos(stbd_wng_tgt_cos)/pitch_deflection_wing;

	//sprintf(oapiDebugString(), "stbd_wng_tgt_cos %0.5f stbd_wing_proc_tgt %0.5f", stbd_wng_tgt_cos, stbd_wing_proc_tgt);

	double dproc = WING_SPEED * step_time;
	//port can
	if (port_can_proc < port_can_proc_tgt)
	{
		port_can_proc = port_can_proc + dproc;
		if(port_can_proc > port_can_proc_tgt)port_can_proc = port_can_proc_tgt;
	}
	else if (port_can_proc > port_can_proc_tgt)
	{
		port_can_proc = port_can_proc - dproc;
		if(port_can_proc < port_can_proc_tgt)port_can_proc = port_can_proc_tgt;
	} 

	//stbd can
	if (stbd_can_proc < stbd_can_proc_tgt)
	{
		stbd_can_proc = stbd_can_proc + dproc;
		if(stbd_can_proc > stbd_can_proc_tgt)stbd_can_proc = stbd_can_proc_tgt;
	}
	else if (stbd_can_proc > stbd_can_proc_tgt)
	{
		stbd_can_proc = stbd_can_proc - dproc;
		if(stbd_can_proc < stbd_can_proc_tgt)stbd_can_proc = stbd_can_proc_tgt;
	}

	//port wing
	if (port_wing_proc < port_wing_proc_tgt)
	{
		port_wing_proc = port_wing_proc + dproc;
		if(port_wing_proc > port_wing_proc_tgt)port_wing_proc = port_wing_proc_tgt;
	}
	else if (port_wing_proc > port_wing_proc_tgt)
	{
		port_wing_proc = port_wing_proc - dproc;
		if(port_wing_proc < port_wing_proc_tgt)port_wing_proc = port_wing_proc_tgt;
	}

	//stbd wing
	if (stbd_wing_proc < stbd_wing_proc_tgt)
	{
		stbd_wing_proc = stbd_wing_proc + dproc;
		if(stbd_wing_proc > stbd_wing_proc_tgt)stbd_wing_proc = stbd_wing_proc_tgt;
	}
	else if (stbd_wing_proc > stbd_wing_proc_tgt)
	{
		stbd_wing_proc = stbd_wing_proc - dproc;
		if(stbd_wing_proc < stbd_wing_proc_tgt)stbd_wing_proc = stbd_wing_proc_tgt;
	}

	//sprintf(oapiDebugString(), "stbd_wng_tgt_cos %0.5f stbd_wing_proc_tgt %0.5f", stbd_wng_tgt_cos, stbd_wing_proc_tgt);


	SetAnimation (anim_port_can, port_can_proc);
	SetAnimation (anim_stbd_can, stbd_can_proc);
	SetAnimation (anim_port_wing, port_wing_proc);
	SetAnimation (anim_stbd_wing, stbd_wing_proc);

	/////// edit airfoils ///////////////////////////////////////////
	VECTOR3 new_pos = port_canard_pos;
	new_pos.x = new_pos.x + (port_can_proc*2); // update pos
	double port_can_area_factor = cos(60*RAD*port_can_proc); // update area
	EditAirfoil (hPortCanard,0x09, new_pos, 0, sqrt(canard_area), (canard_area*port_can_area_factor), aspect_ratio);

	new_pos = stbd_canard_pos;
	new_pos.x = new_pos.x - (stbd_can_proc*2);
	double stbd_can_area_factor = cos(60*RAD*stbd_can_proc);
	EditAirfoil (hStbdCanard,0x09, new_pos, 0, sqrt(canard_area), (canard_area*stbd_can_area_factor), aspect_ratio);
	
	new_pos = port_wing_pos;
	new_pos.x = new_pos.x + (port_wing_proc*2);
	double port_wing_area_factor = cos(70*RAD*port_wing_proc);
	EditAirfoil (hPortWing,0x09, new_pos, 0, sqrt(wing_area), (wing_area*port_wing_area_factor), aspect_ratio);
	
	new_pos = stbd_wing_pos;
	new_pos.x = new_pos.x - (stbd_wing_proc*2);
	double stbd_wing_area_factor = cos(70*RAD*stbd_wing_proc);
	EditAirfoil (hStbdWing,0x09, new_pos, 0, sqrt(wing_area), (wing_area*stbd_wing_area_factor), aspect_ratio);

	double wing_proc_avg = (stbd_wing_proc + port_wing_proc)*0.5;
	double vertical_wing_area = (sin(70*RAD*wing_proc_avg)) * (wing_area*0.333);
	EditAirfoil (hWing_Vertical,0x09, wing_vertical_pos, 0, sqrt(vertical_wing_area), (vertical_wing_area), aspect_ratio);
	

	{
	//// yaw torque fudge
	double total_drag = GetDrag();
	double yaw_force = 0.25*yaw_level*total_drag*(wing_area+canard_area)/(wing_area+canard_area+wing_area+canard_area+hull_area);
	VECTOR3 yaw_torque = _V(yaw_force, 0, 0);
	VECTOR3 aft_attack_point = _V(0,0,-9);
	VECTOR3 fwd_attack_point = _V(0,0,9);
	AddForce(yaw_torque, aft_attack_point);
	AddForce(-yaw_torque, fwd_attack_point);
	}
	//////////////////////////////////////////////

}

void ss_tanker::GetWindVector()
{
	OBJHANDLE hTarget = oapiGetObjectByName(tgtname);
	if (hTarget != NULL) 
	{
		
		OBJHANDLE hGref = GetGravityRef();
	double gref_rad = oapiGetSize(hGref);
	double lng, lat, rad;
	oapiGetEquPos(hTarget,&lng,&lat,&rad);	
	wind_vector_tgt = oapiGetWindVector(hGref,lng,lat,(rad-gref_rad+1400),0,&windspeed_tgt);
	}
	else hTarget = oapiGetBaseByName(GetGravityRef(),tgtname);
	if (hTarget != NULL) 
	{
	OBJHANDLE hGref = GetGravityRef();
	double gref_rad = oapiGetSize(hGref);
	double lng, lat, rad;
	oapiGetBaseEquPos(hTarget,&lng,&lat,&rad);	
	wind_vector_tgt = oapiGetWindVector(hGref,lng,lat,(rad-gref_rad+1400),0,&windspeed_tgt);
	}
	else if (hTarget == NULL)
	{
	wind_vector_tgt = _V(0,0,0);
	windspeed_tgt = 0;
	}
}

void ss_tanker::DoSkydiveTargeting()
{
	OBJHANDLE hTarget = oapiGetObjectByName(tgtname);
	if (hTarget == NULL)
	{
		hTarget = oapiGetBaseByName(GetGravityRef(),tgtname);
		tgt_att_valid = false;
	}
	else //if(tgt_att_valid == true)
	{
	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hTarget);
	DWORD ac = v->AttachmentCount(false);
	if(ac == 0) tgt_att_valid = false;
	else
	{
	for (DWORD j = 0; j < ac; j++)
		{
		ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(false, j);
		const char *id = v->GetAttachmentId (hAtt);
		if (!strnicmp (id, "BST", 3))
			{
			VECTOR3 attpos, attdir, attrot, attpos_hrz;
			v->GetAttachmentParams(hAtt, attpos, attdir, attrot);
			attpos.x = -11.6;
			attpos.z = -4.8;
			v->HorizonRot(attpos,attpos_hrz);
			double v_alt = v->GetAltitude();
			catch_offset = attpos_hrz;
			catch_offset.y = v_alt + attpos_hrz.y;
			tgt_att_valid = true;
			break;
			}
		else if(j == (ac-1))
			{
				tgt_att_valid = false;
			}
		}
	}
	}
	if (hTarget == NULL) 
	{
		strcpy(tgtname, "NULL");
		skydive_sw = 0;
		return;
	}

	if(first_frame_sw == 0)
	{
	first_frame_sw = 1;
	return;
	}


	DeactivateNavmode(1);
	DeactivateNavmode(2);
	DeactivateNavmode(3);
	DeactivateNavmode(4);
	DeactivateNavmode(5);
	DeactivateNavmode(6);
	DeactivateNavmode(7);

	

	double alt_mr = GetAltitude(ALTMODE_MEANRAD);
	VECTOR3 tgt_pos_glob;
	oapiGetGlobalPos(hTarget, &tgt_pos_glob);
	VECTOR3 tgt_pos_loc;
	Global2Local(tgt_pos_glob, tgt_pos_loc);
	VECTOR3 tgt_pos_hrz;
	HorizonRot(tgt_pos_loc, tgt_pos_hrz);

	if(tgt_att_valid == true)
	{
		tgt_pos_hrz.x = tgt_pos_hrz.x + catch_offset.x;
		tgt_pos_hrz.y = tgt_pos_hrz.y + catch_offset.y - 50.55;
		tgt_pos_hrz.z = tgt_pos_hrz.z + catch_offset.z;
	}

	VECTOR3 gspd_vect_loc;
	GetGroundspeedVector(FRAME_LOCAL, gspd_vect_loc);
	VECTOR3 gspd_vect_hrz;
	HorizonRot(gspd_vect_loc, gspd_vect_hrz);

	double gspd_az_hrz = atan2(gspd_vect_hrz.x,gspd_vect_hrz.z);
	if(gspd_az_hrz < 0)gspd_az_hrz = gspd_az_hrz + PI2;
	if(gspd_az_hrz > PI2)gspd_az_hrz = gspd_az_hrz - PI2;
	
	
	//double wind_vector_az = 0;
	//wind_vector_az = atan2(wind_vector_tgt.x, wind_vector_tgt.z);
	//if(wind_vector_az < 0)wind_vector_az = wind_vector_az + PI2;
	//double wind_vector_az_adj = gspd_az_hrz - wind_vector_az;

	VECTOR3 wind_vector_unit = wind_vector_tgt;
	double cos_wgd = 0;
	if(length(wind_vector_tgt) > 0)
	{	
	normalise(wind_vector_unit);

	double tgt_pos_az_hrz = atan2(tgt_pos_hrz.x,tgt_pos_hrz.z);
	if(tgt_pos_az_hrz < 0)tgt_pos_az_hrz = tgt_pos_az_hrz + PI2;
	if(tgt_pos_az_hrz > PI2)tgt_pos_az_hrz = tgt_pos_az_hrz - PI2;

	double wind_az = atan2(wind_vector_unit.x, wind_vector_unit.z);
	if(wind_az < 0)wind_az = wind_az + PI2;
	double wind_gspd_diff = wind_az - gspd_az_hrz;
	cos_wgd = cos(wind_gspd_diff);
	}

	//sprintf(oapiDebugString(), "wvt %0.1f %0.1f %0.1f", wind_vector_tgt.x,wind_vector_tgt.y,wind_vector_tgt.z);

	//double windspd = windspeed_tgt;

	double sin_tgt_angle = length(_V(tgt_pos_hrz.x,0,tgt_pos_hrz.z))/length(tgt_pos_hrz);
	if(abs(tgt_pos_hrz.y) < 2200)sin_tgt_angle = 0;

	VECTOR3 anti_wind_vector_tgt = _V(0,0,0);
	double awvt_factor = (sin_tgt_angle*-abs(cos_wgd)*750);
	//double awvt_factor = (-abs(cos_wgd)*250);
	anti_wind_vector_tgt.x = wind_vector_unit.x*awvt_factor;
	anti_wind_vector_tgt.z = wind_vector_unit.z*awvt_factor;

	//sprintf(oapiDebugString(), "awv %0.1f %0.1f", anti_wind_vector_tgt.x, anti_wind_vector_tgt.z);

	VECTOR3 tgt_pos_hrz_unit = tgt_pos_hrz;
	normalise(tgt_pos_hrz_unit);
	double lateral_offset_factor = -300;
	VECTOR3 tgt_lateral_offset = _V(0,0,0);
	tgt_lateral_offset.x = tgt_pos_hrz_unit.x*lateral_offset_factor;
	tgt_lateral_offset.z = tgt_pos_hrz_unit.z*lateral_offset_factor;

	double tgt_vert_offset = 0;//-600;
	VECTOR3 tgt_vect_hrz;
	tgt_vect_hrz.x = tgt_pos_hrz.x + anti_wind_vector_tgt.x + tgt_lateral_offset.x;
	tgt_vect_hrz.y = tgt_pos_hrz.y + tgt_vert_offset;
	tgt_vect_hrz.z = tgt_pos_hrz.z + anti_wind_vector_tgt.z + tgt_lateral_offset.z;


	double tgt_az_hrz = atan2(tgt_vect_hrz.x,tgt_vect_hrz.z);
	if(tgt_az_hrz < 0)tgt_az_hrz = tgt_az_hrz + PI2;
	if(tgt_az_hrz > PI2)tgt_az_hrz = tgt_az_hrz - PI2;

	//double wind_az = atan2(wind_vector_unit.x, wind_vector_unit.z);
	//if(wind_az < 0)wind_az = wind_az + PI2;
	//double wind_gspd_diff = wind_az - tgt_az_hrz;
	//double cos_wgd = cos(wind_gspd_diff);
	//double windofs = cos_wgd*windspd*8;

	//if(windspeed_tgt > 0) tgt_vect_hrz.y = tgt_vect_hrz.y - windofs;
	//else windofs = 0;
	//windofs = 0;

	VECTOR3 tgt_vect_hrz_unit = tgt_vect_hrz;
	normalise(tgt_vect_hrz_unit);

	VECTOR3 gspd_vect_hrz_unit = gspd_vect_hrz;
	normalise(gspd_vect_hrz_unit);

	VECTOR3 tgt_gspd_norm_unit = crossp(gspd_vect_hrz_unit, tgt_vect_hrz_unit);
	normalise(tgt_gspd_norm_unit);

	double gspd_vert_angle = acos(dotp(gspd_vect_hrz_unit,_V(0,-1,0)));
	double tgt_vert_angle = acos(dotp(tgt_vect_hrz_unit,_V(0,-1,0)));
	double gspd_tgt_angle = acos(dotp(tgt_vect_hrz_unit,gspd_vect_hrz_unit));

	

	if(roll_ofs_init == 0)
	{
		if(tgt_vect_hrz.z > 0) roll_ofs_init = -1;
		else roll_ofs_init = 1;
	}
	if((tgt_az_hrz*DEG) > 180 && (tgt_az_hrz*DEG) < 270) roll_ofs_init = 1;
	else if((tgt_az_hrz*DEG) > 270 && (tgt_az_hrz*DEG) < 360) roll_ofs_init = -1;

	double mode_change_alt = 6000;//6000
	double tgt_disthd = length(_V(tgt_vect_hrz.x,0,tgt_vect_hrz.z));
	double sd_range = tan(gspd_vert_angle)*abs(tgt_vect_hrz.y);
	double sd_ratio = tgt_disthd/sd_range;
	if(sd_ratio > 1)sd_ratio = 1;
	else if(sd_ratio < 0)sd_ratio = 0;
	double tgt_az_off = roll_ofs_init * acos(sd_ratio)*1.25;//1.25 //// course offset
	if(abs(tgt_vect_hrz.y) < mode_change_alt)tgt_az_off = 0;
	if(tgt_vert_angle > (20*RAD))tgt_az_off = 0;

	//sprintf(oapiDebugString(), "tgt_vert_angle %0.1f", tgt_vert_angle*DEG);
 

	double tgt_az_hrz_adj = tgt_az_hrz - gspd_az_hrz + tgt_az_off;
	if(tgt_az_hrz_adj < 0)tgt_az_hrz_adj = tgt_az_hrz_adj + PI2;
	if(tgt_az_hrz_adj > PI2)tgt_az_hrz_adj = tgt_az_hrz_adj - PI2;

	double fp_diff_angle = tgt_vert_angle - (gspd_vert_angle*cos(tgt_az_hrz_adj));

	double roll_adj_angle = (asin(-sin(tgt_az_hrz_adj))*1.5*sin(gspd_vert_angle));
	if(abs(tgt_vect_hrz.y) < mode_change_alt)roll_adj_angle = (asin(-sin(tgt_az_hrz_adj))*2);//1.5
	
	if(roll_adj_angle > (20*RAD))roll_adj_angle = (20*RAD);
	else if(roll_adj_angle < -(20*RAD))roll_adj_angle = -(20*RAD);
	
	double roll_angle = -tgt_az_hrz_adj + roll_adj_angle;

	/////////// wind ////////////////////////////
	double roll_wind_adj = 0;
	double gspd_lateral = length(_V(gspd_vect_hrz.x, 0, gspd_vect_hrz.z));


	VECTOR3 wind_vector = _V(0,0,0);
	double windspeed = 0;
	double wind_vector_az = 0;
	double wind_vector_az_adj = 0;

	if(abs(tgt_vect_hrz.y) < mode_change_alt)
	{
		
		OBJHANDLE hGref = GetGravityRef();
		double gref_rad = oapiGetSize(hGref);
		double lng, lat, rad;
		GetEquPos(lng,lat,rad);	
		wind_vector = oapiGetWindVector(hGref,lng,lat,alt_mr,0,&windspeed);
		wind_vector_az = atan2(wind_vector.x, wind_vector.z);
		if(wind_vector_az < 0)wind_vector_az = wind_vector_az + PI2;
		wind_vector_az_adj = gspd_az_hrz - wind_vector_az;
		roll_wind_adj = (-45*RAD) * sin(wind_vector_az_adj) * (windspeed/length(_V(gspd_vect_hrz.x,gspd_vect_hrz.y,gspd_vect_hrz.z)));
		if(roll_wind_adj > (10*RAD))roll_wind_adj = (10*RAD);
		else if(roll_wind_adj < (-10*RAD))roll_wind_adj = (-10*RAD);
	//	sprintf(oapiDebugString(), "roll_wind_adj %0.1f", roll_wind_adj*DEG);
		roll_angle = roll_angle + roll_wind_adj;

		skydive_sw = 2;

		
	}

	////////// hud reporting
		hud_tgt_az = tgt_az_hrz;
		hud_gspd_az = gspd_az_hrz;
		hud_tgt_range = tgt_disthd;
		hud_ves_range = tan(gspd_vert_angle)*abs(tgt_pos_hrz.y);


	////////////////////////////////////


	 double fpaoa = (90*RAD) - ((tgt_vert_angle/(30*RAD))*(45*RAD));
	double fp_aoa_angle = (fpaoa - (2.5*fp_diff_angle));

	if (fp_aoa_angle < (30*RAD))fp_aoa_angle = (30*RAD);
	else if (fp_aoa_angle > (90*RAD))fp_aoa_angle = (90*RAD);

	if(abs(tgt_vect_hrz.y) > mode_change_alt)fp_aoa_angle = (45*RAD);
	
	
	//sprintf(oapiDebugString(), "roll_adj_angle %0.1f fp_aoa_angle %0.1f gspd_vert_angle %0.1f tgt_vert_angle %0.1f roll_wind_adj %0.1f wind_vector_az %0.1f windspeed %0.1f gspd_lateral %0.1f",roll_adj_angle*DEG, fp_aoa_angle*DEG, gspd_vert_angle*DEG,tgt_vert_angle*DEG, roll_wind_adj*DEG, wind_vector_az*DEG, windspeed, gspd_lateral);
	//sprintf(oapiDebugString(), "tgt_dist %0.1f alt %0.1f gspd_vert_angle %0.1f tgt_vert_angle %0.1f roll_ofs_init %d wind_vector_az %0.1f windspeed %0.1f gspd_lateral %0.1f",length(_V(tgt_pos_hrz.x,0,tgt_pos_hrz.z)), tgt_pos_hrz.y, gspd_vert_angle*DEG,tgt_vert_angle*DEG, roll_ofs_init, wind_vector_az*DEG, windspeed, gspd_lateral);

	VECTOR3 aspd_vect_hrz;
	GetAirspeedVector(FRAME_HORIZON,aspd_vect_hrz);
	VECTOR3 aspd_vect_hrz_unit = aspd_vect_hrz;
	normalise(aspd_vect_hrz_unit);


	VECTOR3 rot_axis_hrz = crossp(_V(0, -1,0),aspd_vect_hrz_unit);
	normalise(rot_axis_hrz);

	MATRIX3 rm = rotm(rot_axis_hrz, fp_aoa_angle);
	VECTOR3 flight_vector_hrz = mul(rm, aspd_vect_hrz_unit);
	normalise(flight_vector_hrz);

	if(roll_angle > PI)roll_angle = roll_angle - PI2;
	else if(roll_angle < -PI)roll_angle = roll_angle + PI2;

	rm = rotm(aspd_vect_hrz_unit, roll_angle);
	flight_vector_hrz = mul(rm, flight_vector_hrz);


	HorizonInvRot(flight_vector_hrz, flight_vector);
	normalise(flight_vector_hrz);
	normalise(flight_vector);

	VECTOR3 norml_gspd_fltvct = crossp(flight_vector_hrz,aspd_vect_hrz_unit);
	normalise(norml_gspd_fltvct);

	VECTOR3 headsup_vector_hrz = crossp(norml_gspd_fltvct, flight_vector_hrz);
	normalise(headsup_vector_hrz);
	HorizonInvRot(headsup_vector_hrz, headsup_vector);

	GoToFlightVectorSD();
	
	////////////////// wing trim /////////////////

	//wing_trim = (fp_aoa_angle - (60*RAD))/(20*RAD);
	wing_trim = GetWTbyAoA(fp_aoa_angle*DEG);
	if(wing_trim > 1)wing_trim = 1;
	else if(wing_trim < 0)wing_trim = 0;

	/////////// start backflip /////////////////

	double effective_gspd = gspd_lateral + (cos(wind_vector_az_adj)*windspeed);
	//sprintf(oapiDebugString(), "effective_gspd %0.1f",effective_gspd);
	//sprintf(oapiDebugString(), "windofs %0.1f",windofs);
	//if(length(_V(tgt_pos_hrz.x,0,tgt_pos_hrz.z)) < ((9*gspd_lateral)+(2*windofs)) && length(tgt_pos_hrz) < 2000)//11
	if(length(_V(tgt_pos_hrz.x,0,tgt_pos_hrz.z)) < ((5*effective_gspd)+(80)) && length(tgt_pos_hrz) < 1800)
	{
		backflip_sw = 1;
		skydive_sw = 0;
		KillAllRCS2();
		//sprintf(oapiDebugString(), "< 1800");
	}
	else if(-tgt_pos_hrz.y < 1000)
	{
		backflip_sw = 1;
		skydive_sw = 0;
		KillAllRCS2();
		//sprintf(oapiDebugString(), " %0.1f", -tgt_pos_hrz.y);
	}
	else backflip_sw = 0;

}

void ss_tanker::GoToFlightVectorSD()
{
	// check if main on
	bool main_is_on = true;
	double mainth_lev = GetThrusterGroupLevel(THGROUP_MAIN);
	if(mainth_lev == 0)main_is_on = false;
	
	// rot rate limit
	double vrot_lim = 7*RAD;
	if(GetAltitude(ALTMODE_MEANRAD) > 50000)vrot_lim = 5*RAD;
	if(main_is_on == true)vrot_lim = 30*RAD;

	// turn off other ap
	DeactivateNavmode(1);
	DeactivateNavmode(2);
	DeactivateNavmode(3);
	DeactivateNavmode(4);
	DeactivateNavmode(5);
	DeactivateNavmode(6);
	DeactivateNavmode(7);

	/////////////////////////////
	double mass = GetMass();

	VECTOR3 current_pmi;
	GetPMI(current_pmi);
	double inertia_x = current_pmi.x*mass;
	double inertia_y = current_pmi.y*mass;
	double inertia_z = current_pmi.z*mass;
	
	double max_rcs = GetThrusterMax(th_rcs[8]);
	double max_torque_x = 1*max_rcs*16;
	double max_torque_y = 1*max_rcs*16;
	double max_torque_z = 2*max_rcs*1;
	
	if (rcs_sw2 == 1)
		{
		max_rcs = GetThrusterMax(th_rcs[16]);
		max_torque_x = 2*max_rcs*16;
		max_torque_y = 2*max_rcs*16;
		max_torque_z = 2*max_rcs*5;
		if(disable_rcs == 1)max_torque_y = 1*max_rcs*16;
		}
	

	

	double max_torque_pup = abs(max_torque_x);
	double max_torque_pdn = -abs(max_torque_x);
	double max_torque_yl = abs(max_torque_y);
	double max_torque_yr = -abs(max_torque_y);
	double max_torque_br = abs(max_torque_z);
	double max_torque_bl = -abs(max_torque_z);


	
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);

	VECTOR3 gspd_vect_loc = _V(0,0,0);
	GetGroundspeedVector(FRAME_LOCAL, gspd_vect_loc);

	VECTOR3 gspd_vect_loc_nm = gspd_vect_loc;
	normalise(gspd_vect_loc_nm);

	normalise(flight_vector);
	VECTOR3 rot_axis = flight_vector;

	double current_ang_x = acos(dotp(_V(0,0,1), gspd_vect_loc_nm));
	double req_ang_x = acos(dotp(flight_vector, gspd_vect_loc_nm));
	double ang_sep_x = req_ang_x - current_ang_x;

	double ang_sep_y = -atan2(flight_vector.x,flight_vector.z);
	if(ang_sep_y > PI)ang_sep_y = ang_sep_y - PI2;
	else if(ang_sep_y < -PI)ang_sep_y = ang_sep_y + PI2;

	VECTOR3 zaxis_fv_norml = crossp(_V(0,0,1),gspd_vect_loc_nm);
	normalise(zaxis_fv_norml);
	headsup_vector = crossp(_V(0,0,1), zaxis_fv_norml);

	normalise(headsup_vector);
	double ang_sep_z = atan2(headsup_vector.x,headsup_vector.y);
	if(ang_sep_z > PI)ang_sep_z = ang_sep_z - PI2;
	else if(ang_sep_z < -PI)ang_sep_z = ang_sep_z + PI2;


	VECTOR3 req_ang_vect;
	double speed_factor = 1;
	req_ang_vect.x = 0.5*speed_factor*ang_sep_x;
	req_ang_vect.y = 0.5*speed_factor*ang_sep_y;
	req_ang_vect.z = 1*speed_factor*ang_sep_z;

	double acc_ratio = 1;
	double acc_ratio_temp = 1;
	
	VECTOR3 req_ang_acc;
	if(abs(req_ang_vect.x) > 1.0*vrot_lim) acc_ratio = 1.0*vrot_lim/abs(req_ang_vect.x);
	req_ang_acc.x = (req_ang_vect.x*acc_ratio) - avel.x;
	if(abs(req_ang_vect.y) > 1.0*vrot_lim) acc_ratio = 1.0*vrot_lim/abs(req_ang_vect.y);
	req_ang_acc.y = (req_ang_vect.y*acc_ratio) - avel.y;
	if(abs(req_ang_vect.z) > vrot_lim) acc_ratio = vrot_lim/abs(req_ang_vect.z);
	req_ang_acc.z = (req_ang_vect.z*acc_ratio) - avel.z;

	
	//double inertia_x = pmi.x * mass;
	//double inertia_y = pmi.y * mass;
	//double inertia_z = pmi.z * mass;

	//sprintf(oapiDebugString(), "max_acc_z_pos %0.1f max_acc_z_neg %0.1f", max_acc_z_pos*DEG, max_acc_z_neg*DEG);

	double max_acc_x_pos = max_torque_pup/inertia_x;
	double max_acc_x_neg = max_torque_pdn/inertia_x;
	double max_acc_y_pos = max_torque_yl/inertia_y;
	double max_acc_y_neg = max_torque_yr/inertia_y;
	double max_acc_z_pos = max_torque_br/inertia_z;
	double max_acc_z_neg = max_torque_bl/inertia_z;

	//sprintf(oapiDebugString(), "max_acc_x_pos %0.3f max_acc_x_neg %0.3f max_acc_y_pos %0.3f max_acc_y_neg %0.3f", max_acc_x_pos*DEG, max_acc_x_neg*DEG, max_acc_y_pos*DEG, max_acc_y_neg*DEG);

	double max_drot_x_pos = max_acc_x_pos*step_time;
	double max_drot_x_neg = max_acc_x_neg*step_time;
	double max_drot_y_pos = max_acc_y_pos*step_time;
	double max_drot_y_neg = max_acc_y_neg*step_time;
	double max_drot_z_pos = max_acc_z_pos*step_time;
	double max_drot_z_neg = max_acc_z_neg*step_time;

	double p_level = 0;
	double y_level = 0;
	double b_level = 0;

	double pth_level = 0;
	double yth_level = 0;
	//double bth_level = 0;


	if(req_ang_acc.x > 0)
	{
		p_level = req_ang_acc.x/max_drot_x_pos;
		pth_level = p_level;
		if(p_level > 1)p_level = 1;
		else if(p_level < 0.1)p_level = 0;
		if(pth_level < 4)pth_level = 0;
		if(pth_level > 4)pth_level = 1;
		if(skydive_sw == 1)pth_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, pth_level);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		//tv_pitch_level = p_level;
		pitch_level = 1.5*p_level;
		if(pitch_level > 1)pitch_level = 1;

		//sprintf(oapiDebugString(), "req_ang_acc.x %0.4f max_drot_x_pos %0.4f p_level %0.4f", req_ang_acc.x*DEG, max_drot_x_pos*DEG, p_level);
	}
	else if(req_ang_acc.x < 0)
	{
		p_level = req_ang_acc.x/max_drot_x_neg;
		pth_level = p_level;
		if(p_level > 1)p_level = 1;
		else if(p_level < 0.1)p_level = 0;
		
		if(pth_level < 4)pth_level = 0;
		if(pth_level > 4)pth_level = 1;
		if(skydive_sw == 1)pth_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP,0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, pth_level);
		//tv_pitch_level = -p_level;
		pitch_level = 1.5*-p_level;
		if(pitch_level < -1)pitch_level = -1;

	}
	else
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP,0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		//tv_pitch_level = 0;
		pitch_level = 0;
	}

	

	if(req_ang_acc.y > 0)
	{
		y_level = req_ang_acc.y/max_drot_y_pos;
		yth_level = y_level;
		if(y_level > 1)y_level = 1;
		else if(y_level < 0.1)y_level = 0;
		if(yth_level < 8)yth_level = 0;
		if(yth_level > 8)yth_level = 1;
		if(skydive_sw == 1)yth_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 1*yth_level);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		//tv_yaw_level = y_level;
		yaw_level = y_level;
	}
	else if(req_ang_acc.y < 0)
	{
		y_level = req_ang_acc.y/max_drot_y_neg;
		yth_level = y_level;
		if(y_level > 1)y_level = 1;
		else if(y_level < 0.1)y_level = 0;
		if(yth_level < 8)yth_level = 0;
		if(yth_level > 8)yth_level = 1;
		if(skydive_sw == 1)yth_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 1*yth_level);
		//tv_yaw_level = -y_level;
		yaw_level = -y_level;
	}
	else
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		//tv_yaw_level = 0;
		yaw_level = 0;
	}


	if(req_ang_acc.z > 0)
	{
		b_level = req_ang_acc.z/max_drot_z_pos;
		if(b_level > 1)b_level = 1;
		else if(b_level < 0.2)b_level = 0;
		//bth_level = b_level;
		//if(bth_level < 0.9)bth_level = 0;
		//if(skydive_sw == 1)bth_level = 0;
		//SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, bth_level);
		//SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		//tv_bank_level = b_level;
		bank_level = -b_level;
	}
	else if(req_ang_acc.z < 0)
	{
		b_level = req_ang_acc.z/max_drot_z_neg;
		if(b_level > 1)b_level = 1;
		else if(b_level < 0.2)b_level = 0;
		//bth_level = b_level;
		//if(bth_level < 0.9)bth_level = 0;
		//if(skydive_sw == 1)bth_level = 0;
		//SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT,0);
		//SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, bth_level);
		//tv_bank_level = -b_level;
		bank_level = b_level;
	}
	else
	{
		//SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT,0);
		//SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		//tv_bank_level = 0;
		bank_level = 0;
	}

	//sprintf(oapiDebugString(), "pitch_level %0.3f yaw_level %0.3f bank_level %0.3f", pitch_level, yaw_level, bank_level);

}


void ss_tanker::DoBackFlip()
{
	VECTOR3 attrot_hrz = _V(0,0,0);

	OBJHANDLE hTarget = oapiGetObjectByName(tgtname);
	if (hTarget == NULL)
	{
		hTarget = oapiGetBaseByName(GetGravityRef(),tgtname);
		tgt_att_valid = false;
	}
	else //if(tgt_att_valid == true)
	{
	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hTarget);
	DWORD ac = v->AttachmentCount(false);
	if(ac == 0) tgt_att_valid = false;
	else
	{
	for (DWORD j = 0; j < ac; j++)
		{
		ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(false, j);
		const char *id = v->GetAttachmentId (hAtt);
		if (!strnicmp (id, "BST", 3))
			{
			VECTOR3 attpos, attdir, attrot, attpos_hrz;
			v->GetAttachmentParams(hAtt, attpos, attdir, attrot);
			attpos.x = -11.6;
			attpos.z = -4.8;
			v->HorizonRot(attpos,attpos_hrz);
			v->HorizonRot(attrot,attrot_hrz);
			double v_alt = v->GetAltitude();
			catch_offset = attpos_hrz;
			catch_offset.y = v_alt + attpos_hrz.y;
			tgt_att_valid = true;
			//sprintf(oapiDebugString(), "catch_offset %0.1f %0.1f %0.1f", catch_offset.x, catch_offset.y,catch_offset.z);
			break;
			}
		else if(j == (ac-1))
			{
				tgt_att_valid = false;
			}
		}
	}
	}

	if (hTarget == NULL) 
	{
		strcpy(tgtname, "NULL");
		backflip_sw = 0;
		return;
	}

	DeactivateNavmode(1);
	DeactivateNavmode(2);
	DeactivateNavmode(3);
	DeactivateNavmode(4);
	DeactivateNavmode(5);
	DeactivateNavmode(6);
	DeactivateNavmode(7);

	double alt_mr = GetAltitude(ALTMODE_MEANRAD);

	VECTOR3 tgt_pos_glob;
	oapiGetGlobalPos(hTarget, &tgt_pos_glob);
	VECTOR3 tgt_pos_loc;
	Global2Local(tgt_pos_glob, tgt_pos_loc);
	VECTOR3 tgt_pos_hrz;
	HorizonRot(tgt_pos_loc, tgt_pos_hrz);

	if(tgt_att_valid == true)
	{
		tgt_pos_hrz.x = tgt_pos_hrz.x + catch_offset.x;
		tgt_pos_hrz.y = tgt_pos_hrz.y + catch_offset.y - 50.55;
		tgt_pos_hrz.z = tgt_pos_hrz.z + catch_offset.z;
	//	sprintf(oapiDebugString(), "tgt_pos_hrz.y %0.1f",tgt_pos_hrz.y);
	}
	////////////// for rolltoazlanding /////////////////
	
	VECTOR3 tgt_pos_loc_roll = _V(0,0,0);
	
	if(tgt_att_valid == true)
	{
	HorizonInvRot(attrot_hrz, tgt_pos_loc_roll);
	az_angle = atan2(-tgt_pos_loc_roll.x,tgt_pos_loc_roll.y);
	if(az_angle < 0)az_angle = az_angle + PI2;
	else if(az_angle > PI2)az_angle = az_angle - PI2;
	if(abs(tgt_pos_hrz.y) < 600)RollToAzLanding();
	}
	////////////////////////////////////////////////

	double tgt_vert_dist = abs(tgt_pos_hrz.y);
	double tgt_lin_dist = length(_V(tgt_pos_hrz.x,0,tgt_pos_hrz.z));
	double tgt_abs_dist = length(tgt_pos_hrz);

	double tgt_vert_offset = 0.6*tgt_pos_hrz.y;
	VECTOR3 tgt_vect_hrz;
	tgt_vect_hrz.x = tgt_pos_hrz.x;
	tgt_vect_hrz.y = tgt_pos_hrz.y;// - tgt_vert_offset;
	tgt_vect_hrz.z = tgt_pos_hrz.z;


	VECTOR3 gspd_vect_loc;
	GetGroundspeedVector(FRAME_LOCAL, gspd_vect_loc);
	VECTOR3 gspd_vect_hrz;
	HorizonRot(gspd_vect_loc, gspd_vect_hrz);

	VECTOR3 tgt_vect_hrz_unit = tgt_vect_hrz;
	normalise(tgt_vect_hrz_unit);

	VECTOR3 gspd_vect_hrz_unit = gspd_vect_hrz;
	normalise(gspd_vect_hrz_unit);

	VECTOR3 tgt_gspd_norm_unit = crossp(gspd_vect_hrz_unit, tgt_vect_hrz_unit);
	normalise(tgt_gspd_norm_unit);

	double gspd_tgt_angle = acos(dotp(gspd_vect_hrz_unit,tgt_vect_hrz_unit));

	double gspd_az_hrz = atan2(gspd_vect_hrz.x,gspd_vect_hrz.z);
	if(gspd_az_hrz < 0)gspd_az_hrz = gspd_az_hrz + PI2;
	if(gspd_az_hrz > PI2)gspd_az_hrz = gspd_az_hrz - PI2;

	double tgt_az_hrz = atan2(tgt_vect_hrz.x,tgt_vect_hrz.z);
	if(tgt_az_hrz < 0)tgt_az_hrz = tgt_az_hrz + PI2;
	if(tgt_az_hrz > PI2)tgt_az_hrz = tgt_az_hrz - PI2;

	double tgt_az_hrz_adj = tgt_az_hrz - gspd_az_hrz;
	if(tgt_az_hrz_adj < 0)tgt_az_hrz_adj = tgt_az_hrz_adj + PI2;
	if(tgt_az_hrz_adj > PI2)tgt_az_hrz_adj = tgt_az_hrz_adj - PI2;
	double cos_taz = cos(tgt_az_hrz_adj);
	double sin_taz = sin(tgt_az_hrz_adj);
	double gspd_lateral = length(_V(gspd_vect_hrz.x,0,gspd_vect_hrz.z));

	double offset = 18.93;
	double alti = -tgt_pos_hrz.y - offset;
	
	double mass = GetMass();
	VECTOR3 weight_vect;
	GetWeightVector(weight_vect);
	double weight = length(weight_vect);
	double gravity = weight/mass;

	double max_thr = 0;
	if(uppermaincfg != 1)
	{
	uppermaincfg = 1;
	ConfigUpperMain();
	}


	if(tgt_vert_dist < 400)
	{
	if(raptor_sl_selection > 1 && GetThrusterGroupLevel(THGROUP_MAIN) < 0.4 && mass < 140000)
		{
		raptor_sl_selection = 1;
		ConfigUpperRaptorSL();
		}
	}
	else if(tgt_vert_dist < 900 && GetThrusterGroupLevel(THGROUP_MAIN) < 0.5 && mass < 150000)
	{
	if(raptor_sl_selection > 2)
		{
		raptor_sl_selection = 2;
		ConfigUpperRaptorSL();
		}
	}
	
	
	max_thr = GetThrusterMax(th_main[1]);

	double max_acc = max_thr/mass;

	double mainth = 0;
	double mode_sw_alt = 400;

	if(tgt_vert_dist > mode_sw_alt) ///////////////// initial flip
	{
	double req_vert_acc = (gspd_vect_hrz.y+50)*1*(gspd_vect_hrz.y+50)/(alti-(mode_sw_alt-20));
	if((gspd_vect_hrz.y+50) > 0)req_vert_acc = -req_vert_acc;
	req_vert_acc = req_vert_acc + gravity;

	//double tts_vert = (alti-(mode_sw_alt-1))/((abs(gspd_vect_hrz.y-60))*0.5);
	double tts_lin = cos_taz*2*(tgt_lin_dist-10)/gspd_lateral;//20
	if(tts_lin < 1.75)tts_lin = 1.75;//1.75

	VECTOR3 req_lin_acc =_V(-gspd_vect_hrz.x/tts_lin, 0, -gspd_vect_hrz.z/tts_lin);


	VECTOR3 req_course_corr = _V((sin_taz*gspd_vect_hrz.z),0,(sin_taz*-gspd_vect_hrz.x));

	if(tgt_lin_dist > 50) req_lin_acc = req_lin_acc + req_course_corr;

	//sprintf(oapiDebugString(), "roty_angle %0.1f",roty_angle*DEG);

	double vert_lin_ratio = req_vert_acc/length(req_lin_acc);
	if(vert_lin_ratio < 0.5)
	{
		req_lin_acc.x = req_lin_acc.x*vert_lin_ratio*2;
		req_lin_acc.z = req_lin_acc.z*vert_lin_ratio*2;
	}

	VECTOR3 req_acc_vect = _V(req_lin_acc.x, req_vert_acc, req_lin_acc.z);
	VECTOR3 req_acc_vect_unit = req_acc_vect;
	normalise(req_acc_vect_unit);

	double req_acc_abs = length(req_acc_vect);
	mainth = req_acc_abs/max_acc;

	double th_lim = mass*40/max_thr;
	//sprintf(oapiDebugString(), "th_lim_init %0.1f",th_lim);
	if(th_lim > 1)th_lim = 1;
	else if(th_lim < 0)th_lim = 0;
	if(mainth > th_lim)mainth = th_lim;

	if(mainth > 1)mainth = 1;
	else if(mainth < 0)mainth = 0;

	double current_mainth = GetThrusterGroupLevel(THGROUP_MAIN);
	if (mainth > current_mainth)
	{
		current_mainth = current_mainth + (0.5*step_time);
		if(current_mainth > mainth)current_mainth = mainth;
	}
	else if (mainth < current_mainth)
	{
		current_mainth = current_mainth - (0.5*step_time);
		if(current_mainth < mainth)current_mainth = mainth;
	}
	//main_throttle_level = mainth;
	mainth = current_mainth;
	SetThrusterGroupLevel(THGROUP_MAIN,mainth);

	HorizonInvRot(req_acc_vect_unit,flight_vector);
	GoToFlightVectorBF();

	//double acc_eng = mainth*max_thr/mass;
	//sprintf(oapiDebugString(), "acc_eng %0.1f",acc_eng);
	}

	if(tgt_vert_dist < mode_sw_alt) //////////// landing
	{
	adcs_reentry_mode = 1;

	VECTOR3 rot_axis = tgt_gspd_norm_unit;
	double rot_angle = -2*abs(gspd_tgt_angle);
	if(rot_angle > (5*RAD)) rot_angle = (5*RAD);
	else if(rot_angle < (-5*RAD)) rot_angle = (-5*RAD);

	if(tgt_vert_dist < 150)
		{
		rot_angle = 0;
		if(legs_status < 2 && tgt_att_valid == false)legs_status = 1;
		}

	if(tgt_vert_dist < 50)
		{
		rot_angle = -2*abs(acos(dotp(gspd_vect_hrz_unit, _V(0,-1,0))));
		if(rot_angle > (2*RAD)) rot_angle = (2*RAD);
		else if(rot_angle < (-2*RAD)) rot_angle = (-2*RAD);
		rot_axis = crossp(gspd_vect_hrz_unit, _V(0,-1,0));
		}
	

	MATRIX3 rot_mat = rotm(rot_axis, rot_angle);
	VECTOR3 flight_vector_hrz = mul(rot_mat, gspd_vect_hrz_unit);

	if(alti < 30) flight_vector_hrz = _V(0,-1,0);

	flight_vector_hrz = -flight_vector_hrz;
	normalise(flight_vector_hrz);

	HorizonInvRot(flight_vector_hrz,flight_vector); // goto fv
	GoToFlightVectorBF();

//////////////////// do throttle ////////////////////////
	VECTOR3 z_local = _V(0,0,1);
	VECTOR3 z_hrz;
	HorizonRot(z_local, z_hrz);
	normalise(z_hrz);
	double pitch_angle_cos = dotp(_V(0,1,0), z_hrz);

	double req_acc = 1*((gspd_vect_hrz.y+20)*(gspd_vect_hrz.y+20))/(alti-100);
	if((gspd_vect_hrz.y+20) > 0)req_acc = -req_acc;
	if(alti < 100)
	{
	req_acc = 1*(((gspd_vect_hrz.y+2)*(gspd_vect_hrz.y+2)))/(alti-1);
	if((gspd_vect_hrz.y+2) > 0)req_acc = -req_acc;
	}
	if(alti < 2)
	{
	req_acc = 0;
	SetADCtrlMode(0);
	}
	else if(mainth < 0)mainth = 0;

	


	req_acc = req_acc + gravity;
	mainth = req_acc/(max_acc*abs(pitch_angle_cos));

	double th_lim = mass*40/max_thr;
	//sprintf(oapiDebugString(), "th_lim %0.1f",th_lim);
	if(th_lim > 1)th_lim = 1;
	else if(th_lim < 0)th_lim = 0;
	if(mainth > th_lim)mainth = th_lim;

	//main_throttle_level = mainth; /////////////////// apply throttle
	double current_mainth = GetThrusterGroupLevel(THGROUP_MAIN);
	if (mainth > current_mainth)
	{
		current_mainth = current_mainth + (0.5*step_time);
		if(current_mainth > mainth)current_mainth = mainth;
	}
	else if (mainth < current_mainth)
	{
		current_mainth = current_mainth - (0.5*step_time);
		if(current_mainth < mainth)current_mainth = mainth;
	}
	
	mainth = current_mainth;
	SetThrusterGroupLevel(THGROUP_MAIN,mainth);
	
	if(GetThrusterGroupLevel(THGROUP_MAIN) == 0)SetThrusterGroupLevel(THGROUP_MAIN,0.5);
	
	if (GroundContact() == true)
		{
		//main_throttle_level = 0;
		//SetThrusterLevel(th_main[1],0);
		SetThrusterGroupLevel(THGROUP_MAIN,0);
		backflip_sw = 0;
		wing_trim = 0;
		KillAllRCS2 ();
		}
	}
	////////////// catch shutdown
	OBJHANDLE hArmAttach = GetAttachmentStatus(arm_attach);
	if(hArmAttach != NULL)
	{
	SetThrusterGroupLevel(THGROUP_MAIN,0);
	backflip_sw = 0;
	wing_trim = 0;
	KillAllRCS2 ();
	}
}

void ss_tanker::DoRetroLand()
{
	VECTOR3 grndspd_vect;
	GetGroundspeedVector(FRAME_HORIZON, grndspd_vect);
	VECTOR3 grndspd_vect_norm = grndspd_vect;
	normalise(grndspd_vect_norm);
	VECTOR3 rot_axis = crossp(_V(0,-1,0),grndspd_vect_norm);
	normalise(rot_axis);
	double descent_angle = acos(dotp(_V(0,-1,0), grndspd_vect_norm));
	VECTOR3 rot_axis_local;
	HorizonInvRot(rot_axis, rot_axis_local);
	headsup_vector = crossp(rot_axis_local, _V(0,0,1));
	
	double rot_angle = 0.2*abs(descent_angle);
	if(rot_angle > (10*RAD)) rot_angle = (10*RAD);
	MATRIX3 rot_mat = rotm(rot_axis, rot_angle);
	VECTOR3 flight_vector_hrz = mul(rot_mat, grndspd_vect_norm);
	flight_vector_hrz = -flight_vector_hrz;
	normalise(flight_vector_hrz);
	HorizonInvRot(flight_vector_hrz,flight_vector);
	GoToFlightVectorBF();

	//main_throttle_level
	VECTOR3 weight_vect;
	GetWeightVector(weight_vect);
	double weight = length(weight_vect);
	double offset = 18.28;
	double alti = GetAltitude(ALTMODE_GROUND) - offset;
	//double max_thr = 3*GetThrusterMax(th_main_top);
	double max_thr = 0;
	if(uppermaincfg == 0)
	{
	max_thr = GetThrusterMax(th_main[1]);
	max_thr = max_thr + GetThrusterMax(th_main[2]);
	}
	else if (uppermaincfg == 1)
	{
	max_thr = GetThrusterMax(th_main[1]);
	}
	else if (uppermaincfg == 2)
	{
	max_thr = GetThrusterMax(th_main[2]);
	}
	double mass = GetMass();
	double gravity = weight/mass;
	double max_acc = max_thr/mass;
	double nominal_acc = 0.3*max_acc;

	if(alti < 100)
	{
		if(legs_status != 2)legs_status = 1;
	}

	double time_to_stop = abs(grndspd_vect.y)/nominal_acc;
	double d_alt = time_to_stop*-grndspd_vect.y*0.5;
	if(d_alt > alti)
	{
		if(retroland_sw == 1)retroland_sw = 2;
	}

	if(retroland_sw == 2)
	{
	VECTOR3 z_local = _V(0,0,1);
	VECTOR3 z_hrz;
	HorizonRot(z_local, z_hrz);
	normalise(z_hrz);
	double pitch_angle_cos = dotp(_V(0,1,0), z_hrz);
	double req_acc = 0.5*((grndspd_vect.y*grndspd_vect.y)-1)/alti;

	req_acc = req_acc + gravity;

	double mainth = req_acc/(max_acc*pitch_angle_cos);
	if(alti < 2) mainth = gravity/(max_acc*pitch_angle_cos);
	else if(mainth < 0)mainth = 0.1;
//	main_throttle_level = mainth;
	SetThrusterGroupLevel(THGROUP_MAIN,mainth);
	
	if (GroundContact() == true)
		{
		//main_throttle_level = 0; //SetThrusterGroupLevel(THGROUP_MAIN, 0);
		SetThrusterGroupLevel(THGROUP_MAIN,0);
		retroland_sw = 0;
		}
	}

	OBJHANDLE hArmAttach = GetAttachmentStatus(arm_attach);
	if(hArmAttach != NULL)
		{
		//main_throttle_level = 0; //SetThrusterGroupLevel(THGROUP_MAIN, 0);
		SetThrusterGroupLevel(THGROUP_MAIN,0);
		retroland_sw = 0;
		}
}

void ss_tanker::GoToFlightVectorBF()
{
	// check if main on
	bool main_is_on = true;
	double mainth_lev = GetThrusterGroupLevel(THGROUP_MAIN);
	if(mainth_lev == 0)main_is_on = false;
	
	// rot rate limit
	double vrot_lim = 10*RAD;
	if(GetAltitude(ALTMODE_MEANRAD) > 50000)vrot_lim = 5*RAD;
	if(main_is_on == true)vrot_lim = mainth_lev*45*RAD;
	if(backflip_sw == 1)vrot_lim = 60*RAD;

	// turn off other ap
	DeactivateNavmode(1);
	DeactivateNavmode(2);
	DeactivateNavmode(3);
	DeactivateNavmode(4);
	DeactivateNavmode(5);
	DeactivateNavmode(6);
	DeactivateNavmode(7);
	/////////////////////////////

	double mass = GetMass();

	VECTOR3 current_pmi;
	GetPMI(current_pmi);
	double inertia_x = current_pmi.x*mass;
	double inertia_y = current_pmi.y*mass;
	double inertia_z = current_pmi.z*mass;
	
	double max_rcs = GetThrusterMax(th_rcs[8]);
	double max_torque_x = 1*max_rcs*16;
	double max_torque_y = 1*max_rcs*16;
	double max_torque_z = 2*max_rcs*1;
	
	if (rcs_sw2 == 1)
		{
		max_rcs = GetThrusterMax(th_rcs[16]);
		max_torque_x = 2*max_rcs*16;
		max_torque_y = 2*max_rcs*16;
		max_torque_z = 2*max_rcs*5;
		if(disable_rcs == 1)max_torque_y = 1*max_rcs*16;
		}
	

	

	double max_torque_pup = abs(max_torque_x);
	double max_torque_pdn = -abs(max_torque_x);
	double max_torque_yl = abs(max_torque_y);
	double max_torque_yr = -abs(max_torque_y);
	double max_torque_br = abs(max_torque_z);
	double max_torque_bl = -abs(max_torque_z);

	//sprintf(oapiDebugString(), "max_torque_pup %0.1f max_torque_pdn %0.1f max_torque_yl %0.1f max_torque_yr %0.1f", max_torque_pup, max_torque_pdn, max_torque_yl, max_torque_yr);
	/*if(main_is_on == true)
	{
	max_rcs = GetThrusterMax(th_rcs_main[0]);
	max_torque_pup = 16.7 * max_rcs * mainth_lev;
	max_torque_pdn = -16.7 * max_rcs * mainth_lev;
	max_torque_yl = 16.7 * max_rcs * mainth_lev;
	max_torque_yr = -16.7 * max_rcs * mainth_lev;
	}*/

	
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);

	normalise(flight_vector);
	VECTOR3 rot_axis = crossp(flight_vector, _V(0,0,1));
	normalise(rot_axis);
	double ang_sep = acos(dotp(_V(0,0,1), flight_vector));

	double req_ang_vel = 0.30*((abs(ang_sep)));
	if(GetAltitude(ALTMODE_MEANRAD) < 50000)req_ang_vel = 1.0*((abs(ang_sep)));
	if(req_ang_vel > vrot_lim) req_ang_vel = vrot_lim;
	

	VECTOR3 req_ang_vect;
	req_ang_vect.x = req_ang_vel*rot_axis.x;
	req_ang_vect.y = req_ang_vel*rot_axis.y;
	req_ang_vect.z = req_ang_vel*rot_axis.z;

	VECTOR3 req_ang_acc;
	req_ang_acc.x = req_ang_vect.x - avel.x;
	req_ang_acc.y = req_ang_vect.y - avel.y;
	req_ang_acc.z = 0;

	VECTOR3 acc_ratio = req_ang_acc;
	normalise(acc_ratio);
	acc_ratio.x = abs(acc_ratio.x);
	acc_ratio.y = abs(acc_ratio.y);

	
	//double inertia_x = pmi.x * mass;
	//double inertia_y = pmi.y * mass;

	double max_acc_x_pos = max_torque_pup/inertia_x;
	double max_acc_x_neg = max_torque_pdn/inertia_x;
	double max_acc_y_pos = max_torque_yl/inertia_y;
	double max_acc_y_neg = max_torque_yr/inertia_y;

	//sprintf(oapiDebugString(), "max_acc_x_pos %0.3f max_acc_x_neg %0.3f max_acc_y_pos %0.3f max_acc_y_neg %0.3f", max_acc_x_pos*DEG, max_acc_x_neg*DEG, max_acc_y_pos*DEG, max_acc_y_neg*DEG);

	double max_drot_x_pos = max_acc_x_pos*step_time;
	double max_drot_x_neg = max_acc_x_neg*step_time;
	double max_drot_y_pos = max_acc_y_pos*step_time;
	double max_drot_y_neg = max_acc_y_neg*step_time;

	double p_level = 0;
	double y_level = 0;

	if(req_ang_acc.x > 0)
	{
		p_level = acc_ratio.x*req_ang_acc.x/max_drot_x_pos;
		if(p_level > 1)p_level = 1;
		else if(p_level < 0.01)p_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, p_level);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		//tv_pitch_level = p_level;
		pitch_level = p_level;

		//sprintf(oapiDebugString(), "req_ang_acc.x %0.4f max_drot_x_pos %0.4f p_level %0.4f", req_ang_acc.x*DEG, max_drot_x_pos*DEG, p_level);
	}
	else if(req_ang_acc.x < 0)
	{
		p_level = acc_ratio.x*req_ang_acc.x/max_drot_x_neg;
		if(p_level > 1)p_level = 1;
		else if(p_level < 0.01)p_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP,0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, p_level);
		//tv_pitch_level = -p_level;
		pitch_level = -p_level;
	}
	else
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP,0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		//tv_pitch_level = 0;
		pitch_level = 0;
	}

	

	if(req_ang_acc.y > 0)
	{
		y_level = acc_ratio.y*req_ang_acc.y/max_drot_y_pos;
		if(y_level > 1)y_level = 1;
		else if(y_level < 0.01)y_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, y_level);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		//tv_yaw_level = y_level;
		yaw_level = y_level;
	}
	else if(req_ang_acc.y < 0)
	{
		y_level = acc_ratio.y*req_ang_acc.y/max_drot_y_neg;
		if(y_level > 1)y_level = 1;
		else if(y_level < 0.01)y_level = 0;
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, y_level);
		//tv_yaw_level = -y_level;
		yaw_level = -y_level;
	}
	else
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		//tv_yaw_level = 0;
		yaw_level = 0;
	}
	pitch_level = 0;
	yaw_level = 0;
	bank_level = 0;
	wing_trim = 0;
}

void ss_tanker::DoReentryNT()
{
	adcs_reentry_mode = 1;

	double alt = GetAltitude(ALTMODE_MEANRAD);

	VECTOR3 vel_rel, pos_rel; 
	GetAirspeedVector(FRAME_LOCAL, vel_rel);
	double speed = length(vel_rel);

	VECTOR3 vel_unit, pos_unit, nml_unit;

	vel_unit = vel_rel;
	normalise(vel_unit);

	HorizonInvRot(_V(0,1,0),pos_rel);
	VECTOR3 vel_pos_nml = crossp(vel_unit, pos_rel);
	
	pos_unit = pos_rel;
	normalise(pos_unit);
	nml_unit = vel_pos_nml;
	normalise(nml_unit);

	double bank_angle = 0;
	double pitch_angle = 70*RAD;
	//if (speed < 500)pitch_angle = 40*RAD;

	/////////////////// set wing trim to pitch angle
	double req_aoa = pitch_angle*DEG;
	//double wtf = 22;
	//if(req_aoa < 58)wtf = 42.5;
	//wing_trim = (req_aoa - 58)*(1/wtf);
	wing_trim = GetWTbyAoA(req_aoa);

	if(alt < 160000)
	{
	if(alt > 140000)bank_angle = 0*RAD;
	else if(alt > 120000)bank_angle = 0*RAD;
	else if(alt > 100000)bank_angle = 45*RAD;
	else if(alt > 80000)bank_angle = -45*RAD;
	else if(alt > 60000)bank_angle = 45*RAD;
	else if(alt > 50000)bank_angle = -45*RAD;
	else if(alt > 40000)bank_angle = 45*RAD;
	else if(alt > 30000)bank_angle = -45*RAD;
	else if(alt > 20000)bank_angle = 0*RAD;
	else bank_angle = 0;
	if(alt < 20000)
		{
		reentry_sw = 0;
		KillAllRCS2();
		adcs_reentry_mode = 0;
		return;
		}
	}

	MATRIX3 rm = rotm(nml_unit, pitch_angle);
	flight_vector = mul(rm,vel_unit);
	rm = rotm(vel_unit,bank_angle);
	flight_vector = mul(rm, flight_vector);
	normalise(flight_vector);

	if(alt < 40000)SetADCtrlMode(7);
	GoToFlightVectorRE();
}

void ss_tanker::DoReentry()
{
	
	double alt = GetAltitude(ALTMODE_MEANRAD);
	
	VECTOR3 vel_loc, pos_loc; 
	GetAirspeedVector(FRAME_LOCAL, vel_loc);
	double speed = length(vel_loc);

	VECTOR3 vel_unit, pos_loc_unit, nml_unit;

	vel_unit = vel_loc;
	normalise(vel_unit);

	HorizonInvRot(_V(0,1,0),pos_loc);
	VECTOR3 vel_pos_nml = crossp(vel_unit, pos_loc);
	
	pos_loc_unit = pos_loc;
	normalise(pos_loc_unit);
	nml_unit = vel_pos_nml;
	normalise(nml_unit);

	double bank_angle = 0;
	double pitch_angle = 70*RAD;

	//////// adc help
	if (speed < 5000)SetADCtrlMode(7);
	
	
	///// targeting
	OBJHANDLE gbod = GetGravityRef();
	OBJHANDLE hTgt = oapiGetObjectByName(tgtname);
	if(hTgt == NULL) hTgt = oapiGetBaseByName(gbod, tgtname);
	if(hTgt == NULL) 
	{
		tgt_exists = false;
		DoReentryNT();
		return;
	}
	tgt_exists = true;
	double gbod_rad = oapiGetSize(gbod);
	VECTOR3 pos_rel,pos_rel_unit, vel_rel, vel_rel_unit;
	GetRelativePos(gbod, pos_rel);
	pos_rel_unit = pos_rel;
	normalise(pos_rel_unit);
	GetGroundspeedVector(FRAME_GLOBAL, vel_rel);
	vel_rel_unit = vel_rel;
	normalise(vel_rel_unit);
	VECTOR3 tgt_pos_rel, tgt_pos_rel_unit;
	oapiGetRelativePos(hTgt, gbod, &tgt_pos_rel);
	tgt_pos_rel_unit = tgt_pos_rel;
	normalise(tgt_pos_rel_unit);

	///////// range
	double tgt_angle = acos(dotp(tgt_pos_rel_unit, pos_rel_unit));
	double tgt_range = tgt_angle*gbod_rad;
	tgt_dist = tgt_range;

	///// generate bank correction
	if(tgt_dist < 5000000)DoEnergyManagement();
	if(bank_correction < -30)bank_correction = -30;
	else if(bank_correction > 30)bank_correction = 30;
	bank_correction = bank_correction - 10;

	VECTOR3 vel_hrzn;
	GetGroundspeedVector(FRAME_HORIZON, vel_hrzn);

	if(vel_hrzn.y > -40)
	{
	reentry_vspd_sw = 1;
	}
	bank_correction = bank_correction*reentry_vspd_sw;
	
	VECTOR3 orb_nml = crossp(vel_rel_unit, pos_rel_unit);
	normalise(orb_nml);
	double cross_range_angle = PI05 - acos(dotp(orb_nml, tgt_pos_rel_unit));
	double cross_range_dist = gbod_rad*cross_range_angle;

	if(tgt_dist < 210000)
	{
	VECTOR3 tgt_pos_glob;
	oapiGetGlobalPos(hTgt, &tgt_pos_glob);
	VECTOR3 tgt_pos_loc;
	Global2Local(tgt_pos_glob, tgt_pos_loc);
	VECTOR3 tgt_pos_hrz;
	HorizonRot(tgt_pos_loc, tgt_pos_hrz);

	double vel_az = atan2(vel_hrzn.x,vel_hrzn.z);
	double tgt_az = atan2(tgt_pos_hrz.x,tgt_pos_hrz.z);
	cross_range_angle = vel_az - tgt_az;
	if(cross_range_angle > PI)cross_range_angle = cross_range_angle - PI2;
	else if(cross_range_angle < -PI)cross_range_angle = cross_range_angle + PI2;

	cross_range_dist = sin(cross_range_angle)*tgt_dist;
	}

	
	
	////// cross range limit ///////////////////////////////////////////////////////

	double cross_range_angle_limit = (0.18*RAD);
	cross_range_angle_limit = cross_range_angle_limit * (1 + sin(bank_correction*RAD)*1.5);

	if(tgt_dist < 300000) cross_range_angle_limit = (0.004*RAD);
	else if(tgt_dist < 210000) cross_range_angle_limit = (5*RAD);

	if(abs(cross_range_angle) > cross_range_angle_limit)
	{
	bank_dir_sw = int(abs(cross_range_angle)/cross_range_angle);
	}

	if(bank_mode_sw == 0)bank_mode_sw = bank_dir_sw;
	else if (bank_mode_sw != bank_dir_sw) bank_mode_sw = 2;

	/////////////////////////////////////////////////////////////////////////////////////////////////////

	double hSpeed = length(_V(vel_hrzn.x,0,vel_hrzn.z));
	double vSpeed = vel_hrzn.y;

	double current_h_acc =  (hSpeed - old_hSpeed)/post_step;
	double current_v_acc = (vSpeed - old_vSpeed)/post_step;
	old_hSpeed = hSpeed;
	old_vSpeed = vSpeed;

	
	double payload_factor = (GetMass() - 145000)*0.002;//100
	if(payload_factor < 0)payload_factor = 0;

	double reentry_time = (tgt_dist-200000)/((hSpeed+2150+payload_factor)/2);

	double h_acc_req = 2*((tgt_dist-200000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
	double h_acc_error = h_acc_req - current_h_acc;

	if(tgt_dist < 250000)
	{
		reentry_time = (tgt_dist-10000)/((hSpeed+400)/2);
		h_acc_req = 2*((tgt_dist-10000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
	}

	double v_speed_req = (20*h_acc_error) + vSpeed_adjust;
	if(v_speed_req > 200)v_speed_req = 200;
	else if(v_speed_req < -200)v_speed_req = -200;


	double v_speed_error = v_speed_req - vSpeed;
	if(abs(v_speed_error) < 1) vSpeed_adjust = (2*h_acc_error);
	double v_acc_req = v_speed_error/20;
	double v_acc_error = v_acc_req - current_v_acc;

	if(alt < 45000)
	{
	v_speed_req = vSpeed + (20*h_acc_error);
	v_speed_error = v_speed_req - vSpeed;
	v_acc_req = v_speed_error/10;
	v_acc_error = v_acc_req - current_v_acc;
	}

	double bank_adjust = 20*abs(v_acc_error);
	if(bank_adjust > 30)bank_adjust = 30;

	double current_bank_angle = DEG*(acos(dotp(nml_unit, _V(-1,0,0))));

	if(v_acc_error > 0.1)bank_angle = current_bank_angle - bank_adjust;
	else if(v_acc_error < -0.1)bank_angle = current_bank_angle + bank_adjust;
	else bank_angle = current_bank_angle;
	
	double min_bank_angle = 15;
	if(abs(cross_range_angle) < (0.25*RAD))min_bank_angle = 0;

	if(tgt_dist < 210000)
	{
		min_bank_angle = 15;
		if(abs(cross_range_angle) < (2*RAD))min_bank_angle = 0;
		
	}

	if (bank_angle < min_bank_angle) bank_angle = min_bank_angle;
	else if (bank_angle > 120) bank_angle = 120;

	hud_tgt_range = tgt_range;
	hud_crossrange_error = cross_range_dist;
	hud_hacc_req = h_acc_req;
	hud_hacc_current = current_h_acc;
	hud_hacc_error = h_acc_error;
	
	bank_angle = bank_angle*bank_dir_sw*RAD;
	if(bank_mode_sw == 2)bank_angle = 0;
	
	//////////////////////////// termination - skydive //////////////////////////////

	if(alt < 10000) // emergency low alt termination
		{
		skydive_sw = 1;
		reentry_sw = 0;
		KillAllRCS2();
		adcs_reentry_mode = 0;
		return;
		}
	
	////////////////////////////////////// aoa reduction

	double ttt = (tgt_dist - (alt*0.5))/hSpeed;

	if(ttt < 15)pitch_angle = 55*RAD;
	else if(ttt < 22)pitch_angle = 60*RAD;
	else if(ttt < 30)pitch_angle = 65*RAD;

	/////////////////// set wing trim to pitch angle
	double req_aoa = pitch_angle*DEG;
	wing_trim = GetWTbyAoA(req_aoa);
	/////////////////////////////////////////////////

	// switch to skydive
	if(GetMachNumber() < 1.8)
	{
		reentry_sw = 0;
		skydive_sw = 1;
		adcs_reentry_mode = 0;
		KillAllRCS2();
		return;
	}

	//sprintf(oapiDebugString(), "ttt %0.2f cur_bank %0.2f bank %0.2f h_acc_req %0.2f h_acc_error %0.2f v_acc_req %0.2f v_acc_error %0.2f v_acc %0.2f cra %0.2f",ttt, current_bank_angle,bank_angle*DEG, h_acc_req,h_acc_error,v_acc_req, v_acc_error,current_v_acc,cross_range_angle*DEG);

	MATRIX3 rm = rotm(nml_unit, pitch_angle);
	flight_vector = mul(rm,vel_unit);
	rm = rotm(vel_unit,bank_angle);
	flight_vector = mul(rm, flight_vector);
	normalise(flight_vector);

	GoToFlightVectorRE();
}

void ss_tanker::DoEnergyManagement()
{
		double altd = GetAltitude();
		VECTOR3 rvelg;
		GetRelativeVel(hGref,rvelg);
		double v2 = length(rvelg);
		v2 = v2*v2;
		double ke = 0.5*v2;
		double pe = mu/(rt+altd);
		double tot_energy = ke+pe;

		double req_energy = 0;
		double energy_diff_abs = 0;
		double energy_diff_ratio = 0;
		int idx = 0;
		double dst_past_tld = 0;
		int upper_step = 0;
		int lower_step = 0;
		double gradient = 0;
		double delta_j = 0;
		VECTOR3 gspd_hrz = _V(0,0,0);
		GetGroundspeedVector(FRAME_HORIZON,gspd_hrz);
		gspd_hrz.y = 0;
		double xz_spd_hrz = length(gspd_hrz);
		//double energy_correct_time = 20;
		double energy_correct_time = 0.1*tgt_dist/xz_spd_hrz;
		int inc_dec = 0;
		//double bcf = 1;
		

	if (tgt_dist < 1000000)
	{
		dst_past_tld = 1000000 - tgt_dist;
		upper_step = (int) (dst_past_tld/step_1000);
		lower_step = upper_step + 1;
		if (lower_step > (ndp_1000-1)) lower_step = upper_step;
		gradient = (J_1000[lower_step] - J_1000[upper_step])/step_1000;
		delta_j = (dst_past_tld - (upper_step*step_1000))*gradient;
		req_energy = J_1000[upper_step] + delta_j;
		energy_diff_abs = req_energy - tot_energy;
		energy_diff_ratio = tot_energy/req_energy;
	}

	else if (tgt_dist < 3000000)
	{
		dst_past_tld = 3000000 - tgt_dist;
		upper_step = (int) (dst_past_tld/step_3000);
		lower_step = upper_step + 1;
		if (lower_step > (ndp_3000-1)) lower_step = upper_step;
		gradient = (J_3000[lower_step] - J_3000[upper_step])/step_3000;
		delta_j = (dst_past_tld - (upper_step*step_3000))*gradient;
		req_energy = J_3000[upper_step] + delta_j;
		energy_diff_abs = req_energy - tot_energy;
		energy_diff_ratio = tot_energy/req_energy;
	}

	else
	{
		dst_past_tld = 5000000 - tgt_dist;
		upper_step = (int) (dst_past_tld/step_5000);
		lower_step = upper_step + 1;
		if (lower_step > (ndp_5000-1)) lower_step = upper_step;
		gradient = (J_5000[lower_step] - J_5000[upper_step])/step_5000;
		delta_j = (dst_past_tld - (upper_step*step_5000))*gradient;
		req_energy = J_5000[upper_step] + delta_j;
		energy_diff_abs = req_energy - tot_energy;
		energy_diff_ratio = tot_energy/req_energy;
	}
	
	if (tot_energy_old == 0)tot_energy_old = tot_energy;
	double curr_energy_dt = (tot_energy - tot_energy_old)/post_step;
	tot_energy_old = tot_energy;
	double req_energy_adjust_dt = energy_diff_abs/energy_correct_time;

	if (dst_past_tld_old == 0)dst_past_tld_old = dst_past_tld;
	double nominal_energy_dt = gradient*(dst_past_tld - dst_past_tld_old)/post_step;
	dst_past_tld_old = dst_past_tld;

	double net_req_energy_dt =  nominal_energy_dt+req_energy_adjust_dt;

	//if (abs(net_req_energy_dt-curr_energy_dt) < 1000)bcf = 0.5;
	//else bcf = 1;

	double diff_abs_ratio = abs(net_req_energy_dt-curr_energy_dt)/abs(curr_energy_dt);

	if (net_req_energy_dt > curr_energy_dt)inc_dec =  -1;
	else if (net_req_energy_dt < curr_energy_dt)inc_dec = 1;

	//digi_dar = 0;
	double new_digi_dar = ((int)(6*diff_abs_ratio))*2;
	if (new_digi_dar == digi_dar_old) digi_dar_check_sw = digi_dar_check_sw + 1;
	else digi_dar_check_sw = 0;
	if (digi_dar_check_sw > 5)digi_dar_check_sw = 5;
	
	if (digi_dar_check_sw == 5) digi_dar = new_digi_dar;
	digi_dar_old = new_digi_dar;

	bank_correction = inc_dec*(0+digi_dar);
	//bank_correction = 0;

	//sprintf(oapiDebugString(), "bank_correction %0.1f net_req_energy_dt %0.4f curr_e_dt %0.4f diff_abs_ratio %0.4f", bank_correction, net_req_energy_dt, curr_energy_dt, diff_abs_ratio);

}

void ss_tanker::DoFBWReentry()
{
	
	double alt = GetAltitude(ALTMODE_MEANRAD);
	
	VECTOR3 vel_loc, pos_loc; 
	GetAirspeedVector(FRAME_LOCAL, vel_loc);
	double speed = length(vel_loc);

	VECTOR3 vel_unit, pos_loc_unit, nml_unit;

	vel_unit = vel_loc;
	normalise(vel_unit);

	HorizonInvRot(_V(0,1,0),pos_loc);
	VECTOR3 vel_pos_nml = crossp(vel_unit, pos_loc);
	
	pos_loc_unit = pos_loc;
	normalise(pos_loc_unit);
	nml_unit = vel_pos_nml;
	normalise(nml_unit);

	//double bank_angle = 0;
	//double pitch_angle = 70*RAD;

	//////// adc help
	//if (speed < 5000)SetADCtrlMode(7);
	
	
	///// targeting
	OBJHANDLE gbod = GetGravityRef();
	OBJHANDLE hTgt = oapiGetObjectByName(tgtname);
	if(hTgt == NULL) hTgt = oapiGetBaseByName(gbod, tgtname);
	if(hTgt == NULL) tgt_exists = false;
	else tgt_exists = true;

	double gbod_rad = oapiGetSize(gbod);
	VECTOR3 pos_rel,pos_rel_unit, vel_rel, vel_rel_unit;
	GetRelativePos(gbod, pos_rel);
	pos_rel_unit = pos_rel;
	normalise(pos_rel_unit);
	GetGroundspeedVector(FRAME_GLOBAL, vel_rel);
	vel_rel_unit = vel_rel;
	normalise(vel_rel_unit);

	
	VECTOR3 tgt_pos_rel = _V(0,0,0);
	VECTOR3	tgt_pos_rel_unit = _V(0,0,0);
	double tgt_angle = 0;
	double tgt_range = 0;

	if(tgt_exists)
	{
	oapiGetRelativePos(hTgt, gbod, &tgt_pos_rel);
	tgt_pos_rel_unit = tgt_pos_rel;
	normalise(tgt_pos_rel_unit);

	///////// range
	tgt_angle = acos(dotp(tgt_pos_rel_unit, pos_rel_unit));
	tgt_range = tgt_angle*gbod_rad;
	tgt_dist = tgt_range;
	}


	VECTOR3 vel_hrzn;
	GetGroundspeedVector(FRAME_HORIZON, vel_hrzn);
	
	double cross_range_angle = 0;
	double cross_range_dist = 0;

	if(tgt_exists)
	{
	VECTOR3 orb_nml = crossp(vel_rel_unit, pos_rel_unit);
	normalise(orb_nml);
	cross_range_angle = PI05 - acos(dotp(orb_nml, tgt_pos_rel_unit));
	cross_range_dist = gbod_rad*cross_range_angle;
	

	if(tgt_dist < 210000)
		{
	VECTOR3 tgt_pos_glob;
	oapiGetGlobalPos(hTgt, &tgt_pos_glob);
	VECTOR3 tgt_pos_loc;
	Global2Local(tgt_pos_glob, tgt_pos_loc);
	VECTOR3 tgt_pos_hrz;
	HorizonRot(tgt_pos_loc, tgt_pos_hrz);

	double vel_az = atan2(vel_hrzn.x,vel_hrzn.z);
	double tgt_az = atan2(tgt_pos_hrz.x,tgt_pos_hrz.z);
	cross_range_angle = vel_az - tgt_az;
	if(cross_range_angle > PI)cross_range_angle = cross_range_angle - PI2;
	else if(cross_range_angle < -PI)cross_range_angle = cross_range_angle + PI2;

	cross_range_dist = sin(cross_range_angle)*tgt_dist;
		}
	}
	
	
	////// cross range limit ///////////////////////////////////////////////////////

	/*double cross_range_angle_limit = (0.18*RAD);
	cross_range_angle_limit = cross_range_angle_limit * (1 + sin(bank_correction*RAD)*1.5);

	if(tgt_dist < 300000) cross_range_angle_limit = (0.004*RAD);
	else if(tgt_dist < 210000) cross_range_angle_limit = (5*RAD);

	if(abs(cross_range_angle) > cross_range_angle_limit)
	{
	bank_dir_sw = int(abs(cross_range_angle)/cross_range_angle);
	}

	if(bank_mode_sw == 0)bank_mode_sw = bank_dir_sw;
	else if (bank_mode_sw != bank_dir_sw) bank_mode_sw = 2;*/

	/////////////////////////////////////////////////////////////////////////////////////////////////////

	double hSpeed = length(_V(vel_hrzn.x,0,vel_hrzn.z));

	double current_h_acc =  (hSpeed - old_hSpeed)/post_step;
	old_hSpeed = hSpeed;

	double payload_factor = (GetMass() - 145000)*0.002;//100
	if(payload_factor < 0)payload_factor = 0;

	double reentry_time = 0;
	double h_acc_req = 0;
	double h_acc_error = 0;

	if(tgt_exists == true)
	{
	reentry_time = (tgt_dist-200000)/((hSpeed+2150+payload_factor)/2);
	h_acc_req = 2*((tgt_dist-200000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
	h_acc_error = h_acc_req - current_h_acc;
	

	if(tgt_dist < 210000)
		{
		reentry_time = (tgt_dist-10000)/((hSpeed+400)/2);
		h_acc_req = 2*((tgt_dist-10000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
		}
	
	}


	hud_tgt_range = tgt_range;
	hud_crossrange_error = cross_range_dist;
	hud_hacc_req = h_acc_req;
	hud_hacc_current = current_h_acc;
	hud_hacc_error = h_acc_error;
	
	


	/////////////////// set wing trim to pitch angle
	wing_trim = GetWTbyAoA(fbw_aoa);
	/////////////////////////////////////////////////

//	sprintf(oapiDebugString(), "bank %0.2f aoa %0.2f h_acc %0.2f h_acc_req %0.2f h_acc_error %0.2f crd %0.2f", fbw_bank, fbw_aoa, current_h_acc, h_acc_req, h_acc_error, cross_range_dist);

	double aoa_angle = fbw_aoa*RAD;
	double fbw_bank_angle = fbw_bank*RAD;

	MATRIX3 rm = rotm(nml_unit, aoa_angle);
	flight_vector = mul(rm,vel_unit);
	rm = rotm(vel_unit,fbw_bank_angle);
	flight_vector = mul(rm, flight_vector);
	normalise(flight_vector);

	GoToFlightVectorRE();
}

void ss_tanker::GoToFlightVectorRE()
{
	DeactivateNavmode(NAVMODE_KILLROT);
	/*if (eng_status == 0)
	{
	reentry_sw = 0;
	KillAllRCS2();
	return;
	}*/
	/*if (rcs_sw2 == 0)
	{
	reentry_sw = 0;
	KillAllRCS2();
	return;
	}*/

	vrot_lim = 30*RAD;
	VECTOR3 fv_unit = flight_vector;
	normalise(fv_unit);
	

	double mass = GetMass();
	VECTOR3 current_pmi;
	GetPMI(current_pmi);
	double inertia_x = current_pmi.x*mass;
	double inertia_y = current_pmi.y*mass;
	double inertia_z = current_pmi.z*mass;
	
	
	double max_rcs = GetThrusterMax(th_rcs[17]);
	double max_torque_x = 2*max_rcs*32;
	double max_torque_y = 2*max_rcs*32;
	if (disable_rcs == 1)
	{
		max_torque_y = 1*max_rcs*32;
	}
	double max_torque_z = 4*max_rcs*5;

	VECTOR3 avel;
	GetAngularVel(avel);

	double max_acc_x = max_torque_x/inertia_x;
	double max_acc_y = max_torque_y/inertia_y;
	double max_acc_z = max_torque_z/inertia_z;

	double stop_angle_x = 0.5*avel.x*(avel.x/max_acc_x);
	double stop_angle_y = 0.5*avel.y*(avel.y/max_acc_y);
	double stop_angle_z = 0.5*avel.z*(avel.z/max_acc_z);

	
	double max_drot_x = max_acc_x*step_time*4;
	double max_drot_y = max_acc_y*step_time*4;
	double max_drot_z = max_acc_z*step_time*4;

	VECTOR3 aspdv;
	GetAirspeedVector(FRAME_LOCAL, aspdv);
	normalise(aspdv);

	VECTOR3 relvel;
	OBJHANDLE gbod = GetGravityRef();
	GetRelativeVel(gbod, relvel);
	double relspeed = length(relvel);
	VECTOR3 relpos;
	GetRelativePos(gbod, relpos);
	double rdist = length(relpos);
	double orb_ang_speed = relspeed/rdist;
	VECTOR3 relpos_hrz = _V(0,1,0);
	VECTOR3 relpos_loc;
	HorizonInvRot(relpos_hrz, relpos_loc);
	normalise(relpos_loc);
	VECTOR3 rot_off = crossp(aspdv, relpos_loc);
	normalise(rot_off);
	rot_off.x = rot_off.x*orb_ang_speed;
	rot_off.y = rot_off.y*orb_ang_speed;
	rot_off.z = rot_off.z*orb_ang_speed;

	VECTOR3 zaxis_vel_nml = crossp(_V(0,0,1),aspdv);
	normalise(zaxis_vel_nml);
	VECTOR3 fv_vel_nml = crossp(fv_unit, aspdv);
	normalise(fv_vel_nml);
	VECTOR3 rot_axis = crossp(fv_vel_nml,zaxis_vel_nml);
	normalise(rot_axis);



	double zaxis_fv_angle = acos(dotp(fv_vel_nml, zaxis_vel_nml));
	if(bank_mode_sw == 2 && zaxis_fv_angle < (30*RAD))bank_mode_sw = 0;

	/*VECTOR3 right_hrz;
	HorizonRot(rot_off,right_hrz);
	normalise(right_hrz);

	VECTOR3 zaxis_vel_nml_hrz;
	HorizonRot(zaxis_vel_nml, zaxis_vel_nml_hrz);
	normalise(zaxis_vel_nml_hrz);

	VECTOR3 fv_vel_nml_hrz;
	HorizonRot(fv_vel_nml,fv_vel_nml_hrz);
	normalise(fv_vel_nml_hrz);

	double fv_zaxis_angle = acos(dotp(fv_vel_nml_hrz,zaxis_vel_nml_hrz));
	double fv_right_angle = acos(dotp(right_hrz,fv_vel_nml_hrz));
	double zaxis_right_angle = acos(dotp(right_hrz,zaxis_vel_nml_hrz));
	if(zaxis_right_angle < PI05 && fv_right_angle < (PI-zaxis_right_angle) && fv_zaxis_angle > zaxis_right_angle)rot_axis = -rot_axis;*/



//	sprintf(oapiDebugString(), "angle %0.3f avel %0.3f %0.3f %0.3f", zaxis_fv_angle*DEG, avel.x*DEG, avel.y*DEG, avel.z*DEG);
	if(zaxis_fv_angle > vrot_lim)zaxis_fv_angle = vrot_lim;

	//if(abs(avel.x) > abs(stop_angle_x))rot_axis.x = 0;
	//if(abs(avel.y) > abs(stop_angle_y))rot_axis.y = 0;
	//if(abs(avel.z) > abs(stop_angle_z))rot_axis.z = 0;

	double pitch_angle = acos(dotp(aspdv,fv_unit));
	double current_pitch_angle = acos(dotp(aspdv,_V(0,0,1)));
	double cos_cpa = cos(current_pitch_angle);
	double sin_cpa = sin(current_pitch_angle);
//	double pitch_error = pitch_angle - current_pitch_angle;
//	if(pitch_error > vrot_lim)pitch_error = vrot_lim;
//	VECTOR3 pe_rot_axis = crossp(_V(0,0,1),aspdv);
//	normalise(pe_rot_axis);

	MATRIX3 rm = rotm(zaxis_vel_nml, -(pitch_angle+PI05));
	headsup_vector = mul(rm, aspdv);
	normalise(headsup_vector);

	double hu_error = acos(dotp(_V(0,1,0), headsup_vector));
//	sprintf(oapiDebugString(), "hu_error angle %0.3f avel %0.3f %0.3f %0.3f", hu_error*DEG, avel.x*DEG, avel.y*DEG, avel.z*DEG);
	if(hu_error > vrot_lim)hu_error = vrot_lim;
	VECTOR3 hu_rot_axis = crossp(headsup_vector, _V(0,1,0));
	normalise(hu_rot_axis);


	VECTOR3 req_acc_hu;
	req_acc_hu.x = hu_rot_axis.x*hu_error*0.2;
	req_acc_hu.y = hu_rot_axis.y*hu_error*0.2;
	req_acc_hu.z = -hu_rot_axis.z*hu_error*0.2;


	VECTOR3 req_acc;
	req_acc.x = (rot_axis.x*zaxis_fv_angle*0.2) - avel.x + rot_off.x;
	req_acc.y = (rot_axis.y*zaxis_fv_angle*0.2) - avel.y + rot_off.y;
	req_acc.z = (-rot_axis.z*zaxis_fv_angle*0.2) + avel.z - rot_off.z;

	req_acc = req_acc+req_acc_hu;

	double thx_lvl = req_acc.x/max_drot_x;
	double thy_lvl = req_acc.y/max_drot_y;
	double thz_lvl = req_acc.z/max_drot_z;
	//sprintf(oapiDebugString(), "slip %0.4f yrot %0.4f reqyv%0.4f", (slipage*DEG),(avel.y*DEG), (reqyv*DEG));

	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;
	if(abs(thx_lvl) < 0.3)thx_lvl = 0;
	pitch_level = 0.5*thx_lvl;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;
	if(abs(thy_lvl) < 0.3)thy_lvl = 0;
	yaw_level = 0.5*thy_lvl;

	if(thz_lvl > 1)thz_lvl = 1;
	else if(thz_lvl < -1)thz_lvl = -1;
	if(abs(thz_lvl) < 0.3)thz_lvl = 0;
	bank_level = 0.5*thz_lvl;


	////////////// x pitch
	if(thx_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
	else if(thx_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	}

	//////////////////////Y YAW
	if(thy_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		}
	else if(thy_lvl < 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	}

	/////////////// z roll
	if(thz_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		}
	else if(thz_lvl < 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, -thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}


}

double ss_tanker::GetAoAbyWT(double wingtrim)
{
	wingtrim = wingtrim*100;
	double est_aoa = 0;
	double wt2 = wingtrim*wingtrim;
	if(adcs_reentry_mode == 1)
	{
	if(wingtrim > 0) est_aoa = 57.5 + (0.5094*wingtrim) - (0.00286*wt2);
	else est_aoa = 57.5 + (0.9030*wingtrim) + (0.00426*wt2);
	}
	else est_aoa = 55.59 + (0.9027*wingtrim) - (0.00857*wt2);

	return est_aoa;
}

double ss_tanker::GetWTbyAoA(double current_aoa)
{
	double est_wt = 0;
	double aoa2 = current_aoa*current_aoa;

	double a = 0;
	double b = 0;
	double c = 0;

//	if(adcs_reentry_mode == 1)
//	{
//	if(current_aoa > 57.5) est_wt = 117.61253 - (5.6410307*current_aoa) + (0.062631731*aoa2);
//	else est_wt =  -148.17197128517893 + (5.3959750637355350*current_aoa) - (0.049027325483437224*aoa2);
//	}
//	else est_wt =  106.52134 - (4.51942766*current_aoa) + (0.046884376*aoa2);

	if(adcs_reentry_mode == 1)
	{
	if(current_aoa > 57.5)
		{
		a = -0.00286;
		b = 0.5094;
		c = 57.5 - current_aoa;
		}
	else
		{
		a = 0.00426;
		b = 0.9030;
		c = 57.5 - current_aoa;
		}
	}
	else
	{
	a = -0.00857;
	b = 0.9027;
	c = 55.59 - current_aoa;
	}

	double b2 = b*b;
	double b24ac = (b2-(4*a*c));
	if(b24ac < 0.0000001)b24ac = 0.0000001;
	est_wt = (-b+(sqrt(b24ac)))/(2*a);

	return est_wt*0.01;
}